
/*!
 * \file kat.c
 * \brief Main routine and banner routine for finesse
 */

/*! \mainpage Finesse - Frequency domain INterferomEter Simulation SoftwarE
 *
 * Version: 0.99.x  (documentation built from $Revision$)
 *
 * \author Andreas Freise
 *
 * FINESSE is a fast interferometer simulation program.  For a given optical
 * setup, it computes the light field amplitudes at every point in the
 * interferometer assuming a steady state.  To do so, the interferometer
 * description is translated into a set of linear equations that are solved
 * numerically. For convenience, a number of standard analyses can be
 * performed automatically by the program, namely computing
 * modulation-demodulation error signals, transfer functions, shot-noise
 * limited sensitivities, and beam shapes.  FINESSE can perform the analysis
 * using the plane-wave approximation or Hermite-Gauss modes. The latter
 * allows computation of the properties of optical systems like telescopes
 * and the effects of mode matching and mirror angular positions. 
 *
 *  @section Copyright notice
 *
 *  This file is part of the interferometer simulation Finesse
 *  http://www.gwoptics.org/finesse
 *
 *  Copyright (C) 1999 onwards Andreas Freise
 *  with parts of the code written by Daniel Brown, Paul Cochrane
 *  and Gerhard Heinzel.
 *
 *  This program is free software; you can redistribute it and/or modify it under
 *  the terms of the GNU General Public License version 3 as published
 *  by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 *  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *  See the GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along with
 *  this library; if not, write to the Free Software Foundation, Inc., 59 Temple Place,
 *  Suite 330, Boston, MA 02111-1307 USA
 */

#include "kat.h"
#include "kat_config.h"
#include "kat_mem.h"
#include "kat_read.h"
#include "kat_io.h"
#include "kat_aux.h"
#include "kat_init.h"
#include "kat_calc.h"
#include "kat_spa.h"
#include "klu.h" 
#include "kat_klu.h"
#ifdef SERVERMODE
#include "kat_server.h"
#endif

#ifdef OSUNIX
#include <fpu_control.h>
#endif

init_variables_t init; //!< Structure to hold initialisation information
interferometer_t inter; //!< Interferometer and simulation variables
options_t options; //!< Options passed to finesse
memory_t mem; //!< Structure to hold memory allocation information
local_var_t vlocal; //!< Unchanged variables for faster function calls
global_var_t vglobal; //!< Structure to hold in, out and gnu filenames
sparse_var_t spa; //!< Variables for sparse matrix routines
klu_sparse_var_t klu; //!< Variables for KLU sparse matrix routines

klu_l_symbolic *Symbolic;
klu_l_numeric *Numeric;
klu_l_common Common;

time_t now; //!< Current time
char mydate[12] = {0}; //!< Date as at time of compilation

char functokens[27] = {0}; //!< Array of function tokens

FILE *fp_log; //!< Log file output file pointer

newfeature_t feature;

//! The main routine

/*!
 * \param argc The number of input parameters
 * \param argv The array of input parameters
 */
int main(int argc, char *argv[]) {

#ifdef OSUNIX
    /*
      This puts the X86 FPU in 64-bit precision mode.  The default
      under Linux is to use 80-bit mode, which produces subtle
      differences from FreeBSD and other systems, eg,
      (int)(1000*atof("0.3")) is 300 in 64-bit mode, 299 in 80-bit
      mode.

      To be tested! Especially the function 'pow' might cause problems.
     */
    fpu_control_t cw;

    _FPU_GETCW(cw);
    cw &= ~_FPU_EXTENDED;
    cw |= _FPU_DOUBLE;
    _FPU_SETCW(cw);
#endif

    FILE *ifp;

    char *s, *s2;
    int basenamelength;
    char errmsg[ERR_MSG_LEN] = {0};

    inter.setup = false;

    strcpy(mydate, MYTIME);
    strcpy(functokens, "abcdfghijklmnopqrstuvwxyz");

    /* **** get options **** */

    // reset options for ...
    options.quiet = 0; // switch off all output
    options.help = 0; // print help
    options.klu = 0; // use SPARSE as default
    options.klu_manual = 0; // overwrite KLU-SPARSE setting manually
    options.check = 0; // do a consistency check of the interferometer matrix
    options.format = 0; // print data to output file with more precision
    options.perl1 = 0; // switch off some output
    options.maxmin = 0; // print max/min values
    options.servermode = 0; // switch between standalone and server mode
    options.portnumber = 0; // port number for client server connections (11000 to 110010)
    options.version = 0; // print version number and quit?
    options.noheader = 0; // do not print header in output file

    bool converting = false;
    
    // process command line arguments
    int num_options = 0;
    while ((argc > 1) && (argv[1 + num_options][0] == '-')) {
        bool found = false;

        if(converting)
            gerror("The --convert option can only be used by itself");
            
        // process -h option (print first help page)
        if (strcasecmp(argv[1 + num_options], "-h") == 0) {
            options.help = 1;
            ++num_options;
            --argc;
            found = true;
        }
            // process -hh option (print second help page)
        else if (strcasecmp(argv[1 + num_options], "-hh") == 0) {
            options.help = 2;
            ++num_options;
            --argc;
            found = true;
        }
          
        else if (strcasecmp(argv[1 + num_options], "-convert") == 0) {
            // this should be the first argument 
            if(num_options != 0)
                gerror("The --convert option can only be used by itself");
            
            converting = true;
            ++num_options;
            --argc;
            found = true;
        }
            // process -v option 
        else if (strcasecmp(argv[1 + num_options], "-v") == 0) {
            options.version = 1;
            ++num_options;
            --argc;
            found = true;
        }
            // process -max option (print max/min values)
        else if (strcasecmp(argv[1 + num_options], "-max") == 0) {
            options.maxmin = 1;
            ++num_options;
            --argc;
            found = true;
        }
            // process -klu option (switch to KLU package)
        else if (strcasecmp(argv[1 + num_options], "-klu") == 0) {
            options.klu = 1;
            options.klu_manual = 1;
            ++num_options;
            --argc;
            found = true;
        }
            // process -sparse option (switch to SPARSE package)
        else if (strcasecmp(argv[1 + num_options], "-sparse") == 0) {
            options.klu = 0;
            options.klu_manual = 1;
            ++num_options;
            --argc;
            found = true;
        }
            // process --perl1 option (switch off some output)
        else if (strcasecmp(argv[1 + num_options], "--perl1") == 0) {
            options.perl1 = 1;
            ++num_options;
            --argc;
            found = true;
        }
            // process --quiet option (switch off all output)
        else if (strcasecmp(argv[1 + num_options], "--quiet") == 0) {
            options.quiet = 1;
            ++num_options;
            --argc;
            found = true;
        }
            // process --noheader option (do not print header in output file)
        else if (strcasecmp(argv[1 + num_options], "--noheader") == 0) {
            options.noheader = 1;
            ++num_options;
            --argc;
            found = true;
        }
#ifdef TEST
            // process --test option (runs unit tests, NOT for user distro)
        else if (strcasecmp(argv[1 + num_options], "--test") == 0) {
            printf("\n##### Running tests #####\n");
            RunAllTests();
            printf("##### Finished tests #####\n");

            // clean up nicely
            fflush(stdout);
            fflush(stderr);
            exit(0);
        }
#endif

            // process --new option (switch on new features)
        else if (strcasecmp(argv[1 + num_options], "--new") == 0) {
            if (options.new == -1) {
                gerror("options --new and --old mutually exclusive\n");
            }
            options.new = 1;
            ++num_options;
            --argc;
            found = true;
        }
            // process --old option (try to switch back to last version)
        else if (strcasecmp(argv[1 + num_options], "--old") == 0) {
            if (options.new == 1) {
                gerror("options --new and --old mutually exclusive\n");
            }
            options.new = -1;
            ++num_options;
            --argc;
            found = true;
        }
            // process -c option (do consistency check of interferometer matrix)
        else if (strcasecmp(argv[1 + num_options], "-c") == 0) {
            options.check = 1;
            ++num_options;
            --argc;
            found = true;
        }
            // process -f option (print data to file with more precision)
        else if (strcasecmp(argv[1 + num_options], "-f") == 0) {
            options.format = 1;
            ++num_options;
            --argc;
            found = true;
        }
            // process --server option (switch into server mode)
        else if (strcasecmp(argv[1 + num_options], "--server") == 0) {
            options.servermode = 1;
            ++num_options;
            --argc;
            // try to read port number 
            if ((num_options < argc) && (argv[1 + num_options][0] != '-')) {
                if (!(my_atoi(argv[1 + num_options], &options.portnumber)) &&
                        options.portnumber >= 11000 &&
                        options.portnumber <= 11010) {
                    found = true;
                    ++num_options;
                    --argc;
                }
            }
            if (!found) {
                banner(stdout, "*", "*", "*");
                gerror("usage: kat --server <portnumber (11000 to 11010)> "
                        "[options] inputfile ...\n");
            }
        }

        // warn that option is not found
        if (!found) {
            banner(stdout, "*", "*", "*");
            gerror("unknown option '%s'\n", argv[1 + num_options]);
        }        
    }

    // print version number
    if (options.version == 1) {
        print_version(stdout);
        my_finish(0);
    }

    // print the first help page
    if (options.help == 1) {
        print_help();
        my_finish(0);
    }

    // print the second help page
    if (options.help == 2) {
        print_help2();
        my_finish(0);
    }

    // switch of max if in servermode
    if (options.maxmin && options.servermode) {
        warn("the option '-max' cannot be used in server mode.\n");
        options.maxmin = 0;
    }
    
    if(!converting){
        // check that the correct number of args are passed in
        if (argc < 2 || argc > 4) {
            banner(stdout, "*", "*", "*");
            strcpy(errmsg, "usage: kat [options] infile [outfile [gnufile]]\n");
            strcat(errmsg, "           or usage: kat [options] basename\n");
            strcat(errmsg, "             (kat -h displays help message)\n");
            gerror(errmsg);
        }
    }else{
        // should only be the .knm file name after the --convert command
        if(argc == 2)
            convert_kmn_file(argv[1 + num_options], "");
        else if(argc == 3)
            convert_kmn_file(argv[1 + num_options], argv[2 + num_options]);
        else 
            gerror("usage: kat -convert kmn_prefix [new_kmn_prefix]\n"
                   "               or usage: kat [options] basename\n"
                   "                 (kat -h displays help message)\n");
        
        // don't do anything after the convert so exit
        exit(0);
    }


    // process the arguments
    if (argc == 2) { // only the basename is specified...
        inter.basename = duplicate_string(argv[num_options + 1]);
        basenamelength = strlen(inter.basename);
        if (basenamelength > 4) {
            s2 = inter.basename + strlen(inter.basename) - 4;
            if (strcasecmp(s2, ".kat") == 0) {
                inter.basename[strlen(inter.basename) - 4] = '\0';
            }
        }

        s2 = inter.basename + strlen(inter.basename) - 1;
        if (*s2 == '.') {
            *s2 = '\0';
        }

        // stripping leading dots and slashes
        while (*inter.basename == '.' || *inter.basename == '/' || *inter.basename == '\\') {
            inter.basename++;
        }

        strcpy(vglobal.input_fname, inter.basename);
        strcat(vglobal.input_fname, ".kat");
        strcpy(vglobal.output_fname, inter.basename);
        strcat(vglobal.output_fname, ".out");
        strcpy(vglobal.gnuplot_fname, inter.basename);
        strcat(vglobal.gnuplot_fname, ".gnu");
        strcpy(vglobal.log_fname, inter.basename);
        strcat(vglobal.log_fname, ".log");
    } else if (argc == 3) { // the input and output file names are specified...
        inter.basename = duplicate_string(argv[num_options + 1]);
        if ((s = strchr(inter.basename, '.')) != NULL) {
            *s = '\0';
        }
        strcpy(vglobal.input_fname, argv[num_options + 1]);
        strcpy(vglobal.output_fname, argv[num_options + 2]);
        strcpy(vglobal.gnuplot_fname, inter.basename);
        strcat(vglobal.gnuplot_fname, ".gnu");
        strcpy(vglobal.log_fname, inter.basename);
        strcat(vglobal.log_fname, ".log");
    } else if (argc == 4) { // input, output and gnuplot file names are given...
        inter.basename = duplicate_string(argv[num_options + 1]);
        if ((s = strchr(inter.basename, '.')) != NULL) {
            *s = '\0';
        }
        strcpy(vglobal.input_fname, argv[num_options + 1]);
        strcpy(vglobal.output_fname, argv[num_options + 2]);
        strcpy(vglobal.gnuplot_fname, argv[num_options + 3]);
        strcpy(vglobal.log_fname, inter.basename);
        strcat(vglobal.log_fname, ".log");
    }


    // this code is repeated in kat_gnu.c void matlabfile(FILE *fp, char *fname)
    // TODO define new string ...
    char matlab_func_str[LINE_LEN];
    strcpy(matlab_func_str, inter.basename);
    int replace_error = replace_string(matlab_func_str, "-", "_", 1, 1);
    if (replace_error == -1) {
        bug_error("could not replace '-' in matlab string");
    }
    strcpy(vglobal.matlab_fname, matlab_func_str);
    strcat(vglobal.matlab_fname, ".m");

    // open the log file
    open_file_to_write_ascii(vglobal.log_fname, &fp_log);

    /*
    if ((fp_log = fopen(vglobal.log_fname, "w")) == NULL) {
      gerror("cannot open '%s' for log output\n", vglobal.log_fname);
    }
     */

    // set up message reporting if perl1/quiet option isn't set
    if (!options.perl1 && !options.quiet) {
        banner(stdout, vglobal.input_fname, vglobal.output_fname,
                vglobal.gnuplot_fname);
    }
    // output the banner to the log file in any case
    if (fp_log != NULL) {
        banner(fp_log, vglobal.input_fname, vglobal.output_fname,
                vglobal.gnuplot_fname);
    }


    // check to see if we can open the input file
    if ((ifp = fopen(vglobal.input_fname, "r")) == NULL) {
        gerror("cannot open '%s'\n", vglobal.input_fname);
    }

    // read in environmental variables
    read_env();

    // read the initialisation file
    read_init();
    initialise_simulation_variables();

    // prescan and reading of input file
    pre_scan(ifp);
    read_file(ifp);
    fclose(ifp);
    fflush(stdout);

    // check whether we should automatically retrace
    if (inter.rebuild & 2 && !inter.retrace_manual) {
        inter.retrace = 1;
        message("Using automatic retrace\n");
    }

    if (inter.tem > 3 && !options.klu_manual) {
        // Use KLU when maxtem>3
        options.klu = 1;
    }

    // print helpful info if required
    if ((inter.tem > 4 && options.check) || inter.tem > 10) {
        message("  ...initialising system, please wait!\n");
    }

    fflush(stdout);

    // set everything up
    setup_system();

    fflush(stdout);

    if (options.klu) {
        message("Using KLU library\n");
    }

    fflush(stdout);

    FILE *fp;
#if DEBUG

    // clean up appendable files
    fp = fopen("dump_sparse_matrix_standard.dat", "w");
    fclose(fp);
    fp = fopen("dump_sparse_matrix_qcorr.dat", "w");
    fclose(fp);
    fp = fopen("dump_solution_standard.dat", "w");
    fclose(fp);
    fp = fopen("dump_solution_qcorr.dat", "w");
    fclose(fp);
    fp = fopen("dump_matrix_standard.dat", "w");
    fclose(fp);
    fp = fopen("dump_matrix_qcorr.dat", "w");
    fclose(fp);
#endif

    // do sparse matrix stuff
    if (options.klu) {
        klu_create();
    } else {
        spa_create();
    }


    if (options.klu) {
        klu_get();
    } else {
        spa_get();
    }

    //spa_order();

    // check the connections in the interferometer
    check_connections();

    if (!options.servermode) {
        make_space();
    }

    // Preorder KLU Matrix
    if (options.klu) {
        // dummy fill for matrix
        spa_fill(inter.light_in_list[0].f0, STANDARD);
        //spa_rhs_clear(STANDARD);
        Symbolic = klu_l_analyze(inter.num_eqns, klu.col_ptr, klu.row_index, &Common);
        Numeric = klu_zl_factor(klu.col_ptr, klu.row_index, klu.value, Symbolic, &Common);
        klu_zl_sort(Symbolic, Numeric, &Common);
    }

    assert(mem.restab != NULL);

    // print matrix to file for looking at the sparsity for example 
    if (inter.printmatrix && !options.quiet) {
        if (options.klu) {
            //--------------------------------------------------------------------------------

            int i;
            message("Rank of n x n (CCS) matrix: n=%d, no. of non-zero elements: %d\n",
                    inter.num_eqns, inter.num_nonzero);

            printf("col_ptr: ");
            for (i = 0; i < inter.num_eqns + 1; i++) {
                printf("%ld ", klu.col_ptr[i]);
            }
            printf("\n");

            printf("row_index: ");
            for (i = 0; i < inter.num_nonzero; i++) {
                printf("%ld ", klu.row_index[i]);
            }
            printf("\n");

            for (i = 0; i < inter.num_nonzero; i++) {
                printf("value: %d - %.2f, %.2f \n", i, klu.value[2 * i], klu.value[2 * i + 1]);
            }
            //--------------------------------------------------------------------------------
        } else {
            spa_fill(inter.light_in_list[0].f0, STANDARD);
            fp = fopen("test.dat", "w");
            fclose(fp);
            spFileMatrix(spa.spamat, "test.dat", "", 0, 1, 0);
        }
    }



    if (inter.x3_axis_is_set) {
        x3_out();
    } else if (options.servermode) {
#ifdef SERVERMODE
        initserver();
#else
        printf("Servermode not implemented in Windows binary\n");
#endif
    } else {
        x12_out();
    }


    //  if (OS == WIN)
    //    spa_destroy(); // Memory leak ???????????????????????????

    if (options.klu) {
        // free KLU sparse matrix
        klu_destroy();
    }

    // TODO need to add more memory freeing but for now here is the mirror stuff
    free_memory_for_mirror_list();

    // be nice and clean up after ourselves
    my_finish(0);
    return 0;
}

//! Prints the banner for Finesse

/*!
 * \param fp The file pointer for direction of the output
 * \param input_fname The input file name
 * \param output_fname The output file name
 * \param gnuplot_fname The name of the gnuplot file
 */
void banner(FILE *fp, char *input_fname, char *output_fname,
        char *gnuplot_fname) {
    now = time(NULL);

    char extra_space[80]="                                                     ";
    int num_spaces=33-strlen(VERSION)-strlen(SVN_REVISION);
    if (num_spaces<0 || num_spaces>20){
        num_spaces=0;
    }
    
    // print the banner if not quiet
    fprintf(fp, "\n"
            "------------------------------------------------------------------------\n"
            "                     FINESSE %s %.*s (build %s)\n"
            "       o_.-=.        Frequency domain INterferomEter Simulation SoftwarE\n"
            "      (\\'\".\\|        %s         http://www.gwoptics.org/finesse/\n"
            "      .>' (_--.\n"
            "   _=/d   ,^\\        Input file %s,\n"
            "  ~~ \\)-'   '        Output file %s,\n"
            "     / |             Gnuplot file %s \n"
            "    '  '                                        %s"
            "------------------------------------------------------------------------\n",
            VERSION, num_spaces, extra_space, SVN_REVISION, mydate,
            input_fname, output_fname, gnuplot_fname,
            asctime(localtime(&now)));
}

/*
 * Local variables:
 * c-indentation-style: bsd
 * c-basic-offset: 2
 * indent-tabs-mode: nil
 * End:
 *
 * vim: tabstop=2 expandtab shiftwidth=2:
 */
