// $Id$

/*!
 * \file kat.h
 * \brief Main header file for Finesse
 * @section Copyright notice
 *
 *  This file is part of the interferometer simulation Finesse
 *  http://www.gwoptics.org/finesse
 *
 *  Copyright (C) 1999 onwards Andreas Freise
 *  with parts of the code written by Daniel Brown, Paul Cochrane
 *  and Gerhard Heinzel.
 *
 *  This program is free software; you can redistribute it and/or modify it under
 *  the terms of the GNU General Public License version 3 as published
 *  by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 *  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *  See the GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along with
 *  this library; if not, write to the Free Software Foundation, Inc., 59 Temple Place,
 *  Suite 330, Boston, MA 02111-1307 USA
 */

#ifndef KAT_H
#define KAT_H

#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <math.h>
#include <time.h>
#include <string.h>
#include <strings.h>
#include <ctype.h>
#include <assert.h>
#include "spMatrix.h"
#include "formulc.h"
#include "kat_constants.h"
#include "md5.h"

/* ================================================================ */



//! information for new features

struct newfeature {
};

//! new feature type
typedef struct newfeature newfeature_t;

//! complex number type

/*!
 * Please note that this is the Finesse implementation of the complex type,
 * NOT the C implementation.  Hence, we use the name complex_t when
 * specifying the type of a variable.
 *
 * \todo Change this to the C complex type?
 */
struct complex {
    double re; //!< real part of a complex number
    double im; //!< imaginary part of a complex number
};

//! complex number type
typedef struct complex complex_t;

typedef enum map_integration_method {
    RIEMANN_SUM_NEW = 1,
    CUBA_CUHRE_SERIAL = 2,
    CUBA_CUHRE_PARA = 3,
    RIEMANN_SUM_OLD = 4
} int_method_t;

//! information about a node within the interferometer

struct node {
    char name[MAX_TOKEN_LEN]; //!< node name
    int connect; //!< is the node connected?
    int io; //!< input(=2) or output(=1) node
    bool q_changed; //!< True if the q value has been changed to what it previously was
    complex_t qx; //!< x component of beam q parameter
    complex_t qy; //!< y component of beam q parameter
    int component_index; //!< component index
    double *n; //!< refractive index of an attached space
    bool q_is_set; //!< is Gaussian beam parameter set?
    bool gnd_node; //!< is this node a GND_NODE?
};

//! node type
typedef struct node node_t;

//! amplitude direction information

struct amp_dir {
    int to; //!< amplitude direction goes towards element
    int fro; //!< amplitude direction goes away from element
};

//! amplitude direction type
typedef struct amp_dir amp_dir_t;

//! information about an ABCD matrix

struct ABCD {
    double A; //!< A element of ABCD matrix
    double B; //!< B element of ABCD matrix
    double C; //!< C element of ABCD matrix
    double D; //!< D element of ABCD matrix
};

//! ABCD matrix type
typedef struct ABCD ABCD_t;

//! memory allocation information relevant for elements in the interferometer

struct memory {
    int num_modulators; //!< number of modulators
    int num_mirrors; //!< number of mirrors
    int num_beamsplitters; //!< number of beam splitters
    int num_spaces; //!< number of free spaces
    int num_lenses; //!< number of lenses
    int num_gratings; //!< number of gratings
    int num_squeezers; //!< number of squeezers
    int num_light_inputs; //!< number of input light fields (i.e. lasers)
    int num_diodes; //!< number of Faraday isolators (diodes)
    int num_signal_inputs; //!< number of signal inputs
    int num_variables; //!< number of dummy parameters (used with xaxis)

    //! number of Gaussian commands 
    /*!
     * comes from the user-defined Gaussian beam parameter
     */
    int num_gauss_cmds;

    int num_light_outputs; //!< number of light outputs

    //! number of beam parameter outputs
    /*!
     * i.e. the beam parameter detector
     */
    int num_beam_param_outputs;

    int num_cavity_param_outputs; //!< number of cavity parameter outputs
    int num_convolution_outputs; //!< number of convolution outputs
    int num_mirror_phase_outputs; //!< number of mirror_phase outputs
    int num_outputs; //!< number of outputs

    //! number of attributes 
    /*!
     * The number of attributes of mirrors, beamsplitters etc.
     */
    int num_attributes;

    int num_cavities; //!< number of cavities
    int num_gnuterm_cmds; //!< number of gnuplot commands
    int num_diff_cmds; //!< number of differentiation commands defined
    int num_constants; //!< number of constants (defined in input file)
    int num_scale_cmds; //!< number of scale commands defined
    int num_nodes; //!< number of nodes
    int num_dump_nodes; //!< number of dump nodes (GND_NODEs)
    int num_reduced_nodes; //!< number of nodes minus number of dump nodes
    int num_components; //!< number of components
    int num_frequencies; //!< number of frequencies

    //! order of higher order modes
    /*!
     * The order of higher order (Hermite-Gauss) modes is defined as max(n+m)
     * for TEM_nm modes
     */
    int hg_mode_order;

    int num_fields; //!< number of field amplitudes
    int num_quadratures; //!< number of quadratures
    int num_pointers; //!< number of pointers to the sparse matrix
    int num_node_names; //!< number of node names
    double *restab; //!< pointer to the output data array

    //! list of all TEM modes
    /*!
     * Usually TEM modes are accessed using n and m as the index but in this
     * case they are indexed by the one number
     */
    int *all_tem;

    char **names; //!< names of everything mentioned in .kat input file
    double *work; //!< workspace for dcuhre function
    int worklen; //!< the length of the workspace
    int num_set_cmds; //!< number of set commands in input file
    int num_put_cmds; //!< number of put commands in input file

    //! number of func commands in input file
    /*!
     * func commands are used for computing formulae in Finesse
     */
    int num_func_cmds;

    int num_locks; //!< number of locks

    //! help check for multiple pointers to same location when building matrices
    int *ilist1;
    //! help check for multiple pointers to same location when building matrices
    int *jlist1;
    //! help check for multiple pointers to same location when building matrices
    int *ilist2;
    //! help check for multiple pointers to same location when building matrices
    int *jlist2;
};

//! memory type
typedef struct memory memory_t;

typedef enum interpolation_method {
    NEAREST = 1,
    LINEAR = 2,
    SPLINE = 3
} interpolation_method_t;

// Structure to store the coupling coefficients for various effects
// such as apertures, tilts etc.

typedef struct mirror_kmn {
    //! coupling coefficient for TEM mode from mode 1 into mode 1
    complex_t **k11;
    //! coupling coefficient for TEM mode from mode 1 into mode 2
    complex_t **k12;
    //! coupling coefficient for TEM mode from mode 2 into mode 1
    complex_t **k21;
    //! coupling coefficient for TEM mode from mode 2 into mode 2
    complex_t **k22;
    bool IsIdentities;
} mirror_kmn_t;

// Stores the q values for calculating the various KNM coefficients for a mirror
// Each q should be calculated using the incoming beam q and transformed using
// ABCD matrix

typedef struct mirror_kmn_q {
    complex_t qxt1_11;
    complex_t qxt2_11;
    complex_t qyt1_11;
    complex_t qyt2_11;

    complex_t qxt1_12;
    complex_t qxt2_12;
    complex_t qyt1_12;
    complex_t qyt2_12;

    complex_t qxt1_21;
    complex_t qxt2_21;
    complex_t qyt1_21;
    complex_t qyt2_21;

    complex_t qxt1_22;
    complex_t qxt2_22;
    complex_t qyt1_22;
    complex_t qyt2_22;
} mirror_kmn_q_t;

typedef struct surface_merged_map {
    unsigned int interpolation_size; //!< Size of interpolation kernel, must be odd number greater than 3
    interpolation_method_t interpolation_method; //!< interpolation method to use: spline=1 or linear=2
    int_method_t integration_method; //!< Integration method to use for kmn calculations
    
    double x0; //!< x-coordinate of optical center of the map (in steps)
    double y0; //!< y-coordinate of optical center of the map (in steps)
    double xstep; //!< the size in x direction of one tile of the grid
    double ystep; //!< the size in y direction of one tile of the grid
    int rows; //!< vertical size of array
    int cols; //!< horizontal size of array

    double **t_phs; //!< transmission phase, to be computed from data of multiple maps
    double **t_abs; //!< transmission amplitude, to be computed from data of multiple maps
    double **r_phs; //!< reflection phase, to be computed from data of multiple maps
    double **r_abs; //!< reflection amplitude, to be computed from data of multiple maps
    double *x;
    double *y; //!< x and y variables for map interpolation, used by GSL interpolation

    mirror_kmn_t kmn; //!< Coupling coefficients in each direction

    char filename[LINE_LEN]; //!< this is the base filename for stored information, it will have .knm and .map appended to it depending on the data
    char name[LINE_LEN]; //!< name given in file
    bool noMapPresent; //!< Boolean flag that states if any maps have been applied
    bool phaseIsOnlyTransmission; //!< True if only a tranismission map was applied
    bool kmn_calculated; //!< True if the kmn have been calculated for the map already

    //config settings
    bool show_kmn_neval;
    bool save_to_file;
    bool save_kmn_matrices;
    bool kmn_save_binary;
    bool save_interp_file;

    // variables below prefix with a _ are the values used to calculate the current
    // kmn values
    complex_t _qx1, _qy1, _qx2, _qy2;
    double _angle;
    double _nr1, _nr2, _lambda;
    double _r_aperture, _xbeta, _ybeta;
    int _maxtem;
    double _relerr,_abserr;
    unsigned int _intmeth, _intermeth, _intersize;
    
    // flags to keep track of messages that the user has been told on per mirror basis
    // this is to stop the same message apperaing every integration for example
    unsigned int usermessages;
} surface_merged_map_t;

//! surface_map: holds the data of a mirror surface map

typedef struct surface_map {
    int type; //!< type of the surface map (phase, absorption, reflectivty)
    bool reflection; //!< affect reflected light?
    bool transmission; //!< affect transmitted light?
    int function; //!< 0:comes from a file, 1: comes from a fucntion
    char filename[LINE_LEN]; //!< file for reading in a numeric phase/amplitde map;
    unsigned char hash[16]; //!< same size as the md5 digest
    char name[LINE_LEN]; //!< name given in file
    double scaling; //!< global scaling coefficient of data values
    double **data; //!< map values to be read from file
    double x0; //!< x-coordinate of optical center of the map (in steps)
    double y0; //!< y-coordinate of optical center of the map (in steps)
    double xstep; //!< the size in x direction of one tile of the grid
    double ystep; //!< the size in y direction of one tile of the grid
    int rows; //!< vertical size of array
    int cols; //!< horizontal size of array
    int has_coefficients; //!< coupling coefficients have been computed (0:no, 1:partly, 2: yes)
    bool has_map_source; //!< map source data is available
    double Loss; //!< Loss derived from mirror R,T
} surface_map_t;

//! information used for 'variable' a method for using dummy parameters

struct variable {
    char name[MAX_TOKEN_LEN]; //!< parameter name
    double param; //!< parameter value
};

//! variable type
typedef struct variable variable_t;


enum MIRROR_USERMESSAGES{
    USING_POLAR=1,
    MAP_TOO_SMALL=2,
    INTERP_SAVE=4,
    KMN_TRANSPOSE=8,
    INT_METHODS=16,
    MAP_RES_TOO_LOW=32
};

//! information pertaining to a mirror

struct mirror {
    char name[MAX_TOKEN_LEN]; //!< mirror name
    /*! 
     * Note that when loss is included that the relationship between the
     * \b power reflectance, the transmittance and the loss is:
     * \f[
     * R + T + Loss = 1
     * \f]
     * which, written in terms of \b amplitude reflectance, transmittance 
     * and loss parameters is:
     * \f[
     * r^2 + t^2 + loss^2 = 1
     * \f]
     */
    double R; //!< mirror power reflectance
    double T; //!< mirror power transmittance
    double Loss; //!< mirror loss

    //! Tuning of the mirror
    /*!
     * It is convenient to split distances between optical elements into two
     * parameters, one macroscopic `length' which is an integer multiple of
     * the wavelength of light used and a microscopic tuning which is the
     * remaining distance between the total distance and the macroscopic
     * length.  The tuning is given as a phase in radians with \f$ 2\pi \f$ 
     * referring to one wavelength.  In Finesse tunings are entered and printed 
     * in degrees.
     */
    double phi;

    double r_aperture; //!< physical size of mirror, infinite if set to 0

    double Rcx; //!< radius of curvature in x plane (tangential)
    double Rcy; //!< radius of curvature in y plane (sagittal)

    mirror_kmn_t kmn_bayer_helms;
    mirror_kmn_t kmn_aperture;
    mirror_kmn_t kmn_curvature;

    int node1_index; //!< index of the connected node
    int node2_index; //!< index of the connected node

    int node1_reduced_index; //!< node index for classical interferometer matrix
    int node2_reduced_index; //!< node index for classical interferometer matrix

    amp_dir_t node1_amp_dir; //!< direction of propagation at node 1
    amp_dir_t node2_amp_dir; //!< direction of propagation at node 2

    //! matrix element of the component in the interferometer matrix
    complex_t ***a11;
    //! matrix element of the component in the interferometer matrix
    complex_t ***a12;
    //! matrix element of the component in the interferometer matrix
    complex_t ***a21;
    //! matrix element of the component in the interferometer matrix
    complex_t ***a22;

    //! Stores the various coupling coefficients for reflection transmission
    mirror_kmn_t kmn;

    //! Transformation matrix for q of mirror; reflection in tangential plane (node 1 -> node1)
    ABCD_t qqr1t;
    //! Transformation matrix for q of mirror; reflection in sagittal plane (node 1 -> node1)
    ABCD_t qqr1s;
    //! Transformation matrix for q of mirror; reflection in tangential plane (node 2 -> node2)
    ABCD_t qqr2t;
    //! Transformation matrix for q of mirror; reflection in sagittal plane (node 2 -> node2)
    ABCD_t qqr2s;
    //! Transformation matrix for q of mirror; transmission in tangential plane
    ABCD_t qqtt;
    //! Transformation matrix for q of mirror; transmission in sagittal plane
    ABCD_t qqts;

    int attribs; //!< bit-coded flag for which attribs have been set
    int maps; //!< bit-coded flag for which maps have been set

    double M; //!< mass of mirror

    double beta_x; //!< misalignment angle in x-plane
    double beta_y; //!< misalignment angle in y-plane

    int rebuild; //!< rebuild ABCD matrix and k_mnmn coefficients?

    double I1; //!< intensity at node 1
    double I2; //!< intensity at node 2

    double carrier_phase; //!< phase of the carrier frequency

    complex_t ***q11; //!< quantum noise matrix element mode 1 -> mode 1
    complex_t ***q12; //!< quantum noise matrix element mode 1 -> mode 2
    complex_t ***q21; //!< quantum noise matrix element mode 2 -> mode 1
    complex_t ***q22; //!< quantum noise matrix element mode 2 -> mode 2

    int map_type_set; //!< Which map types have been set
    surface_map_t * map[MAX_MAPS]; //!< Surface map storage 
    int num_maps; //!< number of surface maps   
    surface_merged_map_t map_merged;
    double angle; //!< angle of the mirror about the z axis
    mirror_kmn_q_t kmn_q; //!< q values for knm calculation
    // flags stating whether a particular Knm should be calculated
    bool K11_calc;
    bool K12_21_calc;
    bool K22_calc;
};

//! mirror type
typedef struct mirror mirror_t;

//! information for a free space element

struct space {
    char name[MAX_TOKEN_LEN]; //!< name of free space element
    double L; //!< length of free space element

    //! matrix element of the component in the interferometer matrix
    complex_t ***a12;
    complex_t ***a21;

    //! coupling coefficient for TEM mode from mode 1 into mode 2
    complex_t **k12;
    //! coupling coefficient for TEM mode from mode 2 into mode 1
    complex_t **k21;

    int node1_index; //!< index of node1
    int node2_index; //!< index of node2

    int node1_reduced_index; //!< node index for classical interferometer matrix
    int node2_reduced_index; //!< node index for classical interferometer matrix

    amp_dir_t node1_amp_dir; //!< direction of propagation at node 1
    amp_dir_t node2_amp_dir; //!< direction of propagation at node 2

    double n; //!< index of refraction
    int rebuild; //!< rebuild the element?
    int attribs; //!< bit-coded flag for which attribs have been set

    ABCD_t qq; //!< Transformation matrix for q of free space

    complex_t ***q12; //!< quantum noise matrix element mode 1 -> mode 2
    complex_t ***q21; //!< quantum noise matrix element mode 2 -> mode 1

    // q[mode index][output][frequency][quadrature]

    double gouy_x; //!< Gouy phase in x direction
    double gouy_y; //!< Gouy phase in y direction
};

//! free space type
typedef struct space space_t;

//! information describing a beam splitter

struct beamsplitter {
    char name[MAX_TOKEN_LEN]; //!< beam splitter name
    /*! 
     * Note that when loss is included that the relationship between the
     * \b power reflectance, the transmittance and the loss is:
     * \f[
     * R + T + Loss = 1
     * \f]
     * which, written in terms of \b amplitude reflectance, transmittance 
     * and loss parameters is:
     * \f[
     * r^2 + t^2 + loss^2 = 1
     * \f]
     */
    double R; //!< beam splitter power reflectance
    double T; //!< beam splitter power transmittance
    double Loss; //!< loss in beam splitter

    double phi; //!< beam splitter tuning \see mirror_t::phi

    double alpha; //!< angle of incidence

    double ir1; //!< refractive index (where???)
    double ir2; //!< refractive index

    double beta_x; //!< misalignment angle in x-plane
    double beta_y; //!< misalignment angle in y-plane

    int node1_index; //!< index of the connected node
    int node2_index; //!< index of the connected node
    int node3_index; //!< index of the connected node
    int node4_index; //!< index of the connected node

    int node1_reduced_index; //!< node index for classical interferometer matrix
    int node2_reduced_index; //!< node index for classical interferometer matrix
    int node3_reduced_index; //!< node index for classical interferometer matrix
    int node4_reduced_index; //!< node index for classical interferometer matrix

    amp_dir_t node1_amp_dir; //!< direction of propagation at node 1
    amp_dir_t node2_amp_dir; //!< direction of propagation at node 2
    amp_dir_t node3_amp_dir; //!< direction of propagation at node 3
    amp_dir_t node4_amp_dir; //!< direction of propagation at node 4

    //! matrix element of the component in the interferometer matrix
    complex_t ***a12;
    //! matrix element of the component in the interferometer matrix
    complex_t ***a21;
    //! matrix element of the component in the interferometer matrix
    complex_t ***a13;
    //! matrix element of the component in the interferometer matrix
    complex_t ***a31;
    //! matrix element of the component in the interferometer matrix
    complex_t ***a24;
    //! matrix element of the component in the interferometer matrix
    complex_t ***a42;
    //! matrix element of the component in the interferometer matrix
    complex_t ***a34;
    //! matrix element of the component in the interferometer matrix
    complex_t ***a43;

    //! coupling coefficient for TEM mode from mode 1 into mode 2
    complex_t **k12;
    //! coupling coefficient for TEM mode from mode 2 into mode 1
    complex_t **k21;
    //! coupling coefficient for TEM mode from mode 1 into mode 3
    complex_t **k13;
    //! coupling coefficient for TEM mode from mode 3 into mode 1
    complex_t **k31;
    //! coupling coefficient for TEM mode from mode 2 into mode 4
    complex_t **k24;
    //! coupling coefficient for TEM mode from mode 4 into mode 2
    complex_t **k42;
    //! coupling coefficient for TEM mode from mode 3 into mode 4
    complex_t **k34;
    //! coupling coefficient for TEM mode from mode 4 into mode 3
    complex_t **k43;

    // tangential ABCD matrices
    //! transformation matrix for q of element; reflection in tangential plane 1
    ABCD_t qqr1t;
    //! transformation matrix for q of element; transmission in tangential plane 1
    ABCD_t qqt1t;
    //! transformation matrix for q of element; reflection in tangential plane 2
    ABCD_t qqr2t;
    //! transformation matrix for q of element; transmission in tangential plane 2
    ABCD_t qqt2t;

    // sagittal ABCD matrices
    //! transformation matrix for q of element; reflection in sagittal plane 1
    ABCD_t qqr1s;
    //! transformation matrix for q of element; transmission in sagittal plane 1
    ABCD_t qqt1s;
    //! transformation matrix for q of element; reflection in sagittal plane 2
    ABCD_t qqr2s;
    //! transformation matrix for q of element; transmission in sagittal plane 2
    ABCD_t qqt2s;

    int attribs; //!< bit-coded flag for which attribs have been set
    double M; //!< mass of beam splitter

    double Rcx; //!< radius of curvature in x-plane (tangential)
    double Rcy; //!< radius of curvature in y-plane (sagittal)

    int rebuild; //!< rebuild the element?

    double I1; //!< intensity at node 1
    double I2; //!< intensity at node 2
    double I3; //!< intensity at node 3
    double I4; //!< intensity at node 4

    double carrier_phase; //!< phase of carrier light field

    complex_t ***q12; //!< quantum noise matrix element mode 1 -> mode 2
    complex_t ***q21; //!< quantum noise matrix element mode 2 -> mode 1
    complex_t ***q13; //!< quantum noise matrix element mode 1 -> mode 3
    complex_t ***q31; //!< quantum noise matrix element mode 3 -> mode 1
    complex_t ***q24; //!< quantum noise matrix element mode 2 -> mode 4
    complex_t ***q42; //!< quantum noise matrix element mode 4 -> mode 2
    complex_t ***q34; //!< quantum noise matrix element mode 3 -> mode 4
    complex_t ***q43; //!< quantum noise matrix element mode 4 -> mode 3
};

//! beam splitter type
typedef struct beamsplitter beamsplitter_t;

//! grating type

struct grating {
    char name[MAX_TOKEN_LEN]; //!< grating name

    int node1_index; //!< index of node1
    int node2_index; //!< index of node2
    int node3_index; //!< index of node3
    int node4_index; //!< index of node4

    int node1_reduced_index; //!< node index for classical interferometer matrix
    int node2_reduced_index; //!< node index for classical interferometer matrix
    int node3_reduced_index; //!< node index for classical interferometer matrix
    int node4_reduced_index; //!< node index for classical interferometer matrix

    amp_dir_t node1_amp_dir; //!< direction of propagation at node 1
    amp_dir_t node2_amp_dir; //!< direction of propagation at node 2
    amp_dir_t node3_amp_dir; //!< direction of propagation at node 3
    amp_dir_t node4_amp_dir; //!< direction of propagation at node 4

    //! matrix element of the component in the interferometer matrix
    complex_t ***a11;
    //! matrix element of the component in the interferometer matrix
    complex_t ***a12;
    //! matrix element of the component in the interferometer matrix
    complex_t ***a13;
    //! matrix element of the component in the interferometer matrix
    complex_t ***a14;
    //! matrix element of the component in the interferometer matrix
    complex_t ***a21;
    //! matrix element of the component in the interferometer matrix
    complex_t ***a22;
    //! matrix element of the component in the interferometer matrix
    complex_t ***a23;
    //! matrix element of the component in the interferometer matrix
    complex_t ***a24;
    //! matrix element of the component in the interferometer matrix
    complex_t ***a31;
    //! matrix element of the component in the interferometer matrix
    complex_t ***a32;
    //! matrix element of the component in the interferometer matrix
    complex_t ***a33;
    //! matrix element of the component in the interferometer matrix
    complex_t ***a34;
    //! matrix element of the component in the interferometer matrix
    complex_t ***a41;
    //! matrix element of the component in the interferometer matrix
    complex_t ***a42;
    //! matrix element of the component in the interferometer matrix
    complex_t ***a43;
    //! matrix element of the component in the interferometer matrix
    complex_t ***a44;

    double d; //!< grating period [nm]
    double alpha; //!< angle of incidence

    // range of allowed grating periods, depends
    // on the type of grating (i.e. how many ports)
    double dmin; //!< minimum of allowed grating periods
    double dmax; //!< maximum of allowed grating periods

    int num_of_ports; //!< number of ports
    double eta[5]; //!< list of efficiencies
    int etaset[5]; //!< have the efficiencies been set?
    /* rho0 is for input-output only. internally it will be mapped as the
     * fourth efficiency, i.e. eta[3] */
    double rho0; //!< another efficiency, see Bunkowski et.al.
    double phi[5]; //!< coupling phases

    int attribs; //!< bit-coded flag for which attribs have been set
    int rebuild; //!< rebuild the element?

    double Rcx; //!< radius of curvature, tangential
    double Rcy; //!< radius of curvature, sagittal

    // tangential ABCD matrices
    //! transformation matrix for q of element; tangential plane 
    ABCD_t **qqt;

    // sagittal ABCD matrices
    //! transformation matrix for q of element; sagittal plane 
    ABCD_t **qqs;

    complex_t ***q11; //!< quantum noise matrix element mode 1 -> mode 1
    complex_t ***q12; //!< quantum noise matrix element mode 1 -> mode 2
    complex_t ***q13; //!< quantum noise matrix element mode 1 -> mode 3
    complex_t ***q14; //!< quantum noise matrix element mode 1 -> mode 4
    complex_t ***q21; //!< quantum noise matrix element mode 2 -> mode 1
    complex_t ***q22; //!< quantum noise matrix element mode 2 -> mode 2
    complex_t ***q23; //!< quantum noise matrix element mode 2 -> mode 3
    complex_t ***q24; //!< quantum noise matrix element mode 2 -> mode 4
    complex_t ***q31; //!< quantum noise matrix element mode 3 -> mode 1
    complex_t ***q32; //!< quantum noise matrix element mode 3 -> mode 2
    complex_t ***q33; //!< quantum noise matrix element mode 3 -> mode 3
    complex_t ***q34; //!< quantum noise matrix element mode 3 -> mode 4
    complex_t ***q41; //!< quantum noise matrix element mode 4 -> mode 1
    complex_t ***q42; //!< quantum noise matrix element mode 4 -> mode 2
    complex_t ***q43; //!< quantum noise matrix element mode 4 -> mode 3
    complex_t ***q44; //!< quantum noise matrix element mode 4 -> mode 4

};

//! grating type
typedef struct grating grating_t;

//! information pertaining to a squeezer

struct squeezer {
    char name[MAX_TOKEN_LEN]; //!< squeezer name

    double squeeze_angle; //!< squeezing angle
    double squeeze_factor; //!< degree of squeezing (0 <= r < Inf)

    int node1_index; //!< index of the connected node
    int node2_index; //!< index of the connected node

    int node1_reduced_index; //!< node index for classical interferometer matrix
    int node2_reduced_index; //!< node index for classical interferometer matrix

    amp_dir_t node1_amp_dir; //!< direction of propagation at node 1
    amp_dir_t node2_amp_dir; //!< direction of propagation at node 2

    //! matrix element of the component in the interferometer matrix
    complex_t ***a11;
    //! matrix element of the component in the interferometer matrix
    complex_t ***a12;
    //! matrix element of the component in the interferometer matrix
    complex_t ***a21;
    //! matrix element of the component in the interferometer matrix
    complex_t ***a22;

    int attribs; //!< bit-coded flag for which attribs have been set

    double carrier_phase; //!< phase of the carrier frequency

    complex_t ***q11; //!< quantum noise matrix element mode 1 -> mode 1
    complex_t ***q12; //!< quantum noise matrix element mode 1 -> mode 2
    complex_t ***q21; //!< quantum noise matrix element mode 2 -> mode 1
    complex_t ***q22; //!< quantum noise matrix element mode 2 -> mode 2
};

//! squeezer type
typedef struct squeezer squeezer_t;

//! input light field information

struct light_in {
    char name[MAX_TOKEN_LEN]; //!< name of input light field
    int node_index; //!< index of the connected node
    int node_reduced_index; //!< node index for classical interferometer matrix
    complex_t a; //!< field amplitude
    double I0; //!< field power

    //! power coefficients
    /*!
     * Used if the command `tem' is used for adding higher order modes to the 
     * input beam.
     */
    complex_t *power_coeff_list;

    double phase; //!< initial phase
    double f0; //!< main frequency

    int attribs; //!< bit-coded flag for which attribs have been set

    double noise_value; //!< laser noise, for use in quantum noise calcn
};

//! input light type
typedef struct light_in light_in_t;

//! output data information

struct output_data {
    char name[MAX_TOKEN_LEN]; //!< name of output data

    double Omax[2]; //!< store max value if -max option used
    double Omax_x[2]; //!< store max value in x-plane if -max option used
    double Omax_y[2]; //!< store max value in y-plane if -max option used
    double Omin[2]; //!< store min value if -max option used
    double Omin_x[2]; //!< store min value in x-plane if -max option used
    double Omin_y[2]; //!< store min value in y-plane if -max option used

    int detector_type; //!< detector type
    int output_type; //!< output type
    double user_defined_scale; //!< user defined scaling factor
    //  complex_t *a0;

    int node_index; //!< index of connected node
    int node_reduced_index; //!< node index for classical interferometer matrix

    bool beam_param_not_set; //!< true if Gauss beam param not set for node

    //! tag `the other beam'
    /*!
     * At each node there are two directions of propagation.  There is one
     * chosen by default, but one can also specify the `other' beam.  This flag
     * says to use the `other' beam.
     */
    int is_second_beam;
    int detector_index; //!< detector index

    complex_t *qx; //!< x component of q Gaussian beam parameter
    complex_t *qy; //!< y component of q Gaussian beam parameter

    // result
    complex_t signal; //!< output signal

    double re; //!< real component
    double im; //!< imaginary component

    double abs; //!< absolute value
    double deg; //!< phase/argument in degrees

    //! \todo is this the right location for the definition of this variable???
    //! quadrature angle to measure; 0=amplitude, 90=phase
    int quadrature_angle;

    bool noplot; //!< plot the output data?
};

//! output data type
typedef struct output_data output_data_t;

//! information about photo- or amplitude-detectors for output light field

struct light_out {
    char name[MAX_TOKEN_LEN]; //!< output light field name
    int num_demods; //!< number of demodulationnum_demods

    double f[MAX_DEMOD + 1]; //!< list of demodulated frequencies
    double phi[MAX_DEMOD + 1]; //!< list of demodulated phases

    //! flag whether phi is given by user, ommitted, or `max' stated instead
    int demod_phase_mode[MAX_DEMOD + 1];
    int sensitivity; //!< whether the pd is used to plot a sensitivity

    // specify certain TEM mode number of amplitude detector
    int field; //!< the kind of field
    bool field_is_set; //!< is the field set?

    // cordinates for beam shape plots
    double x; //!< x coordinate for beam shape plots
    double y; //!< y coordinate for beam shape plots

    double *masking_factor; //!< masking factors (see component "mask")
    bool beam_mask_is_set; //!< masking set?
    int detector_type; //!< detector type (from kat.ini)
    int output_index; //!< the index of the associated output

    // frequency selection for beam shape
    bool freq_is_set; //!< is the frequency set?
    double freq; //!< the frequency for the beam shape
};

//! output light type
typedef struct light_out light_out_t;

//! output beam parameters information

struct beampar_out {
    char name[MAX_TOKEN_LEN]; //!< beam parameter name
    int action; //!< flag for detector mode
    int x; //!< x=1 => horizontal, x=0 => vertical \todo change: orientation
    //! num of free spaces given by user in `bp' cmd for computing a Gouy phase
    int num_spaces;
    int *space_list; //!< free space list
    int output_index; //!< the index of the associated output
};

//! output beam parameters type
typedef struct beampar_out beampar_out_t;

//! output convolution information

struct convolution_out {
    char name[MAX_TOKEN_LEN]; //!< convolution name
    double freq; //!< field frequency (as for the 'beam' detector)
    bool freq_is_set; //!< user defined frequency has been set
    double xsize; //!< horizontal field size in 1/w0x
    double ysize; //!< vertical field size in 1/w0y
    int n; //!< first mode index n for HG_nm, p for LG_pl
    int m; //!< second mode index m for HG_nm, l for LG_pl
    int HGLG; //!< switch between HG modes (1) and LG modes (0)
    double w0x; //!< user defined beam parameter
    double z0x; //!< user defined beam parameter
    double w0y; //!< user defined beam parameter
    double z0y; //!< user defined beam parameter
    bool x_is_set; //!< have user defined parameters for x-direction been set?
    bool y_is_set; //!< have user defined parameters for y-direction been set?

    double *masking_factor; //!< masking factors (see component "mask")
    bool beam_mask_is_set; //!< masking set?
    int detector_type; //!< detector type (from kat.ini)

    int output_index; //!< the index of the associated output
};

//! output convolution type
typedef struct convolution_out convolution_out_t;

//! output mirror phse information

struct mirror_phase_out {
    char name[MAX_TOKEN_LEN]; //!< mirror_phase name
    int output_index; //!< the index of the associated output
};

//! output mirror_phase type
typedef struct mirror_phase_out mirror_phase_out_t;


//! output beam parameters information

struct cavity_par_out {
    char name[MAX_TOKEN_LEN]; //!< cavity name
    int cavity; //!< cavity index
    int action;
    int x; //!< x=1 => horizontal, x=0 => vertical \todo change: orientation
    int output_index; //!< the index of the associated output
};

//! output beam parameters type
typedef struct cavity_par_out cavity_par_out_t;


//! information describing a modulator

struct modulator {
    char name[MAX_TOKEN_LEN]; //!< modulator name

    int node1_index; //!< index of node1
    int node2_index; //!< index of node2

    int node1_reduced_index; //!< node index for classical interferometer matrix
    int node2_reduced_index; //!< node index for classical interferometer matrix

    amp_dir_t node1_amp_dir; //!< direction of propagation at node 1
    amp_dir_t node2_amp_dir; //!< direction of propagation at node 2

    double f; //!< frequency of modulation
    //! order of sidebands
    /*!
     * Each phase modulation creates an infinite number of new frequency
     * components (in pairs with +/- N*f_mod offsets) but only a few are
     * required to get the correct results.  This specifies how many should be
     * used.
     */
    int order;

    double modulation_index; //!< modulation index
    double phase; //!< modulation phase
    int type; //!< modulation type
    //  int mode;
    //! switch to single-sideband mode?
    /*!
     * When this is switched on, only one sideband is produced instead of
     * several pairs.
     */
    int single_sideband_mode;

    //! matrix element of field component in the interferometer matrix
    complex_t ***a12;
    //! matrix element of field component in the interferometer matrix
    complex_t ***a21;

    complex_t *a1; //!< ???
    complex_t *a2; //!< ???

    //! coupling coefficient for TEM mode from mode 1 into mode 2
    complex_t **k12;
    //! coupling coefficient for TEM mode from mode 2 into mode 1
    complex_t **k21;

    ABCD_t qq; //!< transformation matrix for q of modulator

    complex_t ***q12; //!< quantum noise matrix element mode 1 -> mode 2
    complex_t ***q21; //!< quantum noise matrix element mode 2 -> mode 1

    int rebuild; //!< rebuild the modulator?
};

//! modulation type
typedef struct modulator modulator_t;

//! information describing a signal

struct signal {
    char name[MAX_TOKEN_LEN]; //!< signal name

    int *node1_index; //!< index of node1
    int *node2_index; //!< index of node2
    int *node3_index; //!< index of node3
    int *node4_index; //!< index of node4

    int node1_reduced_index; //!< node index for classical interferometer matrix
    int node2_reduced_index; //!< node index for classical interferometer matrix
    int node3_reduced_index; //!< node index for classical interferometer matrix
    int node4_reduced_index; //!< node index for classical interferometer matrix

    int component_index; //!< index of component where signal is injected
    int component_type; //!< the type of component
    int type; //!< the type of signal

    amp_dir_t *node1_amp_dir; //!< direction of propagation at node 1
    amp_dir_t *node2_amp_dir; //!< direction of propagation at node 2
    amp_dir_t *node3_amp_dir; //!< direction of propagation at node 3
    amp_dir_t *node4_amp_dir; //!< direction of propagation at node 4

    double phase; //!< signal phase
    double amplitude; //!< signal amplitude

    complex_t *a1; //!< ???
    complex_t *a2; //!< ???
    complex_t *a3; //!< ???
    complex_t *a4; //!< ???
    complex_t *ao1; //!< ???
    complex_t *ao2; //!< ???
    complex_t *ao3; //!< ???
    complex_t *ao4; //!< ???
};

//! signal type
typedef struct signal signal_t;

//! information describing a diode

struct diode {
    char name[MAX_TOKEN_LEN]; //!< diode name

    int node1_index; //!< index of node1
    int node2_index; //!< index of node2

    int node1_reduced_index; //!< node index for classical interferometer matrix
    int node2_reduced_index; //!< node index for classical interferometer matrix

    double S; //!< ???

    amp_dir_t node1_amp_dir; //!< direction of propagation at node 1
    amp_dir_t node2_amp_dir; //!< direction of propagation at node 2

    //! matrix element of the component in the interferometer matrix
    complex_t ***a12;
    //! matrix element of the component in the interferometer matrix
    complex_t ***a21;

    //! coupling coefficient for TEM mode from mode 1 into mode 2
    complex_t **k12;
    //! coupling coefficient for TEM mode from mode 2 into mode 1
    complex_t **k21;

    ABCD_t qq; //!< transformation matrix for q of diode

    complex_t ***q12; //!< quantum noise matrix element mode 1 -> mode 2
    complex_t ***q21; //!< quantum noise matrix element mode 2 -> mode 1

    int rebuild; //!< rebuild the diode?
};

//! diode type
typedef struct diode diode_t;

//! information describing a lens

struct lens {
    char name[MAX_TOKEN_LEN]; //!< lens name

    int node1_index; //!< index of node1
    int node2_index; //!< index of node2

    int node1_reduced_index; //!< node index for classical interferometer matrix
    int node2_reduced_index; //!< node index for classical interferometer matrix

    double f; //!< focal length
    int rebuild; //!< rebuild the lens?

    amp_dir_t node1_amp_dir; //!< direction of propagation at node 1
    amp_dir_t node2_amp_dir; //!< direction of propagation at node 2

    //! matrix element of the component in the interferometer matrix
    complex_t ***a12;
    //! matrix element of the component in the interferometer matrix
    complex_t ***a21;

    //! coupling coefficient for TEM mode from mode 1 into mode 2
    complex_t **k12;
    //! coupling coefficient for TEM mode from mode 2 into mode 1
    complex_t **k21;

    ABCD_t qq; //!< transformation matrix for q of lens

    complex_t ***q12; //!< quantum noise matrix element mode 1 -> mode 2
    complex_t ***q21; //!< quantum noise matrix element mode 2 -> mode 1
};

//! lens type
typedef struct lens lens_t;

//! a structure of information for attributes of a compnent

struct attr {
    char name[MAX_TOKEN_LEN]; //!< name of attribute
    int component; //!< component index
    int type; //!< the type of component
    double *M; //!< mass
    double *R; //!< radius of curvature
    double *alpha; //!< misalignment angle (elsewhere called beta...)
};

//! attribute type
typedef struct attr attr_t;

//! structure of parameters used when setting an attribute of a component

struct component_param {
    double **xparam; //!< array of parameter values
    int component_index; //!< the component index
    const char *parameter; //!< the name of the parameter
    char *unit; //!< the units of the parameter
    const char *command_string; //!< command string used to set the attribute
    int *lborder; //!< lower boundary of attribute (???)
    int *uborder; //!< upper boundary of attribute (???)
    double *min; //!< minimum value of attribute
    double *max; //!< maximum value of attribute
};

//! component parameter type
typedef struct component_param component_param_t;

//! options passed into program

struct options {
    int help; //!< is the help page asked for?
    int maxmin; //!< is max and min output asked for?
    int perl1; //!< switch off some of the output?
    int quiet; //!< switch off all output?
    int new; //!< is this a new feature of finesse?
    int klu; //!< should we use KLU or SPARSE package?
    int klu_manual; //!< manual setting of optios.klu
    int check; //!< do consistency check of interferometer matrix?
    int format; //!< print data to file with more precision?
    int servermode; //!< server mode flag
    int portnumber; //!< portnumber for client server communication
    int version; //!< print version and quit
    int noheader; //!< do not print header into output file
};

//! options type
typedef struct options options_t;

//! photo detector type

struct pdtype {
    char *name; //!< photo detector name
    double ****cross; //!< beat coefficient on diode between mode n1m1 and n2m2
};

//! photo detector type
typedef struct pdtype pdtype_t;

//! information describing a gnuplot terminal

struct gnuplot_terminal {
    char *name; //!< terminal name
    char *command; //!< terminal command
    char *suffix; //!< output file suffix (e.g. .gif for gif images)
    bool output_is_file; //!< does terminal output to a file?
};

//! gnuplot terminal type
typedef struct gnuplot_terminal gnuplot_terminal_t;

//! cavity type

struct cavity {
    char name[MAX_TOKEN_LEN]; //!< cavity name

    int component1_index; //!< start component for cavity tracing
    int component2_index; //!< stop component for cavity tracing

    int node1_index; //!< start node index
    int node2_index; //!< stop node index

    int stable; //!< is cavity stable?
    complex_t qx; //!< Gaussian beam parameter of eigenmode (horizontal)
    complex_t qy; //!< Gaussian beam parameter of eigenmode (vertical)
    double n; //!< index of refraction at first node
    double finesse; //!< cavity finesse
    double loss; //!< round-trip loss
    double length; //!< optical path length (round-trip)
    double FSR; //!< free spectral range
    double FWHM; //!< full width at half maximum
    double pole; //!< cavity pole (FWHM/2)
};

//! cavity type
typedef struct cavity cavity_t;

//! information describing a Gaussian beam

struct gauss {
    char name[MAX_TOKEN_LEN]; //!< name of Gaussian beam
    int component_index; //!< component index
    int node_index; //!< index at node1

    complex_t qx; //!< complex beam parameter in x-plane
    complex_t qy; //!< complex beam parameter in y-plane

    double wx0; //!< x-component of beam waist
    double wy0; //!< y-component of beam waist
};

//! Gaussian beam type
typedef struct gauss gauss_t;

//! describes an axis; refers to the xaxis, x2axis and x3axis commands

struct xaxis {
    double *xaxis; //!< x-axis
    double minus_xaxis; //!< minus value of x-axis (to be used for mx variables)
    double op; //!< used to store previous values
    char xname[LINE_LEN]; //!< name of axis

    double xmin; //!< minimum extent of axis
    double xmax; //!< maximum extent of axis
    double max; //!< maximum limit for a given axis
    double min; //!< minimum limit for a given axis

    int lborder; //!< x axis lower border
    int uborder; //!< x axis upper border

    int component_idx; //!< Index of component being varied
    int xsteps; //!< the number of steps between xmin and xmax
    int xtype; //!< linear or logarthmic?
};

//! axis type
typedef struct xaxis axis_t;

//! information describing a set command

/*!
 * Set assignes a variable name to an output signal so that it can be usd
 * with func or lock.  Syntax:
 *
 * set component parameter name
 */
struct set_command {
    char name[LINE_LEN]; //!< name of the set
    char component_name[LINE_LEN]; //!< component name
    char parameter_name[LINE_LEN]; //!< parameter name

    double *value; //!< pointer to the output signal

    double max; //!< maximum of set
    double min; //!< minimum of set

    int lborder; //!< lower border of set
    int uborder; //!< upper border of set

    bool is_detector; //!< is `component parameter' a detector or normal component?
};

//! set type
typedef struct set_command set_command_t;

//! describes a put command

/*!
 * Put a variable into a parameter of a component
 */
struct put_command {
    double *target; //!< pointer to component parameter value
    double *source; //!< pointer to the variable value

    char component_name[LINE_LEN]; //!< component name
    char parameter_name[LINE_LEN]; //!< parameter name
    char function_name[LINE_LEN]; //!< function name

    double max; //!< maximum of put
    double min; //!< minimum of put

    int lborder; //!< lower border of put
    int uborder; //!< upper border of put

    //! which mode is put command in?
    /*!
     * 0 = overwriting parameter
     * 1 = adding to parameter
     */
    int mode;
    double startvalue; //!< put start value
};

//! put type
typedef struct put_command put_command_t;



//! describes a tune command

/*!
 * Sets a new value to a parameter of a component as defined by a server-client connection
 */
struct tune {
    double *target; //!< pointer to component parameter value

    char cname[LINE_LEN]; //!< component name
    char pname[LINE_LEN]; //!< parameter name
    double value; //!< parameter value

    double max; //!< maximum 
    double min; //!< minimum 

    int lborder; //!< lower border param value
    int uborder; //!< upper border param value

    double startvalue; //!< original parameter value
};

//! tune type
typedef struct tune tune_t;


//! lock type

struct lock_command {
    char name[LINE_LEN]; //!< lock name
    char iname[LINE_LEN]; //!< variable name that lock should put to zero

    double *input; //!< lock input
    double value; //!< lock value
    double gain; //!< lock gain
    double accuracy; //!< accuracy of lock

    double last1; //!< last lock value (for gradient calculation)
    double last2; //!< second to last lock value (for grad calcn)

    //! lock only the once?
    /*!
     * if once is set, then a lock is performed only for the first point on
     * the x-axis
     */
    int once;
};

//! lock type
typedef struct lock_command lock_command_t;

//! information describing a user-defined function

/*!
 * this is a user-defined function which is parsed by formulc and then
 * evaluated for each point on the x-axis.
 */
struct func_command {
    char name[MAX_TOKEN_LEN]; //!< function name
    char fstring[LINE_LEN]; //!< function string
    double *input[MAX_TOKEN]; //!< list of input values
    char tokens[MAX_TOKEN]; //!< function tokens
    int num_tokens; //!< number of function tokens
    int type[MAX_TOKEN]; //!< array of token types
    int token_number[MAX_TOKEN]; //!< array of token indices (i.e. token index)
    int ninput; //!< number of input values
    formu formula; //!< the formula as parsed by formulc
    double result; //!< function result
    bool is_detector; //!< are any of func inputs a detector output?
    int output; //!< output index
};

//! function type
typedef struct func_command func_command_t;

//! derivative type

struct derivative {
    double *xderiv; //!< derivative in x direction
    char name[LINE_LEN]; //!< derivative function name

    double max; //!< derivative maximum
    double min; //!< derivative minimum

    int lborder; //!< lower border of derivative
    int uborder; //!< upper border of derivative
};

//! derivative type
typedef struct derivative derivative_t;

//! describes the scale command for user-defined rescaled outputs

struct scale {
    double factor; //!< scale factor
    int output; //!< scale output

    //! the type of scaling performed
    /*!
     * the scale can be invoked with a factor or with a keyword.  type stores
     * which keyword or that it was a number.
     */
    int type;
};

//! scale type
typedef struct scale scale_t;

//! simple constants in the input file

struct constant {
    char name[MAX_TOKEN_LEN]; //!< constant name
    char value[LINE_LEN]; //!< constant value
    int name_len; //!< length of constant name
    int value_len; //!< length of constant value
};

//! simple constants in the input file
typedef struct constant constant_t;

//! a collection of flags for the beam analyser

struct beam_analyser {
    bool set; //!< has beam anaylser been defined?
    int x; //!< does analyser look in x-direction?
    int y; //!< does analyser look in y-direction?

    //! should xaxis and x2axis be swapped?
    /*!
     * this can be useful if one axis is the position on the beam analyser.
     * One can then run along the axis without recomputing the light fields.
     * By definition xaxis runs around x2axis, so sometimes one has to swap
     * them to avoid recomputation of fields with beam analysers.
     */
    int swap;
};

//! beam analyser type
typedef struct beam_analyser beam_analyser_t;

//! global variables

struct global_var {
    char input_fname[FNAME_LEN]; //!< input file name
    char output_fname[FNAME_LEN]; //!< output file name
    char gnuplot_fname[FNAME_LEN]; //!< gnuplot file name
    char matlab_fname[FNAME_LEN]; //!< matlab file name
    char log_fname[FNAME_LEN]; //!< log file name
};

//! global variables
typedef struct global_var global_var_t;

//! local variables

struct local_var {
    complex_t *zsolution; //!< solution in z direction
    complex_t *zderiv1; //!< point used to calculate the slope
    complex_t *zderiv2; //!< point used to calculate the slope

    double x_p1; //!< ???
    double x_p2; //!< ???

    complex_t z_in_p1; //!< ???
    complex_t z_out_p1; //!< ???

    complex_t z_in_p2; //!< ???
    complex_t z_out_p2; //!< ???

    complex_t zk1; //!< ???
    int kn1; //!< ???
    int kn2; //!< ???
    complex_t kqx1; //!< ???
    complex_t kqx2; //!< ???
    double kgammax; //!< ???
    double knr; //!< ???

    double natLogsOfN[101]; //!< list of natural logs of N
    int natLogMax; //!< maximum value of n for natural log list

    int *component_trace_list; //!< stores overall component index for beam tracing
    int *trace_n[3]; //!< ???
    int *trace_c1; //!< ???
    int *trace_c2; //!< ???

    double *respt; //!< ???

    int ln1; //!< ???
    int ln2; //!< ???
    int lm1; //!< ???
    int lm2; //!< ???

    complex_t lqx1; //!< ???
    complex_t lqx2; //!< ???
    complex_t lqy1; //!< ???
    complex_t lqy2; //!< ???

    double lcx; //!< ???
    double lcy; //!< ???
    double lsx; //!< ???
    double lsy; //!< ???
    double lnr; //!< ???
};

//! local variables
typedef struct local_var local_var_t;

//! variables for sparse routines

struct sparse_var {
    char *spamat; //!< the sparse matrix
    char *Qspamat; //!< the sparse matrix for quantum noise

    complex_t **diag1; //!< sparse matrix diagonal (temporary)
    complex_t *sprhs1; //!< sparse matrix RHS (temporary)
    complex_t *spsol1; //!< sparse matrix solution (temporary)
    complex_t **diag; //!< sparse matrix diagonal
    complex_t *sprhs; //!< sparse matrix right hand side
    complex_t *spsol; //!< sparse matrix solution

    complex_t **Qdiag1; //!< quantum noise sparse matrix diagonal (temporary)
    complex_t *Qsprhs1; //!< quantum noise sparse matrix RHS (temporary)
    complex_t *Qspsol1; //!< quantum noise sparse matrix solution (temporary)
    complex_t **Qdiag; //!< quantum noise sparse matrix diagonal
    complex_t *Qsprhs; //!< quantum noise sparse matrix right hand side
    complex_t *Qspsol; //!< quantum noise sparse matrix solution
};

//! variables for sparse routines
typedef struct sparse_var sparse_var_t;

//! variables for KLU sparse routines

struct klu_sparse_var {
    /* Adapted from the KLU user guide:
    The matrix is described with four parameters: 
    nz=inter.num_nonzero: integer scaler, number of non-zero elements (a `non-zero'
    element is present in the sparse matrix, however, it can be numerically zero.).
    n=inter.num_eqns: an integer scalar. The matrix is n-by-n. Note that KLU 
    only operates on square matrices. 
    Ap=klu.col_ptr: an integer array of size n+1. The first entry is col_ptr[0]=0, and the last 
    entry nz=col_ptr[n] is equal to the number of entries in the matrix. 
    Ai=klu.row_index: an integer array of size nz = col_ptr[n]. The row indices of 
    entries in column j of A are located in row_index [col_ptr [j] ... col_ptr [j+1]-1]. 
    The matrix is zero-based; row and column indices are in the range 0 to n-1. 
    Ax=klu.value: a double array of size nz for the real case, or 2*nz for the 
    complex case. For the complex  case, the real and imaginary parts are interleaved, 
    compatible with Fortran and the ANSI C99 Complex data type. KLU does not rely on 
    the ANSI C99 data type, however, for portability  reasons. The numerical values 
    in column j of a real matrix are located in Ax [Ap [j] ... Ap [j+1]-1]. For a 
    complex matrix, they appear in Ax [2*Ap [j] ... 2*Ap [j+1]-1], 
    as real/imaginary pairs (the real part appears first, followed by the imaginary part). 
     */

    /*
    From: http://www.cs.utk.edu/~dongarra/etemplates/node374.html
    Assuming we have a nonsymmetric sparse matrix A, we create three vectors: 
    one for floating point numbers (value) and the other two for integers 
    (row_index, col_ptr). The `value' vector stores the values of the nonzero 
    elements of the matrix A as they are traversed in a column-wise fashion. 
    The `col_ind' vector stores the column indexes of the elements in the `value' 
    vector. That is, if value(k)=a_{i,j}, then row_index(k)=i. 
    The col_ptr vector stores the locations in the val vector that start a 
    row; that is, if value(k)=a_{i,j), then col_ptr(j)<= k< col_ptr(j+1). 
    By convention, we define col_ptr(n+1) = nnz+1, where nnz is the number of 
    nonzeros in the matrix A. The storage savings for this approach is significant. 
    Instead of storing n^2 elements, we need only 2nnz+n+1 storage locations.
     */

    double *value; //!< sparse CCS matrix numerical values (length=2*inter.num_nonzero)
    long *row_index; //!< sparse CCS matrix row index        (length=inter.num_nonzero)
    long *col_ptr; //!< sparse CCS matrix column pointer   (length=inter.num_eqns+1)
    double *rhs; //!< ride-hand-side for sparse matrix   (length=2*inter.num_eqns)

};

//! variables for sparse routines
typedef struct klu_sparse_var klu_sparse_var_t;


//! collects all the index variables together to help bounds checking

struct indices {
    int base_index_1; //!< base index for calculation of matrix indices
    int base_index_2; //!< base index for calculation of matrix indices
    int base_index_3; //!< base index for calculation of matrix indices
    int base_index_4; //!< base index for calculation of matrix indices

    int row_index_1; //!< row index for interferometer matrix
    int row_index_2; //!< row index for interferometer matrix
    int row_index_3; //!< row index for interferometer matrix
    int row_index_4; //!< row index for interferometer matrix

    int col_index_1; //!< column index for interferometer matrix
    int col_index_2; //!< column index for interferometer matrix
    int col_index_3; //!< column index for interferometer matrix
    int col_index_4; //!< column index for interferometer matrix
};

//! collects all the index variables together to help bounds checking
typedef struct indices indices_t;

//! interferometer type

struct interferometer {
    char *basename; //!< basename for input and output files

    constant_t *constant_list; //!< list of constants
    int num_constants; //!< number of constants

    mirror_t *mirror_list; //!< list of mirrors
    space_t *space_list; //!< list of free space sections
    beamsplitter_t *bs_list; //!< list of beam splitters
    grating_t *grating_list; //!< list of gratings
    squeezer_t *squeezer_list; //!< list of squeezers
    light_out_t *light_out_list; //!< list of output light beams
    output_data_t *output_data_list; //!< output data list
    beampar_out_t *beampar_out_list; //!< output beam parameters
    cavity_par_out_t *cavity_par_out_list; //!< output cavity parameters
    convolution_out_t *convolution_out_list; //!< output convolution 
    mirror_phase_out_t *mirror_phase_out_list; //!< output mirror phase 
    light_in_t *light_in_list; //!< list of input light sources
    modulator_t *modulator_list; //!< list of modulators
    node_t *node_list; //!< node list
    signal_t *signal_list; //!< signal list
    derivative_t *deriv_list; //!< list of derivatives
    scale_t *scale_list; //!< list of scales
    diode_t *diode_list; //!< diode list
    lens_t *lens_list; //!< list of lenses
    attr_t *attr_list; //!< list of attributes
    gauss_t *gauss_list; //!< Gaussian beam list
    cavity_t *cavity_list; //!< list of cavities
    variable_t *variable_list; //!< list of variables (dummy parameters)

    int num_mirrors; //!< number of mirrors
    int num_beamsplitters; //!< number of beam splitters
    int num_spaces; //!< number of free spaces
    int num_nodes; //!< number of nodes
    int num_light_inputs; //!< number of input light fields
    int num_light_outputs; //!< number of output light fields
    int num_modulators; //!< number of modulators
    int num_signals; //!< number of signals
    int num_diff_cmds; //!< number of derivatives
    int num_outputs; //!< number of elements in output data
    int num_beam_param_outputs; //!< number of output beam parameters
    int num_cavity_param_outputs; //!< number of output cavity parameters
    int num_convolution_outputs; //!< number of output convolution
    int num_mirror_phase_outputs; //!< number of output convolution
    int num_scale_cmds; //!< number of scales
    int num_diodes; //!< number of diodes
    int num_lenses; //!< number of lenses
    int num_attributes; //!< number of attributes
    int num_gauss_cmds; //!< number of Gaussian commands
    int num_cavities; //!< number of cavities
    int num_variables; //!< number of variables
    int num_gratings; //!< number of gratings
    int num_squeezers; //!< number of squeezers
    int num_dump_nodes; //!< number of dump nodes
    int num_reduced_nodes; //!< number of nodes minus number of dump nodes


    int num_eqns; //!< number of equations -> rank of sparse matrix
    int num_nonzero; //!< number of non-zero elemensts in sparse matrix
    int num_quantum_eqns; //!< number of equations for quantum corr.
    int num_quadratures; //!< number of quadratures
    int num_components; //!< number of components
    int num_output_cols; //!< number of output columns
    int time; //!< printing computation time on/off
    int noxaxis; //!< don't run over xaxis, just do one computation
    bool multi; //!< plot multiple surfaces?
    int pause; //!< put "pause -1" in gnuplot file
    int printmatrix; //!< print matrix into file

    //! beam tracing on/off 
    /*!
     * trace is a bit-encoded integer describing the kind of trace to perform.
     *   - 1:   list of TEM modes used
     *   - 2:   cavity eigenvalues and cavity parameters like FWHM, 
     *          FSR optical length and Finesse
     *   - 4:   mode mismatch parameters for the initial setup
     *   - 8:   beam parameters for every node, nodes are listed in
     *          the order found by the tracing algorithm
     *   - 16:  Gouy phases for all spaces
     *   - 32:  coupling coefficients for all components
     *   - 64:  mode matching parameters during calculation, if they change
     *          due to a parameter change, for example by changing a radius of 
     *          curvature.
     *   - 128: nodes found during the cavity tracing
     */
    int trace;
    int grating; //!< verbose output for gratings
    int mycolor; //!< set nice color schemes for Gnuplot
    int width; //!< linewidth for Gnuplot
    int k; //!< which type of coupling coefficient computation

    int tem; //!< maximum order of hermite gauss modes
    bool tem_is_set; //!< maximum order of hermite gauss modes is set!
    int num_fields; //!< number of field amplitudes
    int rebuild; //!< rebuild some k coeffs while tuning parameters
    beam_analyser_t beam; //!< beam analyser settings
    int set_tem_phase_zero; //!< set TEM_00 phase to 0 ?
    int mats; //!< size of matrix

    axis_t x1; //!< x axis
    axis_t x2; //!< y axis
    axis_t x3; //!< z axis

    // ---- new 
    int num_set_cmds; //!< number of sets
    set_command_t *set_list; //!< list of sets

    // --------
    int num_put_cmds; //!< number of puts
    put_command_t *put_list; //!< list of puts
    tune_t *tune_list; //!< list of parameters to be tunen by server-client commands 
    int num_tune_params; //!< number parameter values to be send/received


    int num_locks; //!< number of locks
    lock_command_t *lock_list; //!< list of locks

    int num_func_cmds; //!< number of functions
    func_command_t *function_list; //!< list of functions

    int *func_order; //!< order in which functions should be processed
    int *lock_order; //!< order in which locks should be processed
    bool *is_locked; //!< has element obtained a lock?
    int *breakcount; //!< counter to determine break condition
    int num_locks_and_funcs; //!< the number of locks and functions
    int lock; //!< are there any locks?
    int showiterate; //!< show iteration output
    int startlock; //!< switch off lock when should only be done once

    int startnode; //!< optional starting node for the beam tracing
    int startnode_set; //!< has the command `startnode' been used?

    int retrace; //!< retrace the system?
    int retrace_manual; //!< manual override of the retrace setting?
    int splot; //!< use splot?
    bool x3_axis_is_set; //!< is x3axis set?

    //! use video mode?
    /*!
     * there are two types of output for x3axis: a huge 3D matrix or a set of
     * 2D outputs.  The latter is called video mode because one can make
     * little films this way.  This flag tells which will be th output.
     */
    int video;

    char y1name[MAX_TOKEN_LEN]; //!< label of left hand y axis
    char y2name[MAX_TOKEN_LEN]; //!< label of right hand y axis
    int yaxis[2]; //!< are we using the left and right y axes
    bool ploty2axis; //!< should we plot two y axes?
    int ytype; //!< y-axis log or linear?

    double f0; //!< carrier frequency
    double *f; //!< frequencies
    int *t; //!< is result a transfer function or error signal?
    int num_frequencies; //!< number of frequencies
    int num_non_signal_frequencies; //!< number of non-signal frequencies

    double fsig; //!< signal frequency

    int num_gnuterm_cmds; //!< number of gnuplot terminals
    int *gnuterm; //!< list of gnuplot terminals
    char **gnutermfn; //!< gnuplot terminal function

    char *gnuplotcommand; //!< list of gnuplot plotting commands
    char *matlabcommand; //!< list of matlab commands
    char *matlabplotcommand; //!< list of matlab  plotting commands
    int num_gnuplot_cmds; //!< number of gnuplot plotting commands
    int num_matlabplot_cmds; //!< number of matlab plotting commands
    int num_matlab_cmds; //!< number of matlab commands

    int debug; //!< debug flag

    double scale_factor; //!< plotting scale factor
    int scale_unit; //!< plotting scale units

    double null; //!< the value 0
    double *dnull; //!< pointer to the value 0

    //! what kind of shotnoise to compute (simple, or quantum)
    int shotnoise;

    //! quantum corrections parameter
    /*! 
     * 0 : off
     * 1 : quantum noise (quantum version of shot noise)
     * 2 : radiation pressure noise
     * 3 : both quantum and radiation pressure noise
     */
    int qcorr;

    bool setup; //!< has the setup function been called?

    int kmn; //!< kmn???
};

typedef enum kmn_flags {
    USE_BAYER_HELMS = 0,
    PRINT_COEFFS = 1,
    INT_X_AND_Y = 2,
    INT_X_OR_Y = 4,
    INT_APERTURE = 8,
    INT_CURVATURE = 16,
    INT_X_Y_DISP = 32,
    INT_BAYER_HELMS = 64,
    INT_BAYER_HELMS_OLD = 128
} kmn_flags_t;

//! interferometer type
typedef struct interferometer interferometer_t;

//! intialisation variables from kat.in

struct init_variables {
    char katinifile[LINE_LEN]; //!< path+name of init file read from variable KATINI
    bool katini_filename_set; //!< has filename has been set with KATINI?
    int init_file_is_found; //!< has the initialisation file been found?
    //! how often progress info is output as a percentage
    int percentage_step;
    int num_points_total; //!< total number of points (for progress info)

    double lambda; //!< wavelength of light
    double clight; //!< speed of light
    double eqsmall; //!< used to decide if two values differ
    double small; //!< definition of a "small" number
    double qeff; //!< quantum efficiency
    double epsilon_0_c; //!< permeability of free space
    double deriv_h; //!< Planck's constant for derivations
    double n0; //!< default refractive index

    char gnucommand[LINE_LEN]; //!< gnuplot command
    gnuplot_terminal_t gnuterm[MAX_GNUTERM]; //!< gnuplot terminal
    pdtype_t pdt[MAX_PD_TYPE]; //!< photodetector type
    char *gnutermfn[MAX_GNUTERM]; //!< gnuplot terminal function
    int num_gnuplot_terminals; //!< number of gnuplot terminals
    int winterm; //!< are we using a windows terminal?
    int unixterm; //!< are we using a unix terminal?
    int macosterm; //!< are we using a macos terminal?
    int num_pdtypes; //!< number of photodetector types

    // locking
    int locksteps; //!< number of steps in locking
    //! number of steps for which loop state is checked
    int locktest1;
    //! number of checks before loop gain is increased or decreased
    int locktest2;

    double lockthresholdlow; //!< lower locking threshold
    double lockthresholdhigh; //!< upper locking threshold

    double gainfactor; //!< gain factor
    int autogain; //!< is gain automatic?
    int sequential; //!< is this a sequential lock?
    int autostop; //!< should we stop automatically?

    double gnuversion; //!< the version of gnuplot being used

    int maxintop; //!< maximum number of integration operations in dcuhre
    double abserr; //!< absolute error
    double relerr; //!< relative error
    int maxintcuba;
    int_method_t mapintmethod;
    interpolation_method_t mapinterpmethod;
    int mapinterpsize;
    
    int cuba_numprocs;
    int calc_knm_transpose; //!< if 1 then the transpose element is calculated using phase factor
};

//! intialisation variables
typedef struct init_variables init_variables_t;


/* ==================================================================== */

/* from kat.c */
int main(int argc, char *argv[]);
void banner(FILE *fp, char *input_fname, char *output_fname,
        char *gnuplot_fname);

/* from AllTests.c */
#ifdef TEST
void RunAllTests(void);
#endif

#endif // KAT_H

/*
 * Local variables:
 * c-indentation-style: bsd
 * c-basic-offset: 2
 * indent-tabs-mode: nil
 * End:
 *
 * vim: tabstop=2 expandtab shiftwidth=2:
 */
