// $Id$

/*!
 * \file kat_aa.h
 * \brief Header file for automatic alignment routines in Finesse
 * 
 * @section Copyright notice
 *
 *  This file is part of the interferometer simulation Finesse
 *  http://www.gwoptics.org/finesse
 *
 *  Copyright (C) 1999 onwards Andreas Freise
 *  with parts of the code written by Daniel Brown, Paul Cochrane
 *  and Gerhard Heinzel.
 *
 *  This program is free software; you can redistribute it and/or modify it under
 *  the terms of the GNU General Public License version 3 as published
 *  by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 *  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *  See the GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along with
 *  this library; if not, write to the Free Software Foundation, Inc., 59 Temple Place,
 *  Suite 330, Boston, MA 02111-1307 USA
 */

#ifndef KAT_AA_H
#define KAT_AA_H

#include "kat_optics.h"

// Default settings for interpolation routines
#define DEFAULT_INTERP_SIZE 3;   // must be odd integer > 3
#define DEFAULT_INTERP_METHOD 3; // 1 - spline, 2 - linear, 3 -NN, Nearest Neighbour

//here we define a macro that applies a round off to zero if the number is smaller than the error
//#define INT_VAL(u) ((abs(error[u]*1e-1) < abs(integral[u])) ? integral[u] : 0) * J
#define INT_VAL(u) integral[u] * J
#define DELTA_FUNC(n1,n2,m1,m2) ((n1 == n2) && (m1 == m2) ? ((n1 == m1) ? complex_1 : complex_0) : complex_0)
#define CHECK_MIRROR_KMN_T(A) \
assert(A!=0); \
assert(A->k11!=0); \
assert(A->k12!=0); \
assert(A->k21!=0); \
assert(A->k22!=0);
#define IS_MIRROR_KMN_ALLOCD(A) \
(((A)->k11 != NULL) && ((A)->k12 != NULL) \
&& ((A)->k21 != NULL) && ((A)->k22 != NULL))

mirror_kmn_t tmap;
void allocate_temporary_memory_for_maps();
void allocate_kmn_accel_memory();
complex_t rev_gouy(complex_t k, int n1, int m1, int n2, int m2,
        complex_t qx1, complex_t qx2, complex_t qy1, complex_t qy2);
int set_k_mirror(int i);
int set_k_mirror_phasemap(int i);
int set_k_mirror_phasefunc(int i);
int set_k_beamsplitter(int i);
int set_k_space(int i);
int set_k_lens(int i);
int set_k_modulator(int i);
int set_k_diode(int i);
void set_k_all_components(bool rebuilding);
void mode_match(complex_t q1, complex_t q2, complex_t *k, double *km);
complex_t k_mnmn(int n1, int n2, int m1, int m2,
        complex_t qx1, complex_t qy1, complex_t qx2, complex_t qy2,
        double gamma_x, double gamma_y, double nr);
complex_t k_mnmn_new(int n1, int n2, int m1, int m2,
        complex_t qx1, complex_t qy1, complex_t qx2, complex_t qy2,
        double gamma_x, double gamma_y, double nr);
complex_t S_u(complex_t X, complex_t XS, complex_t F, complex_t FS,
        int m1, int m2);
complex_t S_g(complex_t X, complex_t XS, complex_t F, complex_t FS,
        int m1, int m2);
complex_t k_mm(int m1, int m2, complex_t q1, complex_t q2,
        double gamma, double nr);
int num_eqns_for(int modes);
void get_cavity(void);
int trace_cavity(ABCD_t *Mlocal_x, ABCD_t *Mlocal_y, int component1_index, int node1_index, int component2_index, int node2_index, int *index, int *traced);
void set_ABCD_mirror(int i);
void set_ABCD_beamsplitter(int i);
void set_ABCD_space(int i);
void set_ABCD_lens(int i);
void set_ABCD_grating(int i);
void set_ABCDs(void);
void set_qbase(int component);
void trace_beam(int startnode);
void tracing(int startnode, int start_comp);
int component_matrix(ABCD_t *matrix, int component_index, int node1_index, int node2_index, int tan_sag);
void get_component(int *component_type, int *component_index, int node1_index, int node2_index);
void which_nodes(int i, int current_node, int *num_nodes, int *node_indices);
void which_components(int node_index, int *component1_index, int *component2_index);
char *get_node_name(int node_index);
void set_q(int node_index, int component_index, complex_t qx, complex_t qy);
void check_map_size(surface_map_t *map, complex_t qx1, complex_t qy1, complex_t qx2, complex_t qy2, double nr1, double nr2);
void mirror_kmn_matrix_copy(mirror_kmn_t* src, mirror_kmn_t* dest);
void mirror_kmn_matrix_mult(mirror_kmn_t* A, mirror_kmn_t* B, mirror_kmn_t *result);
void mirror_kmn_alloc(mirror_kmn_t* kmn);
void mirror_kmn_free(mirror_kmn_t* kmn);
void compute_kmn_integral(mirror_t *mirror, complex_t qx1, complex_t qy1,
        complex_t qx2, complex_t qy2, double nr1, double nr2, int integrating, int mismatch, int astigmatism);

void gsl_map_interpol_free_mm(surface_merged_map_t *map);
void gsl_map_interpol_free(surface_map_t *map);
void gsl_map_interpol_init_mm(surface_merged_map_t *map);
void gsl_map_interpol_init(surface_map_t *map);

typedef enum KNM_MIRROR_NODE_DIRECTION {
    K11 = 1,
    K12 = 2,
    K21 = 3,
    K22 = 4
} KNM_MIRROR_NODE_DIRECTION_t;

/*** Contains the parameters needed to do mirror map integration using GSL */
typedef struct knm_map_integration_params {
    int n1, m1, n2, m2;
    u_nm_accel_t *acc_11_nr1_1; //accelerator for calculating k11
    u_nm_accel_t *acc_11_nr1_2;

    u_nm_accel_t *acc_22_nr2_1; //accelerator for calculating k22
    u_nm_accel_t *acc_22_nr2_2;

    u_nm_accel_t *acc_21_nr2_1; //accelerator for calculating k21
    u_nm_accel_t *acc_21_nr1_2;

    u_nm_accel_t *acc_12_nr1_1; //accelerator for calculating k12
    u_nm_accel_t *acc_12_nr2_2;

    mirror_kmn_q_t *kmnq;
    
    surface_merged_map_t *merged_map;
    mirror_t *mirror;
    double nr1, nr2;
    double A, Cr, Ct, Br, Bt;
    double kr, kt, k0;
    double xmin[2], xmax[2];
    // Factors of u_nm that depend on z,n,m,n',m' only, 1c2 part means nr1
    // medium times conjugate of nr2 medium etc.
    complex_t znm1c1;
    complex_t znm1c2;
    complex_t znm2c1;
    complex_t znm2c2;
    KNM_MIRROR_NODE_DIRECTION_t knum;
    double J1, J2;
    double aperture_radius; // 0 if infinite radius
    bool polar_limits_used;
    int integrating;
    double betax, betay;
    bool usingMap;
    bool K11_calc;
    bool K22_calc;
    bool K12_calc;
    bool K21_calc;
} knm_map_integration_params_t;

void get_int_limit(knm_map_integration_params_t *p, int knum);


#endif // KAT_AA_H

/*
 * Local variables:
 * c-indentation-style: bsd
 * c-basic-offset: 2
 * indent-tabs-mode: nil
 * End:
 *
 * vim: tabstop=2 expandtab shiftwidth=2:
 */
