// $Id$

/*!
 * \file kat_gnu.c
 * \brief Routines for organising and writing the output gnuplot file
 * 
 * @section Copyright notice
 *
 *  This file is part of the interferometer simulation Finesse
 *  http://www.gwoptics.org/finesse
 *
 *  Copyright (C) 1999 onwards Andreas Freise
 *  with parts of the code written by Daniel Brown, Paul Cochrane
 *  and Gerhard Heinzel.
 *
 *  This program is free software; you can redistribute it and/or modify it under
 *  the terms of the GNU General Public License version 3 as published
 *  by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 *  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *  See the GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along with
 *  this library; if not, write to the Free Software Foundation, Inc., 59 Temple Place,
 *  Suite 330, Boston, MA 02111-1307 USA
 */


#include "kat.h"
#include "kat_gnu.h"
#include "kat_inline.c"
#include "kat_io.h"
#include "kat_aux.h"

// test
extern interferometer_t inter;
extern init_variables_t init;
extern time_t now;
extern char *mydate;

//! my favourite colour schemes for gnuplot
/*!
 *
 */
static const int lines[3][10] = {
    {0, 1, 2, 3, 4, 5, 6, 7, 8, 9},
    {0, 3, 7, 4, 1, 2, 5, 6, 8, 10},
    {0, 7, 3, 8, 4, 1, 2, 5, 9, 10}
};

//! Convert an output type number into its respective name

/*!
 * \param i type reference
 *
 * \return the output type as a string
 */
char *otypename(int i) {
    switch (i) {
        case OTYPE_RE:
            return "Re";
            break;
        case OTYPE_IM:
            return "Im";
            break;
        case OTYPE_ABS:
            return "Abs";
            break;
        case OTYPE_DEG:
        case OTYPE_DEGP:
        case OTYPE_DEGM:
            return "Phase";
            break;
        case OTYPE_DB:
            return "dB";
            break;
        default:
            bug_error("incorrect output type");
            break;
    }
    return ("error");
}

//! Sets up and writes the gnuplot file

/*!
 * \param step ???
 * \param fp the file pointer
 * \param fname the file name
 */
void gnufile(int step, FILE *fp, char *fname) {
    char xy1, xy2;
    char mytime[30] = {0};
    int col;
    int found;
    char plotname[LINE_LEN] = {0};
    char beamsize[LINE_LEN] = {0};

    //  char ylist[LINE_LEN] = { 0 };
    //  ctemp[10] = { 0 };
    int lt;
    char title[LINE_LEN] = {0};
    int first;

    gnuplot_terminal_t *gp;

    // make sure file pointer isn't null
    assert(fp != NULL);

    // make sure step is greater than or equal to zero
    //!< \todo verify that step should be greater than or equal to zero
    assert(step >= 0);

    if (inter.beam.set && inter.beam.swap) {
        xy1 = 'y';
        xy2 = 'x';
    } else {
        xy1 = 'x';
        xy2 = 'y';
    }

    fputs("reset\n", fp);

    if (inter.x1.xtype == FLOG) {
        fprintf(fp, "set %.1srange[%g:%g]\n", &xy1, inter.x1.xmin / inter.x1.op,
                inter.x1.xmax / inter.x1.op);
    } else {
        fprintf(fp, "set %.1srange[%g:%g]\n", &xy1, inter.x1.xmin - inter.x1.op,
                inter.x1.xmax - inter.x1.op);
    }

    if (inter.x1.xtype == FLOG && inter.x1.xmax / inter.x1.xmin >= 10.0) {
        fprintf(fp, "set logscale %.1s\n", &xy1);
    }

    fprintf(fp, "set %.1slabel \"%s", &xy1, inter.x1.xname);
    fprintf(fp, "\"\n");

    if (inter.splot) {
        if (inter.x2.xtype == FLOG) {
            fprintf(fp, "set %.1srange[%g:%g]\n", &xy2,
                    inter.x2.xmin / inter.x2.op, inter.x2.xmax / inter.x2.op);
        } else {
            fprintf(fp, "set %.1srange[%g:%g]\n", &xy2,
                    inter.x2.xmin - inter.x2.op, inter.x2.xmax - inter.x2.op);
        }

        if (inter.x2.xtype == FLOG && inter.x2.xmax / inter.x2.xmin >= 10.0) {
            fprintf(fp, "set logscale %.1s\n", &xy2);
        }

        fprintf(fp, "set %.1slabel \"%s", &xy2, inter.x2.xname);
        fprintf(fp, "\"\n");

        fprintf(fp, "set zlabel \"%s\"\n", inter.y1name);
        fputs("set mztics 10\n", fp);
    } else {
        fprintf(fp, "set %.1slabel \"%s\"\n", &xy2, inter.y1name);
    }

    // does this work with axis switching ??
    if (inter.yaxis[1]) {
        fprintf(fp, "set y2label \"%s\"\n", inter.y2name);
        fputs("set y2tics nomirror\n", fp);
        fputs("set my2tics 3\n", fp);
        fputs("set ytics nomirror\n", fp);
    }

    fputs("set mxtics 2\n", fp);
    fputs("set mytics 2\n", fp);
    fputs("set zero 0.0\n", fp);

    strcpy(mytime, asctime(localtime(&now)));
    mytime[strlen(mytime) - 1] = '\0';

    if (init.gnuversion > 4.0) {
        fprintf(fp, "set title \"%s                %s\" offset 0, 2\n",
                inter.basename, mytime);
    } else {
        fprintf(fp, "set title \"%s                %s\", 2\n", inter.basename,
                mytime);
    }

    if (inter.x3.xtype != NO_FREQ) {
        if (inter.x3.xtype == FLOG) {
            fprintf(fp, "set label \"%s = %s\" at screen .1, 0.9 \n",
                    inter.x3.xname, double_form(*inter.x3.xaxis / inter.x3.op));
        } else {
            fprintf(fp, "set label \"%s = %s\" at screen .1, 0.9 \n",
                    inter.x3.xname, double_form(*inter.x3.xaxis - inter.x3.op));
        }

        fprintf(fp, "set label \"(%d/%d)\" at screen .8, 0.9 \n", step,
                inter.x3.xsteps);
    }


    // print the beam size for the first beam detector
    /* put into label and kat_init
    if (inter.beam.set) {
      detno = -1;
      int i = 0;

      while (i < inter.num_outputs) {
        if (inter.output_data_list[i].detector_type == BEAM) {
          detno = inter.output_data_list[i].detector_index;
          i = inter.num_outputs;
        }
        i++;
      }
      if (detno == -1) {
        bug_error("detno -1");
      }
      nr = *inter.node_list[inter.output_data_list[detno].node_index].n;
      fprintf(fp, "set label \"wx0 = %sm  wy0 = %sm\" at screen .2, 0.86 \n",
              double_form(w0_size(*inter.output_data_list[detno].qx, nr)),
              double_form(w0_size(*inter.output_data_list[detno].qy, nr)));
      fprintf(fp, "set label \"wx  = %sm  wy  = %sm\" at screen .5, 0.86 \n",
              double_form(w_size(*inter.output_data_list[detno].qx, nr)),
              double_form(w_size(*inter.output_data_list[detno].qy, nr)));
    }
     */


    if (inter.ytype == FLOG) {
        if (inter.splot) {
            fputs("set log z\n", fp);
        } else {
            fputs("set log y\n", fp);
        }

        if (inter.yaxis[1] == OTYPE_IM) {
            fputs("set log y2\n", fp);
        }
    } else {
        if (inter.splot) {
            fputs("set nolog z\n", fp);
        } else {
            fputs("set nolog y\n", fp);
        }

        if (inter.yaxis[1]) {
            fputs("set nolog y2\n", fp);
        }
    }

    /* set first term */
    assert(inter.num_gnuterm_cmds > 0);
    gp = &(init.gnuterm[inter.gnuterm[0]]);
    fputs(gp->command, fp);
    if (gp->output_is_file) {
        fprintf(fp, "set output \"%s\"\n", inter.gnutermfn[0]);
    }

    if (strcasecmp(gp->name, "dumb") == 0) {
        fputs("pause -1 \"Hit return to show graph\"\n", fp);
    }

    if (inter.num_gnuplot_cmds) {
        fputs(inter.gnuplotcommand, fp);
    }

    // 3D plots
    if (inter.splot) {
        if (NOT inter.multi) {
            // fprintf(fp, "set zrange[%g:%g]\n", 99/100.0*inter.total_min, 101/100.0*inter.total_max);
            int i = 0;
            found = 0;

            while (found == 0 && i < inter.num_outputs) {
                if (!inter.output_data_list[i].noplot) {
                    found = 1;
                } else {
                    i++;
                }
            }

            if (found == 0) {
                gerror("no output (without 'noplot') has been found\n");
            }

            strcpy(plotname, inter.output_data_list[i].name);
            if (inter.output_data_list[i].is_second_beam) {
                strcat(plotname, "*");
            }

            output_data_t current_output = inter.output_data_list[i];
            if (current_output.detector_type == BEAM) {
                double nr = *inter.node_list[current_output.node_index].n;
                sprintf(beamsize, " (wx0=%s, wy0=%s)",
                        double_form(w0_size(*current_output.qx, nr)),
                        double_form(w0_size(*current_output.qy, nr)));
                strcat(plotname, beamsize);
            }


            strcat(plotname, " : ");
            strcpy(title, plotname);
            fputs("splot\\\n", fp);

            if (inter.beam.set && inter.beam.swap) {
                fprintf(fp, "\'%s\' using ($2):($1):($%d) title \"%s \"  w l\n",
                        fname, 3 + i, title);
            } else {
                fprintf(fp, "\'%s\' using ($1):($2):($%d) title \"%s \"  w l\n",
                        fname, 3 + i, title);
            }
        }            // ... but newer gnuplot versions allow to plot multiple surfaces at once
        else {
            fputs("splot", fp);
            first = 1;
            col = 3;

            int i;
            for (i = 0; i < inter.num_outputs; i++) {
                strcpy(plotname, "");

                if (inter.num_outputs > 1) {
                    strcat(plotname, inter.output_data_list[i].name);
                }

                if (inter.output_data_list[i].detector_type != FEEDBACK
                        && inter.output_data_list[i].detector_type != UFUNCTION) {
                    strcat(plotname, " ");

                    // check node indices before passing to node_print
                    assert(inter.output_data_list[i].node_index >= 0);
                    assert(inter.output_data_list[i].node_index < inter.num_nodes);

                    strcat(plotname, node_print(inter.output_data_list[i].node_index));

                    if (inter.output_data_list[i].is_second_beam) {
                        strcat(plotname, "*");
                    }
                }

                output_data_t current_output = inter.output_data_list[i];
                if (current_output.detector_type == BEAM) {
                    double nr = *inter.node_list[current_output.node_index].n;
                    sprintf(beamsize, " (wx0=%s, wy0=%s)",
                            double_form(w0_size(*current_output.qx, nr)),
                            double_form(w0_size(*current_output.qy, nr)));
                    strcat(plotname, beamsize);
                }

                strcat(plotname, " : ");
                strcpy(title, plotname);

                if (!inter.output_data_list[i].noplot) {
                    if (!first) {
                        fprintf(fp, ", ");
                    } else {
                        first = 0;
                    }

                    if (inter.beam.set && inter.beam.swap) {
                        fprintf(fp,
                                "\\\n\'%s\' using ($2):($1):($%d) title \"%s \"  w l ",
                                fname, col++, title);
                    } else {
                        fprintf(fp,
                                "\\\n\'%s\' using ($1):($2):($%d) title \"%s \"  w l",
                                fname, col++, title);
                    }

                    if (inter.yaxis[1]) {

                        if (inter.ploty2axis) {
                            strcpy(title, plotname);
                            strcat(title, inter.y2name);
                            fprintf(fp, ", ");

                            if (inter.beam.set && inter.beam.swap) {
                                fprintf(fp,
                                        "\\\n\'%s\' using ($2):($1):($%d) title \"%s \"  w l ",
                                        fname, col++, title);
                            } else {
                                fprintf(fp,
                                        "\\\n\'%s\' using ($1):($2):($%d) title \"%s \"  w l",
                                        fname, col++, title);
                            }
                        } else {
                            col++;
                        }
                    }
                } else {
                    col++;
                }
            }
            fprintf(fp, "\n");
        }
    } else {
        // ordinary 2D plots
        fputs("plot", fp);
        first = 1;
        col = 2;

        lt = 1;

        int i;
        for (i = 0; i < inter.num_outputs; i++) {
            output_data_t current_output = inter.output_data_list[i];
            strcpy(plotname, "");

            if (inter.num_outputs > 1) {
                strcat(plotname, current_output.name);
            }

            if (current_output.detector_type != FEEDBACK &&
                    current_output.detector_type != UFUNCTION) {
                strcat(plotname, " ");

                // if the output beam param is not beam_param_not_set, 
                // append to the plotname
                if (NOT current_output.beam_param_not_set) {
                    // check node indices before passing to node_print()
                    assert(current_output.node_index >= 0);
                    assert(current_output.node_index < inter.num_nodes);

                    strcat(plotname, node_print(current_output.node_index));
                }

                // mark the beam as the "other beam" if necessary
                if (current_output.is_second_beam) {
                    strcat(plotname, "*");
                }
            }

            if (current_output.detector_type == BEAM) {
                double nr = *inter.node_list[current_output.node_index].n;
                sprintf(beamsize, " (wx0=%s, wy0=%s)",
                        double_form(w0_size(*current_output.qx, nr)),
                        double_form(w0_size(*current_output.qy, nr)));
                strcat(plotname, beamsize);
            }

            strcat(plotname, " : ");
            strcpy(title, plotname);

            if (inter.yaxis[1]) {
                strcat(title, inter.y1name);
            }

            if (!inter.output_data_list[i].noplot) {
                if (!first) {
                    fprintf(fp, ", ");
                } else {
                    first = 0;
                }

                fprintf(fp, "\\\n\'%s\' using ($1):($%d) axes x1y%d "
                        "title \"%s \" w l lt %d lw %d",
                        fname, col++, 1, title,
                        (lt++ > 10 ? lt - 1 : lines[inter.mycolor][lt - 1]),
                        inter.width);
            } else {
                col++;
            }

            if (inter.yaxis[1] && inter.ploty2axis) {

                if (!inter.output_data_list[i].noplot) {
                    strcpy(title, plotname);
                    strcat(title, inter.y2name);
                    fprintf(fp, ", \\\n\'%s\' using ($1):($%d) axes x1y%d "
                            "title \"%s \" w l lt %d lw %d",
                            fname, col++, 2, title,
                            (lt++ > 10 ? lt - 1 : lines[inter.mycolor][lt - 1]),
                            inter.width);
                } else {
                    col++;
                }
            }
        }
        fprintf(fp, "\n");
    }

    if ((inter.pause == 1)
            && !gp->output_is_file) {
        fputs("pause -1 \"Hit return to continue\"\n", fp);
    }

    /* set more terminals */
    int k;
    for (k = 1; k < inter.num_gnuterm_cmds; k++) {
        gp = &(init.gnuterm[inter.gnuterm[k]]);
        fputs(gp->command, fp);
        if (gp->output_is_file) {
            fprintf(fp, "set output \"%s\"\n", inter.gnutermfn[k]);
        }
        fputs("replot\n", fp);

        if (inter.debug & 256) {
            message("GNU: gnuc:%s\n", init.gnucommand);
            message("GNU: term:%s\n", gp->name);
        }

        if ((inter.pause == 1)
                && !gp->output_is_file) {
            fputs("pause -1 \"Hit return to continue\"\n", fp);
        }
    }
}


//! Sets up and writes the matlab file, so far this is a quick hack, needs fixing!

/*!
 * \param fp the file pointer
 * \param fname the file name
 */
void matlabfile(FILE *fp, char *fname) {
    char xy1[3], xy2[3], XY1[3], XY2[3];
    char mytime[30] = {0};

    // make sure file pointer isn't null
    assert(fp != NULL);


    // replace - by _ in function and file names so that matlab can see
    // the functions, then replace _ by \\_ in the print strings so that
    // matlab prints this correctly
    char matlab_print_str[LINE_LEN];
    char matlab_func_str[LINE_LEN];
    strcpy(matlab_func_str, inter.basename);
    int replace_error = replace_string(matlab_func_str, "-", "_", 1, 1);
    if (replace_error == -1) {
        bug_error("could not replace '-' in matlab string");
    }
    strcpy(matlab_print_str, matlab_func_str);
    replace_error = replace_string(matlab_print_str, "_", "\\_", 1, 2);
    if (replace_error == -1) {
        bug_error("could not replace '_' in matlab string");
    }

    if (inter.beam.set && inter.beam.swap) {
        strcpy(XY1, "Y");
        strcpy(XY2, "X");
        strcpy(xy1, "y");
        strcpy(xy2, "x");
    } else {
        strcpy(XY1, "X");
        strcpy(XY2, "Y");
        strcpy(xy1, "x");
        strcpy(xy2, "y");
    }

    write_matlab_usage(fp, matlab_func_str);

    if (inter.splot) {
        fprintf(fp, "function [x,y,z] = %s(noplot)\n", matlab_func_str);
    } else {
        fprintf(fp, "function [x,y] = %s(noplot)\n", matlab_func_str);
    }
    fprintf(fp, "\ndata = load(\'%s\');\n", fname);
    fprintf(fp, "[rows,cols]=size(data);\n");

    if (inter.splot) {
        fprintf(fp, "%.2s=data(:,1);\n", XY1);
        fprintf(fp, "%.2s=data(:,2);\n", XY2);
        fprintf(fp, "Z=data(:,3:cols);\n");
        fputs("[x,y,z]=convert3D(X,Y,Z);\n", fp);
    } else {
        fprintf(fp, "x=data(:,1);\n");
        fprintf(fp, "y=data(:,2:cols);\n");
    }


    // print user defined matlab commands
    if (inter.num_matlab_cmds) {
        fprintf(fp, "%%%%%% --- User commands\n");
        fputs(inter.matlabcommand, fp);
        fprintf(fp, "%%%%%% --- \n");
    }

    // prepare plottitle
    strcpy(mytime, asctime(localtime(&now)));
    mytime[strlen(mytime) - 1] = '\0';
    fprintf(fp, "mytitle=\'%s                %s\';\n", matlab_print_str,
            mytime);

    fprintf(fp, "if (nargin==0)\n");
    fprintf(fp, "\nfigure(\'name\','%s');\n", matlab_func_str);

    int subplot;
    subplot = 0;
    if (inter.yaxis[1] && inter.ploty2axis) {
        fprintf(fp, "\nh1=subplot(2,1,1);\n");
        subplot = 1;
    }

    // first plot -----------------------------------------------------------------

    // 3D plots
    if (inter.splot) {

        if (inter.multi) {
            fprintf(fp, "hold on;\n");
        }
        plot3Dmatlab(fp, 0);
        if (inter.multi) {
            fprintf(fp, "hold off;\n");
        }
    } else {
        // ordinary 2D plots
        plot2Dmatlab(fp, 0);
    }

    // set axes to linear or logarithmic
    if (inter.ytype == FLOG) {
        if (inter.splot) {
            fprintf(fp, "set(gca, \'ZScale\', \'log\');\n");
        } else {
            fprintf(fp, "set(gca, \'YScale\', \'log\');\n");
        }

    } else {
        if (inter.splot) {
            fprintf(fp, "set(gca, \'ZScale\', \'lin\');\n");
        } else {
            fprintf(fp, "set(gca, \'YScale\', \'lin\');\n");
        }
    }

    // char for replacing _ in label strings
    char matlab_label[LINE_LEN];

    if (inter.splot) {

        double mmin;
        double mmax;
        double swap;
        if (inter.x2.xtype == FLOG) {
            mmin = inter.x2.xmin / inter.x2.op;
            mmax = inter.x2.xmax / inter.x2.op;
        } else {
            mmin = inter.x2.xmin - inter.x2.op;
            mmax = inter.x2.xmax - inter.x2.op;
        }

        // matlab does not accept XLim with min>max, need to use the 'reverse' keyword
        if (mmin > mmax) {
            SWAP(mmin, mmax);
            fprintf(fp, "set(gca, \'%.2sDir\', \'reverse\');\n", XY2);
        }


        fprintf(fp, "set(gca, \'%.2sLim\', [%g %g]);\n", XY2, mmin, mmax);


        if (inter.x2.xtype == FLOG && inter.x2.xmax / inter.x2.xmin >= 10.0) {
            fprintf(fp, "set(gca, \'%.2sScale\', \'log\');\n", XY2);

        }


        strcpy(matlab_label, inter.x2.xname);
        replace_error = replace_string(matlab_label, "_", "\\_", 1, 2);
        if (replace_error == -1) {
            bug_error("could not replace '_' in matlab string");
        }

        fprintf(fp, "%.2slabel(\'%s\');\n", xy2, matlab_label);

        strcpy(matlab_label, inter.y1name);
        replace_error = replace_string(matlab_label, "_", "\\_", 1, 2);
        if (replace_error == -1) {
            bug_error("could not replace '_' in matlab string");
        }

        fprintf(fp, "zlabel(\'%s\');\n", matlab_label);
        //    fputs("set mztics 10\n", fp);
    } else {

        strcpy(matlab_label, inter.y1name);
        replace_error = replace_string(matlab_label, "_", "\\_", 1, 2);
        if (replace_error == -1) {
            bug_error("could not replace '_' in matlab string");
        }

        fprintf(fp, "%.2slabel(\'%s\');\n", xy2, matlab_label);
    }

    setmatlabaxisprops(fp, XY1, xy1);

    // print user defined matlab commands
    if (inter.num_matlabplot_cmds) {
        fprintf(fp, "%%%%%% --- User plot commands\n");
        fputs(inter.matlabplotcommand, fp);
        fprintf(fp, "%%%%%% --- \n");
    }

    // second plot -----------------------------------------------------------------

    if (subplot) {
        fprintf(fp, "\nh2=subplot(2,1,2);\n");


        // 3D plots
        if (inter.splot) {
            if (inter.multi) {
                fprintf(fp, "hold on;\n");
            }
            plot3Dmatlab(fp, 1);
            if (inter.multi) {
                fprintf(fp, "hold off;\n");
            }
        }
        else {
            // ordinary 2D plots
            plot2Dmatlab(fp, 1);
        }

        strcpy(matlab_label, inter.y2name);
        replace_error = replace_string(matlab_label, "_", "\\_", 1, 2);
        if (replace_error == -1) {
            bug_error("could not replace '_' in matlab string");
        }

        fprintf(fp, "ylabel(\'%s\');\n", matlab_label);

        if (inter.ytype == FLOG) {
            if (inter.yaxis[1] == OTYPE_IM) {
                fprintf(fp, "set(gca, \'YScale\', \'log\');\n");
            }
        } else {
            fprintf(fp, "set(gca, \'YScale\', \'lin\');\n");
        }

        setmatlabaxisprops(fp, XY1, xy1);

        fprintf(fp, "\n");
        // print user defined matlab commands
        if (inter.num_matlabplot_cmds) {
            fprintf(fp, "%%%%%% --- User plot commands\n");
            fputs(inter.matlabplotcommand, fp);
            fprintf(fp, "%%%%%% --- \n");
        }

        // now switch handle to first subplot
        fprintf(fp, "subplot(2,1,1);\n");
    }


    /*
    if (inter.beam.set) {
      detno = -1;
      int i = 0;

      while (i < inter.num_outputs) {
        if (inter.output_data_list[i].detector_type == BEAM) {
          detno = inter.output_data_list[i].detector_index;
          i = inter.num_outputs;
        }
        i++;
      }

      if (detno == -1) {
        bug_error("detno -1");
      }

      nr = *inter.node_list[inter.output_data_list[detno].node_index].n;
		
          // following needs proper x,y coordinates
      fprintf(fp, "mytitle=char(mytitle,\'wx0 = %sm, wy0 = %sm, wx  = %sm, wy  = %sm\');\n",
              double_form(w0_size(*inter.output_data_list[detno].qx, nr)),
              double_form(w0_size(*inter.output_data_list[detno].qy, nr)),
              double_form(w_size(*inter.output_data_list[detno].qx, nr)),
              double_form(w_size(*inter.output_data_list[detno].qy, nr)));
    }
     */


    fprintf(fp, "title(mytitle);\n");
    fprintf(fp, "end\n\n");

    write_nargout_check(fp, inter.splot);

    if (inter.splot) {
        write3Dfunction(fp);
    }

}

//! Writes Matlab help code into .m file

/*!
 * \param fp the file pointer
 */

void write_matlab_usage(FILE *fp, char *matlab_func_str) {
    fprintf(fp, "%%----------------------------------------------------------------\n");
    if (inter.splot) {
        fprintf(fp, "%% function [x,y,z] = %s(noplot)\n", matlab_func_str);
    } else {
        fprintf(fp, "%% function [x,y] = %s(noplot)\n", matlab_func_str);
    }

    fprintf(fp, "%% Matlab function to plot Finesse output data\n");
    fprintf(fp, "%% Usage: \n");
    if (inter.splot) {
        fprintf(fp, "%% [x,y,z] = %s   ", matlab_func_str);
    } else {
        fprintf(fp, "%%   [x,y] = %s   ", matlab_func_str);
    }
    fprintf(fp, " : plots and returns the data\n");
    if (inter.splot) {
        fprintf(fp, "%% [x,y,z] = %s(1)", matlab_func_str);
    } else {
        fprintf(fp, "%%   [x,y] = %s(1)", matlab_func_str);
    }
    fprintf(fp, " : just returns the data\n");
    fprintf(fp, "%%           %s    : just plots the data\n", matlab_func_str);
    fprintf(fp, "%% Created automatically %s", asctime(localtime(&now)));
    fprintf(fp, "%% by Finesse %s (%s), %s\n", VERSION, SVN_REVISION, MYTIME);
    fprintf(fp, "%%----------------------------------------------------------------\n");
}


//! Writes Matlab code into .m file for returning output values

/*!
 * \param fp the file pointer
 * \param is_splot 0 for 2d plot and 1 for 3d plot
 */

void write_nargout_check(FILE *fp, int is_splot) {
    if (is_splot) {
        fprintf(fp, "switch nargout\n");
        fprintf(fp, " case {0}\n");
        fprintf(fp, "  clear x y z;\n");
        fprintf(fp, " case {3}\n");
        fprintf(fp, " otherwise\n");
        fprintf(fp, "  error(\'wrong number of outputs\');\n");
        fprintf(fp, "end\n");
    } else {
        fprintf(fp, "switch nargout\n");
        fprintf(fp, " case {0}\n");
        fprintf(fp, "  clear x y;\n");
        fprintf(fp, " case {2}\n");
        fprintf(fp, " otherwise\n");
        fprintf(fp, "  error(\'wrong number of outputs\');\n");
        fprintf(fp, "end\n");
    }
}

//! Writes Matlab code into .m file for converting 3D data from file into Matlab format for surf

/*!
 * \param fp the file pointer
 */

void write3Dfunction(FILE *fp) {
    fprintf(fp, "\n%%---------------------------------------------------------------\n");
    fprintf(fp, "%% Utility function to convert Finesse 3D data into Matlab format\n");
    fprintf(fp, "function [x,y,M]=convert3D(X,Y,Z)\n");
    fprintf(fp, "[row,col]=size(Z);\n");
    fprintf(fp, "nxp=length(find (Y(:)==Y(1)));\n");
    fprintf(fp, "nyp=row/nxp;\n");
    fprintf(fp, "y=Y(1:nyp);\n");
    fprintf(fp, "x=X(1:nyp:row);\n");
    fprintf(fp, "M=zeros(nyp,nxp,col);\n");
    fprintf(fp, "for i=1:col\n");
    fprintf(fp, "M(:,:,i)=reshape(Z(:,i),nyp,nxp);\n");
    fprintf(fp, "end\n");
}


/*

  // does this work with axis switching ??
  if (inter.yaxis[1]) {
    fprintf(fp, "set y2label \"%s\"\n", inter.y2name);
    fputs("set y2tics nomirror\n", fp);
    fputs("set my2tics 3\n", fp);
    fputs("set ytics nomirror\n", fp);
  }

    if (inter.yaxis[1] == OTYPE_IM) {
      fputs("set log y2\n", fp);
    }

    if (inter.yaxis[1]) {
      fputs("set nolog y2\n", fp);
        }




          if (inter.yaxis[1]) {

            if (inter.ploty2axis) {
              strcpy(title, plotname);
              strcat(title, inter.y2name);
              fprintf(fp, ", ");

              if (inter.beam.set && inter.beam.swap) {
                fprintf(fp, 
                        "\\\n\'%s\' using ($2):($1):($%d) title \"%s \"  w l ", 
                        fname, col++, title);
              }
              else {
                fprintf(fp, 
                        "\\\n\'%s\' using ($1):($2):($%d) title \"%s \"  w l", 
                        fname, col++, title);
              }
            }
            else {
              col++;
            }
          }

      if (inter.yaxis[1] && inter.ploty2axis) {

        if (!inter.output_data_list[i].noplot) {
          strcpy(title, plotname);
          strcat(title, inter.y2name);
          fprintf(fp, ", \\\n\'%s\' using ($1):($%d) axes x1y%d "
                  "title \"%s \" w l lt %d lw %d",
                  fname, col++, 2, title,
                  (lt++ > 10 ? lt - 1 : lines[inter.mycolor][lt - 1]),
                  inter.width);
        }
        else {
          col++;
        }
      }

 */




/*  // Will do this later !!!!
if (inter.x3.xtype != NO_FREQ) {
if (inter.x3.xtype == FLOG) {
  fprintf(fp, "set label \"%s = %s\" at screen .1, 0.9 \n",
          inter.x3.xname, double_form(*inter.x3.xaxis / inter.x3.op));
}
else {
  fprintf(fp, "set label \"%s = %s\" at screen .1, 0.9 \n",
          inter.x3.xname, double_form(*inter.x3.xaxis - inter.x3.op));
}

fprintf(fp, "set label \"(%d/%d)\" at screen .8, 0.9 \n", step,
        inter.x3.xsteps);
}
 */


//!

/*!
 * \todo undocumented
 */
void setmatlabaxisprops(FILE *fp, char *XY1, char *xy1) {

    // char for replacing _ in label strings
    char matlab_label[LINE_LEN];

    double mmin;
    double mmax;
    double swap;
    if (inter.x1.xtype == FLOG) {
        mmin = inter.x1.xmin / inter.x1.op;
        mmax = inter.x1.xmax / inter.x1.op;
    } else {
        mmin = inter.x1.xmin - inter.x1.op;
        mmax = inter.x1.xmax - inter.x1.op;
    }

    // matlab does not accept XLim with min>max, need to use the 'reverse' keyword
    if (mmin > mmax) {
        SWAP(mmin, mmax);
        fprintf(fp, "set(gca, \'%.2sDir\', \'reverse\');\n", XY1);
    }

    fprintf(fp, "set(gca, \'%.2sLim\', [%g %g]);\n", XY1, mmin, mmax);


    if (inter.x1.xtype == FLOG && inter.x1.xmax / inter.x1.xmin >= 10.0) {
        fprintf(fp, "set(gca, \'%.2sScale\', \'log\');\n", XY1);

    }

    strcpy(matlab_label, inter.x1.xname);
    int replace_error = replace_string(matlab_label, "_", "\\_", 1, 2);
    if (replace_error == -1) {
        bug_error("could not replace '_' in matlab string");
    }

    fprintf(fp, "%.2slabel(\'%s\');\n", xy1, matlab_label);

    // some useful things to set:
    fprintf(fp, "grid on;\n");
    if (inter.splot) {
        fprintf(fp, "colorbar;\n");
    }

}


//!

/*!
 * \todo undocumented
 */
void plot2Dmatlab(FILE *fp, int startcol) {
    int first;
    int col;
    char plotname[LINE_LEN] = {0};
    char legendstring[LINE_LEN] = {0};
    char title[LINE_LEN] = {0};
    char beamsize[LINE_LEN] = {0};

    fputs("plot(", fp);
    first = 1;
    //col = 2;
    col = 1;

    strcpy(legendstring, "");

    int i;
    for (i = 0; i < inter.num_outputs; i++) {
        output_data_t current_output = inter.output_data_list[i];
        strcpy(plotname, "");

        if (inter.num_outputs > 1) {
            strcat(plotname, current_output.name);
        }

        if (current_output.detector_type != FEEDBACK &&
                current_output.detector_type != UFUNCTION) {
            strcat(plotname, " ");

            // if the output beam param is not beam_param_not_set, 
            // append to the plotname
            if (NOT current_output.beam_param_not_set) {
                // check node indices before passing to node_print()
                assert(current_output.node_index >= 0);
                assert(current_output.node_index < inter.num_nodes);

                strcat(plotname, node_print(current_output.node_index));
            }

            // mark the beam as the "other beam" if necessary
            if (current_output.is_second_beam) {
                strcat(plotname, "*");
            }
        }

        if (current_output.detector_type == BEAM) {
            double nr = *inter.node_list[current_output.node_index].n;
            sprintf(beamsize, " (wx0=%s, wy0=%s)",
                    double_form(w0_size(*current_output.qx, nr)),
                    double_form(w0_size(*current_output.qy, nr)));
            strcat(plotname, beamsize);
        }

        strcpy(title, plotname);

        //      if (inter.yaxis[1]) {
        //  strcat(title, inter.y1name);
        //}

        if (!inter.output_data_list[i].noplot) {
            if (!first) {
                fprintf(fp, ", ");
                strcat(legendstring, ", ");
            } else {
                first = 0;
            }

            //fprintf(fp, "data(:,1),data(:,%d)",startcol+col++);
            fprintf(fp, "x, y(:,%d)", startcol + col++);
            strcat(legendstring, "\'");
            strcat(legendstring, title);
            strcat(legendstring, "\'");

        } else {
            col++;

        }
        if (inter.yaxis[1]) {
            col++;
        }

    }

    // char for replacing _ in label strings
    char matlab_label[LINE_LEN];

    strcpy(matlab_label, legendstring);
    int replace_error = replace_string(matlab_label, "_", "\\_", 1, 2);
    if (replace_error == -1) {
        bug_error("could not replace '_' in matlab string");
    }


    fprintf(fp, ");\n");
    fprintf(fp, "legend(%s);\n", matlab_label);
}


//! Writes Matlab code to plots 3D data, kind of works but is not optimised yet

/*!
 * 
 */
void plot3Dmatlab(FILE *fp, int startcol) {
    int found;
    int first;
    char plotname[LINE_LEN] = {0};
    char legendstring[LINE_LEN] = {0};
    char title[LINE_LEN] = {0};
    char beamsize[LINE_LEN] = {0};

    int i = 0;
    found = 0;
    first = 1;

    int col_step = 1;
    if (inter.yaxis[1]) {
        col_step = 2;
    }

    for (i = 0; i < inter.num_outputs; i++) {
        strcpy(plotname, inter.output_data_list[i].name);

        if (inter.output_data_list[i].detector_type != FEEDBACK
                && inter.output_data_list[i].detector_type != UFUNCTION) {
            strcat(plotname, " ");

            // check node indices before passing to node_print
            assert(inter.output_data_list[i].node_index >= 0);
            assert(inter.output_data_list[i].node_index < inter.num_nodes);

            strcat(plotname, node_print(inter.output_data_list[i].node_index));

            if (inter.output_data_list[i].is_second_beam) {
                strcat(plotname, "*");
            }
        }

        output_data_t current_output = inter.output_data_list[i];
        if (current_output.detector_type == BEAM) {
            double nr = *inter.node_list[current_output.node_index].n;
            sprintf(beamsize, " (wx0=%s, wy0=%s)",
                    double_form(w0_size(*current_output.qx, nr)),
                    double_form(w0_size(*current_output.qy, nr)));
            strcat(plotname, beamsize);
        }

        strcpy(title, plotname);

        if (!inter.output_data_list[i].noplot) {
            found = 1;
            if (!first) {
                strcat(legendstring, ", ");
            } else {
                first = 0;
            }

            strcat(legendstring, "\'");
            strcat(legendstring, title);
            strcat(legendstring, "\'");

            fputs("surfc(", fp);

            //      printf("i %d colstep %d startcol %d=%d\n",i,col_step, startcol, 1+i*col_step+startcol);

            if (inter.beam.set && inter.beam.swap) {
                fprintf(fp, "y,x,z(:,:,%d),\'EdgeColor\',\'none\');\n", 1 + i * col_step + startcol);
            } else {
                fprintf(fp, "x,y,z(:,:,%d),\'EdgeColor\',\'none\');\n", 1 + i * col_step + startcol);
            }

            // only plot one if multi is not set
            if (!inter.multi) {
                i = inter.num_outputs;
            }

            // end of if (!noplot)
        }
    }

    if (found == 0) {
        gerror("no output (without 'noplot') has been found\n");
    }

    // char for replacing _ in label strings
    char matlab_label[LINE_LEN];
    strcpy(matlab_label, legendstring);
    int replace_error = replace_string(matlab_label, "_", "\\_", 1, 2);
    if (replace_error == -1) {
        bug_error("could not replace '_' in matlab string");
    }


    // in 3D plot sthe legend makes not much sense, comment it out unless 'multi' is set
    //  if (inter.multi) {
    if (!startcol && inter.multi) {
        fprintf(fp, "legend(%s);\n", matlab_label);
    } else {
        fprintf(fp, "%%legend(%s);\n", matlab_label);
    }
}


/*
 * Local variables:
 * c-indentation-style: bsd
 * c-basic-offset: 2
 * indent-tabs-mode: nil
 * End:
 *
 * vim: tabstop=2 expandtab shiftwidth=2:
 */
