
/*!
 * \file kat_io.c
 * \brief Input-output routines.
 * 
 * @section Copyright notice
 *
 *  This file is part of the interferometer simulation Finesse
 *  http://www.gwoptics.org/finesse
 *
 *  Copyright (C) 1999 onwards Andreas Freise
 *  with parts of the code written by Daniel Brown, Paul Cochrane
 *  and Gerhard Heinzel.
 *
 *  This program is free software; you can redistribute it and/or modify it under
 *  the terms of the GNU General Public License version 3 as published
 *  by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 *  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *  See the GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along with
 *  this library; if not, write to the Free Software Foundation, Inc., 59 Temple Place,
 *  Suite 330, Boston, MA 02111-1307 USA
 */


#include "kat.h"
#include "kat_config.h"
#include "kat_io.h"
#include "kat_inline.c"
#include "kat_aux.h"
#include "kat_init.h"
#include "kat_aa.h"
#include <stdarg.h>
#include "md5.h"
#include "base64.h"
#include <sys/types.h> 
#include <sys/stat.h>
#include <unistd.h>

//extern init_variables_t in;
extern interferometer_t inter;
extern global_var_t vglobal;
extern char mydate[];
extern memory_t mem;
extern options_t options;
extern FILE *fp_log;

typedef struct n1n2m1m2 {
    int n1, n2, m1, m2;
} n1n2m1m2_t;

n1n2m1m2_t *nms;

//! Checks if a file exists, returns 1 if so and 0 otherwise

/*!
 * \param filename, string giving the filename
 */
int check_if_file_exists(char *filename) {
    struct stat file_stat;

    if (stat(filename, &file_stat) == 0)
        return (1);
    else
        return (0);
}

//! Renames a file to filename.bak if the file exists

/*!
 * \param filename, string giving the filename
 */
void make_backup_of_file(char *filename) {

    char newname[LINE_LEN];

    if (check_if_file_exists(filename)) {
        strcpy(newname, filename);
        strcat(newname, ".bak");
        rename(filename, newname);
    }
}

//! opens file for writing (making a backup first) and returns handle

/*!
 * \param filename, string giving the filename
 */
void open_file_to_write_ascii(char *filename, FILE **handle) {
    make_backup_of_file(filename);
    *handle = fopen(filename, "w");
    if (*handle == NULL) {
        gerror("Could not open file '%s'\n", filename);
    }
}

void open_file_to_read_ascii(char *filename, FILE **handle) {
    *handle = fopen(filename, "r");
    if (*handle == NULL) {
        gerror("Could not open file '%s'\n", filename);
    }
}

void open_file_to_write_binary(char *filename, FILE **handle) {
    make_backup_of_file(filename);
    *handle = fopen(filename, "wb");
    if (*handle == NULL) {
        gerror("Could not open file '%s'\n", filename);
    }
}

void open_file_to_read_binary(char *filename, FILE **handle) {
    *handle = fopen(filename, "rb");
    if (*handle == NULL) {
        gerror("Could not open file '%s'\n", filename);
    }
}

void _read_merged_map_ascii_data(mirror_t *mirror, int num) {
    assert(mirror != NULL);
    assert((num >= 0) && (num <= 3));

    double **data = NULL; // remove compiler warning
    FILE *file;
    char buf[MAX_MAP_LINE];
    char filenamemap[LINE_LEN];
    int row_count = 0, col_count = 0;
    double val;
    surface_merged_map_t *map = &(mirror->map_merged);

    switch (num) {
        case 0:
            data = map->r_abs;
            sprintf(buf, "%s_r_abs.map", map->filename);
            break;
        case 1:
            data = map->r_phs;
            sprintf(buf, "%s_r_phs.map", map->filename);
            break;
        case 2:
            data = map->t_abs;
            sprintf(buf, "%s_t_abs.map", map->filename);
            break;
        case 3:
            data = map->t_phs;
            sprintf(buf, "%s_t_phs.map", map->filename);
            break;
        default:
            bug_error("%i was passed when the value should only be between 0-3\n", num);
    }

    if (inter.debug && !options.quiet)
        message(" * Reading map file %s...\n", buf);

    strcpy(filenamemap, buf);

    if (!check_if_file_exists(filenamemap))
        gerror("The map file %s is missing. Please delete all other .map and .kmn files\n"
            "    for mirror %s and regenerate them or replace the missing file\n", filenamemap, mirror->name);

    open_file_to_read_ascii(filenamemap, &file);

    while (fgets(buf, MAX_MAP_LINE - 1, file) != NULL) {
        // ensure there is an end of line value, this should make sure that
        // if there happens to be more numbers than can fit in the buffer this
        // should allow the scanf to pick up the end and then the difference
        // in column count should be found
        buf[MAX_MAP_LINE - 1] = '\n';

        if (row_count > map->rows)
            gerror("There are more rows than expected in the map file %s.\n"
                "    Please fix or delete the files and regenerate the map data.\n", filenamemap);

        col_count = 0;

        while (true) {
            // the -1 is here because below we check when only 1 matched was 
            // found using sscanf and then we break, so we would only get here
            // if there were too many columns
            if (col_count > map->cols - 1)
                gerror("There are more columns than expected in the map file %s in row %i,\n"
                    "    Please fix or delete the files and regenerate the map data.\n", filenamemap, row_count + 1);

            int n = sscanf(buf, "%lg %[^\n]s", &val, buf);

            if (n == 2) {
                data[col_count][row_count] = val;
                col_count++;
            } else if (n == 1) {
                data[col_count][row_count] = val;
                col_count++;
                break;
            } else if ((buf[0] == '\n') && (row_count == map->rows))
                break;
            else
                gerror("Problem occured when reading map file %s, please ensure it has not been tampered with.\n"
                    "    If not delete files and regenerate the map data.\n", filenamemap);
        }

        if ((buf[0] == '\n') && (row_count == map->rows))
            break;

        if (col_count != map->cols)
            gerror("There was not the expected amount of columns in the map file %s in row %i.\n"
                "    Please fix or delete the files and regenerate the map data\n", filenamemap, row_count);

        row_count++;
    }

    if (row_count != map->rows)
        gerror("There was not the expected amount of rows in the map file %s.\n"
            "    Please fix or delete the files and regenerate the map data\n", filenamemap);

    fclose(file);
}

void _read_merged_map_binary_data(mirror_t *mirror) {

    FILE *file;
    unsigned char* filebuf;
    unsigned long filesize;
    char buf[LINE_LEN];
    char filename[FILENAME_MAX];
    MD5_CTX ctx;
    surface_merged_map_t *map = &mirror->map_merged;

    sprintf(buf, "%s.map.bin", map->filename);
    strcpy(filename, buf);

    if (!check_if_file_exists(filename))
        gerror("The map file %s is missing. Please delete all .map.bin, .kmn.bin and .kmn files\n"
            "    for mirror %s and regenerate them or replace the missing files\n", filename, mirror->name);

    if (inter.debug && !options.quiet)
        message(" * Reading map file %s...\n", buf);

    MD5Init(&ctx);

    //calculate length that the file should be
    filesize = sizeof (ctx.digest);
    filesize += sizeof (map->rows);
    filesize += sizeof (map->cols);
    filesize += sizeof (double) * 4 * map->rows * map->cols;

    open_file_to_read_binary(filename, &file);

    fseek(file, 0L, SEEK_END);
    unsigned long fs = ftell(file);
    fseek(file, 0L, SEEK_SET);

    if (filesize != fs)
        gerror("The merged map data in %s has a different filesize to what was expected\n"
            "    for the mirror %s.  Please delete all .map.bin, .kmn.bin and .kmn files\n"
            "    for mirror %s and regenerate them or replace with the correct files\n", filename, mirror->name);

#if DEBUG
    filebuf = (unsigned char*) calloc(filesize, sizeof (unsigned char));
#else
    filebuf = (unsigned char*) malloc(filesize);
#endif

    fread(filebuf, sizeof (unsigned char), filesize, file);
    fseek(file, 0L, SEEK_SET); // get back to the start of the file

    MD5Update(&ctx, &filebuf[16], filesize - 16);
    MD5Final(&ctx);

    int rows, cols;
    unsigned long pos = 0;

    // compare checksum
    if (memcmp(ctx.digest, filebuf, sizeof (ctx.digest)))
        gerror("The checksum for file %s did not match the stored checksum value.\n"
            "    Please delete all .map.bin, .kmn.bin and .kmn files for mirror %s\n"
            "    and regenerate them or replace with the correct files\n", filename, mirror->name);

    pos += sizeof (ctx.digest);

    rows = *((int*) &filebuf[pos]);
    pos += sizeof (map->rows);

    cols = *((int*) &filebuf[pos]);
    pos += sizeof (map->cols);

    if (rows != map->rows || cols != map->cols)
        gerror("The merged map data in %s has a different number of rows and columns to\n"
            "    what was expected for the mirror %s. Please delete all .map.bin,\n"
            "    .kmn.bin and .kmn files for mirror %s and regenerate them or\n"
            "     replace with the correct files\n", filename, mirror->name);

    size_t s = sizeof (map->r_abs[0][0]);

    int i;
    for (i = 0; i < map->cols; i++) {
        memcpy(map->r_abs[i], &filebuf[pos], s * rows);
        pos += cols*s;
        EOF_CHECK;
        memcpy(map->r_phs[i], &filebuf[pos], s * rows);
        pos += cols*s;
        EOF_CHECK;
        memcpy(map->t_abs[i], &filebuf[pos], s * rows);
        pos += cols*s;
        EOF_CHECK;
        memcpy(map->t_phs[i], &filebuf[pos], s * rows);
        pos += cols*s;
        EOF_CHECK;
    }

    if (pos != filesize)
        bug_error("Whilst reading the binary merged map we didn't read the entire file.\n");

    free(filebuf);
    fclose(file);
}

void _read_kmn_binary_data(mirror_t *mirror, bool converting) {
    FILE *file;
    unsigned char* filebuf;
    unsigned long filesize;
    char buf[LINE_LEN];
    char filename[FILENAME_MAX];
    MD5_CTX ctx;
    surface_merged_map_t *map = &mirror->map_merged;

    sprintf(buf, "%s.kmn.bin", map->filename);
    strcpy(filename, buf);

    if (!check_if_file_exists(filename))
        gerror("The map file %s is missing. Please delete all .map.bin, .kmn.bin and .kmn files\n"
            "    for mirror %s and regenerate them or replace the missing files\n", filename, mirror->name);

    if (inter.debug && !options.quiet)
        message(" * Reading binary kmn file %s...\n", buf);

    MD5Init(&ctx);
    // this is the num_fields of the simulation that wrote the 
    // data file, not the currrent one!
    int num_fields = (int) (map->_maxtem + 1) * (map->_maxtem + 2) / 2;
    int numkmn = num_fields*num_fields;

    //calculate length that the file should be
    filesize = sizeof (ctx.digest);
    filesize += sizeof (int);
    filesize += numkmn * 4 * (sizeof (complex_t) + sizeof (unsigned char));

    open_file_to_read_binary(filename, &file);

    fseek(file, 0L, SEEK_END);
    unsigned long fs = ftell(file);
    fseek(file, 0L, SEEK_SET);

    if (filesize != fs)
        gerror("The merged map data in %s has a different filesize to what was expected for\n"
            "    the mirror %s.  Please delete all .map.bin, .kmn.bin and .kmn files for\n"
            "    mirror %s and regenerate them or replace with the correct files\n", filename, mirror->name);

#if DEBUG
    filebuf = (unsigned char*) calloc(filesize, sizeof (unsigned char));
#else
    filebuf = (unsigned char*) malloc(filesize);
#endif

    if (filebuf == NULL)
        gerror("Out of memory: Could not allocate enough memory to create binary file buffer.\n");

    // load the whole file into memory
    fread(filebuf, sizeof (unsigned char), filesize, file);
    fseek(file, 0L, SEEK_SET); // get back to the start of the file

    MD5Update(&ctx, &filebuf[16], filesize - 16);
    MD5Final(&ctx);

    unsigned long pos = 0;

    // compare checksum
    if (memcmp(ctx.digest, filebuf, sizeof (ctx.digest)))
        gerror("The checksum for file %s did not match the stored checksum value.\n"
            "    Please delete all .map.bin, .kmn.bin and .kmn files for mirror %s\n"
            "    and regenerate them or replace with the correct files\n", filename, mirror->name);

    pos = sizeof (ctx.digest);

    int numkmn_read = *((int*) &filebuf[pos]);
    pos += sizeof (numkmn_read);

    // if the number of kmn we calculated there would be earlier is not the
    // same as what has just been read there is a problem
    if (numkmn_read != numkmn)
        gerror("The number of kmn's in the file %s was not what was expected.\n"
            "    Please delete all .map.bin, .kmn.bin and .kmn files for mirror %s\n"
            "    and regenerate them or replace with the correct files\n", filename, mirror->name);

#ifdef DEBUG
    if (!IS_MIRROR_KMN_ALLOCD(&(mirror->map_merged.kmn)))
        bug_error(" ! Mirror kmn matric was not allocated\n");
#endif

    int n, m, n1, n2, m1, m2;
    unsigned char _n1, _n2, _m1, _m2;
    size_t cs = sizeof (complex_t);

    for (n = 0; n < num_fields; n++) {
        for (m = 0; m < num_fields; m++) {
            _n1 = *((unsigned char*) &filebuf[pos]);
            pos += 1;
            EOF_CHECK;
            _n2 = *((unsigned char*) &filebuf[pos]);
            pos += 1;
            EOF_CHECK;
            _m1 = *((unsigned char*) &filebuf[pos]);
            pos += 1;
            EOF_CHECK;
            _m2 = *((unsigned char*) &filebuf[pos]);
            pos += 1;
            EOF_CHECK;

            n1 = (int) _n1;
            n2 = (int) _n2;
            m1 = (int) _m1;
            m2 = (int) _m2;

            if (converting) {
                nms[n + num_fields * m].n1 = n1;
                nms[n + num_fields * m].n2 = n2;
                nms[n + num_fields * m].m1 = m1;
                nms[n + num_fields * m].m2 = m2;
            }

            if (((n1 + m1) <= inter.tem) && ((n2 + m2) <= inter.tem)) {
                int f1 = get_field_index_from_tem(n1, m1);
                int f2 = get_field_index_from_tem(n2, m2);

                memcpy(&(map->kmn.k11[f1][f2]), &filebuf[pos], cs);
                pos += cs;
                EOF_CHECK;
                memcpy(&(map->kmn.k12[f1][f2]), &filebuf[pos], cs);
                pos += cs;
                EOF_CHECK;
                memcpy(&(map->kmn.k21[f1][f2]), &filebuf[pos], cs);
                pos += cs;
                EOF_CHECK;
                memcpy(&(map->kmn.k22[f1][f2]), &filebuf[pos], cs);
                pos += cs;
                EOF_CHECK;
            } else {
                pos += cs * 4;
                EOF_CHECK;
            }
        }
    }


    fclose(file);
    free(filebuf);
}

bool read_mirror_kmn_files(mirror_t *mirror, bool converting) {
#define READ(A) \
linenum++;\
fgets(buf, LINE_LEN-1,file); \
if(feof(file)) gerror("Unexpectedly reached the end of the file. Please regenerate the file\n as it could be corrupt or for the wrong version of Finesse\n"); \
((A == 0) ? gerror("Line %i ('%s') could not be read in map file %s\n" , linenum, buf, knmfilename) : NULL)

    FILE *file;

    assert(mirror != NULL);

    surface_merged_map_t *map = &(mirror->map_merged);

    assert(map != NULL);

    char buf[LINE_LEN];
    int linenum = 0;
    char knmfilename[LINE_LEN];

    sprintf(knmfilename, "%s.knm", mirror->map_merged.filename);

    if (!check_if_file_exists(knmfilename)) {
        return false;
    } else {
        open_file_to_read_ascii(knmfilename, &file);

        // read three lines of comments
        READ(sscanf(buf, "%% %s", buf));
        READ(sscanf(buf, "%% %s", buf));
        READ(sscanf(buf, "%% %s", buf));

        int n, mapnum = 0;

        READ(sscanf(buf, "mapnum : %i", &mapnum));
        // if we haven't got the same number of maps recompute

        if (mapnum != mirror->num_maps && !converting) {
            if (inter.debug && !options.quiet)
                message(" * KNM file read: Different number of maps on mirror %s\n", mirror->name);

            return false;
        }

        int mn;
        char mapfile[LINE_LEN];
        char B64maphash[LINE_LEN];
        unsigned char maphash[LINE_LEN];

        // if we are converting the knn file the various map files
        // won't be loaded so no need to check them
        if (!converting) {
            for (n = 0; n < mapnum; n++) {
                READ(sscanf(buf, "map%i : %s %s", &mn, mapfile, B64maphash));
                // if the map is not the same file name then the order could have 
                // changed or different maps have been used.
                if (strcmp(mirror->map[n]->filename, mapfile) != 0) {
                    if (inter.debug && !options.quiet)
                        message(" * KNM file read: Map order on mirror %s is different to knm file\n", mirror->name);

                    return false;
                }

                // make sure it isnt too long
                if (strlen(B64maphash) > 30) {
                    if (inter.debug && !options.quiet)
                        message(" * KNM file read: The Base64 hash key is not correct for file %s\n", mapfile);

                    return false;
                }

                //First need to decode the base64 hash
                if (!decode_base64(maphash, B64maphash)) {
                    if (inter.debug && !options.quiet)
                        message(" * KNM file read: Could not decode the Base64 hash key correctly for file %s\n", mapfile);

                    return false;
                }

                // compare the hash that has been read when reading the maps
                // and that in the file
                if (memcmp((char*) mirror->map[n]->hash, (char*) maphash, 16) != 0) {
                    if (inter.debug && !options.quiet)
                        message(" * KNM file read: Map hash code was not the same for file %s\n", mapfile);

                    return false;
                }
            }
        } else {
            // however if we are converting then we do need to store the map
            // names to write again later, so we fake the maps here
            long currpos;

            while (true) {
                currpos = ftell(file);
                fgets(buf, LINE_LEN - 1, file);

                if (feof(file))
                    gerror("Unexpectedly reached the end of the file.\n Please regenerate the file as it could be corrupt or for the wrong version of Finesse\n");

                // check if the start of the line is a map definition
                if (strcspn(buf, "map") == 0 && strlen(buf) != 0) {
                    sscanf(buf, "map%i : %s", &mn, mapfile);

                    mirror->map[mirror->num_maps] = (surface_map_t *) calloc(1, sizeof (surface_map_t));

                    if (mirror->map[mirror->num_maps] == NULL) {
                        bug_error("memory allocation for surface map failed");
                    }

                    strcpy(mirror->map[mirror->num_maps]->filename, mapfile);

                    mirror->num_maps++;
                } else
                    break;
            }

            //reset to the start of the line so the next read works
            fseek(file, currpos, SEEK_SET);
        }

        READ(sscanf(buf, "integration method : %i", &map->_intmeth));
        READ(sscanf(buf, "interpolation method : %i", &map->_intermeth));
        READ(sscanf(buf, "interpolation size : %i", &map->_intersize));

        // if the interpolation method, size or the integration method used
        // has changed then return false and recompute the map data
        if (map->_intmeth != map->integration_method && !converting) {
            if (inter.debug && !options.quiet)
                message(" * KNM file read: Different integration method used on mirror %s\n", mirror->name);
            return false;
        }
        if (map->_intermeth != map->interpolation_method && !converting) {
            if (inter.debug && !options.quiet)
                message(" * KNM file read: Different interpolation method used on mirror %s\n", mirror->name);
            return false;
        }
        if (map->_intersize != map->interpolation_size && !converting) {
            if (inter.debug && !options.quiet)
                message(" * KNM file read: Different interpolation size used on mirror %s\n", mirror->name);
            return false;
        }

        READ(sscanf(buf, "qx1 : %lg %lg", &(map->_qx1.re), &(map->_qx1.im)));
        READ(sscanf(buf, "qy1 : %lg %lg", &(map->_qy1.re), &(map->_qy1.im)));
        READ(sscanf(buf, "qx2 : %lg %lg", &(map->_qx2.re), &(map->_qx2.im)));
        READ(sscanf(buf, "qy2 : %lg %lg", &(map->_qy2.re), &(map->_qy2.im)));

        READ(sscanf(buf, "nr1 : %lg", &(map->_nr1)));
        READ(sscanf(buf, "nr2 : %lg", &(map->_nr2)));
        READ(sscanf(buf, "lambda : %lg", &(map->_lambda)));
        READ(sscanf(buf, "angle : %lg", &(map->_angle)));
        READ(sscanf(buf, "maxtem : %i", &(map->_maxtem)));
        READ(sscanf(buf, "r_aperture : %lg", &(map->_r_aperture)));
        READ(sscanf(buf, "xbeta : %lg", &(map->_xbeta)));
        READ(sscanf(buf, "ybeta : %lg", &(map->_ybeta)));

        READ(sscanf(buf, "rows : %i", &(map->rows)));
        READ(sscanf(buf, "cols : %i", &(map->cols)));
        READ(sscanf(buf, "xstep : %lg", &(map->xstep)));
        READ(sscanf(buf, "ystep : %lg", &(map->ystep)));
        READ(sscanf(buf, "x0 : %lg", &(map->x0)));
        READ(sscanf(buf, "y0 : %lg", &(map->y0)));

        READ(sscanf(buf, "abserr : %lg", &(map->_abserr)));
        READ(sscanf(buf, "relerr : %lg", &(map->_relerr)));

        if (init.abserr != map->_abserr && !converting) {
            if (inter.debug && !options.quiet)
                message(" * KNM file read: Different absolute error compared to file %s\n", knmfilename);
            return false;
        }

        if (init.relerr != map->_relerr && !converting) {
            if (inter.debug && !options.quiet)
                message(" * KNM file read: Different relative error compared to file %s\n", knmfilename);
            return false;
        }

        int save_binary;
        READ(sscanf(buf, "kmn_binary : %i", &save_binary));

        if (!save_binary && map->kmn_save_binary && !converting)
            gerror("Mirror %s is set to use binary knm and map data files, but the\n"
                "    saved values were generated in ASCII. Please delete the files\n"
                "    and generate them again in binary or switch the mirror's save\n"
                "    mode to ASCII.\n", mirror->name);

        if (save_binary && !map->kmn_save_binary && !converting)
            gerror("Mirror %s is set not to use binary knm and map data files, but the\n"
                "    saved values were generated in binary. Please delete the files and\n"
                "    generate them again in ASCII or switch the mirror's save mode to binary.\n", mirror->name);

        // if we are converting then we need to store the save method
        // so we can read the map files later
        if (converting)
            map->kmn_save_binary = save_binary;

        // Now read the KMN's
        READ(sscanf(buf, "%% -KMN%s", buf));

        int m, n1, m1, n2, m2;
        int num_kmns;

        READ(sscanf(buf, "num_kmns : %i", &num_kmns));

        complex_t t11, t12, t21, t22;

        // allocate some memory to store the n1..m2 for each element in
        // we do this becaues the get_field_index_from_tem function
        // wont work as nothing has been initialised and the map->kmn allocation
        // below
        int num_fields = (int) ((map->_maxtem + 1) * (map->_maxtem + 2) / 2);

        if (converting) {
            nms = (n1n2m1m2_t*) malloc(num_fields * num_fields * sizeof (n1n2m1m2_t));

            // we also have to set inter.tem to use get_field_index_from_tem
            inter.tem = map->_maxtem;
        }

        // allocate memory to store the kmn values in. This allocates
        // enough memory to deal with the current simulation setting
        // of maxtem
        if (!IS_MIRROR_KMN_ALLOCD(&(map->kmn)))
            mirror_kmn_alloc(&(map->kmn));

        if (map->kmn_save_binary)
            _read_kmn_binary_data(mirror, converting);
        else {
            if (inter.debug && !options.quiet)
                message(" * Reading kmn data for mirror %s from %s\n", mirror->name, knmfilename);

            for (n = 0; n < num_fields; n++) {
                for (m = 0; m < num_fields; m++) {
                    READ(sscanf(buf, "%d %d %d %d   %lg %lg %lg %lg %lg %lg %lg %lg"
                            , &n1, &m1, &n2, &m2
                            , &(t11.re), &(t11.im)
                            , &(t22.re), &(t22.im)
                            , &(t12.re), &(t12.im)
                            , &(t21.re), &(t21.im)));

                    if (converting) {
                        nms[n + num_fields * m].n1 = n1;
                        nms[n + num_fields * m].n2 = n2;
                        nms[n + num_fields * m].m1 = m1;
                        nms[n + num_fields * m].m2 = m2;
                    }

                    if (((n1 + m1) <= inter.tem) && ((n2 + m2) <= inter.tem)) {
                        int field1 = get_field_index_from_tem(n1, m1);
                        int field2 = get_field_index_from_tem(n2, m2);

                        map->kmn.k11[field1][field2] = t11;
                        map->kmn.k12[field1][field2] = t12;
                        map->kmn.k21[field1][field2] = t21;
                        map->kmn.k22[field1][field2] = t22;
                    }
                }
            }
        }

        // now we have loaded in the calculated kmns flag it
        map->kmn_calculated = true;

        if (!map->noMapPresent) {
            // allocate memory to store the amplitude and phase data in
            alloc_merged_map_data(map);

            if (map->kmn_save_binary) {
                _read_merged_map_binary_data(mirror);
            } else {
                // Code will fail from here on if it cannot find the correct map files
                _read_merged_map_ascii_data(mirror, 0); // reads r_abs
                _read_merged_map_ascii_data(mirror, 1); // reads r_phs
                _read_merged_map_ascii_data(mirror, 2); // reads t_abs
                _read_merged_map_ascii_data(mirror, 3); // reads t_phs
            }
        }

        // Finally we have successfully loaded the data and various parameters
        // now return true stating that the map merging does not need doing.
        return true;
    }
}

void _write_kmn_binary_data(mirror_t *mirror, bool converting) {
    FILE *file;
    unsigned char* filebuf;
    unsigned long filesize;
    char buf[LINE_LEN];
    char filename[FILENAME_MAX];
    MD5_CTX ctx;
    surface_merged_map_t *map = &mirror->map_merged;

    sprintf(buf, "%s.kmn.bin", map->filename);
    strcpy(filename, buf);

    MD5Init(&ctx);

    int num_fields;

    // if converting we won't have num_fields calculated proprely
    if (converting)
        num_fields = (int) (map->_maxtem + 1) * (map->_maxtem + 2) / 2;
    else
        num_fields = mem.num_fields;

    int numkmn = num_fields*num_fields;

    //calculate length that the file should be
    filesize = sizeof (ctx.digest);
    filesize += sizeof (int);
    // save space for each complex kmn aswell as 4 chars for the n1,n2,m1,m2
    filesize += numkmn * 4 * (sizeof (complex_t) + sizeof (unsigned char));

    filebuf = (unsigned char*) malloc(filesize);

    if (filebuf == NULL)
        bug_error("Out of memory: Could not allocate enough memory to create binary file buffer.");

    open_file_to_write_binary(filename, &file);

    // finally write everything to the file
    if (inter.debug && !options.quiet)
        message(" * Writing kmn binary data to %s...\n", filename);

    //seek to the byte after the md5 checksum as it will be written later
    int pos = sizeof (ctx.digest);

    memcpy(&(filebuf[pos]), &numkmn, sizeof (numkmn));
    pos += sizeof (numkmn);

    int n, m, n1, n2, m1, m2;
    size_t cs = sizeof (complex_t);
    unsigned char _n1, _n2, _m1, _m2;

    for (n = 0; n < num_fields; n++) {
        for (m = 0; m < num_fields; m++) {
            if (converting) {
                n1 = nms[n * num_fields * m].n1;
                n2 = nms[n * num_fields * m].n2;
                m1 = nms[n * num_fields * m].m1;
                m2 = nms[n * num_fields * m].m2;
            } else {
                get_tem_modes_from_field_index(&n1, &m1, n);
                get_tem_modes_from_field_index(&n2, &m2, m);
            }

            _n1 = (unsigned char) n1;
            _m1 = (unsigned char) m1;
            _n2 = (unsigned char) n2;
            _m2 = (unsigned char) m2;

            memcpy(&(filebuf[pos]), &_n1, 1);
            pos += 1;
            memcpy(&(filebuf[pos]), &_n2, 1);
            pos += 1;
            memcpy(&(filebuf[pos]), &_m1, 1);
            pos += 1;
            memcpy(&(filebuf[pos]), &_m2, 1);
            pos += 1;
            memcpy(&(filebuf[pos]), &map->kmn.k11[n][m], cs);
            pos += cs;
            memcpy(&(filebuf[pos]), &map->kmn.k12[n][m], cs);
            pos += cs;
            memcpy(&(filebuf[pos]), &map->kmn.k21[n][m], cs);
            pos += cs;
            memcpy(&(filebuf[pos]), &map->kmn.k22[n][m], cs);
            pos += cs;
        }
    }

    MD5Update(&ctx, &filebuf[sizeof (ctx.digest)], filesize - sizeof (ctx.digest));
    MD5Final(&ctx);
    memcpy(filebuf, ctx.digest, sizeof (ctx.digest));

    fwrite(filebuf, 1, filesize, file);
    fflush(file);
    fclose(file);

    free(filebuf);
}

void _write_merged_map_binary_data(mirror_t *mirror) {

    FILE *file;
    unsigned char* filebuf;
    unsigned long filesize;
    char buf[LINE_LEN];
    char filename[FILENAME_MAX];
    MD5_CTX ctx;
    surface_merged_map_t *map = &mirror->map_merged;

    sprintf(buf, "%s.map.bin", map->filename);
    strcpy(filename, buf);

    MD5Init(&ctx);

    //calculate length that the file should be
    filesize = sizeof (ctx.digest);
    filesize += sizeof (map->rows);
    filesize += sizeof (map->cols);
    filesize += sizeof (double) * 4 * map->rows * map->cols;

#if DEBUG
    filebuf = (unsigned char*) calloc(filesize, sizeof (unsigned char));
#else
    filebuf = (unsigned char*) malloc(filesize);
#endif

    if (filebuf == NULL)
        bug_error("Out of memory: Could not allocate enough memory to create binary file buffer.");

    unsigned long pos = 0;
    pos += sizeof (ctx.digest);

    memcpy(&(filebuf[pos]), &map->rows, sizeof (map->rows));
    pos += sizeof (map->rows);

    memcpy(&(filebuf[pos]), &map->cols, sizeof (map->cols));
    pos += sizeof (map->cols);

    int i;
    size_t s = sizeof (map->r_abs[0][0]);

    if (s != sizeof (map->r_phs[0][0]) || s != sizeof (map->t_abs[0][0]) || s != sizeof (map->t_phs[0][0]))
        bug_error("The byte size of the various merged map elements is not the same!");

    unsigned int bytes = map->rows * s;

    for (i = 0; i < map->cols; i++) {
        memcpy(&(filebuf[pos]), map->r_abs[i], bytes);
        pos += bytes;
        memcpy(&(filebuf[pos]), map->r_phs[i], bytes);
        pos += bytes;
        memcpy(&(filebuf[pos]), map->t_abs[i], bytes);
        pos += bytes;
        memcpy(&(filebuf[pos]), map->t_phs[i], bytes);
        pos += bytes;

        if (pos > filesize)
            bug_error("Tried to write more data than was allocated whilst writing binary map data\n");
    }

    //create a md5 checksum for all the file apart from the checksum itself
    MD5Update(&ctx, &filebuf[16], filesize - 16);
    MD5Final(&ctx);

    // copy the md5 key to file buffer
    memcpy(filebuf, ctx.digest, sizeof (ctx.digest));

    open_file_to_write_binary(filename, &file);
    // finally write everything to the file
    message(" * Writing to merged map binary data to %s...\n", filename);
    fwrite(filebuf, sizeof (filebuf[0]), filesize, file);
    fflush(file);
    fclose(file);

    free(filebuf);
}

void _write_merged_map_ascii_data(mirror_t *mirror, KNM_MIRROR_NODE_DIRECTION_t num) {
    double **data = NULL;
    char buf[LINE_LEN];
    char filenamemap[FILENAME_MAX];
    FILE *file;
    surface_merged_map_t *map = &(mirror->map_merged);

    if (inter.debug && !options.quiet)
        message(" * Writing merged map data for mirror '%s' to file '%s'\n", mirror->name, filenamemap);

    switch (num) {
        case K11:
            data = map->r_abs;
            sprintf(buf, "%s_r_abs.map", map->filename);
            break;
        case K12:
            data = map->r_phs;
            sprintf(buf, "%s_r_phs.map", map->filename);
            break;
        case K21:
            data = map->t_abs;
            sprintf(buf, "%s_t_abs.map", map->filename);
            break;
        case K22:
            data = map->t_phs;
            sprintf(buf, "%s_t_phs.map", map->filename);
            break;
        default:
            bug_error("wrong map data format");
    }

    strcpy(filenamemap, buf);
    open_file_to_write_ascii(filenamemap, &file);

    int i, j;

    for (i = 0; i < map->cols; i++) {
        for (j = 0; j < map->cols; j++) {
            fprintf(file, "%.15g ", data[i][j]);
        }
        fprintf(file, "\n");
    }

    fprintf(file, "\n");

    fclose(file);
}

void write_mirror_kmn_file(mirror_t *mirror, bool converting) {
    assert(mirror != NULL);

    FILE *file;

    assert(mirror != NULL);

    surface_merged_map_t *map = &(mirror->map_merged);

    assert(map != NULL);

    char filename[FILENAME_MAX];

    sprintf(filename, "%s.knm", map->filename);

    if (inter.debug && !options.quiet)
        message(" * Saving kmn information for mirror '%s' in file '%s'\n", mirror->name, filename);

    open_file_to_write_ascii(filename, &file);

    // Start writing header

    time_t now = time(NULL);
    fprintf(file, "%% Written by Finesse(%s)  -  %s", SVN_REVISION, asctime(localtime(&now)));
    fprintf(file, "%% Input file '%s'\n", vglobal.input_fname);
    fprintf(file, "%% --DO NOT CHANGE THIS FILE---------------------------------\n");
    fprintf(file, "mapnum : %i\n", mirror->num_maps);

    int i;
    for (i = 0; i < mirror->num_maps; i++) {
        char *B64_hash;
        B64_hash = encode_base64(sizeof (mirror->map[i]->hash), mirror->map[i]->hash);
        fprintf(file, "map%i : %s %s\n", i, mirror->map[i]->filename, B64_hash);
        free(B64_hash);
    }

    fprintf(file, "integration method : %i\n", (converting) ? map->_intmeth : map->integration_method);
    fprintf(file, "interpolation method : %i\n", (converting) ? map->_intermeth : map->interpolation_method);
    fprintf(file, "interpolation size : %i\n", (converting) ? map->_intersize : map->interpolation_size);
    fprintf(file, "qx1 : %.15g %.15g\n", map->_qx1.re, map->_qx1.im);
    fprintf(file, "qy1 : %.15g %.15g\n", map->_qy1.re, map->_qy1.im);
    fprintf(file, "qx2 : %.15g %.15g\n", map->_qx2.re, map->_qx2.im);
    fprintf(file, "qy2 : %.15g %.15g\n", map->_qy2.re, map->_qy2.im);
    fprintf(file, "nr1 : %.15g\n", map->_nr1);
    fprintf(file, "nr2 : %.15g\n", map->_nr2);
    fprintf(file, "lambda : %.15g\n", map->_lambda);
    fprintf(file, "angle : %.15g\n", map->_angle);
    fprintf(file, "maxtem : %i\n", map->_maxtem);
    fprintf(file, "r_aperture : %.15g\n", map->_r_aperture);
    fprintf(file, "xbeta : %.15g\n", map->_xbeta);
    fprintf(file, "ybeta : %.15g\n", map->_ybeta);
    fprintf(file, "rows : %i\n", map->rows);
    fprintf(file, "cols : %i\n", map->cols);
    fprintf(file, "xstep : %.15g\n", map->xstep);
    fprintf(file, "ystep : %.15g\n", map->ystep);
    fprintf(file, "x0 : %.15g\n", map->x0);
    fprintf(file, "y0 : %.15g\n", map->y0);
    fprintf(file, "abserr : %.15g\n", (converting) ? map->_abserr : init.abserr);
    fprintf(file, "relerr : %.15g\n", (converting) ? map->_relerr : init.relerr);
    fprintf(file, "kmn_binary : %i\n", map->kmn_save_binary);

    // End writing header

    // Start writing kmn

    int num_fields;

    // if converting we won't have num_fields calculated proprely
    if (converting)
        num_fields = (int) (map->_maxtem + 1) * (map->_maxtem + 2) / 2;
    else
        num_fields = mem.num_fields;

    fprintf(file, "%% -KMN\n");
    fprintf(file, "num_kmns : %i\n", num_fields * num_fields);

    if (map->kmn_save_binary)
        _write_kmn_binary_data(mirror, converting);
    else {
        int n, m, n1, m1, n2, m2;
        mirror_kmn_t *kmn = &(map->kmn);

        for (n = 0; n < num_fields; n++) {
            for (m = 0; m < num_fields; m++) {
                if (converting) {
                    n1 = nms[n + num_fields * m].n1;
                    n2 = nms[n + num_fields * m].n2;
                    m1 = nms[n + num_fields * m].m1;
                    m2 = nms[n + num_fields * m].m2;
                } else {
                    get_tem_modes_from_field_index(&n1, &m1, n);
                    get_tem_modes_from_field_index(&n2, &m2, m);
                }

                fprintf(file, "%d %d %d %d   ", n1, m1, n2, m2);
                fprintf(file, "%.15g %.15g ", kmn->k11[n][m].re, kmn->k11[n][m].im);
                fprintf(file, "%.15g %.15g ", kmn->k22[n][m].re, kmn->k22[n][m].im);
                fprintf(file, "%.15g %.15g ", kmn->k12[n][m].re, kmn->k12[n][m].im);
                fprintf(file, "%.15g %.15g\n", kmn->k21[n][m].re, kmn->k21[n][m].im);
            }
        }

        fprintf(file, "\n");
        fclose(file);
    }

    // if no maps are present no need to write them to a file
    if (!map->noMapPresent) {
        if (map->kmn_save_binary) {
            _write_merged_map_binary_data(mirror);
        } else {
            // now we need to write the actual map data
            _write_merged_map_ascii_data(mirror, K11);
            _write_merged_map_ascii_data(mirror, K12);
            _write_merged_map_ascii_data(mirror, K21);
            _write_merged_map_ascii_data(mirror, K22);
        }
    }
}

/**
 * Writes the coupling coefficient matrix to a file in a matrix form. To be used
 * for testing commutation issues.
 * 
 * @param filename Name of output file
 * @param mirror mirror whose kmn to output
 */
void write_kmn_to_matrix_file(char *filename, mirror_kmn_t *kmn) {
    FILE *file;
    int n, m, i;
    complex_t **data = NULL; // removing compiler warning
    char buf[LINE_LEN];

    //if the kmn hasn't been allocated, then it hasn't been used, so no point
    //in writing it to a file
    if (!IS_MIRROR_KMN_ALLOCD(kmn))
        return;

    if (kmn != NULL) {

        for (i = K11; i <= K22; i++) {
            switch (i) {
                case K11: data = kmn->k11;
                    sprintf(buf, "%s_K11.mat", filename);
                    break;
                case K12: data = kmn->k12;
                    sprintf(buf, "%s_K12.mat", filename);
                    break;
                case K21: data = kmn->k21;
                    sprintf(buf, "%s_K21.mat", filename);
                    break;
                case K22: data = kmn->k22;
                    sprintf(buf, "%s_K22.mat", filename);
                    break;
                default: bug_error("Unexpected KNM");
            }

            if (inter.debug && !options.quiet) message(" * Outputing knm matrix to %s\n", buf);

            open_file_to_write_ascii(buf, &file);

            for (n = 0; n < inter.num_fields; n++) {
                for (m = 0; m < inter.num_fields; m++) {
                    complex_t val = data[n][m];

                    fprintf(file, "%.15e%s%.15ei%s", val.re
                            , (val.im > 0) ? "+" : "-"
                            , (val.im > 0) ? val.im : -val.im
                            , (m == inter.num_fields - 1) ? "" : " ");
                }

                fprintf(file, "\n");
            }

            fclose(file);
        }
    } else {
        warn("Could not write kmn to output file as kmn variables were not set\n");
    }
}


//! Print the progress of the simulation to stderr

/*!
 * \param num_points the total number of points in the simulation
 * \param current_point the current point
 */
void print_progress(int num_points, int current_point) {
    int progress_percentage = ((100 * (current_point - 1)) / num_points + 1);
    fprintf(stderr, PROG_STRING3, progress_percentage);
    fflush(stderr);
}


//! Print the progress of the simulation andd estimated remianing time to stderr

/*!
 * \param num_points the total number of points in the simulation
 * \param current_point the current point
 */
void print_progress_and_time(int num_points, int current_point, time_t starttime) {
    int progress_percentage = ((100 * (current_point - 1)) / num_points + 1);
    double estimated_time = ((num_points - current_point) / (double) current_point) * difftime(time(NULL), starttime);
    char message_str[ERR_MSG_LEN];
    sprint_time(message_str, estimated_time);
    fprintf(stderr, PROG_STRING5, progress_percentage, message_str);
    //fprintf(stderr, PROG_STRING3, progress_percentage);
    fflush(stderr);
}




//! Return a character representation of the node at the given index 

/*!
 * \param node_index the node index
 */
const char *node_print(int node_index) {
    static char gndstring[] = "dump";

    // sanity check input
    assert(node_index >= 0);
    assert(node_index < inter.num_nodes);

    if (node_index < inter.num_nodes) {
        return duplicate_string(inter.node_list[node_index].name);
    }
    gerror("node_print: node undefined\n");

    return gndstring; /* only to silence compiler warning */
}

//! Error reporting routine with bug report information

/*!
 * \param error_msg the error string to print
 * \param function_name the name of the function where error occurred
 * \param source_fname the name of the file in which error occurred
 * \param line_number the line number of the file where error occurred
 */
void _bug_error(const char *function_name, const char *source_fname,
        int line_number, const char *error_msg, ...) {
    va_list argptr; // the pointer to the argument
    char format_str[ERR_MSG_LEN]; // to hold the error message format string

    if( strlen(error_msg) > sizeof(format_str) )
        bug_error("Message string is too long for buffer.\n");
    
    // turn on bold text
    bold_on(stderr);

    // make the format string
    sprintf(format_str,
            "\n\nBug detected!\n"
            "In function %s() of file %s at line %d.\n"
            "Error message: %s\n\n"
            "  Please report the version of Finesse you are using and the input\n"
            "  file that caused the bug at www.gwoptics.org/finesse/bug/. Thank you.\n",
            function_name, source_fname, line_number, error_msg);

    // variable argument stuff
    // point to the first unnamed arge (after the last named arg)
    va_start(argptr, error_msg);
    // print the error message
    vfprintf(stderr, format_str, argptr);
    va_end(argptr); // clean up

    bold_off(stderr);

    // output to the log file as well
    if (fp_log != NULL) {
        va_start(argptr, error_msg);
        vfprintf(fp_log, format_str, argptr);
        va_end(argptr); // clean up
    }

    my_finish(1);
}


//! Generic error reporting routine with variable argument list

/*!
 * \param error_msg mandatory error string, works similarly to printf
 *
 * \todo untested
 */
void gerror(const char *error_msg, ...) {
    va_list argptr; // the pointer to the argument

    // turn on bold, and report the error
    bold_on(stderr);
    fprintf(stderr, "\n*** Error: \n");

    // the variable argument stuff
    // point to the first unnamed arg (after the last named arg)
    va_start(argptr, error_msg);
    vfprintf(stderr, error_msg, argptr);
    va_end(argptr); // clean up

    // turn off bold
    bold_off(stderr);

    // outpuut to the log file as well
    if (fp_log != NULL) {
        va_start(argptr, error_msg);
        vfprintf(fp_log, error_msg, argptr);
        va_end(argptr); // clean up
    }

    my_finish(1);
}

//! Generic error reporting routine with variable argument list

/*! 
 * As gerror but without stopping the program.
 *
 */
void server_gerror(const char *error_msg, ...) {
    va_list argptr; // the pointer to the argument

    // turn on bold, and report the error
    bold_on(stderr);
    fprintf(stderr, "\n*** Error: \n");

    // the variable argument stuff
    // point to the first unnamed arg (after the last named arg)
    va_start(argptr, error_msg);
    vfprintf(stderr, error_msg, argptr);
    va_end(argptr); // clean up

    // turn off bold
    bold_off(stderr);

    // outpuut to the log file as well
    if (fp_log != NULL) {
        va_start(argptr, error_msg);
        vfprintf(fp_log, error_msg, argptr);
        va_end(argptr); // clean up
    }
    fflush(stderr);
    fflush(fp_log);

    if (NOT options.servermode) {
        my_finish(1);
    }
}


//! Set bold font weight on

/*!
 * \param fp the file pointer
 */
void bold_on(FILE *fp) {
    // make sure the file pointer is not null
    assert(fp != NULL);

    // print ANSI code for BOLD
    if (OS == UNIX) {
        fputs("\x1B[1m", fp);
    }
}

//! Set bold font weight off

/*!
 * \param fp the file pointer
 */
void bold_off(FILE *fp) {
    // make sure the file pointer is not null
    assert(fp != NULL);

    // print ANSI code for RESET ALL
    if (OS == UNIX) {
        fputs("\x1B[0m", fp);
    }
}

//! Base message reporting routine for the warn() and message() functions.

/*!
 * This routine forms the basis of the warn() and message() functions.  They
 * are just wrappers for this routine, and are defined as macros with
 * variable argument lists.  The message_type parameter can be either 
 * WARNING or MESSAGE.
 *
 * \param message_string the message string to print
 * \param message_type the type of message one wants to print
 */
void _user_message(const char *message_string, int message_type, ...) {
    va_list argptr; // the pointer to the argument
    char format_str[ERR_MSG_LEN] = {0};
    FILE *fp_output;

    if( strlen(message_string) > sizeof(format_str) )
        bug_error("Message string is too long for buffer.\n");
    
    // if we have a warning (i.e. have used warn()) then make stderr the
    // output stream, if we have just a plain message, use stdout as the
    // output stream, otherwise, barf
    fp_output = stdout; // assign a default output stream

    if (message_type == WARNING) {
        fp_output = stderr;
        // prepend '** ' to WARNING strings
        strcpy(format_str, "** ");
    } else if (message_type == MESSAGE) {
        // prepend ' ' to MESSAGE strings
        strcpy(format_str, " ");
    } else {
        gerror("Incorrect message type passed to _user_message()\n");
    }
     
    strcat(format_str, message_string);

    // the variable argument stuff
    if (!options.quiet) {

        // put bold on if is a warning
        if (message_type == WARNING) {
            bold_on(fp_output);
        }

        // the variable argument stuff
        // point to the first unnamed arg (after the last named arg)
        va_start(argptr, message_type);
        vfprintf(fp_output, format_str, argptr);
        va_end(argptr); // clean up

        // switch bold back off if required
        if (message_type == WARNING) {
            bold_off(fp_output);
        }
    }

    // send the messages to the log file as well
    if (fp_log != NULL) {
        va_start(argptr, message_type);
        vfprintf(fp_log, format_str, argptr);
        va_end(argptr); // clean up
    }
}

//! Print debugging information to standard output

/*!
 * \param function_name Function name where debug message is used
 * \param message_string The string to be printed in the message
 *
 * \todo untested
 */
void _debug_msg(const char *function_name, const char *message_string, ...) {
    if (!(inter.debug & 65536))
        return;

    va_list argptr; // the pointer to the argument
    char format_str[BUFF_LEN]; //<! \todo the value used here should be \#defined

    if( strlen(message_string) > sizeof(format_str) )
        bug_error("Message string is too long for buffer.\n");
        
    // prepend the function name to the format string
    sprintf(format_str, "%s(): %s", function_name, message_string);

    // the variable argument stuff
    // point to the first unnamed arg (after the last named arg)
    va_start(argptr, message_string);
    vprintf(format_str, argptr);
    va_end(argptr); // clean up
}

//! Flush stdout and stderr and exit program

/*!
 * \param return_value the return value to return from the program as a whole
 */
void my_finish(int return_value) {
#ifndef TEST
    fflush(stdout);
    fflush(stderr);
    exit(return_value);

    // close the log file nicely
    if (fp_log != NULL) {
        fclose(fp_log);
    }
#else
    message("Returned with return code: %d\n", return_value);
#endif
}


//! Print version number

/*!
 *
 */
void print_version(FILE *file) {
    fprintf(file, "Finesse %s (%s), %s\n", VERSION, SVN_REVISION, mydate);
}



//! Print the second help screen

/*!
 *
 */
void print_help2() {
    printf(
            "------------------------------------------------------------------------\n"
            "  FINESSE %s   - Help Screen (2) -                    %s\n"
            "------------------------------------------------------------------------\n",
            VERSION, mydate);
    printf(
            " ** Some conventions:\n"
            "  names (for components and nodes) must be less than 15 characters long\n"
            "  angles of incidence, phases and tunings are given in [deg]\n"
            "  (a tuning of 360 deg corresponds to a position change of lambda)\n"
            "  misalignment angles are given in [rad]\n"
            " ** Geometrical conventions:\n"
            "  tangential plane: x, z (index n), saggital plane: y, z (index m)\n"
            "  xbeta refers to a rotation in the x, z plane, i.e. around the y-axis\n"
            "  R<0 when the center of the respective sphere is down beam\n"
            "  (the beam direction is defined locally through the node order:\n"
            "  i.e. mirror: node1 -> node2, beam splitter: node1 -> node3)\n"
            "  beam parameter z<0 when waist position is down beam\n"
            " ** trace n: `n' bit coded word, produces the following output:\n"
            "     trace 1:   list of TEM modes used\n"
            "     trace 2:   cavity eigenvalues and cavity parameters like FWHM, \n"
            "                FSR optical length and Finesse\n"
            "     trace 4:   mode mismatch parameters for the initial setup\n"
            "     trace 8:   beam parameters for every node, nodes are listed in\n"
            "                the order found by the tracing algorithm\n"
            "     trace 16:  Gouy phases for all spaces\n"
            "     trace 32:  coupling coefficients for all components\n"
            "     trace 64:  mode matching parameters during calculation, if\n"
            "                they change due to a parameter change, for example\n"
            "                by changing a radius of curvature.\n"
            "     trace 128: nodes found during the cavity tracing\n"
            " ** phase 0-7: also bit coded, i.e. 3 means `1 and 2'\n"
            "     phase 1:   phase of coupling coefficients k_00 set 0\n"
            "     phase 2:   Gouy phase of TEM_00 set to 0\n"
            "     phase 4:   `ad name n m f' yields amplitude without Gouy phase\n"
            "     (default: phase 3)\n"
            " ** kmn n: `n' bit coded word, set computation of coupling coeffs.:\n"
            "     kmn 0  : analytic solution of all effects used\n"
            "     kmn 1  : verbose, i.e. print coupling coefficients\n"
            "     kmn 2  : numerical intergration if x and y misalignment is set\n"
            "     kmn 4  : numerical intergration if x or y misalignment is set\n"
            "     kmn 8  : calculates aperture kmn by integration\n"
            "     kmn 16 : calculates curvature kmn by integration\n"
            "     kmn 32 : calculates bayer-helms kmn by integration\n"
            "     (default: kmn 0)\n"
            " ** bp, possible parameters for this detector:\n"
            "     w  : beam radius\n"
            "     w0 : waist radius\n"
            "     z  : distance to waist\n"
            "     zr : Rayleigh range\n"
            "     g  : Gouy phase\n"
            "     r  : Radius of curvature (phase front) in meters\n"
            "     q  : Gaussian beam parameter\n"
            " ** isol S, suppression given in dB:\n"
            "     amplitude coefficient computed as 10^-(S/20)\n"
            /*
              " ** grating types are defined via their number of ports:\n"
              "     - 2: 1st order Littrow (eta_0, eta_1)\n"
              "     - 3: 2nd order Littrow (eta_0, eta_1, eta_2, rho_0)\n"
              "     - 4: not Littrow (eta_0, eta_1, eta_2, eta_3)\n"
             */
            /* 
               "debug n: `n' bit coded word, every bit gives various debugging output\n"
               "     x\n"
               "scale factor : factor can be also ampere/meter/deg\n"
               "     debug 1:   list of internal input parameters\n"
               "     x\n"
               "     x\n"
               "attr ... : one to three per line Rc, Rx, Ry but xbeta, ybeta \n"
               "     `...\n"
               "  \n"
             */
            " ** maxtem O/off : O=n+m (TEM_nm) the order of TEM modes, \n"
            "     `off' switches the TEM modus off explicitly\n"
            " ** conf:\n"
            "    This command is used to configure settings of a component\n"
            "    that are not physical simulation parameters.\n"
            "    that can be set for each component.\n"
            "    - conf  component integration_method integer\n"
            "      sets the numerical integration method. Cuba refers to a\n"
            "      self-adapting routine which is faster but less robust\n"
            "      1 - Riemann Sum\n"
            "      2 - Cubature - Serial\n"
            "      3 - Cubature - Parallel (default)\n"
            "    - conf component interpolation_method integer\n"
            "      set the interpolation method for the numerical integration of\n"
            "      surface maps, the (use NN for maps with sharp edges)\n"
            "      1 - Nearest Neighbour (default)\n"
            "      2 - Linear\n"
            "      3 - Spline\n"
            "    - conf component interpolation_size integer\n"
            "      sets the size of the interpolation kernel, must be odd and > 0\n"
            "    - conf component show_kmn_neval : 0/1\n"
            "      Shows the number of integrand evaluations used for the map integration.\n"
            "    - conf component save_kmn_matrices : 0/1\n"
            "      If true the kmn matrices are saved to .mat files\n"
            "      for distortion, merged map and the final result\n"
            "    - conf component save_kmn_binary : 0/1\n"
            "      If true the kmn and merged map data is stored in a\n"
            "      binary format rather than ascii. See -convert option\n"
            "      for kat in -h for converting between the 2 formats\n"
            "    - conf component save_interp_file : 0/1\n"
            "      If true then for each kmn calculated a file is written\n"
            "      to outputting each interpolated point. The output file\n"
            "      will have 4 columns: x,y,A,phi. So for each integrand\n"
            "      evaluation the interpolated point is plotted\n"
            );
}

//! Print help screen

/*!
 *
 */
void print_help() {
    printf(
            "------------------------------------------------------------------------\n"
            "  FINESSE %s     - Help Screen -                      %s\n"
            "------------------------------------------------------------------------\n",
            VERSION, mydate);
    printf(
            "** Usage (1) kat [options] infile [outfile [gnufile]] \n"
            "   or    (2) kat [options] basename\n"
            "   or    (3) kat -convert in_file [out_file]\n"
            " in (2) e.g. basename 'test' means input filename :"
            " 'test.kat', output\n filename : 'test.out' and Gnuplot batch file"
            " name : 'test.gnu'.\n"
            "** Support :\n"
            " Visit the user forums at http://www.gwoptics.org/finesse/forums/\n"
            "** Available options :\n"
            " -v : prints version number and build date\n"
            " -h : prints this help (-hh prints second help screen)\n"
            " -c : check consistency of interferometer matrix\n"
            " -max : prints max/min\n"
            " --server : starts Finesse in server mode\n"
            " --noheader : suppresses header information in output data files\n"
            " --perl1 : suppresses printing of banner\n"
            " --quiet : suppresses almost all screen outputs\n"
            " -convert: See usage example (3) above. Used to convert kmn files\n"
            "           between text and binary formats"
            //  " --new : switch on newly introduced features\n"
            " -sparse, -klu : switch to SPARSE or KLU library respectively\n"
            "** Available interferometer components :\n"
            " l name P f [phase] node                          - laser\n"
            " m name R T phi node1 node2                       - mirror\n"
            "(or: m1 name T Loss phi ...          \n"
            "     m2 name R Loss phi ... )        \n"
            " s name L [n] node1 node2                         - space\n"
            " bs name R T phi alpha node1 node2 node3 node4    - beamsplitter\n"
            "(or: bs1 name T Loss phi ... \n"
            "     bs2 name R Loss phi ... )             \n"
#ifdef DEVELOP
            " gr[n] name d node1 node2 [node3 [node4]]         - grating\n"
#endif 
            " isol name S node1 node2                          - isolator\n"
            " mod name f midx order am/pm [phase] node1 node2  - modulator\n"
            " lens f node1 node2                               - thin lens\n"
            //  " sq name factor angle node1 node2                 - squeezer\n"
            "** Detectors :\n"
            " pd[n] name [f1 [phase1 [f2... ]]] node[*]        - photodetector [mixer]\n"
            " pdS[n] name [f1 phase1 [f2... ]] node[*]         - sensitivity\n"
            " pdN[n] name [f1 phase1 [f2... ]] node[*]         - norm. photodetector\n"
            " ad name [n m] f node[*]                          - amplitude detector\n"
            " shot name node[*]                                - shot noise\n"
            " bp name x/y parameter node[*]                    - plots beam parameters\n"
            " cp cavity_name x/y parameter                     - plots cavity parameters\n"
            " gouy name x/y space-list                         - plots gouy phase\n"
            " beam name [f] node[*]                            - plots beam shape\n"
#ifdef DEVELOP
            " conv name [f] HG/LG n m xsize ysize [w0x z0x [w0y z0y]] node[*]\n"
            "                                                  -  computes a convolution\n"
#endif 
            " qshot name num_demod f [phase] node[*]           - quantum shotnoise detector\n"
            " qshotS name num_demod f [phase] node[*]          - quantum shotnoise sens.\n"
            /*
            " qnoised name qnoise_type num_demods f1 phase1 [f2 phase2 [...]] node[*]\n"
            "                                                  - quantum noise detector\n"
             */
            "** Available commands :\n"
            " fsig name component [type] f phase [amp]         - signal\n"
            " tem input n m factor phase                       - input power in TEMs\n"
            " mask detector n m factor                         - mode mask for outputs\n"
            " pdtype detector type-name                        - set detector type\n"
            " attr component M value Rcx/y value x/ybeta value - attributes of m/bs\n"
            "  (alignment angles beta in [rad])\n"
            " map component filename                           - read mirror map file\n"
            " kmn mirror_name filename_prefix                  - save coefficients to file\n"
            " maxtem order                                     - TEM order: n+m<=order\n"
            " gauss name component node w0 z [wy0 zy]          - set q parameter\n"
            " gauss* name component node q [qy] (q as 'z z_R') - set q parameter\n"
            " cav name component1 node component2 node         - trace beam in cavity\n"
            " startnode node                                   - startnode of trace\n"
            " retrace [off]                                    - re-trace beam on/off\n"
            " phase 0-7  (default: 3)                          - change Gouy phases\n"
            "  (1: phi(00)=0, 2: gouy(00)=0, 4: switch ad phase)\n"
            " conf component_name setting value                - configures component."
            " Use -hh option for more details.\n"
            "** Plot and Output related commands :\n"
            " xaxis[*] component param. lin/log min max steps  - parameter to tune\n"
            " x2axis[*] component param. lin/log min max steps - second x-axis for\n"
            "                                                    3D plot\n"
            " noxaxis                                          - ignore xaxis commands\n"
            " const name value                                 - constant $name\n"
            " variable name value                              - variable $name\n"
            " set name component parameter                     - variable $name\n"
            " func name = function-string                      - function $name\n"
            " lock[*] name $var gain accuracy                  - lock: make $var to 0\n"
            " put component parameter $var/$x1/$x2             - updates parameter\n"
            " noplot output                                    - no plot for 'output'\n"
            " trace verbosity                                  - verbose tracing\n"
            " yaxis [lin/log] abs:deg/db:deg/re:im/abs/db/deg  - y-axis definition\n"
            " scale factor [output]                            - y-axis rescaling\n"
            " diff component parameter                         - differentiation\n"
            " deriv_h value                                    - step size for diff\n"
            "** Auxiliary plot commands :\n"
            " gnuterm terminal [filename]                      - Gnuplot terminal\n"
            " pause                                            - pauses after plotting\n"
            " multi                                            - plots all surfaces\n"
            "                                                    save/load kmn file\n"
            " GNUPLOT \\ ... \\ END                              - set of extra commands\n"
            "                                                    for plotting.\n");
}

//! Prints max and min values of all output traces

/*!
 *
 * \todo refactor for readability
 * \todo untested
 */
void max_min(void) {
    int naxes;

    if (inter.yaxis[1]) {
        naxes = 1;
    } else {
        naxes = 0;
    }

    int j;
    for (j = 0; j < inter.num_outputs; j++) {
        output_data_t *output_data;
        output_data = &(inter.output_data_list[j]);

        /*
          if (inter.sum)
          j=inter.num_light_outputs;
         */

        message("%s ", output_data->name);
        if (naxes) {
            message("(y1) ");
        }

        if (inter.x3_axis_is_set) {
            if (inter.x3.xtype == FLOG) {
                message("x3 %.10g ", *inter.x3.xaxis / inter.x3.op);
            } else {
                message("x3 %.10g ", *inter.x3.xaxis - inter.x3.op);
            }
        }

        message("max ");

        message("x %.10g ", output_data->Omax_x[0]);
        if (inter.splot) {
            message("y %.10g ", output_data->Omax_y[0]);
        }
        message(": %.10g", output_data->Omax[0]);

        message(", min ");

        message("x %.10g ", output_data->Omin_x[0]);
        if (inter.splot) {
            message("y %.10g ", output_data->Omin_y[0]);
        }
        message(": %.10g", output_data->Omin[0]);
        message("\n");

        if (naxes) {
            message("\n%s (y2) ", output_data->name);
            message("max ");
            message("x %.10g ", output_data->Omax_x[1]);

            if (inter.splot) {
                message("y %.10g ", output_data->Omax_y[1]);
            }
            message(": %.10g", output_data->Omax[1]);

            message(", min ");
            message("x %.10g ", output_data->Omin_x[1]);

            if (inter.splot) {
                message("y %.10g ", output_data->Omin_y[1]);
            }
            message(": %.10g", output_data->Omin[1]);

            message("\n");
        }
    }
}

void convert_kmn_file(const char *filename, const char *newfilename) {

    message(" * Converting kmn file type\n");

    char fname[FILENAME_MAX];
    strcpy(fname, filename);

    mirror_t mirror;
    unsigned int i = strcspn(filename, ".knm");
    // check if we have found a file extension
    if (i < strlen(fname)) {
        // end the string before the .knm
        // as we add it later
        fname[i + 2] = '\0';
    }

    strcpy(mirror.map_merged.filename, fname);
    // need a mirror name for the error messages, pick m1 here?
    strcpy(mirror.name, "m1");

    message(" * Opening kmn file...\n");
    if (!read_mirror_kmn_files(&mirror, true))
        gerror("An error occurred whilst reading the kmn file.\n Please check"
            " it exists and has been correctly generated.\n");

    if (strlen(newfilename) != 0) {
        unsigned int i = strcspn(newfilename, ".knm");
        strcpy(fname, newfilename);
        // check if we have found a file extension
        if (i < strlen(fname)) {
            // end the string before the .knm
            // as we add it later
            fname[i + 2] = '\0';
        }

        strcpy(mirror.map_merged.filename, fname);
        message(" * Saving using new file name...\n");
    }

    // invert the save type
    mirror.map_merged.kmn_save_binary = !mirror.map_merged.kmn_save_binary;

    if (mirror.map_merged.kmn_save_binary)
        message(" * Converting from ASCII to binary format\n");
    else
        message(" * Converting from binary to ASCII format\n");

    write_mirror_kmn_file(&mirror, true);

    message(" * kmn files have now been converted and saved with prefix %s", mirror.map_merged.filename);
}

/*
 * Local variables:
 * c-indentation-style: bsd
 * c-basic-offset: 2
 * indent-tabs-mode: nil
 * End:
 *
 * vim: tabstop=2 expandtab shiftwidth=2:
 */
