// $Id$

/*!
 * \file kat_klu.c
 * \brief KLU sparse matrix library interface routines.
 *
 * These routines also set up the interferometer matrix.
 * 
 * @section Copyright notice
 *
 *  This file is part of the interferometer simulation Finesse
 *  http://www.gwoptics.org/finesse
 *
 *  Copyright (C) 1999 onwards Andreas Freise
 *  with parts of the code written by Daniel Brown, Paul Cochrane
 *  and Gerhard Heinzel.
 *
 *  This program is free software; you can redistribute it and/or modify it under
 *  the terms of the GNU General Public License version 3 as published
 *  by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 *  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *  See the GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along with
 *  this library; if not, write to the Free Software Foundation, Inc., 59 Temple Place,
 *  Suite 330, Boston, MA 02111-1307 USA
 */


#include "kat.h"
#include "kat_klu.h"
#include "kat_inline.c"
#include "kat_spa.h"
#include "kat_check.h"
#include "kat_io.h"
#include "kat_aux.h"
#include "klu.h" 

extern const complex_t complex_i;
extern const complex_t complex_1;
extern const complex_t complex_m1;
extern const complex_t complex_0;

//extern init_variables_t in;
extern interferometer_t inter;
extern options_t options;
extern memory_t mem;
extern sparse_var_t spa;
extern klu_sparse_var_t klu;

extern klu_l_symbolic *Symbolic;
extern klu_l_numeric *Numeric;
extern klu_l_common Common;

long *temp_col_ptr;

//! Allocate memory for KLU sparse matrix

/*! in Compressed Column format (CCS)
 *
 */
void klu_create(void) {
    int i;

    // create the standard interferometer matrix
    debug_msg("Creating standard interferometer matrix\n");

    // Allocate vectors for storing matrix in CCS format. See kat.h for
    // a description of the format

    // allocate column pointers
    if ((klu.col_ptr = (long *) malloc((inter.num_eqns + 1) * sizeof (long))) == NULL) {
        bug_error("cannot allocate for sparse matrix (3)");
    }

    if ((temp_col_ptr = (long *) malloc((inter.num_eqns + 1) * sizeof (long))) == NULL) {
        bug_error("cannot allocate for sparse matrix (4)");
    }

    // set column pointers/counters to zero
    for (i = 0; i < inter.num_eqns + 1; i++) {
        klu.col_ptr[i] = 0;
        temp_col_ptr[i] = 0;
    }

    // allocate ride-hand-side vector (again using doube for complex values)
    if ((klu.rhs = (double *) malloc(2 * inter.num_eqns * sizeof (double))) == NULL) {
        bug_error("cannot allocate for sparse matrix (5)");
    }

    // point solution vector to klu_rhs
    spa.spsol = (complex_t *) klu.rhs;

    // create the quantum noise interferometer matrix
    if (inter.qcorr) {
        /* NOT YET
        debug_msg("Creating quantum interferometer matrix\n");
        spa.Qspamat = spCreate(inter.num_quantum_eqns, 1, &err);
        //report_sparse_matrix_error(err, 2);

        if (spa.Qspamat == NULL) {
          bug_error("quantum corrections sparse matrix is NULL!");
        }

        // report some memory-related information if asked
        if (inter.debug & 512) {
          message("Memory needed for quantum interferometer matrix is about");
          kbytes = 20.0 * inter.num_quantum_eqns 
            + 12.0 * inter.num_quantum_eqns * inter.num_quantum_eqns;

          if (kbytes > 1048576.0) {
            kbytes = kbytes / 1048576.0;
            unit_string = munit;
          }
          else {
            kbytes = kbytes / 1024.0;
            unit_string = kunit;
          }
          message(" %.4g %s,\n", kbytes, unit_string);
        }
         */
    }

    // set default parameters for KLU solving
    klu_l_defaults(&Common);

    /* overwrite Defaults:

    tol: partial pivoting tolerance. If the diagonal entry has a magnitude greater than or equal 
         to tol times the largest magnitude of entries in the pivot column, then the diagonal entry is 
         chosen. Default value: 0.001. 
    ordering: which fill-reducing ordering to use: 0 for AMD, 1 for COLAMD, 2 for a user- 
         provided permutation P and Q (or a natural ordering if P and Q are NULL), or 3 for the 
         user order function. Default: 0 (AMD). 
    scale: whether or not the matrix should be scaled. If scale < 0, then no scaling is performed 
         and the input matrix is not checked for errors. If scale >= 0, the input matrix is check for 
         errors. If scale=0, then no scaling is performed. If scale=1, then each row of A is divided 
         by the sum of the absolute values in that row. If scale=2, then each row of A is divided by 
         the maximum absolute value in that row. Default: 2. 
    btf: if nonzero, then BTF is used to permute the input matrix into block upper triangular 
         form. This step is skipped if Common.btf is zero. Default: 1. 
    maxwork: sets an upper limit on the amount of work performed in btf maxtrans to 
         maxwork*nnz(A). If the limit is reached, a partial zero-free diagonal might be found. This has 
         no effect on whether or not the matrix can be factorized, since the matrix can be factorized 
         with no BTF pre-ordering at all. This option provides a tradeoff between the effectiveness of 
         the BTF ordering and the cost to compute it. A partial result can result in fewer, and larger, 
         blocks in the BTF form, resulting to more work required to factorize the matrix. No limit is 
         enforced if maxwork <= 0. Default: 0. 
    user order: a pointer to a function that can be provided by the application that uses KLU, 
         to redefine the fill-reducing ordering used by KLU for each diagonal block. The int and 
         UF long prototypes must be as follows: 

     */

    Common.ordering = 0;
    Common.scale = 2;
    Common.btf = 1;
    Common.maxwork = 0;




}


//! Destroy a KLU sparse matrix

/*!
 *
 *
 * \todo untested
 */

void klu_destroy(void) {
    free(klu.value);
    free(klu.row_index);
    free(klu.col_ptr);
    free(klu.rhs);
    if (inter.qcorr) {
        /* not yet
        spDestroy((char *) spa.Qspamat);
         */
    }
}


//! connect pointers from optical components to sparese matrix elements

/*!
 *
 */
void klu_get(void) {
    int i;
    char *unit_string;
    char munit[7] = "Mbytes";
    char kunit[7] = "Kbytes";
    double kbytes;

    // number of equations per node (e.g. 2 for TEM 00)
    // num_eqs_per_node= 2 * number of fields or =(inter.tem+1)*(inter.tem+2);
    int num_eqns_per_node = 2 * inter.num_fields;

    // 1. count elements per column
    // a) diagonal elements 
    for (i = 0; i < inter.num_eqns; i++) {
        klu.col_ptr[i]++;
    }
    // b) optical couplings


    get_mirror_matrix_elements(num_eqns_per_node, COUNT);
    get_beamsplitter_matrix_elements(num_eqns_per_node, COUNT);
    get_modulator_matrix_elements(num_eqns_per_node, COUNT);
    get_space_matrix_elements(num_eqns_per_node, COUNT);
    get_squeezer_matrix_elements(num_eqns_per_node, COUNT);
    get_lens_matrix_elements(num_eqns_per_node, COUNT);
    get_diode_matrix_elements(num_eqns_per_node, COUNT);
    get_grating_matrix_elements(num_eqns_per_node, COUNT);

    /*
    for (i=0; i<inter.num_eqns; i++) {
      printf("col_ptr: %d - %ld\n",i,klu.col_ptr[i]);
    }
     */

    // Now col_ptr contains how many elemens per column, need to
    // accumalte values in col_ptr to represent the normal format
    long temp_index1, temp_index2;
    temp_index1 = 0;
    temp_index2 = 0;


    for (i = 0; i < inter.num_eqns + 1; i++) {
        temp_index1 = klu.col_ptr[i];
        klu.col_ptr[i] = temp_index2;
        temp_col_ptr[i] = temp_index2;
        temp_index2 += temp_index1;
    }

    // set number of non-zero elements
    inter.num_nonzero = klu.col_ptr[inter.num_eqns];

    // allocate double vector for storing complex matrix non-zero elements
    if ((klu.value = (double *) malloc(2 * inter.num_nonzero * sizeof (double))) == NULL) {
        bug_error("cannot allocate for sparse matrix (1)");
    }
    // allocate respective row_index vector
    if ((klu.row_index = (long *) malloc(inter.num_nonzero * sizeof (long))) == NULL) {
        bug_error("cannot allocate for sparse matrix (2)");
    }

    // report some memory-related information if asked
    if (inter.debug & 512) {
        message("Rank of n x n (CCS) matrix: n=%d, no. of non-zero elements: %d\n",
                inter.num_eqns, inter.num_nonzero);

        message("Memory needed for interferometer matrix (CCS format) is about");
        kbytes = sizeof (long) * (inter.num_eqns + 1.0) * inter.num_nonzero * (2 * sizeof (double) + sizeof (long));

        if (kbytes > 1048576.0) {
            kbytes = kbytes / 1048576.0;
            unit_string = munit;
        } else {
            kbytes = kbytes / 1024.0;
            unit_string = kunit;
        }
        message(" %.4g %s,\n", kbytes, unit_string);
    }


    // fill diagonal with 1s
    for (i = 0; i < inter.num_eqns; i++) {
        temp_index1 = temp_col_ptr[i]++;
        klu.row_index[temp_index1] = i;
        klu.value[2 * temp_index1] = 1.0; // real part
        klu.value[2 * temp_index1 + 1] = 0.0; // imaginary part
    }

    // reset connect count for all nodes
    reset_connect_count();

    // Now get pointers into matrix

    get_mirror_matrix_elements(num_eqns_per_node, GETELEMENT);
    get_beamsplitter_matrix_elements(num_eqns_per_node, GETELEMENT);
    get_modulator_matrix_elements(num_eqns_per_node, GETELEMENT);
    get_space_matrix_elements(num_eqns_per_node, GETELEMENT);
    get_squeezer_matrix_elements(num_eqns_per_node, GETELEMENT);
    get_lens_matrix_elements(num_eqns_per_node, GETELEMENT);
    get_diode_matrix_elements(num_eqns_per_node, GETELEMENT);
    get_grating_matrix_elements(num_eqns_per_node, GETELEMENT);

}


//! Gets and returns the pointer to KLU sparse matrix element

/*!
 * \param i row matrix index
 * \param j column matrix index
 *
 * \return the pointer to the matrix element value referenced by i and j
 *
 */
complex_t *get_klu_matrix_elem(int row_index, int col_index, int matrix_type) {
    long temp_index;
    complex_t *p;

    // The following adds to the CSS matrix as we run through the
    // components. So see how to access CSS matrices, have a look at the
    // function cs_triplet in the CSparse package.

    p = NULL;
    switch (matrix_type) {
        case STANDARD:
            temp_index = temp_col_ptr[col_index]++;
            klu.row_index[temp_index] = (long) row_index;
            //  printf("temp %ld, i %d, j %d\n",temp_index,row_index,col_index);
            p = (complex_t *)&(klu.value[2 * temp_index]);
            break;
        case QCORR:
            break;
        default:
            bug_error("no such type");
    }
    return p;
}


/*
int klutest (void) 
{ 
  int n = 5 ; 
  int ldim = 5;
  long Ap [ ] = {0, 2, 5, 9, 10, 12} ; 
  long Ai [ ] = { 0, 1, 0, 2, 4, 1, 2, 3, 4, 2, 1, 4} ; 
  double Ax [ ] = {2., 0, 3., 0, 3., 0, -1., 0, 4., 0, 4., 0, -3., 0, 1., 0, 2., 0, 2., 0, 6., 0, 1.,0} ; 
  double b [ ] = {8., 0, 45., 0, -3., 0, 3., 0, 19., 0} ; 
  complex_t *z1;

    klu_l_symbolic *Symbolic ; 
    klu_l_numeric *Numeric ; 
    klu_l_common Common ; 
    int i ; 
    klu_l_defaults (&Common) ; 
    Symbolic = klu_l_analyze (n, Ap, Ai, &Common) ; 
    Numeric = klu_zl_factor (Ap, Ai, Ax, Symbolic, &Common) ; 
    klu_zl_solve (Symbolic, Numeric, ldim, 1, b, &Common) ; 
    for (i = 0 ; i < 2*n ; i++) printf ("x [%d] = %g\n", i, b [i]) ; 
  Ax[1]=1.5;

  //testing if we can acces the double storage of complex numbers using casts
  z1=(complex_t *)&(Ax[3]);
  z1->im=0.5;
    //for (i = 0 ; i < 12; i++) printf ("A [%d] = %g\n", i, Ax [i]) ; 
  // works!


  klu_zl_refactor (Ap, Ai, Ax, Symbolic, Numeric, &Common) ; 
    klu_zl_solve (Symbolic, Numeric, ldim, 1, b, &Common) ; 
    for (i = 0 ; i < 2*n ; i++) printf ("x [%d] = %g\n", i, b [i]) ; 
    klu_l_free_symbolic (&Symbolic, &Common) ; 
    klu_zl_free_numeric (&Numeric, &Common) ; 
  //	for (i = 0 ; i < 2*n ; i++) printf ("x [%d] = %g\n", i, b [i]) ; 
    return (0) ; 
} 
 */

/*
 * Local variables:
 * c-indentation-style: bsd
 * c-basic-offset: 2
 * indent-tabs-mode: nil
 * End:
 *
 * vim: tabstop=2 expandtab shiftwidth=2:
 */

