// $Id$

/*!
 * \file kat_mem.c
 * \brief Memory allocation routines for components used in simulation
 * 
 * @section Copyright notice
 *
 *  This file is part of the interferometer simulation Finesse
 *  http://www.gwoptics.org/finesse
 *
 *  Copyright (C) 1999 onwards Andreas Freise
 *  with parts of the code written by Daniel Brown, Paul Cochrane
 *  and Gerhard Heinzel.
 *
 *  This program is free software; you can redistribute it and/or modify it under
 *  the terms of the GNU General Public License version 3 as published
 *  by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 *  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *  See the GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along with
 *  this library; if not, write to the Free Software Foundation, Inc., 59 Temple Place,
 *  Suite 330, Boston, MA 02111-1307 USA
 */


#include "kat.h"
#include "kat_mem.h"
#include "kat_read.h"
#include "kat_inline.c"
#include "kat_io.h"
#include "kat_aux.h"
#include "kat_init.h"
#include "kat_dump.h"
#include "kat_aa.h"

extern options_t options;
extern init_variables_t init;
extern interferometer_t inter;
extern memory_t mem;
extern local_var_t vlocal;
extern sparse_var_t spa;
extern char functokens[];
extern int endcomp; //!< Flag for the end of the computation???
extern int lastparam; //!< Flag for the last parameter???
extern int *t_s; //!< Type_of_Signal list [frequency]
extern int *transfer;
extern double *f_s; //!< Frequency value list [frequency]
extern complex_t ***a_s; //!< Amplitude list [field] [output] [frequency]
//! Quantum noise list [field] [detector] [frequency] [equation]
extern complex_t ****qnoise;

//! Pre scan of input file

/*! 
 * Scan the input file first and allocate memory for the various components
 * and the interferometer matrix
 *
 * \param fp the file pointer
 *
 * \todo refactor for readability
 *
 * \todo break up this routine into smaller pieces!!!
 */
void pre_scan(FILE *fp) {
    char s0[LINE_LEN], *s;
    double kbytes;

    // check that the file pointer is non-null
    assert(fp != NULL);

    endcomp = 0;
    lastparam = -1;

    long int bytes = 0;
    long int bytes1 = 0;
    long int bytes2 = 0;

    mem.num_modulators = 0;
    mem.num_mirrors = 0;
    mem.num_beamsplitters = 0;
    mem.num_spaces = 0;
    mem.num_lenses = 0;
    mem.num_gratings = 0;
    mem.num_squeezers = 0;
    mem.num_light_inputs = 0;
    mem.num_diodes = 0;
    mem.num_signal_inputs = 0;
    mem.num_gauss_cmds = 0;
    mem.num_light_outputs = 0;
    mem.num_outputs = 0;
    mem.num_beam_param_outputs = 0;
    mem.num_attributes = 0;
    mem.num_cavities = 0;
    mem.num_gnuterm_cmds = 0;
    mem.num_diff_cmds = 0;
    mem.num_constants = 0;
    mem.num_scale_cmds = 0;
    mem.hg_mode_order = 0;
    mem.num_node_names = 0;
    mem.num_nodes = 1; // why is this set to 1 ???
    mem.num_dump_nodes = 0;
    mem.num_components = 0;
    mem.num_func_cmds = 0;
    mem.num_put_cmds = 0;
    mem.num_locks = 0;
    mem.num_set_cmds = 0;

    // Check for replacement of init variables
    // TODO: move deriv_h here from kat_read and maybe add more params
    while (fgets(s0, LINE_LEN - 1, fp) != NULL) {
        s = s0 + strspn(s0, WHITESPACE); /* skip whitespace */

        if (prepare_line(s, 1)) {
            continue;
        }

        if (strncasecmp(s, "lambda", 5) == 0) {
            read_lambda(s);
        }
    }

    rewind(fp);

    // Count number of variables first
    while (fgets(s0, LINE_LEN - 1, fp) != NULL) {
        s = s0 + strspn(s0, WHITESPACE); /* skip whitespace */

        if (prepare_line(s, 1)) {
            continue;
        }

        if (string_matches(s, "gnuplot")) {
            read_init_dummy(fp, 0);
        } else if (string_matches(s, "/*")) {
            check_for_star(s);
            read_init_dummy(fp, 2);
        } else if (string_matches(s, "const")) {
            mem.num_constants++;
        } else if (string_matches(s, "var ")) {
            gerror("line '%s':\n"
                    "the 'var' command has been renamed to 'const' in version 0.99\n",
                    s);
        }
    }

    // Allocate memory for variables
    inter.constant_list =
            (constant_t *) calloc(mem.num_constants + 1, sizeof (constant_t));
    if (inter.constant_list == NULL) {
        gerror("Memeory allocation error no. 0\n");
    }
    bytes += (mem.num_constants + 1) * sizeof (constant_t);

    rewind(fp);

    // Read in variables
    while (fgets(s0, LINE_LEN - 1, fp) != NULL) {
        s = s0 + strspn(s0, WHITESPACE); /* skip whitespace */

        if (prepare_line(s, 1)) {
            continue;
        }

        if (string_matches(s, "gnuplot")) {
            read_init_dummy(fp, 0);
        } else if (string_matches(s, "/*")) {
            read_init_dummy(fp, 1);
        } else if (string_matches(s, "const")) {
            read_const(s);
        } else if (string_matches(s, "xparam")
                || string_matches(s, "x2param")
                || string_matches(s, "x3param")) {
            gerror("line '%s':\nthe 'xparam' command has been replaced "
                    "in version 0.99, see the manual\n"
                    "for the new commands 'set', 'func' and 'put'\n", s);
        }
    }

    // Sort variables so that the longest names are first
    sort_variable_names();

    // then check for debug level
    rewind(fp);
    while (fgets(s0, LINE_LEN - 1, fp) != NULL) {
        s = s0 + strspn(s0, WHITESPACE); /* skip whitespace */

        if (prepare_line(s, 1)) {
            continue;
        }

        insert_constants_values(s);

        if (string_matches(s, "gnuplot")) {
            read_init_dummy(fp, 0);
        } else if (string_matches(s, "/*")) {
            read_init_dummy(fp, 1);
        } else if (string_matches(s, "debug")) {
            read_debug(s);
        }
    }

    if (options.quiet) {
        inter.debug = 0;
    } else if (inter.debug & 1) {
        message("\n\n-------debug output--------\n");
        dump_constants(stdout);
    }

    if (inter.debug & 512) {
        message("Scanning input file:\n");
        fflush(stdout);
    }

    // count number of components

    rewind(fp);

    int modulation = 0;
    while (fgets(s0, LINE_LEN - 1, fp) != NULL) {
        s = s0 + strspn(s0, WHITESPACE); /* skip whitespace */

        if (prepare_line(s, 1)) {
            continue;
        }

        insert_constants_values(s);

        if (string_matches(s, "gnuplot")) {
            read_init_dummy(fp, 0);
        } else if (string_matches(s, "/*")) {
            read_init_dummy(fp, 1);
        } else if (string_matches(s, "mod")) {
            mem.num_modulators++;
            mem.num_node_names++;
            mem.num_nodes = mem.num_nodes + 2;
            modulation += get_modulation_order(s);
        } else if (string_matches(s, "m* ") ||
                string_matches(s, "m1 ") ||
                string_matches(s, "m2 ") ||
                string_matches(s, "m ") ||
                string_matches(s, "mirror ") ||
                string_matches(s, "mirror* ") ||
                string_matches(s, "mirror1 ") ||
                string_matches(s, "mirror2 ")) {
            mem.num_mirrors++;
            mem.num_node_names++;
            mem.num_nodes = mem.num_nodes + 2;
        } else if (string_matches(s, "bs* ") ||
                string_matches(s, "bs2 ") ||
                string_matches(s, "bs1 ") ||
                string_matches(s, "bs ") ||
                string_matches(s, "beamsplitter* ") ||
                string_matches(s, "beamsplitter2 ") ||
                string_matches(s, "beamsplitter1 ") ||
                string_matches(s, "beamsplitter ")) {
            mem.num_beamsplitters++;
            mem.num_node_names++;
            mem.num_nodes = mem.num_nodes + 4;
        } else if (string_matches(s, "gr") &&
                NOT string_matches(s, "grating")) {
            int num = s[2];

            if (num != 9 && num != 32) {
                if (num < 50 || num > 54) {
                    num = 2;
                } else {
                    num -= 48;
                }
            } else {
                num = 2;
            }
            mem.num_gratings++;
            mem.num_node_names++;
            mem.num_nodes = mem.num_nodes + num;
        } else if (string_matches(s, "sq ") ||
                string_matches(s, "squeezer ")) {
            mem.num_squeezers++;
            mem.num_node_names++;
            mem.num_nodes = mem.num_nodes + 2;
        } else if (string_matches(s, "s ") ||
                string_matches(s, "space ")) {
            mem.num_spaces++;
            mem.num_node_names++;
            mem.num_nodes = mem.num_nodes + 2;
        } else if (string_matches(s, "lens ")) {
            mem.num_lenses++;
            mem.num_node_names++;
            mem.num_nodes = mem.num_nodes + 2;
        } else if (string_matches(s, "l ") ||
                string_matches(s, "laser ") ||
                string_matches(s, "light ")) {
            mem.num_light_inputs++;
            mem.num_node_names++;
            mem.num_nodes++;
        }            // if we have a dump, this is now counted as a "real" node
            // \todo this is a waste of time, only the first token is read, hence
            // dump is never read
        else if (string_matches(s, "dump")) {
            mem.num_dump_nodes++;
            mem.num_node_names++;
            mem.num_nodes++;
        } else if (string_matches(s, "isol ") ||
                string_matches(s, "diode ")) {
            mem.num_diodes++;
            mem.num_node_names++;
            mem.num_nodes = mem.num_nodes + 2;
        } else if (string_matches(s, "fsig ")) {
            mem.num_signal_inputs++;
            mem.num_node_names++;
        } else if (string_matches(s, "variable ")) {
            mem.num_variables++;
        } else if (string_matches(s, "gauss ") ||
                string_matches(s, "gauss* ")) {
            mem.num_gauss_cmds++;
            mem.num_node_names++;
        } else if (string_matches(s, "shot ") ||
                string_matches(s, "qshot ") ||
                string_matches(s, "qshotS ") ||
                string_matches(s, "pdS") ||
                string_matches(s, "pdN") ||
                string_matches(s, "pd") ||
                string_matches(s, "ad ") ||
                string_matches(s, "qnoised ") ||
                string_matches(s, "beam ")) {
            mem.num_node_names++;
            mem.num_light_outputs++;
            mem.num_outputs++;

            // if using qnoised make sure we set up the QN memory by setting
            // inter.qcorr here (this feels dodgy, as we can switch it off later)
            if (string_matches(s, "qnoised ")) {
                set_qnoise_type(QN_ONLY);
            }

            // if we have a 'shot' detector, we must set inter.shotnoise
            // appropriately
            if (string_matches(s, "shot ")) {
                inter.shotnoise = 4;
            }
        } else if (string_matches(s, "gouy ") ||
                string_matches(s, "bp ")) {
            mem.num_node_names++;
            mem.num_beam_param_outputs++;
            mem.num_outputs++;
        } else if (string_matches(s, "cp ")) {
            mem.num_node_names++;
            mem.num_cavity_param_outputs++;
            mem.num_outputs++;
        } else if (string_matches(s, "conv ")) {
            mem.num_node_names++;
            mem.num_convolution_outputs++;
            mem.num_outputs++;
        } else if (string_matches(s, "mp ")) {
            mem.num_node_names++;
            mem.num_mirror_phase_outputs++;
            mem.num_outputs++;
        } else if (string_matches(s, "attr ")) {
            mem.num_attributes++;
            mem.num_node_names++;
        } else if (string_matches(s, "cav ") ||
                string_matches(s, "cavity ")) {
            mem.num_cavities++;
            mem.num_node_names++;
        } else if (string_matches(s, "gnuterm")) {
            mem.num_gnuterm_cmds++;
        } else if (string_matches(s, "diff")) {
            mem.num_diff_cmds++;
        } else if (string_matches(s, "scale")) {
            mem.num_scale_cmds++;
        } else if (string_matches(s, "lock")) {
            mem.num_node_names++;
            mem.num_locks++;
            mem.num_outputs++;
        } else if (string_matches(s, "put")) {
            mem.num_node_names++;
            mem.num_put_cmds++;
        } else if (string_matches(s, "func")) {
            mem.num_node_names++;
            mem.num_func_cmds++;
            mem.num_outputs++;
        } else if (string_matches(s, "set")) {
            mem.num_set_cmds++;
            mem.num_node_names++;
        }            // read maxtem, order of TEM modes
        else if (string_matches(s, "maxtem ")) {
            read_maxtem(s);
        }            /*
    // read specifier for coupling coefficient computation
    else if (string_matches(s, "kmn ")) {
      read_kmn(s);
    }
*/
            // if noise is used, read in and set the relevant inter.shotnoise value
        else if (string_matches(s, "noise")) {
            read_noise(s);
        }
    }

    if (inter.debug & 512) {
        message("  using TEM up to n+m=%d\n", mem.hg_mode_order);
        fflush(stdout);
    }

    // if init file has been found, read in pdtypes now
    // (now the maximum TEM order is known!)
    if (init.init_file_is_found) {
        read_init_pdtype();
    }

    if (inter.debug & 1024) {
        message("pdtypes:\n");
        dump_pdtype(stdout);
        fflush(stdout);
    }

    // compute number of derived paramters
    // ------------------------------------------------------------
    // number of names = number of names so far plus all node names
    mem.num_node_names += mem.num_nodes;
    // number of electric fields per node per frequency 
    // (a function of the order of HG modes)
    mem.num_fields = (int) (mem.hg_mode_order + 1) * (mem.hg_mode_order + 2) / 2;

    // the number of quadratures
    mem.num_quadratures = 2;

    // number of components, used for allocating memory for component numbers
    mem.num_components = mem.num_mirrors + mem.num_spaces + mem.num_beamsplitters
            + mem.num_modulators + mem.num_light_inputs
            + mem.num_diodes + mem.num_lenses
            + mem.num_gratings + mem.num_squeezers;

    debug_msg("kat_mem(): found %d components in interferometer\n",
            mem.num_components);

    mem.num_frequencies = mem.num_light_inputs * (1 + modulation);

    // if signal frequencies are present: number of frequencies triples
    if (mem.num_signal_inputs > 0) {
        mem.num_frequencies *= 3;
    }

    // number of matrix pointers (=2*no.of.nodes*no.of.fields)^2
    mem.num_pointers = 2 * mem.num_nodes * mem.num_fields
            * 2 * mem.num_nodes * mem.num_fields;

    // calculate the number of reduced nodes
    mem.num_reduced_nodes = mem.num_nodes - mem.num_dump_nodes;

    if (inter.debug & 512) {
        message("  %d mirrors, ", mem.num_mirrors);
        message("%d beamsplitters, ", mem.num_beamsplitters);
        message("%d spaces, ", mem.num_spaces);
        message("%d modulators\n", mem.num_modulators);
        message("  %d isolators, ", mem.num_diodes);
        message("%d lenses, ", mem.num_lenses);
        message("%d gratings, ", mem.num_gratings);
        message("%d squeezers, ", mem.num_squeezers);
        message("%d lasers, ", mem.num_light_inputs);
        message("%d detectors\n", mem.num_outputs);
        message("  %d attributes, ", mem.num_attributes);
        message("%d gauss, ", mem.num_gauss_cmds);
        message("%d cavities, ", mem.num_cavities);
        message("%d constants\n", mem.num_constants);
        message("%d variables\n", mem.num_variables);
        message("  %d scales, ", mem.num_scale_cmds);
        message("  %d puts, ", mem.num_put_cmds);
        message("  %d locks, ", mem.num_locks);
        message("  %d functions, ", mem.num_func_cmds);
        message("  %d sets, ", mem.num_set_cmds);
        message("%d differentiations, ", mem.num_diff_cmds);
        message("%d gnuterms\n", mem.num_gnuterm_cmds);
        message("  derived paramters:\n");
        message("  %d fields, ", mem.num_fields);
        message("  %d frequencies, ", mem.num_frequencies);
        message("  %d nodes, ", mem.num_nodes);
        message("  %d dump nodes\n", mem.num_dump_nodes);
        message("  %d reduced nodes\n", mem.num_reduced_nodes);
        message("  %d names, ", mem.num_node_names);
        message("  %d pointers\n", mem.num_pointers);
        fflush(stdout);
    }

    // alloctate memeory for reading input file, print 
    // possible allocation failure or size of allocated memory

    int err = allocate_memory(&bytes, &bytes1, &bytes2);
    if (err != 0) {
        warn("Memeory allocation error no. %d, after succesful\n", err);
        warn("allocation of \n");
        bytes += bytes1 + bytes2;

        if (bytes > 1048576) {
            kbytes = ((double) bytes) / 1048576.;
            warn("(%.3g Mbytes).\n", kbytes);
        } else {
            kbytes = ((double) bytes) / 1024.;
            warn("(%.3g Kbytes).\n", kbytes);
        }
        my_finish(1);
    }
}

//! Allocate memeory for components and variables needed for reading input file

/*!
 * \param bytes ???
 * \param bytes1 ???
 * \param bytes2 ???
 *
 * \todo separate memory allocation subsections into separate subroutines
 *
 * \todo options.check branch is untested
 */
int allocate_memory(long int *bytes, long int *bytes1, long int *bytes2) {
    double kbytes, kbytes1;
    char *s, *s1;
    char munit[7] = "Mbytes";
    char kunit[7] = "Kbytes";

    // allocate memory for gnuterm commands
    int error_code = allocate_memory_for_gnuterms(bytes);
    if (error_code != 0) {
        return error_code;
    }

    // allocate memory for carrier frequencies
    error_code = allocate_memory_for_frequency_list(bytes);
    if (error_code != 0) {
        return error_code;
    }

    // allocate memory for mirrors
    error_code = allocate_memory_for_mirror_list(bytes);
    if (error_code != 0) {
        return error_code;
    }

    // allocate memory for beam splitters
    error_code = allocate_memory_for_beamsplitter_list(bytes);
    if (error_code != 0) {
        return error_code;
    }

    // allocate memory for gratings
    error_code = allocate_memory_for_grating_list(bytes);
    if (error_code != 0) {
        return error_code;
    }

    // allocate memory for squeezers
    error_code = allocate_memory_for_squeezer_list(bytes);
    if (error_code != 0) {
        return error_code;
    }

    // allocate memory for spaces
    error_code = allocate_memory_for_space_list(bytes);
    if (error_code != 0) {
        return error_code;
    }

    // allocate memory for input light fields
    error_code = allocate_memory_for_light_input_list(bytes);
    if (error_code != 0) {
        return error_code;
    }

    // allocate memory for output light fields
    error_code = allocate_memory_for_detector_list(bytes);
    if (error_code != 0) {
        return error_code;
    }

    // allocate memory for outputs
    error_code = allocate_memory_for_output_data_list(bytes);
    if (error_code != 0) {
        return error_code;
    }

    // allocate memory for output beam parameter
    error_code = allocate_memory_for_beam_param_list(bytes);
    if (error_code != 0) {
        return error_code;
    }
    // allocate memory for output cavity parameter
    error_code = allocate_memory_for_cavity_param_list(bytes);
    if (error_code != 0) {
        return error_code;
    }

    // allocate memory for output convolution
    error_code = allocate_memory_for_convolution_list(bytes);
    if (error_code != 0) {
        return error_code;
    }

    // allocate memory for output mirror phase
    error_code = allocate_memory_for_mirror_phase_list(bytes);
    if (error_code != 0) {
        return error_code;
    }

    // allocate memory for modulators
    error_code = allocate_memory_for_modulator_list(bytes);
    if (error_code != 0) {
        return error_code;
    }

    // allocate memory for frequency signals
    error_code = allocate_memory_for_signal_list(bytes);
    if (error_code != 0) {
        return error_code;
    }

    // allocate memory for a derivative
    error_code = allocate_memory_for_derivative_list(bytes);
    if (error_code != 0) {
        return error_code;
    }

    // allocate memory for scale commands
    error_code = allocate_memory_for_scale_cmd_list(bytes);
    if (error_code != 0) {
        return error_code;
    }

    // allocate memory for diodes
    error_code = allocate_memory_for_diode_list(bytes);
    if (error_code != 0) {
        return error_code;
    }

    // allocate memory for variables
    error_code = allocate_memory_for_variable_list(bytes);
    if (error_code != 0) {
        return error_code;
    }
    // allocate memory for lenses
    error_code = allocate_memory_for_lens_list(bytes);
    if (error_code != 0) {
        return error_code;
    }

    // allocate memory for attributes
    if (mem.num_attributes) {
        inter.attr_list = (attr_t *) calloc(mem.num_attributes + 1, sizeof (attr_t));
        if (inter.attr_list == NULL) {
            return (29);
        }
        *bytes += (mem.num_attributes + 1) * sizeof (attr_t);
    }

    // allocate memory for a gauss command
    if (mem.num_gauss_cmds) {
        inter.gauss_list =
                (gauss_t *) calloc(mem.num_gauss_cmds + 1, sizeof (gauss_t));
        if (inter.gauss_list == NULL) {
            return (30);
        }
        *bytes += (mem.num_gauss_cmds + 1) * sizeof (gauss_t);
    }

    // allocate memory for a cavity
    if (mem.num_cavities) {
        inter.cavity_list = (cavity_t *)
                calloc(mem.num_cavities + 1, sizeof (cavity_t));
        if (inter.cavity_list == NULL) {
            return (31);
        }
        *bytes += (mem.num_cavities + 1) * sizeof (cavity_t);
    }

    // allocate memory for a node
    if (mem.num_nodes) {
        inter.node_list = (node_t *) calloc(mem.num_nodes + 1, sizeof (node_t));
        if (inter.node_list == NULL) {
            return (32);
        }
        *bytes += (mem.num_nodes + 1) * sizeof (node_t);
    }

    // allocate memory for the node names
    if (mem.num_node_names) {
        mem.names = (char **) calloc(mem.num_node_names + 1, sizeof (char *));
        if (mem.names == NULL) {
            return (33);
        }
        *bytes += (mem.num_node_names + 1) * sizeof (char *);
    }

    if (mem.num_components) {
        vlocal.component_trace_list = (int *) calloc(mem.num_components + 1, sizeof (int));
        if (vlocal.component_trace_list == NULL) {
            return (34);
        }
        *bytes += (mem.num_components + 1) * sizeof (int);
    }

    if (mem.num_nodes) {
        vlocal.trace_c1 = (int *) calloc((mem.num_nodes + 1)*2, sizeof (int));
        if (vlocal.trace_c1 == NULL) {
            return (98);
        }
        *bytes += (mem.num_nodes + 1) * 2 * sizeof (int);

        vlocal.trace_c2 = (int *) calloc(mem.num_nodes + 1, sizeof (int));
        if (vlocal.trace_c2 == NULL) {
            return (97);
        }
        *bytes += (mem.num_nodes + 1) * sizeof (int);
    }

    int i;
    for (i = 0; i < mem.num_outputs; i++) {
        inter.output_data_list[i].Omax[0] = -1e20;
        inter.output_data_list[i].Omax[1] = -1e20;
        inter.output_data_list[i].Omin[0] = 1e20;
        inter.output_data_list[i].Omin[1] = 1e20;
    }

    if (mem.num_light_outputs) {
        transfer = (int *) calloc(mem.num_light_outputs + 1, sizeof (int));
        if (transfer == NULL) {
            return (35);
        }
        *bytes += (mem.num_light_outputs + 1) * sizeof (int);
    }

    // allocate memory for puts
    if (mem.num_put_cmds) {
        inter.put_list = (put_command_t *) calloc(mem.num_put_cmds + 1, sizeof (put_command_t));
        if (inter.put_list == NULL) {
            return (89);
        }
        *bytes += (mem.num_put_cmds + 1) * sizeof (put_command_t);
    }

    // allocate memory for locks
    error_code = allocate_memory_for_locks(bytes);
    if (error_code != 0) {
        return error_code;
    }

    // allocate memory for functions
    error_code = allocate_memory_for_function_list(bytes);
    if (error_code != 0) {
        return error_code;
    }

    if (mem.num_func_cmds || mem.num_locks) {
        inter.is_locked =
                (bool *) calloc(mem.num_func_cmds + mem.num_locks + 2, sizeof (bool));
        if (inter.is_locked == NULL) {
            return (87);
        }
    }

    // allocate memory for sets
    if (mem.num_set_cmds) {
        inter.set_list = (set_command_t *) calloc(mem.num_set_cmds + 1, sizeof (set_command_t));
        if (inter.set_list == NULL) {
            return (85);
        }
        *bytes += (mem.num_set_cmds + 1) * sizeof (set_command_t);
    }

    t_s = (int *) calloc(mem.num_frequencies + 1, sizeof (int));
    f_s = (double *) calloc(mem.num_frequencies + 1, sizeof (double));

    if (f_s == NULL || t_s == NULL) {
        return (36);
    }
    *bytes1 += (mem.num_frequencies + 1) * sizeof (int);
    *bytes1 += (mem.num_frequencies + 1) * sizeof (double);

    // allocate memory for the amplitude list
    error_code = allocate_memory_for_amplitude_list(bytes1);
    if (error_code != 0) {
        return error_code;
    }

    // allocate memory for the quantum noise list
    error_code = allocate_memory_for_quantum_amplitude_list(bytes1);
    if (error_code != 0) {
        return error_code;
    }

    // TODO this memory is for the old Fortan numerical integration and 
    // can be removed once we fully moved to the new integration routine

    if (inter.tem_is_set && inter.kmn > 1) {
        mem.worklen = floor((init.maxintop - 65) / 65.0 * 8.0 + 43);
        mem.work = (double *) calloc(mem.worklen + 1, sizeof (double));

        if (mem.work == NULL) {
            return (96);
        }

        *bytes1 += (mem.worklen + 1) * sizeof (double);
    } else {
        mem.worklen = 0;
    }

    // allocate memory for some temporary complex variales in kat_calc
    vlocal.zsolution =
            (complex_t *) calloc(mem.num_outputs + 1, sizeof (complex_t));

    if (vlocal.zsolution == NULL) {
        return (41);
    }
    *bytes1 += (mem.num_outputs + 1) * sizeof (complex_t);

    vlocal.zderiv1 =
            (complex_t *) calloc(mem.num_outputs + 1, sizeof (complex_t));

    if (vlocal.zderiv1 == NULL) {
        return (42);
    }
    *bytes1 += (mem.num_outputs + 1) * sizeof (complex_t);

    vlocal.zderiv2 =
            (complex_t *) calloc(mem.num_outputs + 1, sizeof (complex_t));

    if (vlocal.zderiv2 == NULL) {
        return (43);
    }
    *bytes1 += (mem.num_outputs + 1) * sizeof (complex_t);

    // allocate memory for sparse algorithm
    debug_msg("Allocating memory for matrices\n");
    // sparse matrix diagonal
    spa.diag1 = (complex_t **) calloc(mem.num_reduced_nodes
            * (mem.hg_mode_order + 1)
            * (mem.hg_mode_order + 2) + 1, sizeof (complex_t *));

    if (spa.diag1 == NULL) {
        return (45);
    }
    *bytes2 += (mem.num_reduced_nodes
            * (mem.hg_mode_order + 1)
            * (mem.hg_mode_order + 2) + 1) * sizeof (complex_t *);
    spa.diag = spa.diag1 + 1;

    // sparse matrix RHS
    spa.sprhs1 = (complex_t *) calloc(mem.num_reduced_nodes
            * (mem.hg_mode_order + 1)
            * (mem.hg_mode_order + 2) + 1, sizeof (complex_t));

    if (spa.sprhs1 == NULL) {
        return (46);
    }
    *bytes2 += (mem.num_reduced_nodes
            * (mem.hg_mode_order + 1)
            * (mem.hg_mode_order + 2) + 1) * sizeof (complex_t);
    spa.sprhs = spa.sprhs1 + 1;

    // sparse matrix solution
    spa.spsol1 = (complex_t *) calloc(mem.num_reduced_nodes
            * (mem.hg_mode_order + 1)
            * (mem.hg_mode_order + 2) + 1, sizeof (complex_t));

    if (spa.spsol1 == NULL) {
        return (47);
    }
    *bytes2 += (mem.num_reduced_nodes
            * (mem.hg_mode_order + 1)
            * (mem.hg_mode_order + 2) + 1)
            * sizeof (complex_t);
    spa.spsol = spa.spsol1 + 1;

    // allocate memory for sparse algorithm quantum noise stuff
    if (inter.qcorr) {
        // quantum version of sparse matrix diagonal
        debug_msg("Allocating memory for quantum noise matrices\n");
        spa.Qdiag1 =
                (complex_t **) calloc(mem.num_nodes
                * (mem.hg_mode_order + 1)
                * (mem.hg_mode_order + 2) + 1, sizeof (complex_t *));

        if (spa.Qdiag1 == NULL) {
            return (45);
        }
        *bytes2 += (mem.num_nodes
                * (mem.hg_mode_order + 1)
                * (mem.hg_mode_order + 2) + 1) * sizeof (complex_t *);
        spa.Qdiag = spa.Qdiag1 + 1;

        // quantum verison of sparse matrix RHS
        spa.Qsprhs1 =
                (complex_t *) calloc(mem.num_nodes
                * (mem.hg_mode_order + 1)
                * (mem.hg_mode_order + 2) + 1, sizeof (complex_t));

        if (spa.Qsprhs1 == NULL) {
            return (46);
        }
        *bytes2 += (mem.num_nodes
                * (mem.hg_mode_order + 1)
                * (mem.hg_mode_order + 2) + 1) * sizeof (complex_t);
        spa.Qsprhs = spa.Qsprhs1 + 1;

        // quantum version of sparse matrix solution
        spa.Qspsol1 =
                (complex_t *) calloc(mem.num_nodes
                * (mem.hg_mode_order + 1)
                * (mem.hg_mode_order + 2) + 1, sizeof (complex_t));

        if (spa.Qspsol1 == NULL) {
            return (47);
        }
        *bytes2 += (mem.num_nodes
                * (mem.hg_mode_order + 1)
                * (mem.hg_mode_order + 2) + 1) * sizeof (complex_t);
        spa.Qspsol = spa.Qspsol1 + 1;
    }

    // check options
    if (options.check) {
        mem.ilist1 = (int *) calloc(mem.num_pointers + 1, sizeof (int));

        if (mem.ilist1 == NULL) {
            return (48);
        }
        *bytes2 += (mem.num_pointers + 1) * sizeof (int);

        mem.ilist2 = (int *) calloc(mem.num_pointers + 1, sizeof (int));

        if (mem.ilist2 == NULL) {
            return (49);
        }
        *bytes2 += (mem.num_pointers + 1) * sizeof (int);

        mem.jlist1 = (int *) calloc(mem.num_pointers + 1, sizeof (int));

        if (mem.jlist1 == NULL) {
            return (50);
        }
        *bytes2 += (mem.num_pointers + 1) * sizeof (int);

        mem.jlist2 = (int *) calloc(mem.num_pointers + 1, sizeof (int));

        if (mem.jlist2 == NULL) {
            return (51);
        }
        *bytes2 += (mem.num_pointers + 1) * sizeof (int);
    }

    *bytes1 += *bytes2;

    if (inter.debug & 512) {
        message("Initial memory allocated:\n");

        if (*bytes > 1048576) {
            kbytes = ((double) *bytes) / 1048576.;
            s = munit;
        } else {
            kbytes = ((double) *bytes) / 1024.;
            s = kunit;
        }

        if (*bytes1 > 1048576) {
            kbytes1 = ((double) *bytes1) / 1048576.;
            s1 = munit;
        } else {
            kbytes1 = ((double) *bytes1) / 1024.;
            s1 = kunit;
        }

        message("  optical setup=%.4g %s, internal variables=%.4g %s.\n", kbytes,
                s, kbytes1, s1);
        fflush(stdout);
    }

    return (0);
}

//! Allocate memory for gnuplot commands

/*!
 * \todo gnuterms -> gnuterm_commands???
 *
 * \param bytes number of bytes so far allocated???
 */
int allocate_memory_for_gnuterms(long int *bytes) {
    inter.gnuterm = (int *) calloc(mem.num_gnuterm_cmds + 1, sizeof (int));
    inter.gnutermfn =
            (char **) calloc(mem.num_gnuterm_cmds + 1, sizeof (char *));

    if (inter.gnuterm == NULL || inter.gnutermfn == NULL) {
        return (1);
    }
    *bytes += (mem.num_gnuterm_cmds + 1) * sizeof (int)
            + (mem.num_gnuterm_cmds + 1) * sizeof (char *);

    return 0;
}

//! Allocate memory for carrier frequency list

/*!
 * \param bytes number of bytes so far allocated???
 */
int allocate_memory_for_frequency_list(long int *bytes) {
    if (mem.num_frequencies) {
        inter.f = (double *) calloc(mem.num_frequencies + 1, sizeof (double));
        inter.t = (int *) calloc(mem.num_frequencies + 1, sizeof (int));

        if (inter.f == NULL || inter.t == NULL) {
            return (2);
        }
        *bytes += (mem.num_frequencies + 1) * sizeof (double)
                + (mem.num_frequencies + 1) * sizeof (int);
        inter.f[0] = NO_FREQ;
    }
    return 0;
}

void free_memory_for_mirror_list() {
    int i;
    mirror_t *mirror;

    // this is a temporary map that was used to 
    // calculate kmn's
    if (IS_MIRROR_KMN_ALLOCD(&tmap))
        mirror_kmn_free(&tmap);

    for (i = 0; i < mem.num_mirrors; i++) {
        mirror = &inter.mirror_list[i];
        // here we check to see if any of the kmn matrices were allocated memory
        // during the kmn calculation process
        if (IS_MIRROR_KMN_ALLOCD(&mirror->kmn_aperture)) mirror_kmn_free(&(mirror->kmn_aperture));
        if (IS_MIRROR_KMN_ALLOCD(&mirror->map_merged.kmn)) mirror_kmn_free(&(mirror->map_merged.kmn));
        if (IS_MIRROR_KMN_ALLOCD(&mirror->kmn_bayer_helms)) mirror_kmn_free(&(mirror->kmn_bayer_helms));
    }
}

//! Allocate memory for mirrors

/*!
 * \param bytes number of bytes so far allocated???
 */
int allocate_memory_for_mirror_list(long int *bytes) {
    if (mem.num_mirrors) {
        inter.mirror_list =
                (mirror_t *) calloc(mem.num_mirrors + 1, sizeof (mirror_t));

        if (inter.mirror_list == NULL) {
            return (3);
        }
        *bytes += (mem.num_mirrors + 1) * sizeof (mirror_t);

        int i;
        for (i = 0; i < mem.num_mirrors; i++) {
            // field amplitudes
            inter.mirror_list[i].a11 = (complex_t ***) calloc(mem.num_fields + 1, sizeof (complex_t **));
            inter.mirror_list[i].a12 = (complex_t ***) calloc(mem.num_fields + 1, sizeof (complex_t **));
            inter.mirror_list[i].a21 = (complex_t ***) calloc(mem.num_fields + 1, sizeof (complex_t **));
            inter.mirror_list[i].a22 = (complex_t ***) calloc(mem.num_fields + 1, sizeof (complex_t **));

            // coupling coefficients
            inter.mirror_list[i].kmn.k11 =
                    (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));
            inter.mirror_list[i].kmn.k12 =
                    (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));
            inter.mirror_list[i].kmn.k21 =
                    (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));
            inter.mirror_list[i].kmn.k22 =
                    (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));

            if (inter.mirror_list[i].a11 == NULL ||
                    inter.mirror_list[i].a12 == NULL ||
                    inter.mirror_list[i].a21 == NULL ||
                    inter.mirror_list[i].a22 == NULL ||

                    inter.mirror_list[i].kmn.k11 == NULL ||
                    inter.mirror_list[i].kmn.k12 == NULL ||
                    inter.mirror_list[i].kmn.k21 == NULL ||
                    inter.mirror_list[i].kmn.k22 == NULL) {
                return (4);
            }
            *bytes += 4 * (mem.num_fields + 1) * sizeof (complex_t **);
            *bytes += 4 * (mem.num_fields + 1) * sizeof (complex_t *);

            int field_index;
            for (field_index = 0; field_index < mem.num_fields; field_index++) {
                // field amplitudes
                inter.mirror_list[i].a11[field_index] = (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));
                inter.mirror_list[i].a12[field_index] = (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));
                inter.mirror_list[i].a21[field_index] = (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));
                inter.mirror_list[i].a22[field_index] = (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));

                // coupling coefficients
                inter.mirror_list[i].kmn.k11[field_index] =
                        (complex_t *) calloc(mem.num_fields + 1, sizeof (complex_t));
                inter.mirror_list[i].kmn.k12[field_index] =
                        (complex_t *) calloc(mem.num_fields + 1, sizeof (complex_t));
                inter.mirror_list[i].kmn.k21[field_index] =
                        (complex_t *) calloc(mem.num_fields + 1, sizeof (complex_t));
                inter.mirror_list[i].kmn.k22[field_index] =
                        (complex_t *) calloc(mem.num_fields + 1, sizeof (complex_t));

                if (inter.mirror_list[i].a11[field_index] == NULL ||
                        inter.mirror_list[i].a12[field_index] == NULL ||
                        inter.mirror_list[i].a21[field_index] == NULL ||
                        inter.mirror_list[i].a22[field_index] == NULL ||

                        inter.mirror_list[i].kmn.k11[field_index] == NULL ||
                        inter.mirror_list[i].kmn.k12[field_index] == NULL ||
                        inter.mirror_list[i].kmn.k21[field_index] == NULL ||
                        inter.mirror_list[i].kmn.k22[field_index] == NULL) {
                    return (5);
                }
                *bytes += 4 * (mem.num_fields + 1) * sizeof (complex_t *);
                *bytes += 4 * (mem.num_fields + 1) * sizeof (complex_t);
            }
        }

        // quantum noise stuff
        if (inter.qcorr) {
            int i;
            for (i = 0; i < mem.num_mirrors; i++) {
                inter.mirror_list[i].q11 =
                        (complex_t ***) calloc(mem.num_fields + 1, sizeof (complex_t **));
                inter.mirror_list[i].q12 =
                        (complex_t ***) calloc(mem.num_fields + 1, sizeof (complex_t **));
                inter.mirror_list[i].q21 =
                        (complex_t ***) calloc(mem.num_fields + 1, sizeof (complex_t **));
                inter.mirror_list[i].q22 =
                        (complex_t ***) calloc(mem.num_fields + 1, sizeof (complex_t **));

                if (inter.mirror_list[i].q11 == NULL ||
                        inter.mirror_list[i].q12 == NULL ||
                        inter.mirror_list[i].q21 == NULL ||
                        inter.mirror_list[i].q22 == NULL) {
                    return (4);
                }

                *bytes += 4 * (mem.num_fields + 1) * sizeof (complex_t **);

                int field_index;
                for (field_index = 0; field_index < mem.num_fields; field_index++) {
                    inter.mirror_list[i].q11[field_index] =
                            (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));
                    inter.mirror_list[i].q12[field_index] =
                            (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));
                    inter.mirror_list[i].q21[field_index] =
                            (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));
                    inter.mirror_list[i].q22[field_index] =
                            (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));

                    if (inter.mirror_list[i].q11[field_index] == NULL ||
                            inter.mirror_list[i].q12[field_index] == NULL ||
                            inter.mirror_list[i].q21[field_index] == NULL ||
                            inter.mirror_list[i].q22[field_index] == NULL) {
                        return (4);
                    }
                    *bytes += 4 * (mem.num_fields + 1) * sizeof (complex_t *);

                }
            }
        }
    }

    return 0;
}

//! Allocate memory for beam splitters

/*!
 * \param bytes number of bytes so far allocated???
 */
int allocate_memory_for_beamsplitter_list(long int *bytes) {
    if (mem.num_beamsplitters) {
        inter.bs_list = (beamsplitter_t *) calloc(mem.num_beamsplitters + 1,
                sizeof (beamsplitter_t));

        if (inter.bs_list == NULL) {
            return (6);
        }
        *bytes += (mem.num_beamsplitters + 1) * sizeof (beamsplitter_t);

        int i;
        for (i = 0; i < mem.num_beamsplitters; i++) {
            // field amplitudes
            inter.bs_list[i].a12 =
                    (complex_t ***) calloc(mem.num_fields + 1, sizeof (complex_t **));
            inter.bs_list[i].a21 =
                    (complex_t ***) calloc(mem.num_fields + 1, sizeof (complex_t **));
            inter.bs_list[i].a13 =
                    (complex_t ***) calloc(mem.num_fields + 1, sizeof (complex_t **));
            inter.bs_list[i].a31 =
                    (complex_t ***) calloc(mem.num_fields + 1, sizeof (complex_t **));
            inter.bs_list[i].a24 =
                    (complex_t ***) calloc(mem.num_fields + 1, sizeof (complex_t **));
            inter.bs_list[i].a42 =
                    (complex_t ***) calloc(mem.num_fields + 1, sizeof (complex_t **));
            inter.bs_list[i].a34 =
                    (complex_t ***) calloc(mem.num_fields + 1, sizeof (complex_t **));
            inter.bs_list[i].a43 =
                    (complex_t ***) calloc(mem.num_fields + 1, sizeof (complex_t **));

            // coupling coefficients
            inter.bs_list[i].k12 =
                    (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));
            inter.bs_list[i].k21 =
                    (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));
            inter.bs_list[i].k13 =
                    (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));
            inter.bs_list[i].k31 =
                    (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));
            inter.bs_list[i].k24 =
                    (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));
            inter.bs_list[i].k42 =
                    (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));
            inter.bs_list[i].k34 =
                    (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));
            inter.bs_list[i].k43 =
                    (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));

            if (inter.bs_list[i].a12 == NULL ||
                    inter.bs_list[i].a21 == NULL ||
                    inter.bs_list[i].a13 == NULL ||
                    inter.bs_list[i].a31 == NULL ||
                    inter.bs_list[i].a24 == NULL ||
                    inter.bs_list[i].a42 == NULL ||
                    inter.bs_list[i].a34 == NULL ||
                    inter.bs_list[i].a43 == NULL ||

                    inter.bs_list[i].k12 == NULL ||
                    inter.bs_list[i].k21 == NULL ||
                    inter.bs_list[i].k13 == NULL ||
                    inter.bs_list[i].k31 == NULL ||
                    inter.bs_list[i].k24 == NULL ||
                    inter.bs_list[i].k42 == NULL ||
                    inter.bs_list[i].k34 == NULL ||
                    inter.bs_list[i].k43 == NULL) {
                return (7);
            }

            *bytes += 8 * (mem.num_fields + 1) * sizeof (complex_t **);
            *bytes += 8 * (mem.num_fields + 1) * sizeof (complex_t *);

            int j;
            for (j = 0; j < mem.num_fields; j++) {
                // field amplitudes
                inter.bs_list[i].a12[j] =
                        (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));
                inter.bs_list[i].a21[j] =
                        (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));
                inter.bs_list[i].a13[j] =
                        (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));
                inter.bs_list[i].a31[j] =
                        (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));
                inter.bs_list[i].a24[j] =
                        (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));
                inter.bs_list[i].a42[j] =
                        (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));
                inter.bs_list[i].a34[j] =
                        (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));
                inter.bs_list[i].a43[j] =
                        (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));

                // coupling coefficients
                inter.bs_list[i].k12[j] =
                        (complex_t *) calloc(mem.num_fields + 1, sizeof (complex_t));
                inter.bs_list[i].k21[j] =
                        (complex_t *) calloc(mem.num_fields + 1, sizeof (complex_t));
                inter.bs_list[i].k13[j] =
                        (complex_t *) calloc(mem.num_fields + 1, sizeof (complex_t));
                inter.bs_list[i].k31[j] =
                        (complex_t *) calloc(mem.num_fields + 1, sizeof (complex_t));
                inter.bs_list[i].k24[j] =
                        (complex_t *) calloc(mem.num_fields + 1, sizeof (complex_t));
                inter.bs_list[i].k42[j] =
                        (complex_t *) calloc(mem.num_fields + 1, sizeof (complex_t));
                inter.bs_list[i].k34[j] =
                        (complex_t *) calloc(mem.num_fields + 1, sizeof (complex_t));
                inter.bs_list[i].k43[j] =
                        (complex_t *) calloc(mem.num_fields + 1, sizeof (complex_t));

                if (inter.bs_list[i].a12[j] == NULL ||
                        inter.bs_list[i].a21[j] == NULL ||
                        inter.bs_list[i].a13[j] == NULL ||
                        inter.bs_list[i].a31[j] == NULL ||
                        inter.bs_list[i].a24[j] == NULL ||
                        inter.bs_list[i].a42[j] == NULL ||
                        inter.bs_list[i].a34[j] == NULL ||
                        inter.bs_list[i].a43[j] == NULL ||

                        inter.bs_list[i].k12[j] == NULL ||
                        inter.bs_list[i].k21[j] == NULL ||
                        inter.bs_list[i].k13[j] == NULL ||
                        inter.bs_list[i].k31[j] == NULL ||
                        inter.bs_list[i].k24[j] == NULL ||
                        inter.bs_list[i].k42[j] == NULL ||
                        inter.bs_list[i].k34[j] == NULL ||
                        inter.bs_list[i].k43[j] == NULL) {
                    return (8);
                }

                *bytes += 8 * (mem.num_fields + 1) * sizeof (complex_t *);
                *bytes += 8 * (mem.num_fields + 1) * sizeof (complex_t);

            }
        }

        // quantum noise
        if (inter.qcorr) {
            int i;
            for (i = 0; i < mem.num_beamsplitters; i++) {
                inter.bs_list[i].q12 =
                        (complex_t ***) calloc(mem.num_fields + 1, sizeof (complex_t **));
                inter.bs_list[i].q21 =
                        (complex_t ***) calloc(mem.num_fields + 1, sizeof (complex_t **));
                inter.bs_list[i].q13 =
                        (complex_t ***) calloc(mem.num_fields + 1, sizeof (complex_t **));
                inter.bs_list[i].q31 =
                        (complex_t ***) calloc(mem.num_fields + 1, sizeof (complex_t **));
                inter.bs_list[i].q24 =
                        (complex_t ***) calloc(mem.num_fields + 1, sizeof (complex_t **));
                inter.bs_list[i].q42 =
                        (complex_t ***) calloc(mem.num_fields + 1, sizeof (complex_t **));
                inter.bs_list[i].q34 =
                        (complex_t ***) calloc(mem.num_fields + 1, sizeof (complex_t **));
                inter.bs_list[i].q43 =
                        (complex_t ***) calloc(mem.num_fields + 1, sizeof (complex_t **));

                if (inter.bs_list[i].q12 == NULL ||
                        inter.bs_list[i].q21 == NULL ||
                        inter.bs_list[i].q13 == NULL ||
                        inter.bs_list[i].q31 == NULL ||
                        inter.bs_list[i].q24 == NULL ||
                        inter.bs_list[i].q42 == NULL ||
                        inter.bs_list[i].q34 == NULL ||
                        inter.bs_list[i].q43 == NULL) {
                    return (7);
                }
                *bytes += 8 * (mem.num_fields + 1) * sizeof (complex_t **);

                int j;
                for (j = 0; j < mem.num_fields; j++) {
                    inter.bs_list[i].q12[j] =
                            (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));
                    inter.bs_list[i].q21[j] =
                            (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));
                    inter.bs_list[i].q13[j] =
                            (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));
                    inter.bs_list[i].q31[j] =
                            (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));
                    inter.bs_list[i].q24[j] =
                            (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));
                    inter.bs_list[i].q42[j] =
                            (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));
                    inter.bs_list[i].q34[j] =
                            (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));
                    inter.bs_list[i].q43[j] =
                            (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));

                    if (inter.bs_list[i].q12[j] == NULL ||
                            inter.bs_list[i].q21[j] == NULL ||
                            inter.bs_list[i].q13[j] == NULL ||
                            inter.bs_list[i].q31[j] == NULL ||
                            inter.bs_list[i].q24[j] == NULL ||
                            inter.bs_list[i].q42[j] == NULL ||
                            inter.bs_list[i].q34[j] == NULL ||
                            inter.bs_list[i].q43[j] == NULL) {
                        return (7);
                    }
                    *bytes += 8 * (mem.num_fields + 1) * sizeof (complex_t *);

                }
            }
        }
    }

    return 0;
}

//! Allocate memory for gratings

/*!
 * \param bytes number of bytes so far allocated???
 *
 * \todo qcorr branch is untested
 */
int allocate_memory_for_grating_list(long int *bytes) {
    if (mem.num_gratings) {
        inter.grating_list =
                (grating_t *) calloc(mem.num_gratings + 1, sizeof (grating_t));
        if (inter.grating_list == NULL) {
            return (106);
        }
        *bytes += (mem.num_gratings + 1) * sizeof (grating_t);

        // field amplitudes
        int i;
        for (i = 0; i < mem.num_gratings; i++) {
            inter.grating_list[i].a11 =
                    (complex_t ***) calloc(mem.num_fields + 1, sizeof (complex_t **));
            inter.grating_list[i].a12 =
                    (complex_t ***) calloc(mem.num_fields + 1, sizeof (complex_t **));
            inter.grating_list[i].a13 =
                    (complex_t ***) calloc(mem.num_fields + 1, sizeof (complex_t **));
            inter.grating_list[i].a14 =
                    (complex_t ***) calloc(mem.num_fields + 1, sizeof (complex_t **));
            inter.grating_list[i].a21 =
                    (complex_t ***) calloc(mem.num_fields + 1, sizeof (complex_t **));
            inter.grating_list[i].a22 =
                    (complex_t ***) calloc(mem.num_fields + 1, sizeof (complex_t **));
            inter.grating_list[i].a23 =
                    (complex_t ***) calloc(mem.num_fields + 1, sizeof (complex_t **));
            inter.grating_list[i].a24 =
                    (complex_t ***) calloc(mem.num_fields + 1, sizeof (complex_t **));
            inter.grating_list[i].a31 =
                    (complex_t ***) calloc(mem.num_fields + 1, sizeof (complex_t **));
            inter.grating_list[i].a32 =
                    (complex_t ***) calloc(mem.num_fields + 1, sizeof (complex_t **));
            inter.grating_list[i].a33 =
                    (complex_t ***) calloc(mem.num_fields + 1, sizeof (complex_t **));
            inter.grating_list[i].a34 =
                    (complex_t ***) calloc(mem.num_fields + 1, sizeof (complex_t **));
            inter.grating_list[i].a41 =
                    (complex_t ***) calloc(mem.num_fields + 1, sizeof (complex_t **));
            inter.grating_list[i].a42 =
                    (complex_t ***) calloc(mem.num_fields + 1, sizeof (complex_t **));
            inter.grating_list[i].a43 =
                    (complex_t ***) calloc(mem.num_fields + 1, sizeof (complex_t **));
            inter.grating_list[i].a44 =
                    (complex_t ***) calloc(mem.num_fields + 1, sizeof (complex_t **));

            if (inter.grating_list[i].a11 == NULL ||
                    inter.grating_list[i].a12 == NULL ||
                    inter.grating_list[i].a21 == NULL ||
                    inter.grating_list[i].a22 == NULL ||
                    inter.grating_list[i].a13 == NULL ||
                    inter.grating_list[i].a23 == NULL ||
                    inter.grating_list[i].a31 == NULL ||
                    inter.grating_list[i].a32 == NULL ||
                    inter.grating_list[i].a33 == NULL ||
                    inter.grating_list[i].a14 == NULL ||
                    inter.grating_list[i].a41 == NULL ||
                    inter.grating_list[i].a24 == NULL ||
                    inter.grating_list[i].a42 == NULL ||
                    inter.grating_list[i].a34 == NULL ||
                    inter.grating_list[i].a43 == NULL ||
                    inter.grating_list[i].a44 == NULL) {
                return (107);
            }
            *bytes += 16 * (mem.num_fields + 1) * sizeof (complex_t **);

            int j;
            for (j = 0; j < mem.num_fields; j++) {
                inter.grating_list[i].a11[j] =
                        (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));
                inter.grating_list[i].a12[j] =
                        (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));
                inter.grating_list[i].a13[j] =
                        (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));
                inter.grating_list[i].a14[j] =
                        (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));
                inter.grating_list[i].a21[j] =
                        (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));
                inter.grating_list[i].a22[j] =
                        (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));
                inter.grating_list[i].a23[j] =
                        (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));
                inter.grating_list[i].a24[j] =
                        (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));
                inter.grating_list[i].a31[j] =
                        (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));
                inter.grating_list[i].a32[j] =
                        (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));
                inter.grating_list[i].a33[j] =
                        (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));
                inter.grating_list[i].a34[j] =
                        (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));
                inter.grating_list[i].a41[j] =
                        (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));
                inter.grating_list[i].a42[j] =
                        (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));
                inter.grating_list[i].a43[j] =
                        (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));
                inter.grating_list[i].a44[j] =
                        (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));

                if (inter.grating_list[i].a11[j] == NULL ||
                        inter.grating_list[i].a12[j] == NULL ||
                        inter.grating_list[i].a21[j] == NULL ||
                        inter.grating_list[i].a22[j] == NULL ||
                        inter.grating_list[i].a13[j] == NULL ||
                        inter.grating_list[i].a23[j] == NULL ||
                        inter.grating_list[i].a31[j] == NULL ||
                        inter.grating_list[i].a32[j] == NULL ||
                        inter.grating_list[i].a33[j] == NULL ||
                        inter.grating_list[i].a14[j] == NULL ||
                        inter.grating_list[i].a41[j] == NULL ||
                        inter.grating_list[i].a24[j] == NULL ||
                        inter.grating_list[i].a42[j] == NULL ||
                        inter.grating_list[i].a34[j] == NULL ||
                        inter.grating_list[i].a43[j] == NULL ||
                        inter.grating_list[i].a44[j] == NULL) {
                    return (108);
                }
                *bytes += 16 * (mem.num_fields + 1) * sizeof (complex_t *);
            }

            // ABCD matrices (4 by 4 matrix to store matrixes from 
            // each node to each node for up to 4 ports)
            inter.grating_list[i].qqt = (ABCD_t **) calloc(4, sizeof (ABCD_t *));
            if (inter.grating_list[i].qqt == NULL)
                return (108);
            for (j = 0; j < 4; j++) {
                inter.grating_list[i].qqt[j] = (ABCD_t *) calloc(4, sizeof (ABCD_t));
                if (inter.grating_list[i].qqt[j] == NULL)
                    return (108);
            }
            inter.grating_list[i].qqs = (ABCD_t **) calloc(4, sizeof (ABCD_t *));
            if (inter.grating_list[i].qqs == NULL)
                return (108);
            for (j = 0; j < 4; j++) {
                inter.grating_list[i].qqs[j] = (ABCD_t *) calloc(4, sizeof (ABCD_t));
                if (inter.grating_list[i].qqs[j] == NULL)
                    return (108);
            }
        }

        // quantum noise
        if (inter.qcorr) {
            int i;
            for (i = 0; i < mem.num_gratings; i++) {
                inter.grating_list[i].q11 =
                        (complex_t ***) calloc(mem.num_fields + 1, sizeof (complex_t **));
                inter.grating_list[i].q12 =
                        (complex_t ***) calloc(mem.num_fields + 1, sizeof (complex_t **));
                inter.grating_list[i].q13 =
                        (complex_t ***) calloc(mem.num_fields + 1, sizeof (complex_t **));
                inter.grating_list[i].q14 =
                        (complex_t ***) calloc(mem.num_fields + 1, sizeof (complex_t **));
                inter.grating_list[i].q21 =
                        (complex_t ***) calloc(mem.num_fields + 1, sizeof (complex_t **));
                inter.grating_list[i].q22 =
                        (complex_t ***) calloc(mem.num_fields + 1, sizeof (complex_t **));
                inter.grating_list[i].q23 =
                        (complex_t ***) calloc(mem.num_fields + 1, sizeof (complex_t **));
                inter.grating_list[i].q24 =
                        (complex_t ***) calloc(mem.num_fields + 1, sizeof (complex_t **));
                inter.grating_list[i].q31 =
                        (complex_t ***) calloc(mem.num_fields + 1, sizeof (complex_t **));
                inter.grating_list[i].q32 =
                        (complex_t ***) calloc(mem.num_fields + 1, sizeof (complex_t **));
                inter.grating_list[i].q33 =
                        (complex_t ***) calloc(mem.num_fields + 1, sizeof (complex_t **));
                inter.grating_list[i].q34 =
                        (complex_t ***) calloc(mem.num_fields + 1, sizeof (complex_t **));
                inter.grating_list[i].q41 =
                        (complex_t ***) calloc(mem.num_fields + 1, sizeof (complex_t **));
                inter.grating_list[i].q42 =
                        (complex_t ***) calloc(mem.num_fields + 1, sizeof (complex_t **));
                inter.grating_list[i].q43 =
                        (complex_t ***) calloc(mem.num_fields + 1, sizeof (complex_t **));
                inter.grating_list[i].q44 =
                        (complex_t ***) calloc(mem.num_fields + 1, sizeof (complex_t **));

                if (inter.grating_list[i].q11 == NULL ||
                        inter.grating_list[i].q12 == NULL ||
                        inter.grating_list[i].q21 == NULL ||
                        inter.grating_list[i].q22 == NULL ||
                        inter.grating_list[i].q13 == NULL ||
                        inter.grating_list[i].q23 == NULL ||
                        inter.grating_list[i].q31 == NULL ||
                        inter.grating_list[i].q32 == NULL ||
                        inter.grating_list[i].q33 == NULL ||
                        inter.grating_list[i].q14 == NULL ||
                        inter.grating_list[i].q41 == NULL ||
                        inter.grating_list[i].q24 == NULL ||
                        inter.grating_list[i].q42 == NULL ||
                        inter.grating_list[i].q34 == NULL ||
                        inter.grating_list[i].q43 == NULL ||
                        inter.grating_list[i].q44 == NULL) {
                    return (107);
                }
                *bytes += 16 * (mem.num_fields + 1) * sizeof (complex_t **);

                int j;
                for (j = 0; j < mem.num_fields; j++) {
                    inter.grating_list[i].q11[j] =
                            (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));
                    inter.grating_list[i].q12[j] =
                            (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));
                    inter.grating_list[i].q13[j] =
                            (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));
                    inter.grating_list[i].q14[j] =
                            (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));
                    inter.grating_list[i].q21[j] =
                            (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));
                    inter.grating_list[i].q22[j] =
                            (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));
                    inter.grating_list[i].q23[j] =
                            (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));
                    inter.grating_list[i].q24[j] =
                            (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));
                    inter.grating_list[i].q31[j] =
                            (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));
                    inter.grating_list[i].q32[j] =
                            (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));
                    inter.grating_list[i].q33[j] =
                            (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));
                    inter.grating_list[i].q34[j] =
                            (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));
                    inter.grating_list[i].q41[j] =
                            (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));
                    inter.grating_list[i].q42[j] =
                            (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));
                    inter.grating_list[i].q43[j] =
                            (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));
                    inter.grating_list[i].q44[j] =
                            (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));

                    if (inter.grating_list[i].q11[j] == NULL ||
                            inter.grating_list[i].q12[j] == NULL ||
                            inter.grating_list[i].q21[j] == NULL ||
                            inter.grating_list[i].q22[j] == NULL ||
                            inter.grating_list[i].q13[j] == NULL ||
                            inter.grating_list[i].q23[j] == NULL ||
                            inter.grating_list[i].q31[j] == NULL ||
                            inter.grating_list[i].q32[j] == NULL ||
                            inter.grating_list[i].q33[j] == NULL ||
                            inter.grating_list[i].q14[j] == NULL ||
                            inter.grating_list[i].q41[j] == NULL ||
                            inter.grating_list[i].q24[j] == NULL ||
                            inter.grating_list[i].q42[j] == NULL ||
                            inter.grating_list[i].q34[j] == NULL ||
                            inter.grating_list[i].q43[j] == NULL ||
                            inter.grating_list[i].q44[j] == NULL) {
                        return (107);
                    }
                    *bytes += 16 * (mem.num_fields + 1) * sizeof (complex_t *);

                }
            }
        }
    }

    return 0;
}

//! Allocate memory for squeezers

/*!
 * \param bytes number of bytes so far allocated???
 *
 * \todo untested
 */
int allocate_memory_for_squeezer_list(long int *bytes) {
    if (mem.num_squeezers) {
        inter.squeezer_list =
                (squeezer_t *) calloc(mem.num_squeezers + 1, sizeof (squeezer_t));

        if (inter.squeezer_list == NULL) {
            return (3);
        }
        *bytes += (mem.num_squeezers + 1) * sizeof (squeezer_t);

        int i;
        for (i = 0; i < mem.num_squeezers; i++) {
            // field amplitudes
            inter.squeezer_list[i].a11 =
                    (complex_t ***) calloc(mem.num_fields + 1, sizeof (complex_t **));
            inter.squeezer_list[i].a12 =
                    (complex_t ***) calloc(mem.num_fields + 1, sizeof (complex_t **));
            inter.squeezer_list[i].a21 =
                    (complex_t ***) calloc(mem.num_fields + 1, sizeof (complex_t **));
            inter.squeezer_list[i].a22 =
                    (complex_t ***) calloc(mem.num_fields + 1, sizeof (complex_t **));

            if (inter.squeezer_list[i].a11 == NULL ||
                    inter.squeezer_list[i].a12 == NULL ||
                    inter.squeezer_list[i].a21 == NULL ||
                    inter.squeezer_list[i].a22 == NULL) {
                return (4);
            }
            *bytes += 4 * (mem.num_fields + 1) * sizeof (complex_t **);

            int j;
            for (j = 0; j < mem.num_fields; j++) {
                // field amplitudes
                inter.squeezer_list[i].a11[j] =
                        (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));
                inter.squeezer_list[i].a12[j] =
                        (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));
                inter.squeezer_list[i].a21[j] =
                        (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));
                inter.squeezer_list[i].a22[j] =
                        (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));

                if (inter.squeezer_list[i].a11[j] == NULL ||
                        inter.squeezer_list[i].a12[j] == NULL ||
                        inter.squeezer_list[i].a21[j] == NULL ||
                        inter.squeezer_list[i].a22[j] == NULL) {
                    return (5);
                }
                *bytes += 4 * (mem.num_fields + 1) * sizeof (complex_t *);
            }
        }

        // quantum noise stuff
        if (inter.qcorr) {
            int i;
            for (i = 0; i < mem.num_squeezers; i++) {
                inter.squeezer_list[i].q11 =
                        (complex_t ***) calloc(mem.num_fields + 1, sizeof (complex_t **));
                inter.squeezer_list[i].q12 =
                        (complex_t ***) calloc(mem.num_fields + 1, sizeof (complex_t **));
                inter.squeezer_list[i].q21 =
                        (complex_t ***) calloc(mem.num_fields + 1, sizeof (complex_t **));
                inter.squeezer_list[i].q22 =
                        (complex_t ***) calloc(mem.num_fields + 1, sizeof (complex_t **));

                if (inter.squeezer_list[i].q11 == NULL ||
                        inter.squeezer_list[i].q12 == NULL ||
                        inter.squeezer_list[i].q21 == NULL ||
                        inter.squeezer_list[i].q22 == NULL) {
                    return (4);
                }

                *bytes += 4 * (mem.num_fields + 1) * sizeof (complex_t **);

                int j;
                for (j = 0; j < mem.num_fields; j++) {
                    inter.squeezer_list[i].q11[j] =
                            (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));
                    inter.squeezer_list[i].q12[j] =
                            (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));
                    inter.squeezer_list[i].q21[j] =
                            (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));
                    inter.squeezer_list[i].q22[j] =
                            (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));

                    if (inter.squeezer_list[i].q11[j] == NULL ||
                            inter.squeezer_list[i].q12[j] == NULL ||
                            inter.squeezer_list[i].q21[j] == NULL ||
                            inter.squeezer_list[i].q22[j] == NULL) {
                        return (4);
                    }
                    *bytes += 4 * (mem.num_fields + 1) * sizeof (complex_t *);

                }
            }
        }
    }

    return 0;
}

//! Allocate memory for spaces

/*!
 * \param bytes number of bytes so far allocated???
 *
 * \todo qcorr branch is untested
 */
int allocate_memory_for_space_list(long int *bytes) {
    if (mem.num_spaces) {
        inter.space_list = (space_t *) calloc(mem.num_spaces + 1, sizeof (space_t));
        if (inter.space_list == NULL) {
            return (9);
        }
        *bytes += (mem.num_spaces + 1) * sizeof (space_t);

        // field amplitudes
        int i;
        for (i = 0; i < mem.num_spaces; i++) {
            inter.space_list[i].a12 = (complex_t ***) calloc(mem.num_fields + 1, sizeof (complex_t **));
            inter.space_list[i].a21 = (complex_t ***) calloc(mem.num_fields + 1, sizeof (complex_t **));

            inter.space_list[i].k12 =
                    (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));
            inter.space_list[i].k21 =
                    (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));

            if (inter.space_list[i].a12 == NULL ||
                    inter.space_list[i].a21 == NULL ||

                    inter.space_list[i].k12 == NULL ||
                    inter.space_list[i].k21 == NULL) {
                return (10);
            }
            *bytes += 2 * (mem.num_fields + 1) * sizeof (complex_t **);
            *bytes += 2 * (mem.num_fields + 1) * sizeof (complex_t *);

            int j;
            for (j = 0; j < mem.num_fields; j++) {
                inter.space_list[i].a12[j] = (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));
                inter.space_list[i].a21[j] = (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));

                inter.space_list[i].k12[j] =
                        (complex_t *) calloc(mem.num_fields + 1, sizeof (complex_t));
                inter.space_list[i].k21[j] =
                        (complex_t *) calloc(mem.num_fields + 1, sizeof (complex_t));

                if (inter.space_list[i].a12[j] == NULL ||
                        inter.space_list[i].a21[j] == NULL ||

                        inter.space_list[i].k12[j] == NULL ||
                        inter.space_list[i].k21[j] == NULL) {
                    return (11);
                }
                *bytes += 2 * (mem.num_fields + 1) * sizeof (complex_t *);
                *bytes += 2 * (mem.num_fields + 1) * sizeof (complex_t);
            }
        }

        // quantum noise
        if (inter.qcorr) {
            int i;
            for (i = 0; i < mem.num_spaces; i++) {
                inter.space_list[i].q12 =
                        (complex_t ***) calloc(mem.num_fields + 1, sizeof (complex_t **));
                inter.space_list[i].q21 =
                        (complex_t ***) calloc(mem.num_fields + 1, sizeof (complex_t **));

                if (inter.space_list[i].q12 == NULL ||
                        inter.space_list[i].q21 == NULL) {
                    return (10);
                }
                *bytes += 2 * (mem.num_fields + 1) * sizeof (complex_t **);

                int j;
                for (j = 0; j < mem.num_fields; j++) {
                    inter.space_list[i].q12[j] =
                            (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));
                    inter.space_list[i].q21[j] =
                            (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));

                    if (inter.space_list[i].q12[j] == NULL ||
                            inter.space_list[i].q21[j] == NULL) {
                        return (10);
                    }
                    *bytes += 2 * (mem.num_fields + 1) * sizeof (complex_t *);

                }
            }
        }
    }
    return 0;
}

//! Allocate memory for light inputs

/*!
 * \param bytes number of bytes so far allocated???
 */
int allocate_memory_for_light_input_list(long int *bytes) {
    if (mem.num_light_inputs) {
        inter.light_in_list =
                (light_in_t *) calloc(mem.num_light_inputs + 1, sizeof (light_in_t));

        if (inter.light_in_list == NULL) {
            return (12);
        }
        *bytes += (mem.num_light_inputs + 1) * sizeof (light_in_t);

        int i;
        for (i = 0; i < mem.num_light_inputs; i++) {
            inter.light_in_list[i].power_coeff_list =
                    (complex_t *) calloc(mem.num_fields + 1, sizeof (complex_t));

            if (inter.light_in_list[i].power_coeff_list == NULL) {
                return (13);
            }
            *bytes += (mem.num_fields + 1) * sizeof (complex_t);
        }
    }
    return 0;
}

//! Allocate memory for detectors

/*!
 * \param bytes number of bytes so far allocated???
 */
int allocate_memory_for_detector_list(long int *bytes) {
    if (mem.num_light_outputs) {
        inter.light_out_list =
                (light_out_t *) calloc(mem.num_light_outputs + 1, sizeof (light_out_t));
        if (inter.light_out_list == NULL) {
            return (14);
        }
        *bytes += (mem.num_light_outputs + 1) * sizeof (light_out_t);

        int i;
        for (i = 0; i < mem.num_light_outputs; i++) {
            inter.light_out_list[i].masking_factor =
                    (double *) calloc(mem.num_fields + 1, sizeof (double));

            if (inter.light_out_list[i].masking_factor == NULL) {
                return (15);
            }
            *bytes += (mem.num_fields + 1) * sizeof (double);
        }
    }
    return 0;
}

//! Allocate memory for output data

/*!
 * \param bytes number of bytes so far allocated???
 */
int allocate_memory_for_output_data_list(long int *bytes) {
    if (mem.num_outputs) {
        inter.output_data_list =
                (output_data_t *) calloc(mem.num_outputs + 1, sizeof (output_data_t));
        if (inter.output_data_list == NULL) {
            return (80);
        }
        *bytes += (mem.num_outputs + 1) * sizeof (output_data_t);

        /*
          for (i=0;i<mem.num_outputs;i++)
          {
          inter.output_data_list[i].a0=(complex*)malloc((size_t)(mem.num_fields+1)*sizeof(complex_t));
          if (inter.output_data_list[i].a0==NULL)
          return (81);
         *bytes+=(mem.num_fields+1)*sizeof(complex_t);
          }
         */
    }
    return 0;
}

//! Allocate memory for beam parameter outputs

/*!
 * \param bytes number of bytes so far allocated???
 */
int allocate_memory_for_beam_param_list(long int *bytes) {
    if (mem.num_beam_param_outputs) {
        inter.beampar_out_list = (beampar_out_t *)
                calloc(mem.num_beam_param_outputs + 1, sizeof (beampar_out_t));
        if (inter.beampar_out_list == NULL) {
            return (82);
        }
        *bytes += (mem.num_beam_param_outputs + 1) * sizeof (beampar_out_t);
    }
    return 0;
}
//! Allocate memory for cavity parameter outputs

/*!
 */
int allocate_memory_for_cavity_param_list(long int *bytes) {
    if (mem.num_cavity_param_outputs) {
        inter.cavity_par_out_list = (cavity_par_out_t *)
                calloc(mem.num_cavity_param_outputs + 1, sizeof (cavity_par_out_t));
        if (inter.cavity_par_out_list == NULL) {
            return (182);
        }
        *bytes += (mem.num_cavity_param_outputs + 1) * sizeof (cavity_par_out_t);
    }
    return 0;
}

//! Allocate memory for convolution outputs

/*!
 */
int allocate_memory_for_convolution_list(long int *bytes) {
    if (mem.num_convolution_outputs) {
        inter.convolution_out_list = (convolution_out_t *)
                calloc(mem.num_convolution_outputs + 1, sizeof (convolution_out_t));
        if (inter.convolution_out_list == NULL) {
            return (982);
        }
        *bytes += (mem.num_convolution_outputs + 1) * sizeof (convolution_out_t);
    }
    return 0;
}

//! Allocate memory for mirror_phase outputs

/*!
 */
int allocate_memory_for_mirror_phase_list(long int *bytes) {
    if (mem.num_mirror_phase_outputs) {
        inter.mirror_phase_out_list = (mirror_phase_out_t *)
                calloc(mem.num_mirror_phase_outputs + 1, sizeof (mirror_phase_out_t));
        if (inter.mirror_phase_out_list == NULL) {
            return (983);
        }
        *bytes += (mem.num_mirror_phase_outputs + 1) * sizeof (mirror_phase_out_t);
    }
    return 0;
}

//! Allocate memory for modulators

/*!
 * \param bytes number of bytes so far allocated???
 *
 * \todo qcorr branch is untested
 */
int allocate_memory_for_modulator_list(long int *bytes) {
    if (mem.num_modulators) {
        inter.modulator_list = (modulator_t *)
                calloc(mem.num_modulators + 1, sizeof (modulator_t));
        if (inter.modulator_list == NULL) {
            return (16);
        }
        *bytes += (mem.num_modulators + 1) * sizeof (modulator_t);

        // field amplitudes
        int i;
        for (i = 0; i < mem.num_modulators; i++) {
            inter.modulator_list[i].a12 =
                    (complex_t ***) calloc(mem.num_fields + 1, sizeof (complex_t **));
            inter.modulator_list[i].a21 =
                    (complex_t ***) calloc(mem.num_fields + 1, sizeof (complex_t **));

            // coupling coefficients
            inter.modulator_list[i].k12 =
                    (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));
            inter.modulator_list[i].k21 =
                    (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));

            inter.modulator_list[i].a1 =
                    (complex_t *) calloc(mem.num_fields + 1, sizeof (complex_t));
            inter.modulator_list[i].a2 =
                    (complex_t *) calloc(mem.num_fields + 1, sizeof (complex_t));

            if (inter.modulator_list[i].a12 == NULL ||
                    inter.modulator_list[i].a21 == NULL ||

                    inter.modulator_list[i].a2 == NULL ||
                    inter.modulator_list[i].a1 == NULL ||

                    inter.modulator_list[i].k12 == NULL ||
                    inter.modulator_list[i].k21 == NULL) {
                return (17);
            }
            *bytes += 2 * (mem.num_fields + 1) * sizeof (complex_t **);
            *bytes += 2 * (mem.num_fields + 1) * sizeof (complex_t *);
            *bytes += 2 * (mem.num_fields + 1) * sizeof (complex_t);

            int j;
            for (j = 0; j < mem.num_fields; j++) {
                inter.modulator_list[i].a12[j] =
                        (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));
                inter.modulator_list[i].a21[j] =
                        (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));

                inter.modulator_list[i].k12[j] =
                        (complex_t *) calloc(mem.num_fields + 1, sizeof (complex_t));
                inter.modulator_list[i].k21[j] =
                        (complex_t *) calloc(mem.num_fields + 1, sizeof (complex_t));

                if (inter.modulator_list[i].a12[j] == NULL ||
                        inter.modulator_list[i].a21[j] == NULL ||
                        inter.modulator_list[i].k12[j] == NULL ||
                        inter.modulator_list[i].k21[j] == NULL) {
                    return (18);
                }
                *bytes += 2 * (mem.num_fields + 1) * sizeof (complex_t *);
                *bytes += 2 * (mem.num_fields + 1) * sizeof (complex_t);
            }
        }

        // quantum noise
        if (inter.qcorr) {
            int i;
            for (i = 0; i < mem.num_modulators; i++) {
                inter.modulator_list[i].q12 =
                        (complex_t ***) calloc(mem.num_fields + 1, sizeof (complex_t **));
                inter.modulator_list[i].q21 =
                        (complex_t ***) calloc(mem.num_fields + 1, sizeof (complex_t **));

                if (inter.modulator_list[i].q12 == NULL ||
                        inter.modulator_list[i].q21 == NULL) {
                    return (17);
                }
                *bytes += 2 * (mem.num_fields + 1) * sizeof (complex_t **);

                int j;
                for (j = 0; j < mem.num_fields; j++) {
                    inter.modulator_list[i].q12[j] =
                            (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));
                    inter.modulator_list[i].q21[j] =
                            (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));

                    if (inter.modulator_list[i].q12[j] == NULL ||
                            inter.modulator_list[i].q21[j] == NULL) {
                        return (17);
                    }
                    *bytes += 2 * (mem.num_fields + 1) * sizeof (complex_t *);

                }
            }
        }
    }
    return 0;
}

//! Allocate memory for signals

/*!
 * \param bytes number of bytes so far allocated???
 */
int allocate_memory_for_signal_list(long int *bytes) {
    if (mem.num_signal_inputs) {
        inter.signal_list = (signal_t *)
                calloc(mem.num_signal_inputs + 1, sizeof (signal_t));
        if (inter.signal_list == NULL) {
            return (19);
        }
        *bytes += (mem.num_signal_inputs + 1) * sizeof (signal_t);

        int i;
        for (i = 0; i < mem.num_signal_inputs; i++) {
            inter.signal_list[i].a1 =
                    (complex_t *) calloc(mem.num_fields + 1, sizeof (complex_t));
            inter.signal_list[i].a2 =
                    (complex_t *) calloc(mem.num_fields + 1, sizeof (complex_t));
            inter.signal_list[i].a3 =
                    (complex_t *) calloc(mem.num_fields + 1, sizeof (complex_t));
            inter.signal_list[i].a4 =
                    (complex_t *) calloc(mem.num_fields + 1, sizeof (complex_t));

            inter.signal_list[i].ao1 =
                    (complex_t *) calloc(mem.num_fields + 1, sizeof (complex_t));
            inter.signal_list[i].ao2 =
                    (complex_t *) calloc(mem.num_fields + 1, sizeof (complex_t));
            inter.signal_list[i].ao3 =
                    (complex_t *) calloc(mem.num_fields + 1, sizeof (complex_t));
            inter.signal_list[i].ao4 =
                    (complex_t *) calloc(mem.num_fields + 1, sizeof (complex_t));

            if (inter.signal_list[i].a1 == NULL ||
                    inter.signal_list[i].a2 == NULL ||
                    inter.signal_list[i].a3 == NULL ||
                    inter.signal_list[i].a4 == NULL ||

                    inter.signal_list[i].ao1 == NULL ||
                    inter.signal_list[i].ao2 == NULL ||
                    inter.signal_list[i].ao3 == NULL ||
                    inter.signal_list[i].ao4 == NULL) {
                return (20);
            }
            *bytes += 8 * (mem.num_fields + 1) * sizeof (complex_t);
        }
    }
    return 0;
}

//! Allocate memory for derivatives

/*!
 * \param bytes number of bytes so far allocated???
 */
int allocate_memory_for_derivative_list(long int *bytes) {
    if (mem.num_diff_cmds) {
        inter.deriv_list = (derivative_t *) calloc(mem.num_diff_cmds + 1, sizeof (derivative_t));
        if (inter.deriv_list == NULL) {
            return (21);
        }
        *bytes += (mem.num_diff_cmds + 1) * sizeof (derivative_t);
    }
    return 0;
}

//! Allocate memory for scale commands

/*!
 * \param bytes number of bytes so far allocated???
 */
int allocate_memory_for_scale_cmd_list(long int *bytes) {
    if (mem.num_scale_cmds) {
        inter.scale_list =
                (scale_t *) calloc(mem.num_scale_cmds + 1, sizeof (scale_t));
        if (inter.scale_list == NULL) {
            return (22);
        }
        *bytes += (mem.num_scale_cmds + 1) * sizeof (scale_t);
    }
    return 0;
}

//! Allocate memory for variable commands

/*!
 * \param bytes number of bytes so far allocated???
 */
int allocate_memory_for_variable_list(long int *bytes) {
    if (mem.num_variables) {
        inter.variable_list =
                (variable_t *) calloc(mem.num_variables + 1, sizeof (variable_t));
        if (inter.variable_list == NULL) {
            return (22); // TODO fix double number
        }
        *bytes += (mem.num_variables + 1) * sizeof (variable_t);
    }
    return 0;
}


//! Allocate memory for diodes

/*!
 * \param bytes number of bytes so far allocated???
 *
 * \todo qcorr branch is untested
 */
int allocate_memory_for_diode_list(long int *bytes) {
    if (mem.num_diodes) {
        inter.diode_list = (diode_t *) calloc(mem.num_diodes + 1, sizeof (diode_t));
        if (inter.diode_list == NULL) {
            return (23);
        }

        *bytes += (mem.num_diodes + 1) * sizeof (diode_t);

        // field amplitudes
        int i;
        for (i = 0; i < mem.num_diodes; i++) {
            inter.diode_list[i].a12 =
                    (complex_t ***) calloc(mem.num_fields + 1, sizeof (complex_t **));
            inter.diode_list[i].a21 =
                    (complex_t ***) calloc(mem.num_fields + 1, sizeof (complex_t **));

            // coupling coefficients
            inter.diode_list[i].k12 =
                    (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));
            inter.diode_list[i].k21 =
                    (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));

            if (inter.diode_list[i].a12 == NULL ||
                    inter.diode_list[i].a21 == NULL ||

                    inter.diode_list[i].k12 == NULL ||
                    inter.diode_list[i].k21 == NULL) {
                return (24);
            }

            *bytes += 2 * (mem.num_fields + 1) * sizeof (complex_t **);
            *bytes += 2 * (mem.num_fields + 1) * sizeof (complex_t *);

            int j;
            for (j = 0; j < mem.num_fields; j++) {
                inter.diode_list[i].a12[j] =
                        (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));
                inter.diode_list[i].a21[j] =
                        (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));

                inter.diode_list[i].k12[j] =
                        (complex_t *) calloc(mem.num_fields + 1, sizeof (complex_t));
                inter.diode_list[i].k21[j] =
                        (complex_t *) calloc(mem.num_fields + 1, sizeof (complex_t));

                if (inter.diode_list[i].a12[j] == NULL ||
                        inter.diode_list[i].a21[j] == NULL ||

                        inter.diode_list[i].k12[j] == NULL ||
                        inter.diode_list[i].k21[j] == NULL) {
                    return (25);
                }

                *bytes += 2 * (mem.num_fields + 1) * sizeof (complex_t *);
                *bytes += 2 * (mem.num_fields + 1) * sizeof (complex_t);
            }
        }

        // quantum noise
        if (inter.qcorr) {
            int i;
            for (i = 0; i < mem.num_diodes; i++) {
                inter.diode_list[i].q12 =
                        (complex_t ***) calloc(mem.num_fields + 1, sizeof (complex_t **));
                inter.diode_list[i].q21 =
                        (complex_t ***) calloc(mem.num_fields + 1, sizeof (complex_t **));

                if (inter.diode_list[i].q12 == NULL ||
                        inter.diode_list[i].q21 == NULL) {
                    return (24);
                }
                *bytes += 2 * (mem.num_fields + 1) * sizeof (complex_t **);

                int j;
                for (j = 0; j < mem.num_fields; j++) {
                    inter.diode_list[i].q12[j] =
                            (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));
                    inter.diode_list[i].q21[j] =
                            (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));

                    if (inter.diode_list[i].q12[j] == NULL ||
                            inter.diode_list[i].q21[j] == NULL) {
                        return (24);
                    }
                    *bytes += 2 * (mem.num_fields + 1) * sizeof (complex_t *);

                }
            }
        }
    }
    return 0;
}

//! Allocate memory for lenses

/*!
 * \param bytes number of bytes so far allocated???
 *
 * \todo qcorr branch is untested
 */
int allocate_memory_for_lens_list(long int *bytes) {
    if (mem.num_lenses) {
        inter.lens_list = (lens_t *) calloc(mem.num_lenses + 1, sizeof (lens_t));
        if (inter.lens_list == NULL) {
            return (26);
        }
        *bytes += (mem.num_lenses + 1) * sizeof (lens_t);

        // field amplitudes
        int i;
        for (i = 0; i < mem.num_lenses; i++) {
            inter.lens_list[i].a12 =
                    (complex_t ***) calloc(mem.num_fields + 1, sizeof (complex_t **));
            inter.lens_list[i].a21 =
                    (complex_t ***) calloc(mem.num_fields + 1, sizeof (complex_t **));

            // coupling coefficients
            inter.lens_list[i].k12 =
                    (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));
            inter.lens_list[i].k21 =
                    (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));

            if (inter.lens_list[i].a12 == NULL ||
                    inter.lens_list[i].a21 == NULL ||

                    inter.lens_list[i].k12 == NULL ||
                    inter.lens_list[i].k21 == NULL) {
                return (27);
            }
            *bytes += 2 * (mem.num_fields + 1) * sizeof (complex_t **);
            *bytes += 2 * (mem.num_fields + 1) * sizeof (complex_t *);

            int j;
            for (j = 0; j < mem.num_fields; j++) {
                inter.lens_list[i].a12[j] =
                        (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));
                inter.lens_list[i].a21[j] =
                        (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));

                inter.lens_list[i].k12[j] =
                        (complex_t *) calloc(mem.num_fields + 1, sizeof (complex_t));
                inter.lens_list[i].k21[j] =
                        (complex_t *) calloc(mem.num_fields + 1, sizeof (complex_t));

                if (inter.lens_list[i].a12[j] == NULL ||
                        inter.lens_list[i].a21[j] == NULL ||

                        inter.lens_list[i].k12[j] == NULL ||
                        inter.lens_list[i].k21[j] == NULL) {
                    return (28);
                }

                *bytes += 2 * (mem.num_fields + 1) * sizeof (complex_t *);
                *bytes += 2 * (mem.num_fields + 1) * sizeof (complex_t);
            }
        }

        // quantum noise
        if (inter.qcorr) {
            int i;
            for (i = 0; i < mem.num_lenses; i++) {
                inter.lens_list[i].q12 =
                        (complex_t ***) calloc(mem.num_fields + 1, sizeof (complex_t **));
                inter.lens_list[i].q21 =
                        (complex_t ***) calloc(mem.num_fields + 1, sizeof (complex_t **));

                if (inter.lens_list[i].q12 == NULL ||
                        inter.lens_list[i].q21 == NULL) {
                    return (27);
                }
                *bytes += 2 * (mem.num_fields + 1) * sizeof (complex_t **);

                int j;
                for (j = 0; j < mem.num_fields; j++) {
                    inter.lens_list[i].q12[j] =
                            (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));
                    inter.lens_list[i].q21[j] =
                            (complex_t **) calloc(mem.num_fields + 1, sizeof (complex_t *));

                    if (inter.lens_list[i].q12[j] == NULL ||
                            inter.lens_list[i].q21[j] == NULL) {
                        return (27);
                    }
                    *bytes += 2 * (mem.num_fields + 1) * sizeof (complex_t *);

                }
            }
        }
    }
    return 0;
}

//! Allocate memory for locks

/*!
 * \param bytes number of bytes so far allocated???
 */
int allocate_memory_for_locks(long int *bytes) {
    if (mem.num_locks) {
        inter.lock_list = (lock_command_t *) calloc(mem.num_locks + 1, sizeof (lock_command_t));
        if (inter.lock_list == NULL) {
            return (88);
        }

        *bytes += (mem.num_locks + 1) * sizeof (lock_command_t);

        inter.lock_order = (int *) calloc(mem.num_locks + 1, sizeof (int));
        if (inter.lock_order == NULL) {
            return (88);
        }

        inter.breakcount = (int *) calloc(mem.num_locks + 1, sizeof (int));
        if (inter.breakcount == NULL) {
            return (86);
        }
    }
    return 0;
}

//! Allocate memory for functions

/*!
 * \param bytes number of bytes so far allocated???
 */
int allocate_memory_for_function_list(long int *bytes) {
    if (mem.num_func_cmds) {
        inter.function_list = (func_command_t *)
                calloc(mem.num_func_cmds + 1, sizeof (func_command_t));

        if (inter.function_list == NULL) {
            return (87);
        }
        *bytes += (mem.num_func_cmds + 1) * sizeof (func_command_t);

        inter.func_order = (int *) calloc(mem.num_func_cmds + 1, sizeof (int));
        if (inter.func_order == NULL) {
            return (87);
        }
    }
    return 0;
}

//! Allocate memory for signal amplitudes

/*!
 * \param bytes1 number of bytes so far allocated???
 */
int allocate_memory_for_amplitude_list(long int *bytes1) {
    if (mem.num_fields) {
        a_s = (complex_t ***) calloc(mem.num_fields + 1, sizeof (complex_t **));

        if (a_s == NULL) {
            return (37);
        }
        *bytes1 += (mem.num_fields + 1) * sizeof (complex_t **);

        int i;
        for (i = 0; i < mem.num_fields; i++) {
            a_s[i] = (complex_t **)
                    calloc(mem.num_light_outputs + 1, sizeof (complex_t *));

            if (a_s[i] == NULL) {
                return (38);
            }
            *bytes1 += (mem.num_light_outputs + 1) * sizeof (complex_t *);

            int j;
            for (j = 0; j < mem.num_light_outputs; j++) {
                a_s[i][j] = (complex_t *)
                        calloc(mem.num_frequencies + 1, sizeof (complex_t));
                if (a_s[i][j] == NULL) {
                    return (39);
                }
                *bytes1 += (mem.num_frequencies + 1) * sizeof (complex_t);
            }
        }

        mem.all_tem = (int *) calloc((mem.num_fields + 1)*2, sizeof (int));
        if (mem.all_tem == NULL) {
            return (99);
        }

        *bytes1 += (mem.num_fields + 1) * 2 * sizeof (int);
    }
    return 0;
}

//! Allocate memory for quantum noise amplitudes

/*!
 * \param bytes1 number of bytes so far allocated???
 */
int allocate_memory_for_quantum_amplitude_list(long int *bytes1) {
    if (inter.qcorr && mem.num_fields) {
        qnoise = (complex_t ****) calloc(mem.num_fields + 1, sizeof (complex_t ***));

        if (qnoise == NULL) {
            return (41);
        }
        *bytes1 += (mem.num_fields + 1) * sizeof (complex_t ***);

        int i;
        for (i = 0; i < mem.num_fields; i++) {
            qnoise[i] = (complex_t ***)
                    calloc(mem.num_light_outputs + 1, sizeof (complex_t **));

            if (qnoise[i] == NULL) {
                return (42);
            }
            *bytes1 += (mem.num_light_outputs + 1) * sizeof (complex_t **);

            int j;
            for (j = 0; j < mem.num_light_outputs; j++) {
                qnoise[i][j] = (complex_t **)
                        calloc(mem.num_frequencies + 1, sizeof (complex_t *));

                if (qnoise[i][j] == NULL) {
                    return (43);
                }
                *bytes1 += (mem.num_frequencies + 1) * sizeof (complex_t *);

                int k;
                for (k = 0; k < mem.num_frequencies; k++) {
                    qnoise[i][j][k] = (complex_t *)
                            calloc(2 * mem.num_nodes + 1, sizeof (complex_t));

                    if (qnoise[i][j][k] == NULL) {
                        return (44);
                    }
                    *bytes1 += (2 * mem.num_nodes + 1) * sizeof (complex_t);
                }
            }
        }
    }
    return 0;
}

//! Make the memory for solution data

/*!
 *
 */
void make_space(void) {
    if (mem.restab != NULL) {
        bug_error("memory allocation failure (1)");
    }

    if ((mem.restab = calloc(inter.num_output_cols * (inter.x2.xsteps + 1) *
            (inter.x1.xsteps + 1), sizeof (double))) == NULL) {
        gerror("make_space: memory allocation failure (2)\n");
    }
}

void read_lambda(const char *command_string) {
    char command_name[MAX_TOKEN_LEN] = {0};
    char value_string[MAX_TOKEN_LEN] = {0};
    double value;
    char rest_string[REST_STRING_LEN] = {0};


    int num_vars_read = sscanf(command_string, "%s %s %80s",
            command_name, value_string, rest_string);
    int num_vars_expected = 2;
    if ((num_vars_read < num_vars_expected) || atod(value_string, &value)) {
        gerror("line '%s':\nexpected 'lambda value'\n", command_string);
    }
    warn("overwriting default wavelength with lambda=%sm\n", double_form(value));
    init.lambda = value;
    inter.f0 = init.clight / init.lambda;
}

/*
 * Local variables:
 * c-indentation-style: bsd
 * c-basic-offset: 2
 * indent-tabs-mode: nil
 * End:
 *
 * vim: tabstop=2 expandtab shiftwidth=2:
 */
