// $Id$

/*!
 * \file kat_quant.c
 * \brief Quantum noise and radiation pressure specific routines
 * 
 * @section Copyright notice
 *
 *  This file is part of the interferometer simulation Finesse
 *  http://www.gwoptics.org/finesse
 *
 *  Copyright (C) 1999 onwards Andreas Freise
 *  with parts of the code written by Daniel Brown, Paul Cochrane
 *  and Gerhard Heinzel.
 *
 *  This program is free software; you can redistribute it and/or modify it under
 *  the terms of the GNU General Public License version 3 as published
 *  by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 *  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *  See the GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along with
 *  this library; if not, write to the Free Software Foundation, Inc., 59 Temple Place,
 *  Suite 330, Boston, MA 02111-1307 USA
 */


#include "kat.h"
#include "kat_inline.c"
#include "kat_quant.h"
#include "kat_spa.h"
#include "kat_io.h"
#include "kat_aux.h"
#include "kat_calc.h"
#include "kat_aa.h"

extern const complex_t complex_i;
extern const complex_t complex_1;
extern const complex_t complex_m1;
extern const complex_t complex_0;

//extern init_variables_t in;
extern interferometer_t inter;
extern sparse_var_t spa;

//! Quantum noise list [field] [detector] [frequency] [equation]
extern complex_t ****qnoise;

extern double *f_s; //!< Frequency value list [frequency]
extern complex_t ***a_s; //!< Amplitude list [field] [output] [frequency]
extern int *t_s; //!< Type_of_Signal list [frequency]

//! Calculate quantum noise corrections for one point on x-axis

/*!
 *
 * \todo signals branch is untested
 * \todo modulation branch is untested
 */
void quantum_point(void) {
    double f, f1, f2;
    light_in_t *light_in;

    //! \todo build a frequency list
    // what should the frequency list look like?
    // how does one build it?
    // what does one look for?
    // where does one look?
    //   - the input light list to start with
    //   - what about sidebands?
    // but there's already a frequency list built for data_point()...

    // initialise frequency index
    int frequency_index = 0;

    // loop over all carriers
    int input_index;
    for (input_index = 0;
            input_index < inter.num_light_inputs;
            input_index++) {
        // grab the current input light field
        light_in = &(inter.light_in_list[input_index]);

        // get the carrier freq and assign to temporary freq variable for later use
        f = light_in->f0;
        f1 = f;

        // loop over the nodes
        int node_index;
        for (node_index = 0; node_index < inter.num_nodes; node_index++) {

            // fill the sparse matrix at the given frequency and clear the RHS vector
            spa_fill(f1, QCORR);
            spa_rhs_clear(QCORR);

            // get the noise at the given node index for all *inputs*
            const double input_node_noise = get_input_noise_at(node_index);

            debug_msg("input node_noise at node %s = %g\n",
                    get_node_name(node_index), input_node_noise);

            // fill the RHS vector for all fields
            int field_index;
            for (field_index = 0; field_index < inter.num_fields; field_index++) {
                // calculate the correct element index
                int rhs_element_index = node_index * 2 * inter.num_fields + 1 +
                        2 * field_index;

                // check that rhs_element_index doesn't get too big
                assert(rhs_element_index < 2 * inter.num_eqns);

                // assign the noise value to the relevant element of the RHS vector
                spa_rhs(rhs_element_index, co(input_node_noise, 0.0), QCORR);
            }

#if DEBUG
            dump_matrix(QCORR, "stdout");
#endif

            // factor the matrix and solve it
            spa_factor(QCORR);
            spa_sol(QCORR);

#ifdef DEBUG
            dump_rhs_vector(QCORR, "stdout");
            dump_solution_vector(QCORR, "stdout");
#endif

            // save the solution for later calculations
            store_quantum_solution(f, frequency_index, 2 * node_index, CARRIER_FIELD);

            // fill the sparse matrix at the given frequency and clear the RHS vector
            spa_fill(f1, QCORR);
            spa_rhs_clear(QCORR);

            // get the noise at the given node index for all *outputs*
            const double output_node_noise = get_output_noise_at(node_index);

            debug_msg("output node_noise at node %s = %g\n",
                    get_node_name(node_index), output_node_noise);

            // fill the RHS vector for all fields
            for (field_index = 0; field_index < inter.num_fields; field_index++) {
                // calculate the correct element index
                int rhs_element_index = node_index * 2 * inter.num_fields +
                        2 * field_index;

                // check that rhs_element_index doesn't get too big
                assert(rhs_element_index < 2 * inter.num_eqns);

                // assign the noise value to the relevant element of the RHS vector
                spa_rhs(rhs_element_index, co(output_node_noise, 0.0), QCORR);
            }

#if DEBUG
            dump_matrix(QCORR, "stdout");
#endif

            // factor the matrix and solve it
            spa_factor(QCORR);
            spa_sol(QCORR);

#ifdef DEBUG
            dump_rhs_vector(QCORR, "stdout");
            dump_solution_vector(QCORR, "stdout");
#endif

            // save the solution for later calculations
            store_quantum_solution(f, frequency_index, 2 * node_index + 1, CARRIER_FIELD);

        } // nodes

        frequency_index++;

        // \todo do I need to have some kind of f_s and t_s list as in data_point()?

        // fill modulators with carrier
        fill_modulators();

        // if signals are defined, solve the system for each new frequency
        if (inter.num_signals) {
            // fill the signals with the carrier light
            fill_signals();

            int upper_lower = 1; // start with the upper sideband
            // loop over the upper and then the lower sidebands
            int sideband_index;
            for (sideband_index = 0; sideband_index < 2; sideband_index++) {
                // calculate the sideband frequency; at this point f1 is the carrier
                // frequency
                f = f1 + upper_lower * inter.fsig;

                // calculate the noise at each node and solve the system
                int node_index;
                for (node_index = 0; node_index < inter.num_nodes; node_index++) {

                    double node_noise = get_input_noise_at(node_index);

                    // fill sparse matrix at the given freq and clear the RHS vector
                    spa_fill(f, QCORR);
                    spa_rhs_clear(QCORR);

                    // fill the RHS vector with the signals
                    quantum_signals_rhs(CARRIER_FIELD, 0, node_noise);

                    // factor and solve the matrix
                    spa_factor(QCORR);
                    spa_sol(QCORR);

                    // save the solution for later
                    store_quantum_solution(f, frequency_index, node_index,
                            CARRIER_SIGNAL);

                } // nodes

                // increment the frequency index
                frequency_index++;

                upper_lower = -1; // now choose the lower sideband

            } // sidebands
        }

        // for each modulation frequency about the carrier, solve the system
        int mod_index;
        for (mod_index = 0; mod_index < inter.num_modulators; mod_index++) {
            modulator_t *modulator;
            modulator = &(inter.modulator_list[mod_index]);

            // loop over each modulation order
            int mod_order_index;
            for (mod_order_index = 0;
                    mod_order_index < modulator->order;
                    mod_order_index++) {
                int mod_upper_lower = 1; // start at the upper modulation sideband

                // loop over each modulation sideband
                int mod_sideband_index;
                for (mod_sideband_index = 0;
                        mod_sideband_index < 2;
                        mod_sideband_index++) {
                    // calculate the relevant frequency
                    f = f1 + mod_upper_lower * (mod_order_index + 1) * modulator->f;
                    f2 = f; // keep a copy of the frequency for later use

                    // loop over all the nodes, get the noise at the given node, set
                    // the RHS vector elements, solve the system and save the results
                    int node_index;
                    for (node_index = 0; node_index < inter.num_nodes; node_index++) {

                        double node_noise = get_input_noise_at(node_index);

                        // fill the matrix at the given frequency
                        spa_fill(f, QCORR);

                        // clear the RHS vector
                        spa_rhs_clear(QCORR);

                        // assign the RHS vector elements for all fields
                        int field_index;
                        for (field_index = 0;
                                field_index < inter.num_fields;
                                field_index++) {
                            // determine the RHS vector indices to assign for the two nodes
                            // of the modulator
                            int node1_vector_index =
                                    2 * inter.num_fields * modulator->node1_index
                                    + modulator->node1_amp_dir.fro + 2 * field_index;
                            int node2_vector_index =
                                    2 * inter.num_fields * modulator->node2_index
                                    + modulator->node2_amp_dir.fro + 2 * field_index;

                            // assign the RHS elements
                            spa_rhs(node1_vector_index, co(node_noise, 0.0), QCORR);
                            spa_rhs(node2_vector_index, co(node_noise, 0.0), QCORR);

                        } // fields

                        // factor the matrix and solve it
                        spa_factor(QCORR);
                        spa_sol(QCORR);

                        // store the solution for the current modulation frequency
                        store_quantum_solution(f, frequency_index, node_index,
                                MODULATOR_FIELD);

                    } // nodes

                    // increment the frequency index
                    frequency_index++;

                    // skip the second sideband in SSB-mode
                    if (modulator->single_sideband_mode == ON) {
                        mod_sideband_index++;
                    }

                    // since quantum correlations are on then fill the mirrors and beam
                    // splitters with the modulation sideband intensities.
                    // This is also done in data_point(); does it need to be done
                    // there???
                    fill_mirrors();

                    // if signals are defined, solve the system for each new frequency
                    if (inter.num_signals) {
                        // fill the signals with the modulation sidebands
                        fill_signals();
                        int upper_lower = 1; // start with the upper sideband

                        // now loop over each sideband and set the RHS vector elements
                        int sideband_index;
                        for (sideband_index = 0; sideband_index < 2; sideband_index++) {
                            f = f2 + upper_lower * inter.fsig; // calculate the frequency

                            // loop over all the nodes, get the noise for the given node,
                            // set the RHS vector elements, solve the system and store the
                            // results
                            int node_index;
                            for (node_index = 0;
                                    node_index < inter.num_nodes;
                                    node_index++) {

                                double node_noise = get_input_noise_at(node_index);

                                // fill IFO matrix for the given frequency and clear RHS vector
                                spa_fill(f, QCORR);
                                spa_rhs_clear(QCORR);

                                // set the RHS vector elements
                                quantum_signals_rhs(MODULATOR_FIELD, mod_index, node_noise);

                                // factor and solve the matrix
                                spa_factor(QCORR);
                                spa_sol(QCORR);

                                // save the solution for later
                                store_quantum_solution(f, frequency_index, node_index,
                                        MODULATOR_SIGNAL);

                            } // nodes

                            // increment the frequency index
                            frequency_index++;

                            // now choose the lower signal sideband
                            upper_lower = -1;

                        } // signal sidebands
                    }

                    // now choose the lower modulation sideband
                    mod_upper_lower = -1;

                } // modulation sidebands
            } // modulation orders
        } // modulations

    } // light inputs (carriers)

}

//! Gets the amount of *input* quantum noise at the given node index

/*!
 * \param node_index the node index at which to calculate the noise value
 *
 * \todo "if light_input_noise > UNIT_VACUUM" branch untested
 * \todo noise at component1_index branches untested
 */
double get_input_noise_at(int node_index) {
    // sanity check on input
    assert(node_index >= 0);
    assert(node_index < inter.num_nodes);

    // extract the node
    node_t *node;
    node = &(inter.node_list[node_index]);
    debug_msg("node name = %s\n", node->name);

    // what components are we connected to?
    int component1_index, component2_index;
    which_components(node_index, &component1_index, &component2_index);

    // check that the component indices are what we expect
    assert(component1_index >= 0);
    assert(component1_index < inter.num_components);

    assert(component2_index >= 0 || component2_index == NOT_FOUND);
    assert(component2_index < inter.num_components);

    // what kinds of components do we have?
    int component1_type = get_component_type(component1_index);
    int component2_type = get_component_type(component2_index);

    double node_noise = 0.0;

    // process light input options
    if (component2_type == LIGHT_INPUT) {
        node_noise = get_light_input_noise_at(component2_index);
    } else if (component2_index == NOT_FOUND &&
            component1_type != LIGHT_INPUT) {
        debug_msg("sitting at a dump or detector-connected node\n");

        // at a dump or detector-connected node the noise is the UNIT_VACUUM
        node_noise = UNIT_VACUUM;
    } else if (component2_index != NOT_FOUND &&
            component1_type == LIGHT_INPUT) {
        node_noise = get_light_input_noise_at(component1_index);
    } else if (component2_index == NOT_FOUND &&
            component1_type == LIGHT_INPUT) {
        double light_input_noise = get_light_input_noise_at(component1_index);

        // choose the noise to be the greater of the laser noise or the vacuum
        if (light_input_noise > UNIT_VACUUM) {
            node_noise = light_input_noise;
        } else {
            node_noise = UNIT_VACUUM;
        }
    } else {
        // just for completeness, set the node noise for the trivial case
        node_noise = 0.0;
    }

    return sqrt(node_noise);
}

//! Gets the amount of *output* quantum noise at the given node index

/*!
 * \param node_index the node index at which to calculate the noise value
 *
 * \todo MODULATOR branches untested
 * \todo all but one BEAMSPLITTER branch untested (component1_type is tested)
 */
double get_output_noise_at(int node_index) {
    // sanity check on input
    assert(node_index >= 0);
    assert(node_index < inter.num_nodes);

    // extract the node
    node_t *node;
    node = &(inter.node_list[node_index]);
    debug_msg("node name = %s\n", node->name);

    // what components are we connected to?
    int component1_index, component2_index;
    which_components(node_index, &component1_index, &component2_index);

    // check that the component indices are what we expect
    assert(component1_index >= 0);
    assert(component1_index < inter.num_components);

    assert(component2_index >= 0 || component2_index == NOT_FOUND);
    assert(component2_index < inter.num_components);

    // what kinds of components do we have?
    int component1_type = get_component_type(component1_index);
    int component2_type = get_component_type(component2_index);

    double node_noise = 0.0;

    // process noise contributed by loss in mirrors
    if (component1_type == MIRROR &&
            component2_type == MIRROR) {
        node_noise = get_noise_at_mirror(component1_index);
        node_noise += get_noise_at_mirror(component2_index);
    } else if (component1_type == MIRROR) {
        node_noise = get_noise_at_mirror(component1_index);
    } else if (component2_type == MIRROR) {
        node_noise = get_noise_at_mirror(component2_index);
    } else {
        // just for completeness, set the node noise for the trivial case
        node_noise = 0.0;
    }

    // process noise contributed by loss in beamsplitters
    if (component1_type == BEAMSPLITTER &&
            component2_type == BEAMSPLITTER) {
        node_noise += get_noise_at_beamsplitter(component1_index);
        node_noise += get_noise_at_beamsplitter(component2_index);
    } else if (component1_type == BEAMSPLITTER) {
        node_noise += get_noise_at_beamsplitter(component1_index);
    } else if (component2_type == BEAMSPLITTER) {
        node_noise += get_noise_at_beamsplitter(component2_index);
    } else {
        // just for completeness, set the node noise for the trivial case
        node_noise += 0.0;
    }

    // process noise contributed by loss in modulators
    if (component1_type == MODULATOR &&
            component2_type == MODULATOR) {
        //node_noise += get_noise_at_modulator(component1_index);
        //node_noise += get_noise_at_modulator(component2_index);
    } else if (component1_type == MODULATOR) {
        //node_noise += get_noise_at_modulator(component1_index);
    } else if (component2_type == MODULATOR) {
        node_noise += get_noise_at_modulator(component2_index);
    } else {
        // just for completeness, set the node noise for the trivial case
        node_noise += 0.0;
    }

    return sqrt(node_noise);
}

//! Get the noise at the mirror at the given component index

/*!
 * \param component_index the overall component index where the mirror is
 *
 * \return the noise at the mirror
 */
double get_noise_at_mirror(int component_index) {
    int mirror_index = get_component_index(component_index);

    assert(mirror_index >= 0);

    mirror_t *mirror;
    mirror = &inter.mirror_list[mirror_index];

    // calculate the amount of noise from the loss
    double loss = mirror->Loss;
    double noise_at_mirror = UNIT_VACUUM * loss;

    return noise_at_mirror;
}

//! Get the noise at the beamsplitter at the given component index

/*!
 * \param component_index the overall component index where the beamsplitter is
 *
 * \return the noise at the beamsplitter
 */
double get_noise_at_beamsplitter(int component_index) {
    int beamsplitter_index = get_component_index(component_index);

    assert(beamsplitter_index >= 0);

    beamsplitter_t *beamsplitter;
    beamsplitter = &inter.bs_list[beamsplitter_index];

    // calculate the amount of noise from the loss
    double loss = beamsplitter->Loss;
    double noise_at_beamsplitter = UNIT_VACUUM * loss;

    return noise_at_beamsplitter;
}

//! Get the noise at the modulator at the given component index

/*!
 * \param component_index the overall component index where the modulator is
 *
 * \return the noise at the modulator
 *
 * \todo untested
 */
double get_noise_at_modulator(int component_index) {
    int modulator_index = get_component_index(component_index);

    assert(modulator_index >= 0);

    modulator_t *modulator;
    modulator = &inter.modulator_list[modulator_index];

    // calculate the amount of noise from the loss
    double mod_index = modulator->modulation_index;
    message("modulator->modulation_index = %g\n", modulator->modulation_index);
    double loss = 1.0 - dbesj0(mod_index) * dbesj0(mod_index);
    double noise_at_modulator = UNIT_VACUUM * loss;

    message("loss = %g\n", loss);
    message("noise_at_modulator = %g\n", noise_at_modulator);

    return noise_at_modulator;
}

//! Get the noise value for the given light input

/*!
 * \param component_index the overall component index where the light input is
 *
 * \return the noise value
 */
double get_light_input_noise_at(int component_index) {
    // find the relevant input light index
    int light_in_index = get_component_index(component_index);
    // get the noise component of the light input
    light_in_t *light_in;
    light_in = &inter.light_in_list[light_in_index];
    debug_msg("light input name: %s\n", light_in->name);
    debug_msg("light input noise value: %g\n", light_in->noise_value);

    return light_in->noise_value;
}

//! Store the solution of the last solve operation for quantum noise

/*!
 * \param frequency_index The index of the currently found frequencies
 * \param frequency The current frequency value
 * \param equation_index Index of the current node
 * \param field_type The type of field; CARRIER_FIELD, MODULATOR_FIELD, CARRIER_SIGNAL,
 * MODULATOR_SIGNAL
 *
 * \todo debug & 4 branch untested
 */
void store_quantum_solution(double frequency, int frequency_index,
        int equation_index, int field_type) {
    // sanity checks on inputs
    assert(frequency_index >= 0);
    assert(frequency_index < inter.num_frequencies);
    assert(equation_index >= 0);
    assert(equation_index < inter.num_quantum_eqns);
    assert(field_type == CARRIER_FIELD ||
            field_type == CARRIER_SIGNAL ||
            field_type == MODULATOR_FIELD ||
            field_type == MODULATOR_SIGNAL);

    // store the values for given field, output, frequency and node
    int detector_index;
    for (detector_index = 0; detector_index < inter.num_light_outputs; detector_index++) {
        // get the output data object for the carrier
        output_data_t carrier_output =
                inter.output_data_list[inter.light_out_list[detector_index].output_index];
        int field_index;
        for (field_index = 0; field_index < inter.num_fields; field_index++) {
            // determine the index within Qspsol to get
            int solution_index = 2 * inter.num_fields * carrier_output.node_index
                    + carrier_output.is_second_beam + 2 * field_index;
            debug_msg("solution_index = %d\n", solution_index);

            // check that the solution index is in the right bounds
            assert(solution_index >= 0);
            assert(solution_index < inter.num_quantum_eqns);

            // assign to qnoise the relevant element from Qspsol
            qnoise[field_index][detector_index][frequency_index][equation_index] =
                    spa.Qspsol[solution_index];
            debug_msg("spa.Qspsol[solution_index] = %g : %g\n",
                    spa.Qspsol[solution_index].re, spa.Qspsol[solution_index].im);
        }
    }

    // if debugging is on, output the values assigned, and for what
    // parameters
    // \todo check that the debugging output here is producing the
    // correct values
    if (inter.debug & 4) {
        int detector_index;
        for (detector_index = 0;
                detector_index < inter.num_light_outputs;
                detector_index++) {
            int field_index;
            for (field_index = 0; field_index < inter.num_fields; field_index++) {
                switch (field_type) {
                        // carrier information
                    case CARRIER_FIELD:
                        message("mode %d carrier no %d freq %s : %s\n",
                                field_index, detector_index, freq_form(frequency),
                                complex_form15(
                                qnoise[field_index][detector_index][frequency_index][equation_index])
                                );
                        break;
                        // signal sidebands about the carrier
                    case CARRIER_SIGNAL:
                        message("mode %d signal1 no %d freq %s : %s\n",
                                field_index, detector_index, freq_form(frequency),
                                complex_form15(
                                qnoise[field_index][detector_index][frequency_index][equation_index])
                                );
                        break;
                        // modulation sidebands about the carrier
                    case MODULATOR_FIELD:
                        message("mode %d modulator no %d freq %s : %s\n",
                                field_index, detector_index, freq_form(frequency),
                                complex_form15(
                                qnoise[field_index][detector_index][frequency_index][equation_index])
                                );
                        break;
                        // signal sidebands about the modulation sidebands
                    case MODULATOR_SIGNAL:
                        message("mode %d signal2 no %d freq %s : %s\n",
                                field_index, detector_index, freq_form(frequency),
                                complex_form15(
                                qnoise[field_index][detector_index][frequency_index][equation_index])
                                );
                        break;
                    default:
                        bug_error("Incorrect field type %d\n", field_type);
                }
            }
        }
    }
    debug_msg("Finished storing quantum solution for node %s\n",
            get_node_name(equation_index % 2));
    debug_msg("##########################################\n");
}

//! Compute the RHS vector elements for signal modulation

/*
 * This routine is specifically for quantum noise RHS vector element
 * assignment.
 *
 * \param carrier Are the signals about the CARRIER_FIELD or the
 * MODULATOR_FIELD?
 * \param modulator_index Index of modulator.  Used only if carrier ==
 * MODULATOR_FIELD
 * \param node_noise Noise amount at the current node
 *
 * \todo untested
 */
void quantum_signals_rhs(int carrier, int modulator_index, double node_noise) {
    // declare the signal variable
    signal_t *signal;

    // declare the nodes of the signal
    node_t sig_node1, sig_node2, sig_node3, sig_node4;

    // check that the carrier argument came through correctly
    assert(carrier == CARRIER_FIELD || carrier == MODULATOR_FIELD);
    assert(modulator_index >= 0);
    // inter.num_modulators can be zero, in which case the upper bound check
    // will fail incorrectly, so if it *is* nonzero, run the assert, otherwise
    // continue
#ifndef NDEBUG
    if (inter.num_modulators > 0) {
        assert(modulator_index < inter.num_modulators);
    }
#endif

    // loop over all signals
    int signal_index;
    for (signal_index = 0; signal_index < inter.num_signals; signal_index++) {
        // grab the current signal
        signal = &(inter.signal_list[signal_index]);

        // depending upon the component type associated with this signal, handle
        // the RHS vector element assignment appropriately
        switch (signal->component_type) {
            case BEAMSPLITTER:
                sig_node1 = inter.node_list[*(signal->node1_index)];
                sig_node2 = inter.node_list[*(signal->node2_index)];
                sig_node3 = inter.node_list[*(signal->node3_index)];
                sig_node4 = inter.node_list[*(signal->node4_index)];

                // check the kind of signal we have and handle appropriately
                if (signal->type == SIG_AMP ||
                        signal->type == SIG_PHS) {
                    // loop over all fields and assign the RHS vector elements
                    int field_index;
                    for (field_index = 0;
                            field_index < inter.num_fields;
                            field_index++) {
                        spa_rhs(2 * inter.num_fields * *(signal->node1_index)
                                + signal->node1_amp_dir->fro + 2 * field_index,
                                co(node_noise, 0.0), QCORR);

                        spa_rhs(2 * inter.num_fields * *(signal->node2_index)
                                + signal->node2_amp_dir->fro + 2 * field_index,
                                co(node_noise, 0.0), QCORR);

                        spa_rhs(2 * inter.num_fields * *(signal->node3_index)
                                + signal->node3_amp_dir->fro + 2 * field_index,
                                co(node_noise, 0.0), QCORR);

                        spa_rhs(2 * inter.num_fields * *(signal->node4_index)
                                + signal->node4_amp_dir->fro + 2 * field_index,
                                co(node_noise, 0.0), QCORR);
                    }
                } else if (signal->type == SIG_X ||
                        signal->type == SIG_Y) {
                    bug_error("beamsplitter x,y: modulation type not implemented");
                } else {
                    bug_error("Incorrect signal type for beamsplitter: %s\n",
                            get_signal_type_str(signal->type));
                }
                break;
            case MIRROR:
                sig_node1 = inter.node_list[*(signal->node1_index)];
                sig_node2 = inter.node_list[*(signal->node2_index)];

                // check the kind of signal we have and handle appropriately
                if (signal->type == SIG_AMP ||
                        signal->type == SIG_PHS ||
                        signal->type == SIG_X ||
                        signal->type == SIG_Y) {
                    // loop over all fields and assign the RHS vector elements
                    int field_index;
                    for (field_index = 0;
                            field_index < inter.num_fields;
                            field_index++) {
                        spa_rhs(2 * inter.num_fields * *(signal->node1_index)
                                + signal->node1_amp_dir->fro + 2 * field_index,
                                co(node_noise, 0.0), QCORR);
                        spa_rhs(2 * inter.num_fields * *(signal->node2_index)
                                + signal->node2_amp_dir->fro + 2 * field_index,
                                co(node_noise, 0.0), QCORR);
                    }
                } else {
                    bug_error("Incorrect signal type for mirror: %s\n",
                            get_signal_type_str(signal->type));
                }
                break;
            case SPACE:
                sig_node1 = inter.node_list[*(signal->node1_index)];
                sig_node2 = inter.node_list[*(signal->node2_index)];

                // check the kind of signal we have and handle appropriately
                if (signal->type == SIG_AMP ||
                        signal->type == SIG_PHS) {
                    // can't handle amplitude modulation just yet
                    if (signal->type == SIG_AMP) {
                        gerror("Amplitude modulation of signals is not yet implemented\n");
                    }

                    // loop over all fields and assign the RHS vector elements
                    int field_index;
                    for (field_index = 0;
                            field_index < inter.num_fields;
                            field_index++) {
                        spa_rhs(2 * inter.num_fields * *(signal->node1_index)
                                + signal->node1_amp_dir->fro + 2 * field_index,
                                co(node_noise, 0.0), QCORR);
                        spa_rhs(2 * inter.num_fields * *(signal->node2_index)
                                + signal->node2_amp_dir->fro + 2 * field_index,
                                co(node_noise, 0.0), QCORR);
                    }
                } else {
                    bug_error("Incorrect signal type for space: %s\n",
                            get_signal_type_str(signal->type));
                }
                break;
            case LIGHT_INPUT:
                sig_node1 = inter.node_list[*(signal->node1_index)];

                // input modulation can only be around carrier, so check that we
                // have the carrier, otherwise barf
                if (carrier == CARRIER_FIELD) {

                    // check the kind of signal we have and handle appropriately
                    if (signal->type == SIG_AMP ||
                            signal->type == SIG_PHS ||
                            signal->type == SIG_FRQ) {

                        // if we have amplitude or phase modulation, barf as we can't
                        // handle this yet
                        if (signal->type == SIG_AMP ||
                                signal->type == SIG_PHS) {
                            bug_error("Amplitude or phase modulation of the input "
                                    "not yet implemented");
                        }

                        // loop over all fields and assign the RHS vector elements
                        int field_index;
                        for (field_index = 0;
                                field_index < inter.num_fields;
                                field_index++) {
                            spa_rhs(2 * inter.num_fields * *(signal->node1_index) + 1 +
                                    2 * field_index, co(node_noise, 0.0), QCORR);
                        }

                    } else {
                        bug_error("Incorrect signal type for light input: %s\n",
                                get_signal_type_str(signal->type));
                    }
                }
                break;
            case MODULATOR:
                sig_node1 = inter.node_list[*(signal->node1_index)];
                sig_node2 = inter.node_list[*(signal->node2_index)];

                // check the kind of signal we have and handle appropriately
                if (signal->type == SIG_PHS ||
                        signal->type == SIG_FRQ) {
                    // no phase noise sidebands on carrier
                    if (carrier != CARRIER_FIELD) {
                        // make sure that we are at the right modulator
                        if (signal->component_index == modulator_index) {

                            // loop over all fields and assign the RHS vector elements
                            int field_index;
                            for (field_index = 0;
                                    field_index < inter.num_fields;
                                    field_index++) {
                                spa_rhs(2 * inter.num_fields * *(signal->node1_index)
                                        + signal->node1_amp_dir->fro + 2 * field_index,
                                        co(node_noise, 0.0), QCORR);
                                spa_rhs(2 * inter.num_fields * *(signal->node2_index)
                                        + signal->node2_amp_dir->fro + 2 * field_index,
                                        co(node_noise, 0.0), QCORR);
                            }
                        }
                    }
                } else if (signal->type == SIG_AMP) {
                    bug_error("Signal amplitude modulation of modulations "
                            "not implemented yet");
                } else {
                    bug_error("Incorrect signal type for modulator: %s\n",
                            get_signal_type_str(signal->type));
                }
                break;
            default:
                bug_error("Incorrect component type in signal: %s\n",
                        get_component_type_name(signal->component_type));
        }

    }

}

//! Calculate the quantum noise at detector

/*!
 * \param detector_index detector index
 * \param num_demods number of demodulations of output fields
 */
complex_t quantum_noise_detector(
        int detector_index,
        int num_demods) {
    // sanity check on input
    assert(detector_index >= 0);
    assert(detector_index < inter.num_light_outputs);

    assert(num_demods >= 0);
    assert(num_demods < MAX_DEMOD);

    complex_t mixed_amplitude[MAX_MIX + 1] = {complex_0};

    light_out_t *light_output;
    light_output = &(inter.light_out_list[detector_index]);

    complex_t amplitude = complex_0;

    //!\todo now wrap demodulation around here...
    //!\todo test that output now makes sense...
    double fmix[MAX_MIX + 1][MAX_DEMOD + 1] = {
        { 0},
        { 0}
    };
    int num_mixes = create_mix_table(num_demods, fmix);
    // calculate output for every freq mix
    int mix_index;
    for (mix_index = 1; mix_index <= num_mixes; mix_index++) {
        double f_ref = 0.0;

        // calculate the current mix frequency from the mix table and the
        // current light output's demodulation frequency
        //! \todo could make this loop into get_reference_freq()
        int demod_index;
        for (demod_index = 1; demod_index <= num_demods; demod_index++) {
            f_ref += fmix[mix_index][demod_index] * light_output->f[demod_index];
        }

        amplitude = get_quantum_noise_amplitude(detector_index);

        // report on the sum for the given mix frequency if required
        if (inter.debug & 8) {
            message("pd2(5) Sum: z%d %s\n", mix_index,
                    complex_form15(mixed_amplitude[mix_index]));
        }
    }

    debug_msg("amplitude = %s\n", complex_form(amplitude));

    // find out the classical power falling on the quantum detector
    // this is to scale the output of the quantum detector
    double classical_power = get_classical_power(detector_index);

    double scaled_amplitude = 0.0;
    // really basic (and probably quite incorrect) way of scaling the output
    if (classical_power == 0.0) {
        // there is no laser, so just return the amplitude as-is
        scaled_amplitude = zmod(amplitude);
    } else {
        // otherwise reduce the returned amplitude with respect to the power
        debug_msg("scaling to classical power temporarily switched off\n");
        //scaled_amplitude = zmod(amplitude)/classical_power;
        scaled_amplitude = zmod(amplitude);
    }

    debug_msg("classical_power = %f\n", classical_power);
    debug_msg("scaled_amplitude = %f\n", scaled_amplitude);

    // return the final calculated value
    return co(scaled_amplitude, 0.0);
}

//! Get the current quantum noise amplitude

complex_t get_quantum_noise_amplitude(int detector_index) {

    double amplitude = 0.0;
    complex_t phase = complex_0;
    // add up all the contributions to the quantum field at the detector
    int freq_index = 0;
    int equation_index;
    for (equation_index = 0;
            equation_index < inter.num_quantum_eqns;
            equation_index++) {
        int field_index;
        for (field_index = 0; field_index < inter.num_fields; field_index++) {
            // get the current value of the quantum noise list
            complex_t current_qnoise =
                    qnoise[field_index][detector_index][freq_index][equation_index];

            amplitude += zabs(current_qnoise);

            // the final phase will be phase component of the sum of all the 
            // complex noise values
            phase = z_pl_z(phase, current_qnoise);

            // if debugging output has been asked for, provide it
            if (inter.debug & 8) {
                message("qnoised q[%d][%d][%d][%d]=%s, cumulative amplitude=%g\n",
                        field_index, detector_index, freq_index, equation_index,
                        complex_form15(current_qnoise), amplitude);
            }
        }
    }

    return co(amplitude, 0.0);
}

//! Calculates the quantum shotnoise

/*!
 * This uses Jan Harms' calculation of the shotnoise spectral density
 *
 * \todo expand upon the calculation of the shotnoise spectral density
 *
 * \param detector_index the index of the photodetector
 * \param num_demods the number of demodulations to perform
 *
 * \return the shotnoise level
 */
double quantum_shotnoise(int detector_index, int num_demods) {
    // sanity check the inputs
    assert(detector_index >= 0);
    assert(detector_index < inter.num_light_outputs);

    assert(num_demods >= 0);
    assert(num_demods < MAX_DEMOD);

    light_out_t *light_output;
    light_output = &(inter.light_out_list[detector_index]);

    //!< \todo fmix -> freq_mix_table ???

    // calculation of the number of frequencies which aren't signal
    // frequencies would be a good idea here
    //!< \todo add num_non_signal_frequencies to interferometer structure
    //eventually ???
    inter.num_non_signal_frequencies = get_num_non_signal_frequencies(t_s);

    //!< \todo move memory allocation stuff into kat_mem.c

    // allocate memory for the stripped frequency list
    double *stripped_frequencies;
    stripped_frequencies = calloc(inter.num_non_signal_frequencies, sizeof (double));
    if (stripped_frequencies == NULL) {
        gerror("Error allocating memory for stripped frequency list\n");
    }

    strip_signal_frequencies(f_s, t_s, stripped_frequencies);

    // set up the frequency table
    double **demod_freq_matrix;
    demod_freq_matrix = (double **)
            calloc(inter.num_non_signal_frequencies, sizeof (double *));
    if (demod_freq_matrix == NULL) {
        gerror("Error allocating memory for demodulation frequency matrix\n");
    }
    int freq_index;
    for (freq_index = 0; freq_index < inter.num_non_signal_frequencies; freq_index++) {
        demod_freq_matrix[freq_index] = (double *)
                calloc(MAX_QUANTUM_MIX, sizeof (double));
        if (demod_freq_matrix[freq_index] == NULL) {
            gerror("Error allocating memory for demodulation frequency matrix\n");
        }
    }

    // allocate memory for the base frequency list
    double *base_freqs;
    base_freqs = (double *) calloc(inter.num_non_signal_frequencies, sizeof (double));
    if (base_freqs == NULL) {
        gerror("Error allocating memory for base frequency list\n");
    }

    // allocate memory for the base amplitude list
    complex_t *base_amplitudes;
    base_amplitudes =
            (complex_t *) calloc(inter.num_non_signal_frequencies, sizeof (complex_t));
    if (base_amplitudes == NULL) {
        gerror("Error allocating memory for base amplitude list\n");
    }

    // allocate memory for the index table
    int *index_table;
    index_table = (int *) calloc(inter.num_non_signal_frequencies, sizeof (int));
    if (index_table == NULL) {
        gerror("Error allocating memory for index table\n");
    }

    build_frequency_index_table(index_table, stripped_frequencies);

    build_base_frequency_list(base_freqs, stripped_frequencies, index_table);

    // get frequency-unsorted amplitude list
    complex_t *unsorted_amplitudes;
    unsorted_amplitudes =
            (complex_t *) calloc(inter.num_frequencies, sizeof (complex_t));
    for (freq_index = 0; freq_index < inter.num_frequencies; freq_index++) {
        unsorted_amplitudes[freq_index] = a_s[0][detector_index][freq_index];
    }

    // allocate memory for the stripped amplitude list
    complex_t *stripped_amplitudes;
    stripped_amplitudes = calloc(inter.num_non_signal_frequencies, sizeof (complex_t));
    if (stripped_amplitudes == NULL) {
        gerror("Error allocating memory for stripped amplitude list\n");
    }

    strip_signal_amplitudes(unsorted_amplitudes, t_s, stripped_amplitudes);

    //!< \todo what happens for multiple fields????  Ask Jan.

    build_base_amplitude_list(base_amplitudes, stripped_amplitudes, index_table);

    // grab the demodulation frequencies
    double demod_freqs[MAX_DEMOD + 1];
    double demod_phases[MAX_DEMOD + 1];
    // we don't want the last demodulation as it's unphysical, so reducing the
    // number of demodulations here by one
    num_demods = num_demods - 1;

    int demod_index;
    for (demod_index = 0; demod_index < num_demods; demod_index++) {
        demod_freqs[demod_index] =
                inter.light_out_list[detector_index].f[demod_index + 1];
        demod_phases[demod_index] =
                inter.light_out_list[detector_index].phi[demod_index + 1];
    }

    build_demod_frequency_matrix(demod_freq_matrix, base_freqs, demod_freqs,
            num_demods);

    double shotnoise_spectral_density =
            get_shotnoise_spectral_density(demod_freq_matrix, base_amplitudes,
            demod_phases, num_demods);

    // scale the shotnoise_spectral_density appropriately (by \hbar \omega_0)
    // this scaling is temporarily added in place of the 'scale qshot'
    // functionality
    //shotnoise_spectral_density *= H_PLANCK*inter.f0;

    return shotnoise_spectral_density;
}

//! Gets the number of non signal frequencies in the simulation

/*!
 * \param t_s list of signal types
 */
int get_num_non_signal_frequencies(int *t_s) {
    int num_non_signal_frequencies = inter.num_frequencies;
    int freq_index;
    for (freq_index = 0; freq_index < inter.num_frequencies; freq_index++) {
        if (is_signal(t_s[freq_index])) {
            num_non_signal_frequencies--;
        }
    }
    return num_non_signal_frequencies;
}

//! Removes any signal frequencies from the frequency list

/*!
 * \param f_s list of frequencies to strip
 * \param t_s list of signal types
 * \param stripped_frequencies returned list of stripped frequencies
 */
void strip_signal_frequencies(
        double *f_s,
        int *t_s,
        double *stripped_frequencies) {
    int non_signal_freq_index = 0;
    int freq_index;
    for (freq_index = 0; freq_index < inter.num_frequencies; freq_index++) {
        if (NOT is_signal(t_s[freq_index])) {
            assert(non_signal_freq_index < inter.num_non_signal_frequencies);
            stripped_frequencies[non_signal_freq_index] = f_s[freq_index];
            non_signal_freq_index++;
        }
    }

}

//! Removes any signal amplitudes from the amplitude list

/*!
 * \param unsorted_amplitudes list of unsorted amplitudes to be stripped
 * \param t_s list of signal types
 * \param stripped_amplitudes returned list of stripped, unsorted amplitudes
 */
void strip_signal_amplitudes(
        complex_t *unsorted_amplitudes,
        int *t_s,
        complex_t *stripped_amplitudes) {
    int non_signal_freq_index = 0;
    int freq_index;
    for (freq_index = 0; freq_index < inter.num_frequencies; freq_index++) {
        if (NOT is_signal(t_s[freq_index])) {
            assert(non_signal_freq_index < inter.num_non_signal_frequencies);
            stripped_amplitudes[non_signal_freq_index] =
                    unsorted_amplitudes[freq_index];
            non_signal_freq_index++;
        }
    }

}

//! Calculates the amplitude sum as part of the quantum shotnoise calculation

/*!
 * Performs the inner amplitude sum calculation as part of the quantum
 * shotnoise calculation and returns it to the calling routine. 
 *
 * \param field_index_outer index of the outer field summation loop
 * \param field_index_inner index of the inner field summation loop
 * \param detector_index index of the current photodetector
 * \param f_ref reference frequency
 *
 * \return the value of the amplitude sum
 *
 * \todo untested
 */
complex_t get_noise_amplitude_sum(
        int field_index_outer,
        int field_index_inner,
        int detector_index,
        double f_ref
        ) {
    // sanity check on inputs
    assert(field_index_outer >= 0);
    assert(field_index_outer < inter.num_fields);

    assert(field_index_inner >= 0);
    assert(field_index_inner < inter.num_fields);

    assert(detector_index >= 0);
    assert(detector_index < inter.num_light_outputs);

    // the vacuum noise level
    complex_t vacuum_noise_field = {UNIT_VACUUM, 0.0};

    // the internal sum of amplitudes used as part of the calculation of the
    // output photodetector signal
    complex_t amplitude_sum = complex_0;

    // look at all frequencies, and if the difference between the signal
    // frequencies is equal to the reference frequency calculate one of the
    // internal sums used for calculating the DC signal
    int freq_index_outer;
    for (freq_index_outer = 0;
            freq_index_outer < inter.num_frequencies;
            freq_index_outer++) {
        int freq_index_inner;
        for (freq_index_inner = 0;
                freq_index_inner < inter.num_frequencies;
                freq_index_inner++) {

            // should change this to use 'teq' as in get_amplitude sum #######################################
            double freq_diff = f_s[freq_index_outer] - f_s[freq_index_inner];

            // want to know what frequencies we currently have
            printf("f_s[%d](outer) = %g\n", freq_index_outer, f_s[freq_index_outer]);
            printf("f_s[%d](inner) = %g\n", freq_index_outer, f_s[freq_index_outer]);

            printf("t_s[%d](outer) = %d\n", freq_index_outer, t_s[freq_index_outer]);
            printf("t_s[%d](inner) = %d\n\n", freq_index_outer, t_s[freq_index_outer]);

            if (
                    NOT both_fields(t_s[freq_index_outer], t_s[freq_index_inner]) &&
                    NOT both_signals(t_s[freq_index_outer], t_s[freq_index_inner])
                    ) {

                if (eq(freq_diff, f_ref)) {
                    // now we want to calculate:
                    // a_m s_n^* + a_m v_n^* + s_m a_n^* + v_m a_n^*
                    // this will actually only calculate a_m s_n^* + a_m v_n^*,
                    // however since we are looping over all frequencies, the last
                    // two terms (s_m a_n^* + v_m a_n^*) will actually get picked up
                    // in a later pass through the nested for loops

                    complex_t a_m__s_n_star = // a_m s_n^*
                            z_by_zc(a_s[field_index_outer][detector_index][freq_index_outer],
                            a_s[field_index_inner][detector_index][freq_index_inner]);

                    // add term to cumulative sum
                    amplitude_sum = z_pl_z(amplitude_sum, a_m__s_n_star);
                }
            }

            // need to collect all relevant frequencies.  These are where there
            // are fields, where there are modulations, where there are signals,
            // and differences between them.  This gives a list of frequencies
            // at which some will be signals, and some will be amplitudes, but
            // at *all* will be vacuum
            // 
            // but need to make a check when have squeezed light (hence,
            // correlations)

            //if ( eq( freq_diff, f_ref) ) {
            // make sure not multiplying a vacuum by a signal (value should be 
            // neglected)
            if (NOT is_signal(t_s[freq_index_inner])) {
                complex_t a_m__v_n_star = // a_m v_n^*
                        z_by_zc(a_s[field_index_inner][detector_index][freq_index_inner],
                        vacuum_noise_field);
                // add term to cumulative sum
                amplitude_sum = z_pl_z(amplitude_sum, a_m__v_n_star);
                message("**** amplitude_sum = %g + %g i\n",
                        amplitude_sum.re, amplitude_sum.im);
            }
            //}


        } // inner sum
    } // outer sum

    return amplitude_sum;
}


//! Determines if both kinds of signals are fields

/*!
 * Boolean function to determine if both of signal types are both fields.
 * Returns true when they both are fields, false otherwise.
 *
 * \param signal_type_left one signal type
 * \param signal_type_right another signal type
 *
 * \return true when both signal types are fields, false otherwise
 *
 * \todo untested
 */
bool both_fields(int signal_type_left, int signal_type_right) {
    if ((signal_type_left == CARRIER_FIELD || signal_type_left == MODULATOR_FIELD) &&
            (signal_type_right == CARRIER_FIELD || signal_type_right == MODULATOR_FIELD)) {
        return true;
    } else {
        return false;
    }
}

//! Determines if both kinds of fields are "signal fields"

/*!
 * Boolean function to determine if both of signal types are both signal 
 * fields, e.g. gravitational wave signals.  Returns true when they both 
 * are signal fields, false otherwise.
 *
 * \param signal_type_left one signal type
 * \param signal_type_right another signal type
 *
 * \return true when both signal types are signal fields, false otherwise
 *
 * \todo untested
 */
bool both_signals(int signal_type_left, int signal_type_right) {
    if ((signal_type_left == SIGNAL_FIELD) &&
            (signal_type_right == SIGNAL_FIELD)) {
        return true;
    } else {
        return false;
    }
}

//! Determines if kind of field is a signal field

/*!
 * Returns true when the signal type is a basic SIGNAL_FIELD, a CARRIER_SIGNAL or
 * a MODULATOR_SIGNAL
 *
 * \param signal_type the type of signal, from the t_s[] array
 */
bool is_signal(int signal_type) {
    if (signal_type == SIGNAL_FIELD ||
            signal_type == CARRIER_SIGNAL ||
            signal_type == MODULATOR_SIGNAL) {
        return true;
    } else {
        return false;
    }
}

//! Calculates the field amplitude for the current mixing frequency

/*!
 * \param detector_index the detector index
 * \param f_ref the current reference frequency
 * \param light_output the light output structure
 *
 * \return the field amplitude for the current mixing frequency
 *
 * \todo untested
 */
complex_t get_mixed_noise_amplitude(
        int detector_index,
        double f_ref,
        light_out_t *light_output) {
    // sanity check inputs
    assert(detector_index >= 0);
    assert(detector_index < inter.num_light_outputs);

    // user-defined photodetector
    complex_t total_amplitude_sum = complex_0;
    if (light_output->detector_type > -1) {
        // not 100% sure that we need to handle user-defined photodetectors with
        // the quantum shotnoise detector, so if get to here, barf
        gerror("User-defined quantum shotnoise detectors not yet implemented\n");

        int field_index_outer;
        for (field_index_outer = 0;
                field_index_outer < inter.num_fields;
                field_index_outer++) {
            int field_index_inner;
            for (field_index_inner = 0;
                    field_index_inner < inter.num_fields;
                    field_index_inner++) {

                complex_t amplitude_sum = get_noise_amplitude_sum(
                        field_index_outer, field_index_inner, detector_index, f_ref);

                total_amplitude_sum = z_pl_z(total_amplitude_sum, amplitude_sum);
            }
        }
    }        // default photodetector
    else {
        int field_index;
        for (field_index = 0; field_index < inter.num_fields; field_index++) {

            complex_t amplitude_sum = get_noise_amplitude_sum(
                    field_index, field_index, detector_index, f_ref);

            total_amplitude_sum = z_pl_z(total_amplitude_sum, amplitude_sum);
        }
    }

    message("total amplitude sum = %g + %gi\n",
            total_amplitude_sum.re, total_amplitude_sum.im);
    message("               norm = %g\n", zabs(total_amplitude_sum));

    return total_amplitude_sum;
}

//! Build the demodulation frequency matrix and return it

/*!
 * \param demod_freq_matrix Demodulation frequency matrix to return
 * \param base_freqs Array of base frequencies in system (carrier and mods)
 * \param demod_freqs Array of demodulation frequencies
 * \param num_demods Number of demodulations
 */
void build_demod_frequency_matrix(
        double **demod_freq_matrix,
        const double *base_freqs,
        const double *demod_freqs,
        const int num_demods) {
    // the frequency table holds the frequencies to be used for the
    // demodulations in order to calculate the output noise "signal" (the
    // quantum shotnoise spectral density).
    // It has the same number of rows as frequencies within the system
    // (carriers + modulations) and has the same number of columns as
    // 2**num_demods
    // Special care has to be taken in calculation of the noise spectral
    // density when some frequencies coincide, so these elements need to be
    // identified and then the amplitude values corresponding to these
    // elements used differently to the case where all frequencies are
    // different

    int freq_index;
    for (freq_index = 0; freq_index < inter.num_non_signal_frequencies; freq_index++) {

        int permute_index;
        for (permute_index = 0;
                permute_index < npow(2, num_demods);
                permute_index++) {

            double freq = 0.0;
            int demod_index;
            for (demod_index = 0; demod_index < num_demods; demod_index++) {

                int factor = permute_index / npow(2, num_demods - (demod_index + 1)) + 1;
                int sign = npow(-1, factor);
                freq += sign * demod_freqs[demod_index];
            }

            demod_freq_matrix[freq_index][permute_index] =
                    base_freqs[freq_index] + freq;
        }
    }

    //!< \todo return from here one day!!!
}

//! Makes the (sorted) list of base frequencies in the system

/*!
 * Base frequencies are considered to be carrier frequencies and modulation
 * sidebands
 *
 * \param base_freqs Array of frequencies to calculate and return
 * \param f_s Unsorted list of frequencies to sort
 * \param index_table Table of indices used to sort the frequencies
 */
void build_base_frequency_list(
        double *base_freqs,
        double *f_s,
        int *index_table) {

    // now make the rank table
    int *rank_table;
    rank_table = (int *) calloc(inter.num_non_signal_frequencies, sizeof (int));
    if (rank_table == NULL) {
        gerror("Error allocating memory for rank table\n");
    }

    int freq_index;
    for (freq_index = 0; freq_index < inter.num_non_signal_frequencies; freq_index++) {
        rank_table[index_table[freq_index]] = freq_index;
    }

    // now use the rank table to sort the frequencies
    for (freq_index = 0; freq_index < inter.num_non_signal_frequencies; freq_index++) {
        base_freqs[freq_index] = f_s[rank_table[freq_index]];
    }

    //!< \todo return from here one day!!!
}

//! Makes the (frequency-sorted) list of base amplitudes in the system

/*!
 * Base frequencies are considered to be carrier frequencies and modulation
 * sidebands
 *
 * \param base_amplitudes Array of frequencies to calculate and return
 * \param unsorted_amplitudes List of input unsorted amplitudes to be sorted
 * \param index_table Table of array indices used to sort the amplitudes
 */
void build_base_amplitude_list(
        complex_t *base_amplitudes,
        complex_t *unsorted_amplitudes,
        int *index_table) {

    // now make the rank table
    int *rank_table;
    rank_table = (int *) calloc(inter.num_non_signal_frequencies, sizeof (int));
    if (rank_table == NULL) {
        gerror("Error allocating memory for rank table\n");
    }

    int freq_index;
    for (freq_index = 0; freq_index < inter.num_non_signal_frequencies; freq_index++) {
        rank_table[index_table[freq_index]] = freq_index;
    }

    // now use the rank table to sort the amplitudes
    for (freq_index = 0; freq_index < inter.num_non_signal_frequencies; freq_index++) {
        base_amplitudes[freq_index] = unsorted_amplitudes[rank_table[freq_index]];
    }

    //!< \todo return from here one day!!!
}

//! Calculates the shotnoise spectral density

/*!
 * \param demod_freq_matrix Table of demodulated frequencies
 * \param base_amplitudes List of base amplitudes in system (carrier + mods)
 * \param demod_phases List of demodulation phases
 * \param num_demods Number of demodulations
 */
double get_shotnoise_spectral_density(
        double **demod_freq_matrix,
        complex_t *base_amplitudes,
        double *demod_phases,
        int num_demods) {
#define DEBUG_MSGS 0
    int max_num_mixes = npow(2, num_demods);

    // set up the match table
    bool **has_matched;
    has_matched = (bool **)
            calloc(inter.num_non_signal_frequencies, sizeof (bool *));
    if (has_matched == NULL) {
        gerror("Error allocating memory for match table\n");
    }
    int freq_index;
    for (freq_index = 0; freq_index < inter.num_non_signal_frequencies; freq_index++) {
        has_matched[freq_index] = (bool *)
                calloc(MAX_QUANTUM_MIX, sizeof (bool));
        if (has_matched[freq_index] == NULL) {
            gerror("Error allocating memory for match table\n");
        }
    }

    // initialise the match table
    for (freq_index = 0; freq_index < inter.num_non_signal_frequencies; freq_index++) {
        int mix_index;
        for (mix_index = 0; mix_index < max_num_mixes; mix_index++) {
            has_matched[freq_index][mix_index] = false;
        }
    }

#if DEBUG_MSGS
    // print out the demodulation frequency matrix temporarily
    for (freq_index = 0; freq_index < inter.num_non_signal_frequencies; freq_index++) {
        int mix_index;
        for (mix_index = 0; mix_index < max_num_mixes; mix_index++) {
            printf("demod_freq_matrix[%d][%d] = %g\n",
                    freq_index, mix_index, demod_freq_matrix[freq_index][mix_index]);
        }
    }
#endif

    // set up the demod phase table
    double *demod_phase_matrix;
    demod_phase_matrix = (double *) calloc(max_num_mixes, sizeof (double));
    if (demod_phase_matrix == NULL) {
        gerror("Error allocating memory for demodulation phase matrix\n");
    }

    // initialise the demod phase table
    int mix_index;
    for (mix_index = 0; mix_index < max_num_mixes; mix_index++) {
        demod_phase_matrix[mix_index] = 0.0;
    }

    // work out the relevant demodulation phase
    int demod_index;
    for (demod_index = 0; demod_index < num_demods; demod_index++) {
        int mix_index;
        for (mix_index = 0; mix_index < max_num_mixes; mix_index++) {
            int factor = floor(mix_index / npow(2, num_demods - demod_index - 1));
            int sign = npow(-1, factor);
            demod_phase_matrix[mix_index] += sign * demod_phases[demod_index];
#if DEBUG_MSGS
            printf("demod_phases[%d] = %g, sign = %d\n",
                    demod_index, demod_phases[demod_index], sign);
            printf("demod_phase_matrix[%d] = %g\n",
                    mix_index, demod_phase_matrix[mix_index]);
#endif
        }
    }

    // need to look for matches in the demod frequency matrix before we can
    // calculate the shotnoise spectral density properly
    double shotnoise_spectral_density = 0.0;
    for (freq_index = 0; freq_index < inter.num_non_signal_frequencies; freq_index++) {
        int mix_index;
        for (mix_index = 0; mix_index < max_num_mixes; mix_index++) {
            // reset the amplitude for the current term to be later squared
            complex_t term_amplitude = complex_0;

            // don't test values which have already matched
            if (has_matched[freq_index][mix_index]) {
                continue;
            } else {
                // add the current value, with its relevant demodulation phase to
                // the amplitude for the current term to be squared
                term_amplitude = z_pl_z(
                        term_amplitude,
                        z_by_ph(
                        base_amplitudes[freq_index], demod_phase_matrix[mix_index]
                        )
                        );
#if DEBUG_MSGS
                printf("\n");
                printf("base_amplitudes[%d] = (%g, %g)\n",
                        freq_index, base_amplitudes[freq_index].re,
                        base_amplitudes[freq_index].im);
                printf("demod_phase_matrix[%d] = %g\n",
                        mix_index, demod_phase_matrix[mix_index]);
                printf("First term amplitude = (%g, %g)\n",
                        term_amplitude.re, term_amplitude.im);
#endif
            }

            int test_freq_index;
            for (test_freq_index = 0;
                    test_freq_index < inter.num_non_signal_frequencies;
                    test_freq_index++) {
                int test_mix_index;
                for (test_mix_index = 0;
                        test_mix_index < max_num_mixes;
                        test_mix_index++) {
                    // don't want to test the reference value against itself
                    if (test_freq_index == freq_index &&
                            test_mix_index == mix_index) {
                        continue;
                    }

                    // check if frequencies match
                    if (eq(demod_freq_matrix[freq_index][mix_index],
                            demod_freq_matrix[test_freq_index][test_mix_index])) {
                        has_matched[test_freq_index][test_mix_index] = true;
                        term_amplitude =
                                z_pl_z(term_amplitude,
                                z_by_ph(
                                base_amplitudes[test_freq_index],
                                demod_phase_matrix[test_mix_index]
                                )
                                );
#if DEBUG_MSGS
                        printf("base_amplitudes[%d] = (%g, %g)\n",
                                test_freq_index,
                                base_amplitudes[test_freq_index].re,
                                base_amplitudes[test_freq_index].im);
                        printf("demod_phase_matrix[%d] = %g\n",
                                test_mix_index, demod_phase_matrix[test_mix_index]);
                        printf("Second term amplitude = (%g, %g)\n",
                                term_amplitude.re, term_amplitude.im);
                        printf("Matching frequency %g at test_freq_index = %d, "
                                "test_mix_index = %d\n",
                                demod_freq_matrix[test_freq_index][test_mix_index],
                                test_freq_index, test_mix_index);
#endif
                    }
                }
            }

            shotnoise_spectral_density += zabs(term_amplitude);
#if DEBUG_MSGS
            printf("spectral density = %g\n", shotnoise_spectral_density);
            printf("demod_freq_matrix[%d][%d] = %g\n", freq_index, mix_index,
                    demod_freq_matrix[freq_index][mix_index]);
            printf("demod_phases[%d] = %g\n", mix_index, demod_phases[mix_index]);
#endif

        }
    }

    return shotnoise_spectral_density;
}

//! Makes the index table for sorting the frequency list

/*!
 * \todo convert to C array indexing conventions (currently uses Fortran)
 * \todo reimplement: shouldn't use NumRec code verbatim
 *
 * \todo main else branch is untested
 */
void build_frequency_index_table(int *index_table, double *f_s) {
    // allocate memory for the fortran index table
    int *fortran_index_table;
    fortran_index_table = (int *) calloc(inter.num_non_signal_frequencies + 1, sizeof (int));
    if (fortran_index_table == NULL) {
        gerror("Error allocating memory for fortran index table\n");
    }

    // initialise the fortran index table
    int freq_index;
    for (freq_index = 1; freq_index <= inter.num_non_signal_frequencies; freq_index++) {
        fortran_index_table[freq_index] = freq_index;
    }

    //! Swap two integer values
#define ISWAP(a,b) int itemp=(a);(a)=(b);(b)=itemp;

    unsigned long int M = 7;
    unsigned long int i, j, k;
    unsigned long int l = 1;
    unsigned long int ir = inter.num_non_signal_frequencies;
    unsigned long int indext;
    int jstack = 0;
    int *istack;
    istack = (int *) calloc(50, sizeof (int));
    if (istack == NULL) {
        gerror("Error initialising istack\n");
    }
    double a;
    while (1) {
        if (ir - l < M) {
            for (j = l + 1; j <= ir; j++) {
                indext = fortran_index_table[j];
                a = f_s[indext - 1];
                for (i = j - 1; i >= l; i--) {
                    if (f_s[fortran_index_table[i] - 1] <= a) {
                        break;
                    }
                    fortran_index_table[i + 1] = fortran_index_table[i];
                }
                fortran_index_table[i + 1] = indext;
            }
            if (jstack == 0) {
                break;
            }
            ir = istack[jstack--];
            l = istack[jstack--];
        } else {
            k = (l + ir) >> 1;
            ISWAP(fortran_index_table[k], fortran_index_table[l + 1]);
            if (f_s[fortran_index_table[l] - 1] > f_s[fortran_index_table[ir] - 1]) {
                ISWAP(fortran_index_table[l], fortran_index_table[ir]);
            }
            if (f_s[fortran_index_table[l + 1] - 1] > f_s[fortran_index_table[ir] - 1]) {
                ISWAP(fortran_index_table[l + 1], fortran_index_table[ir]);
            }
            if (f_s[fortran_index_table[l] - 1] > f_s[fortran_index_table[l + 1] - 1]) {
                ISWAP(fortran_index_table[l], fortran_index_table[l + 1]);
            }
            i = l + 1;
            j = ir;
            indext = fortran_index_table[l + 1];
            a = f_s[indext - 1];

            while (1) {
                do {
                    i++;
                } while (f_s[fortran_index_table[i] - 1] < a);
                do {
                    j--;
                } while (f_s[fortran_index_table[j] - 1] > a);
                if (j < i) {
                    break;
                }
                ISWAP(fortran_index_table[i], fortran_index_table[j]);
            }

            fortran_index_table[l + 1] = fortran_index_table[j];
            fortran_index_table[j] = indext;
            jstack += 2;
            if (jstack > 50) {
                gerror("Stack too small for index table creation\n");
            }
            if (ir - i + 1 >= j - 1) {
                istack[jstack] = ir;
                istack[jstack - 1] = i;
                ir = j - 1;
            } else {
                istack[jstack] = j - 1;
                istack[jstack - 1] = l;
                l = i;
            }
        }
    }

    // set the index table so that array indexing works as expected
    for (freq_index = 0; freq_index < inter.num_non_signal_frequencies; freq_index++) {
        index_table[freq_index] = fortran_index_table[freq_index + 1] - 1;
    }

    //!< \todo return from here one day!!!
}

/*
 * Local variables:
 * c-indentation-style: bsd
 * c-basic-offset: 2
 * indent-tabs-mode: nil
 * End:
 *
 * vim: tabstop=2 expandtab shiftwidth=2:
 */
