// $Id$

/*!
 * \file kat_spa.c
 * \brief Sparse matrix library interface routines.
 *
 * These routines also set up the interferometer matrix.
 * 
 * @section Copyright notice
 *
 *  This file is part of the interferometer simulation Finesse
 *  http://www.gwoptics.org/finesse
 *
 *  Copyright (C) 1999 onwards Andreas Freise
 *  with parts of the code written by Daniel Brown, Paul Cochrane
 *  and Gerhard Heinzel.
 *
 *  This program is free software; you can redistribute it and/or modify it under
 *  the terms of the GNU General Public License version 3 as published
 *  by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 *  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *  See the GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along with
 *  this library; if not, write to the Free Software Foundation, Inc., 59 Temple Place,
 *  Suite 330, Boston, MA 02111-1307 USA
 */

#include "kat.h"
#include "kat_inline.c"
#include "kat_spa.h"
#include "kat_klu.h"
#include "klu.h"
#include "kat_check.h"
#include "kat_io.h"
#include "kat_aux.h"

const complex_t complex_i = {0.0, 1.0}; //!<  sqrt(-1) or 0 + i
const complex_t complex_1 = {1.0, 0.0}; //!<  1 in complex space: 1 + 0i
const complex_t complex_m1 = {-1.0, 0.0}; //!< -1 in complex space: -1 + 0i
const complex_t complex_0 = {0.0, 0.0}; //!<  0 in complex space: 0 + 0i

//extern init_variables_t in;
extern interferometer_t inter;
extern options_t options;
extern memory_t mem;
extern sparse_var_t spa;
extern klu_sparse_var_t klu;

extern klu_l_symbolic *Symbolic;
extern klu_l_numeric *Numeric;
extern klu_l_common Common;


static int maxrow1 = 0; //!< maximum number of rows in SPARSE matrix

//! Create a sparse matrix

/*!
 *
 */
void spa_create(void) {
    int err;
    char *unit_string;
    char munit[7] = "Mbytes";
    char kunit[7] = "Kbytes";
    double kbytes;

    // create the standard interferometer matrix
    debug_msg("Creating standard interferometer matrix\n");
    spa.spamat = spCreate(inter.num_eqns, 1, &err);
    report_sparse_matrix_error(err, 1);
    if (spa.spamat == NULL) {
        bug_error("sparse matrix is NULL!");
    }

    // report some memory-related information if asked
    if (inter.debug & 512) {
        message("Memory needed for interferometer matrix is about");
        kbytes = 20.0 * inter.num_eqns + 12.0 * inter.num_eqns * inter.num_eqns;

        if (kbytes > 1048576.0) {
            kbytes = kbytes / 1048576.0;
            unit_string = munit;
        } else {
            kbytes = kbytes / 1024.0;
            unit_string = kunit;
        }
        message(" %.4g %s,\n", kbytes, unit_string);
    }

    // create the quantum noise interferometer matrix
    if (inter.qcorr) {
        debug_msg("Creating quantum interferometer matrix\n");
        spa.Qspamat = spCreate(inter.num_quantum_eqns, 1, &err);
        report_sparse_matrix_error(err, 2);

        if (spa.Qspamat == NULL) {
            bug_error("quantum corrections sparse matrix is NULL!");
        }

        // report some memory-related information if asked
        if (inter.debug & 512) {
            message("Memory needed for quantum interferometer matrix is about");
            kbytes = 20.0 * inter.num_quantum_eqns
                    + 12.0 * inter.num_quantum_eqns * inter.num_quantum_eqns;

            if (kbytes > 1048576.0) {
                kbytes = kbytes / 1048576.0;
                unit_string = munit;
            } else {
                kbytes = kbytes / 1024.0;
                unit_string = kunit;
            }
            message(" %.4g %s,\n", kbytes, unit_string);
        }

    }
}

//! Destroy a sparse matrix

/*!
 *
 *
 * \todo untested
 */
void spa_destroy(void) {
    spDestroy((char *) spa.spamat);
    if (inter.qcorr) {
        spDestroy((char *) spa.Qspamat);
    }
}

int nijlist1; //!< One list of nij values
int nijlist2; //!< Another list of nij values

complex_t *get_matrix_elem(int i, int j, int matrix_type) {
    complex_t *p;

    if (options.klu) {
        p = get_klu_matrix_elem(i, j, matrix_type);
    } else {
        p = get_sparse_matrix_elem(i, j, matrix_type);
    }

    return p;
}


//! Gets and returns a sparse matrix element

/*!
 * \param i row matrix index
 * \param j column matrix index
 * \param matrix_type the type of matrix to get (STANDARD or QCORR)
 *
 * \return the matrix element value referenced by i and j
 *
 * \todo options.check branches untested
 */
complex_t *get_sparse_matrix_elem(int i, int j, int matrix_type) {
    double *p;
    char text[80] = {0};

    // sanity checks on inputs
    assert(matrix_type == STANDARD || matrix_type == QCORR);

    // i and j can't be less than zero or greater than the number of equations
    // (which depends upon the type)
#ifndef NDEBUG
    switch (matrix_type) {
        case STANDARD:
            assert(i >= 0);
            assert(i < inter.num_eqns);
            assert(j >= 0);
            assert(j < inter.num_eqns);
        case QCORR:
            assert(i >= 0);
            assert(i < inter.num_quantum_eqns);
            assert(j >= 0);
            assert(j < inter.num_quantum_eqns);
    }
#endif

    switch (matrix_type) {
        case STANDARD:
            /* check for multiple pointers */
            if (options.check) {
                int k;
                for (k = 0; k < nijlist1; k++) {
                    if (mem.ilist1[k] == i && mem.jlist1[k] == j) {
                        sprintf(text, "i=%d j=%d", i, j);
                        warn("Warning: Several pointers to same location (%s)\n", text);
                        //      printf("%d ",nijlist1);
                    }
                }
                assert(nijlist1 < mem.num_pointers);
                mem.ilist1[nijlist1] = i;
                mem.jlist1[nijlist1] = j;
            }
            ++nijlist1;

            if (i > maxrow1) {
                maxrow1 = i;
            }

            p = spGetElement(spa.spamat, i + 1, j + 1);
            if (p == NULL) {
                gerror("get_matrix_elem()\n");
            }
            return (complex_t *) p;
            break;
        case QCORR:
            if (options.check) {
                int k;
                for (k = 0; k < nijlist2; k++) {
                    if (mem.ilist2[k] == i && mem.jlist2[k] == j) {
                        sprintf(text, "i=%d j=%d", i, j);
                        warn("Warning: Several pointers to same location (%s)\n", text);
                    }
                }
                assert(nijlist2 < mem.num_pointers);
                mem.ilist2[nijlist2] = i;
                mem.jlist2[nijlist2] = j;
            }
            ++nijlist2;

            // make sure the matrix is there


            p = spGetElement(spa.Qspamat, i + 1, j + 1);
            if (p == NULL) {
                gerror("get_matrix_elem() Q\n");
            }
            return (complex_t *) p;
            break;
        default:
            bug_error("no such type");
    }
    return NULL; //silence compiler Warning
}

//! Solve a KLU sparse matrix

/*!
 * \param matrix_type the type of matrix to solve (standard or quantum)
 */
void spa_klu_sol(int matrix_type) {
    // sanity checks on input
    assert(matrix_type == STANDARD || matrix_type == QCORR);

    switch (matrix_type) {
        case STANDARD:
            klu_zl_solve(Symbolic, Numeric, inter.num_eqns, 1, klu.rhs, &Common);
            break;
        case QCORR:
            break;
        default:
            bug_error("no such type");
    }
}

//! Solve a SPARSE sparse matrix

/*!
 * \param matrix_type the type of matrix to solve (standard or quantum)
 */
void spa_sparse_sol(int matrix_type) {
    // sanity checks on input
    assert(matrix_type == STANDARD || matrix_type == QCORR);

    switch (matrix_type) {
        case STANDARD:
            spSolve(spa.spamat, (double *) (spa.sprhs - 1), (double *) (spa.spsol - 1));
            break;
        case QCORR:
            spSolve(spa.Qspamat, (double *) (spa.Qsprhs - 1), (double *) (spa.Qspsol - 1));
            break;
        default:
            bug_error("no such type");
    }
}

//! Solve a sparse matrix

/*!
 * \param matrix_type the type of matrix to solve (standard or quantum)
 */
void spa_sol(int matrix_type) {
    // sanity checks on input
    assert(matrix_type == STANDARD || matrix_type == QCORR);

    if (options.klu) {
        spa_klu_sol(matrix_type);
    } else {
        spa_sparse_sol(matrix_type);
    }
}

//! Solve a transposed sparse matrix

/*!
 * \param matrix_type the type of matrix to solve (standard or quantum)
 *
 * \todo untested
 */
void spa_tsol(int matrix_type) {
    // sanity checks on input
    assert(matrix_type == STANDARD || matrix_type == QCORR);

    switch (matrix_type) {
        case STANDARD:
            spSolveTransposed(spa.spamat, (double *) (spa.sprhs - 1),
                    (double *) (spa.spsol - 1));
            break;
        case QCORR:
            spSolveTransposed(spa.Qspamat, (double *) (spa.Qsprhs - 1),
                    (double *) (spa.Qspsol - 1));
            break;
        default:
            bug_error("no such type");
    }
}

//! Get the sparse (interferometer) matrix

/*!
 *
 */
void spa_get(void) {
    nijlist1 = 0;
    nijlist2 = 0;

    // number of equations per node (e.g. 2 for TEM 00)
    // num_eqs_per_node= 2 * number of fields or =(inter.tem+1)*(inter.tem+2);
    int num_eqns_per_node = 2 * inter.num_fields;

    //printf("num_eqns_per_node = %d\n", num_eqns_per_node);
    //printf("inter.num_eqns = %d\n", inter.num_eqns);

    // set up the diagonal of the field amplitudes matrix
    int i;
    for (i = 0; i < inter.num_eqns; i++) {
        spa.diag[i] = get_matrix_elem(i, i, STANDARD);
    }

    // set up the diagonal of the quantum noise matrix
    if (inter.qcorr) {
        for (i = 0; i < inter.num_quantum_eqns; i++) {
            spa.Qdiag[i] = get_matrix_elem(i, i, QCORR);
        }
    }

    get_mirror_matrix_elements(num_eqns_per_node, GETELEMENT);
    get_beamsplitter_matrix_elements(num_eqns_per_node, GETELEMENT);
    get_modulator_matrix_elements(num_eqns_per_node, GETELEMENT);
    get_space_matrix_elements(num_eqns_per_node, GETELEMENT);
    get_squeezer_matrix_elements(num_eqns_per_node, GETELEMENT);
    get_lens_matrix_elements(num_eqns_per_node, GETELEMENT);
    get_diode_matrix_elements(num_eqns_per_node, GETELEMENT);
    get_grating_matrix_elements(num_eqns_per_node, GETELEMENT);

    inter.mats = nijlist1;

    if (inter.debug & 512) {
        message("size of n x n matrix: n=%d, no. of non-zero elements: %d\n",
                maxrow1 + 1, nijlist1 + 1);
    }
}

//! Get field amplitude and quantum noise matrix elements for mirrors

/*!
 * \param num_eqns_per_node The number of equations per interferometer node
 * \param mode Switches between modes: 0=COUNT prescan for CCS storage
 *             1=GETELEMENT get pointer to matrix element
 */
void get_mirror_matrix_elements(int num_eqns_per_node, int mode) {
    // get the field amplitude and quantum noise matrix elements for
    // the mirrors
    int mirror_index;
    for (mirror_index = 0; mirror_index < inter.num_mirrors; mirror_index++) {
        mirror_t *mirror;
        mirror = &(inter.mirror_list[mirror_index]);
        debug_msg("\n");
        debug_msg("Component = Mirror\n");
        debug_msg("Classical...\n");

        node_t node1 = inter.node_list[mirror->node1_index];
        node_t node2 = inter.node_list[mirror->node2_index];

        int is_node1_connected = node1.connect;
        int is_node2_connected = node2.connect;

        mirror->node1_amp_dir.fro = is_node1_connected;
        mirror->node1_amp_dir.to = ((is_node1_connected) ? 0 : 1);
        mirror->node2_amp_dir.fro = is_node2_connected;
        mirror->node2_amp_dir.to = ((is_node2_connected) ? 0 : 1);

        // calculate the base index values to the interferometer matrix
        int base_index_1 = mirror->node1_reduced_index * num_eqns_per_node
                + is_node1_connected;
        int base_index_2 = mirror->node2_reduced_index * num_eqns_per_node
                + is_node2_connected;

        debug_msg("is_node1_connected? = %d, is_node2_connected? = %d\n",
                is_node1_connected, is_node2_connected);
        debug_msg("base_index_1 = %d, base_index_2 = %d\n",
                base_index_1, base_index_2);

        // field amplitudes
        int j;
        for (j = 0; j < inter.num_fields; j++) { // outputs
            int col_index_1 = base_index_1 + 2 * j + 1 - 2 * is_node1_connected;
            int col_index_2 = base_index_2 + 2 * j + 1 - 2 * is_node2_connected;
            debug_msg("col_index_1 = %d, col_index_2 = %d\n",
                    col_index_1, col_index_2);

            int k;
            for (k = 0; k < inter.num_fields; k++) { // frequencies
                int row_index_1 = base_index_1 + 2 * k;
                int row_index_2 = base_index_2 + 2 * k;
                debug_msg("row_index_1 = %d, row_index_2 = %d\n",
                        row_index_1, row_index_2);

                check_mirror_indices(mirror, STANDARD);

                if (mode == COUNT) {
                    update_element_count(mirror->node1_index, mirror->node2_index, col_index_1, col_index_2, MIRROR);
                } else {
                    if (NOT node1.gnd_node) {
                        // tell SPARSE which elements are non-null
                        mirror->a11[j][k] = get_matrix_elem(row_index_1, col_index_1, STANDARD); // TEST 020710 from indeices 1 t 2
                    } else {
                        mirror->a11[j][k] = NULL;
                    }

                    if (NOT node2.gnd_node) {
                        // tell SPARSE which elements are non-null
                        mirror->a22[j][k] = get_matrix_elem(row_index_2, col_index_2, STANDARD); // TEST 020710 from indeices 2 t 1
                    } else {
                        mirror->a22[j][k] = NULL;
                    }

                    if ((NOT node1.gnd_node) &&
                            (NOT node2.gnd_node)) {
                        // tell SPARSE which elements are non-null
                        mirror->a21[j][k] = get_matrix_elem(row_index_1, col_index_2, STANDARD);
                        mirror->a12[j][k] = get_matrix_elem(row_index_2, col_index_1, STANDARD);
                    } else {
                        mirror->a12[j][k] = NULL;
                        mirror->a21[j][k] = NULL;
                    }
                }
            }
        }

        // quantum noise
        if (inter.qcorr) {
            base_index_1 = mirror->node1_index * num_eqns_per_node
                    + is_node1_connected;
            base_index_2 = mirror->node2_index * num_eqns_per_node
                    + is_node2_connected;

            debug_msg("\n");
            debug_msg("Quantum noise...\n");
            debug_msg("is_node1_connected? = %d, is_node2_connected? = %d\n",
                    is_node1_connected, is_node2_connected);
            debug_msg("base_index_1 = %d, base_index_2 = %d\n",
                    base_index_1, base_index_2);

            int j;
            for (j = 0; j < inter.num_fields; j++) { // outputs
                int col_index_1 = base_index_1 + 2 * j + 1 - 2 * is_node1_connected;
                int col_index_2 = base_index_2 + 2 * j + 1 - 2 * is_node2_connected;
                debug_msg("col_index_1 = %d, col_index_2 = %d\n",
                        col_index_1, col_index_2);

                int k;
                for (k = 0; k < inter.num_fields; k++) { // frequencies
                    int row_index_1 = base_index_1 + 2 * k;
                    int row_index_2 = base_index_2 + 2 * k;
                    debug_msg("row_index_1 = %d, row_index_2 = %d\n",
                            row_index_1, row_index_2);

                    check_mirror_indices(mirror, QCORR);

                    // tell SPARSE which elements are non-null
                    mirror->q11[j][k] =
                            get_matrix_elem(row_index_1, col_index_1, QCORR);

                    // tell SPARSE which elements are non-null
                    mirror->q22[j][k] =
                            get_matrix_elem(row_index_2, col_index_2, QCORR);

                    // tell SPARSE which elements are non-null
                    mirror->q21[j][k] =
                            get_matrix_elem(row_index_1, col_index_2, QCORR);
                    mirror->q12[j][k] =
                            get_matrix_elem(row_index_2, col_index_1, QCORR);
                }
            }
        }

        // increment the connectivity information
        if (NOT node1.gnd_node) {
            ++inter.node_list[mirror->node1_index].connect;
        }
        if (NOT node2.gnd_node) {
            ++inter.node_list[mirror->node2_index].connect;
        }
    }
}

//! Get field amplitude and quantum noise matrix elements for beam splitter

/*!
 * \param num_eqns_per_node The number of equations per interferometer node
 * \param mode Switches between modes: 0=COUNT prescan for CCS storage
 *             1=GETELEMENT get pointer to matrix element
 */
void get_beamsplitter_matrix_elements(int num_eqns_per_node, int mode) {
    // get the field amplitude and quantum noise matrices for
    // the beam splitters
    int i;
    for (i = 0; i < inter.num_beamsplitters; i++) {
        debug_msg("\n");
        debug_msg("Component = Beam splitter\n");
        debug_msg("Classical...\n");

        beamsplitter_t *bs;
        bs = &(inter.bs_list[i]);

        node_t node1 = inter.node_list[bs->node1_index];
        node_t node2 = inter.node_list[bs->node2_index];
        node_t node3 = inter.node_list[bs->node3_index];
        node_t node4 = inter.node_list[bs->node4_index];

        int is_node1_connected = node1.connect;
        int is_node2_connected = node2.connect;
        int is_node3_connected = node3.connect;
        int is_node4_connected = node4.connect;

        bs->node1_amp_dir.fro = is_node1_connected;
        bs->node1_amp_dir.to = ((is_node1_connected) ? 0 : 1);
        bs->node2_amp_dir.fro = is_node2_connected;
        bs->node2_amp_dir.to = ((is_node2_connected) ? 0 : 1);
        bs->node3_amp_dir.fro = is_node3_connected;
        bs->node3_amp_dir.to = ((is_node3_connected) ? 0 : 1);
        bs->node4_amp_dir.fro = is_node4_connected;
        bs->node4_amp_dir.to = ((is_node4_connected) ? 0 : 1);

        // calculate the base index values to the interferometer matrix
        int base_index_1 = bs->node1_reduced_index * num_eqns_per_node
                + is_node1_connected;
        int base_index_2 = bs->node2_reduced_index * num_eqns_per_node
                + is_node2_connected;
        int base_index_3 = bs->node3_reduced_index * num_eqns_per_node
                + is_node3_connected;
        int base_index_4 = bs->node4_reduced_index * num_eqns_per_node
                + is_node4_connected;

        debug_msg("is_node1_connected? = %d, is_node2_connected? = %d, "
                "is_node3_connected? = %d, is_node4_connected? = %d\n",
                is_node1_connected, is_node2_connected,
                is_node3_connected, is_node4_connected);
        debug_msg("base_index_1 = %d, base_index_2 = %d, "
                "base_index_3 = %d, base_index_4 = %d\n",
                base_index_1, base_index_2, base_index_3, base_index_4);

        // field amplitudes
        int j;
        for (j = 0; j < inter.num_fields; j++) { // outputs
            int col_index_1 = base_index_1 + 2 * j + 1 - 2 * is_node1_connected;
            int col_index_2 = base_index_2 + 2 * j + 1 - 2 * is_node2_connected;
            int col_index_3 = base_index_3 + 2 * j + 1 - 2 * is_node3_connected;
            int col_index_4 = base_index_4 + 2 * j + 1 - 2 * is_node4_connected;

            int k;
            for (k = 0; k < inter.num_fields; k++) { // frequencies
                int row_index_1 = base_index_1 + 2 * k;
                int row_index_2 = base_index_2 + 2 * k;
                int row_index_3 = base_index_3 + 2 * k;
                int row_index_4 = base_index_4 + 2 * k;

                check_beamsplitter_indices(bs, STANDARD);


                if (mode == COUNT) {
                    update_element_count(bs->node1_index, bs->node2_index, col_index_1, col_index_2, SPACE);
                    update_element_count(bs->node1_index, bs->node3_index, col_index_1, col_index_3, SPACE);
                    update_element_count(bs->node2_index, bs->node4_index, col_index_2, col_index_4, SPACE);
                    update_element_count(bs->node3_index, bs->node4_index, col_index_3, col_index_4, SPACE);
                } else {
                    if (NOT node1.gnd_node &&
                            NOT node2.gnd_node) {
                        // tell SPARSE which elements are non-null
                        bs->a21[j][k] = get_matrix_elem(row_index_1, col_index_2, STANDARD);
                        bs->a12[j][k] = get_matrix_elem(row_index_2, col_index_1, STANDARD);
                    } else {
                        bs->a12[j][k] = NULL;
                        bs->a21[j][k] = NULL;
                    }

                    if (NOT node1.gnd_node &&
                            NOT node3.gnd_node) {
                        // tell SPARSE which elements are non-null
                        bs->a31[j][k] = get_matrix_elem(row_index_1, col_index_3, STANDARD);
                        bs->a13[j][k] = get_matrix_elem(row_index_3, col_index_1, STANDARD);
                    } else {
                        bs->a13[j][k] = NULL;
                        bs->a31[j][k] = NULL;
                    }

                    if (NOT node2.gnd_node &&
                            NOT node4.gnd_node) {
                        // tell SPARSE which elements are non-null
                        bs->a42[j][k] = get_matrix_elem(row_index_2, col_index_4, STANDARD);
                        bs->a24[j][k] = get_matrix_elem(row_index_4, col_index_2, STANDARD);
                    } else {
                        bs->a24[j][k] = NULL;
                        bs->a42[j][k] = NULL;
                    }

                    if (NOT node3.gnd_node &&
                            NOT node4.gnd_node) {
                        // tell SPARSE which elements are non-null
                        bs->a43[j][k] = get_matrix_elem(row_index_3, col_index_4, STANDARD);
                        bs->a34[j][k] = get_matrix_elem(row_index_4, col_index_3, STANDARD);
                    } else {
                        bs->a34[j][k] = NULL;
                        bs->a43[j][k] = NULL;
                    }
                }
            }
        }

        // quantum noise
        if (inter.qcorr) {
            base_index_1 = bs->node1_index * num_eqns_per_node
                    + is_node1_connected;
            base_index_2 = bs->node2_index * num_eqns_per_node
                    + is_node2_connected;
            base_index_3 = bs->node3_index * num_eqns_per_node
                    + is_node3_connected;
            base_index_4 = bs->node4_index * num_eqns_per_node
                    + is_node4_connected;

            debug_msg("\n");
            debug_msg("Quantum noise...\n");
            debug_msg("is_node1_connected? = %d, is_node2_connected? = %d, "
                    "is_node3_connected? = %d, is_node4_connected? = %d\n",
                    is_node1_connected, is_node2_connected,
                    is_node3_connected, is_node4_connected);
            debug_msg("base_index_1 = %d, base_index_2 = %d, "
                    "base_index_3 = %d, base_index_4 = %d\n",
                    base_index_1, base_index_2, base_index_3, base_index_4);

            int j;
            for (j = 0; j < inter.num_fields; j++) { // outputs
                int col_index_1 = base_index_1 + 2 * j + 1 - 2 * is_node1_connected;
                int col_index_2 = base_index_2 + 2 * j + 1 - 2 * is_node2_connected;
                int col_index_3 = base_index_3 + 2 * j + 1 - 2 * is_node3_connected;
                int col_index_4 = base_index_4 + 2 * j + 1 - 2 * is_node4_connected;
                debug_msg("col_index_1 = %d, col_index_2 = %d, "
                        "col_index_3 = %d, col_index_4 = %d\n",
                        col_index_1, col_index_2,
                        col_index_3, col_index_4);

                int k;
                for (k = 0; k < inter.num_fields; k++) { // frequencies
                    int row_index_1 = base_index_1 + 2 * k;
                    int row_index_2 = base_index_2 + 2 * k;
                    int row_index_3 = base_index_3 + 2 * k;
                    int row_index_4 = base_index_4 + 2 * k;
                    debug_msg("row_index_1 = %d, row_index_2 = %d, "
                            "row_index_3 = %d, row_index_4 = %d\n",
                            row_index_1, row_index_2,
                            row_index_3, row_index_4);

                    check_beamsplitter_indices(bs, QCORR);

                    // tell SPARSE which elements are non-null
                    bs->q21[j][k] =
                            get_matrix_elem(row_index_1, col_index_2, QCORR);
                    bs->q12[j][k] =
                            get_matrix_elem(row_index_2, col_index_1, QCORR);

                    // tell SPARSE which elements are non-null
                    bs->q31[j][k] =
                            get_matrix_elem(row_index_1, col_index_3, QCORR);
                    bs->q13[j][k] =
                            get_matrix_elem(row_index_3, col_index_1, QCORR);

                    // tell SPARSE which elements are non-null
                    bs->q42[j][k] =
                            get_matrix_elem(row_index_2, col_index_4, QCORR);
                    bs->q24[j][k] =
                            get_matrix_elem(row_index_4, col_index_2, QCORR);

                    // tell SPARSE which elements are non-null
                    bs->q43[j][k] =
                            get_matrix_elem(row_index_3, col_index_4, QCORR);
                    bs->q34[j][k] =
                            get_matrix_elem(row_index_4, col_index_3, QCORR);
                }
            }
        }

        // increment the connectivity information
        if (NOT node1.gnd_node) {
            ++inter.node_list[bs->node1_index].connect;
        }

        if (NOT node2.gnd_node) {
            ++inter.node_list[bs->node2_index].connect;
        }

        if (NOT node3.gnd_node) {
            ++inter.node_list[bs->node3_index].connect;
        }

        if (NOT node4.gnd_node) {
            ++inter.node_list[bs->node4_index].connect;
        }

    }
}

//! Get field amplitude and quantum noise modulator elements for modulator

/*!
 * \param num_eqns_per_node The number of equations per interferometer node
 * \param mode Switches between modes: 0=COUNT prescan for CCS storage
 *             1=GETELEMENT get pointer to matrix element
 *
 * \todo qcorr branch untested
 */
void get_modulator_matrix_elements(int num_eqns_per_node, int mode) {
    // get the field amplitude and quantum noise matrices for
    // the modulators
    int i;
    for (i = 0; i < inter.num_modulators; i++) {
        debug_msg("\n");
        debug_msg("Component = Modulator\n");
        debug_msg("Classical...\n");

        modulator_t *modulator;
        modulator = &(inter.modulator_list[i]);

        node_t node1 = inter.node_list[modulator->node1_index];
        node_t node2 = inter.node_list[modulator->node2_index];

        int is_node1_connected = node1.connect;
        int is_node2_connected = node2.connect;

        modulator->node1_amp_dir.fro = is_node1_connected;
        modulator->node1_amp_dir.to = ((is_node1_connected) ? 0 : 1);
        modulator->node2_amp_dir.fro = is_node2_connected;
        modulator->node2_amp_dir.to = ((is_node2_connected) ? 0 : 1);

        // calculate the base index values to the interferometer matrix
        int base_index_1 = modulator->node1_reduced_index * num_eqns_per_node
                + is_node1_connected;
        int base_index_2 = modulator->node2_reduced_index * num_eqns_per_node
                + is_node2_connected;

        debug_msg("is_node1_connected? = %d, is_node2_connected? = %d\n",
                is_node1_connected, is_node2_connected);
        debug_msg("base_index_1 = %d, base_index_2 = %d\n",
                base_index_1, base_index_2);

        // field amplitudes
        int j;
        for (j = 0; j < inter.num_fields; j++) { // outputs
            int col_index_1 = base_index_1 + 2 * j + 1 - 2 * is_node1_connected;
            int col_index_2 = base_index_2 + 2 * j + 1 - 2 * is_node2_connected;

            int k;
            for (k = 0; k < inter.num_fields; k++) { // frequencies
                int row_index_1 = base_index_1 + 2 * k;
                int row_index_2 = base_index_2 + 2 * k;

                check_modulator_indices(modulator, STANDARD);

                if (mode == COUNT) {
                    update_element_count(modulator->node1_index, modulator->node2_index, col_index_1, col_index_2, SPACE);
                } else {
                    if (NOT node1.gnd_node &&
                            NOT node2.gnd_node) {
                        // tell SPARSE which elements are non-null
                        modulator->a21[j][k] = get_matrix_elem(row_index_1, col_index_2,
                                STANDARD);
                        modulator->a12[j][k] = get_matrix_elem(row_index_2, col_index_1,
                                STANDARD);
                    } else {
                        modulator->a21[j][k] = NULL;
                        modulator->a12[j][k] = NULL;
                    }
                }
            }
        }

        // quantum noise
        if (inter.qcorr) {
            base_index_1 = modulator->node1_index * num_eqns_per_node
                    + is_node1_connected;
            base_index_2 = modulator->node2_index * num_eqns_per_node
                    + is_node2_connected;

            debug_msg("\n");
            debug_msg("Quantum noise...\n");
            debug_msg("is_node1_connected? = %d, is_node2_connected? = %d\n",
                    is_node1_connected, is_node2_connected);
            debug_msg("base_index_1 = %d, base_index_2 = %d\n",
                    base_index_1, base_index_2);

            int j;
            for (j = 0; j < inter.num_fields; j++) { // outputs
                int col_index_1 = base_index_1 + 2 * j + 1 - 2 * is_node1_connected;
                int col_index_2 = base_index_2 + 2 * j + 1 - 2 * is_node2_connected;
                debug_msg("col_index_1 = %d, col_index_2 = %d\n",
                        col_index_1, col_index_2);

                int k;
                for (k = 0; k < inter.num_fields; k++) { // frequencies
                    int row_index_1 = base_index_1 + 2 * k;
                    int row_index_2 = base_index_2 + 2 * k;
                    debug_msg("row_index_1 = %d, row_index_2 = %d\n",
                            row_index_1, row_index_2);

                    check_modulator_indices(modulator, QCORR);

                    // tell SPARSE which elements are non-null
                    modulator->q21[j][k] =
                            get_matrix_elem(row_index_1, col_index_2, QCORR);
                    modulator->q12[j][k] =
                            get_matrix_elem(row_index_2, col_index_1, QCORR);
                }
            }
        }

        // increment the connectivity information
        if (NOT node1.gnd_node &&
                NOT node2.gnd_node) {
            ++inter.node_list[modulator->node1_index].connect;
            ++inter.node_list[modulator->node2_index].connect;
        }
    }

}

//! Get field amplitude and quantum noise space elements

/*!
 * \param num_eqns_per_node The number of equations per interferometer node
 * \param mode Switches between modes: 0=COUNT prescan for CCS storage
 *             1=GETELEMENT get pointer to matrix element
 *
 * \todo qcorr branch untested
 */
void get_space_matrix_elements(int num_eqns_per_node, int mode) {
    // get the field amplitude and quantum noise matrices for
    // the spaces
    int i;
    for (i = 0; i < inter.num_spaces; i++) {
        debug_msg("\n");
        debug_msg("Component = Space\n");
        debug_msg("Classical...\n");

        space_t *space;
        space = &(inter.space_list[i]);

        node_t node1 = inter.node_list[space->node1_index];
        node_t node2 = inter.node_list[space->node2_index];

        int is_node1_connected = node1.connect;
        int is_node2_connected = node2.connect;

        space->node1_amp_dir.fro = is_node1_connected;
        space->node1_amp_dir.to = ((is_node1_connected) ? 0 : 1);
        space->node2_amp_dir.fro = is_node2_connected;
        space->node2_amp_dir.to = ((is_node2_connected) ? 0 : 1);

        // calculate the base index values to the interferometer matrix
        int base_index_1 = space->node1_reduced_index * num_eqns_per_node
                + is_node1_connected;
        int base_index_2 = space->node2_reduced_index * num_eqns_per_node
                + is_node2_connected;

        debug_msg("is_node1_connected? = %d, is_node2_connected? = %d\n",
                is_node1_connected, is_node2_connected);
        debug_msg("base_index_1 = %d, base_index_2 = %d\n",
                base_index_1, base_index_2);

        // field amplitudes
        int j;
        for (j = 0; j < inter.num_fields; j++) { // outputs
            int col_index_1 = base_index_1 + 2 * j + 1 - 2 * is_node1_connected;
            int col_index_2 = base_index_2 + 2 * j + 1 - 2 * is_node2_connected;
            debug_msg("col_index_1 = %d, col_index_2 = %d\n",
                    col_index_1, col_index_2);

            int k;
            for (k = 0; k < inter.num_fields; k++) { // frequencies
                int row_index_1 = base_index_1 + 2 * k;
                int row_index_2 = base_index_2 + 2 * k;
                debug_msg("row_index_1 = %d, row_index_2 = %d\n",
                        row_index_1, row_index_2);

                check_space_indices(space, STANDARD);

                if (mode == COUNT) {
                    update_element_count(space->node1_index, space->node2_index, col_index_1, col_index_2, SPACE);
                } else {
                    if (NOT node1.gnd_node &&
                            NOT node2.gnd_node) {
                        // tell SPARSE which elements are non-null
                        space->a21[j][k] = get_matrix_elem(row_index_1, col_index_2, STANDARD);
                        space->a12[j][k] = get_matrix_elem(row_index_2, col_index_1, STANDARD);
                    } else {
                        space->a21[j][k] = NULL;
                        space->a12[j][k] = NULL;
                    }
                }
            }
        }

        // quantum noise
        if (inter.qcorr) {
            base_index_1 = space->node1_index * num_eqns_per_node
                    + is_node1_connected;
            base_index_2 = space->node2_index * num_eqns_per_node
                    + is_node2_connected;

            debug_msg("\n");
            debug_msg("Quantum noise...\n");
            debug_msg("is_node1_connected? = %d, is_node2_connected? = %d\n",
                    is_node1_connected, is_node2_connected);
            debug_msg("base_index_1 = %d, base_index_2 = %d\n",
                    base_index_1, base_index_2);

            int j;
            for (j = 0; j < inter.num_fields; j++) { // outputs
                int col_index_1 = base_index_1 + 2 * j + 1 - 2 * is_node1_connected;
                int col_index_2 = base_index_2 + 2 * j + 1 - 2 * is_node2_connected;
                debug_msg("col_index_1 = %d, col_index_2 = %d\n",
                        col_index_1, col_index_2);

                int k;
                for (k = 0; k < inter.num_fields; k++) { // frequencies
                    int row_index_1 = base_index_1 + 2 * k;
                    int row_index_2 = base_index_2 + 2 * k;
                    debug_msg("row_index_1 = %d, row_index_2 = %d\n",
                            row_index_1, row_index_2);

                    check_space_indices(space, QCORR);

                    // tell SPARSE which elements are non-null
                    space->q21[j][k] =
                            get_matrix_elem(row_index_1, col_index_2, QCORR);
                    space->q12[j][k] =
                            get_matrix_elem(row_index_2, col_index_1, QCORR);
                }
            }
        }
        // increment the connectivity information
        if ((NOT node1.gnd_node) && (NOT node2.gnd_node)) {
            ++inter.node_list[space->node1_index].connect;
            ++inter.node_list[space->node2_index].connect;
        }
    }
}

//! Get field amplitude and quantum noise squeezer elements

/*!
 * \param num_eqns_per_node The number of equations per interferometer node
 * \param mode Switches between modes: 0=COUNT prescan for CCS storage
 *             1=GETELEMENT get pointer to matrix element
 *
 * \todo untested
 */
void get_squeezer_matrix_elements(int num_eqns_per_node, int mode) {
    // get the field amplitude and quantum noise matrices for
    // the squeezers
    int i;
    for (i = 0; i < inter.num_squeezers; i++) {
        debug_msg("\n");
        debug_msg("Component = Squeezer\n");
        debug_msg("Classical...\n");

        squeezer_t *squeezer;
        squeezer = &(inter.squeezer_list[i]);

        node_t node1 = inter.node_list[squeezer->node1_index];
        node_t node2 = inter.node_list[squeezer->node2_index];

        int is_node1_connected = node1.connect;
        int is_node2_connected = node2.connect;

        squeezer->node1_amp_dir.fro = is_node1_connected;
        squeezer->node1_amp_dir.to = ((is_node1_connected) ? 0 : 1);
        squeezer->node2_amp_dir.fro = is_node2_connected;
        squeezer->node2_amp_dir.to = ((is_node2_connected) ? 0 : 1);

        // calculate the base index values to the interferometer matrix
        int base_index_1 = squeezer->node1_reduced_index * num_eqns_per_node
                + is_node1_connected;
        int base_index_2 = squeezer->node2_reduced_index * num_eqns_per_node
                + is_node2_connected;

        debug_msg("is_node1_connected? = %d, is_node2_connected? = %d\n",
                is_node1_connected, is_node2_connected);
        debug_msg("base_index_1 = %d, base_index_2 = %d\n",
                base_index_1, base_index_2);

        // field amplitudes
        int j;
        for (j = 0; j < inter.num_fields; j++) { // outputs
            int col_index_1 = base_index_1 + 2 * j + 1 - 2 * is_node1_connected;
            int col_index_2 = base_index_2 + 2 * j + 1 - 2 * is_node2_connected;
            debug_msg("col_index_1 = %d, col_index_2 = %d\n",
                    col_index_1, col_index_2);

            int k;
            for (k = 0; k < inter.num_fields; k++) { // frequencies
                int row_index_1 = base_index_1 + 2 * k;
                int row_index_2 = base_index_2 + 2 * k;
                debug_msg("row_index_1 = %d, row_index_2 = %d\n",
                        row_index_1, row_index_2);

                check_squeezer_indices(squeezer, STANDARD);

                if (mode == COUNT) {
                    update_element_count(squeezer->node1_index, squeezer->node2_index, col_index_1, col_index_2, SPACE);
                } else {
                    if (NOT node1.gnd_node) {
                        squeezer->a11[j][k] = get_matrix_elem(row_index_1, col_index_1,
                                STANDARD);
                    } else {
                        squeezer->a11[j][k] = NULL;
                    }

                    if (NOT node2.gnd_node) {
                        squeezer->a22[j][k] = get_matrix_elem(row_index_2, col_index_2,
                                STANDARD);
                    } else {
                        squeezer->a22[j][k] = NULL;
                    }

                    if (NOT node1.gnd_node &&
                            NOT node2.gnd_node) {
                        // tell SPARSE which elements are non-null
                        squeezer->a21[j][k] = get_matrix_elem(row_index_1, col_index_2,
                                STANDARD);
                        squeezer->a12[j][k] = get_matrix_elem(row_index_2, col_index_1,
                                STANDARD);
                    } else {
                        squeezer->a21[j][k] = NULL;
                        squeezer->a12[j][k] = NULL;
                    }
                }
            }
        }

        // quantum noise
        if (inter.qcorr) {
            base_index_1 = squeezer->node1_index * num_eqns_per_node
                    + is_node1_connected;
            base_index_2 = squeezer->node2_index * num_eqns_per_node
                    + is_node2_connected;

            debug_msg("\n");
            debug_msg("Quantum noise...\n");
            debug_msg("is_node1_connected? = %d, is_node2_connected? = %d\n",
                    is_node1_connected, is_node2_connected);
            debug_msg("base_index_1 = %d, base_index_2 = %d\n",
                    base_index_1, base_index_2);

            int j;
            for (j = 0; j < inter.num_fields; j++) { // outputs
                int col_index_1 = base_index_1 + 2 * j + 1 - 2 * is_node1_connected;
                int col_index_2 = base_index_2 + 2 * j + 1 - 2 * is_node2_connected;
                debug_msg("col_index_1 = %d, col_index_2 = %d\n",
                        col_index_1, col_index_2);

                int k;
                for (k = 0; k < inter.num_fields; k++) { // frequencies
                    int row_index_1 = base_index_1 + 2 * k;
                    int row_index_2 = base_index_2 + 2 * k;
                    debug_msg("row_index_1 = %d, row_index_2 = %d\n",
                            row_index_1, row_index_2);

                    check_squeezer_indices(squeezer, QCORR);

                    // tell SPARSE which elements are non-null
                    squeezer->q11[j][k] =
                            get_matrix_elem(row_index_1, col_index_1, QCORR);
                    squeezer->q21[j][k] =
                            get_matrix_elem(row_index_1, col_index_2, QCORR);
                    squeezer->q12[j][k] =
                            get_matrix_elem(row_index_2, col_index_1, QCORR);
                    squeezer->q22[j][k] =
                            get_matrix_elem(row_index_2, col_index_2, QCORR);
                }
            }
        }

        // increment the connectivity information
        if (NOT node1.gnd_node &&
                NOT node2.gnd_node) {
            ++inter.node_list[squeezer->node1_index].connect;
            ++inter.node_list[squeezer->node2_index].connect;
        }
    }

}

//! Get field amplitude and quantum noise lens elements

/*!
 * \param num_eqns_per_node The number of equations per interferometer node
 * \param mode Switches between modes: 0=COUNT prescan for CCS storage
 *             1=GETELEMENT get pointer to matrix element
 *
 * \todo qcorr branch untested
 */
void get_lens_matrix_elements(int num_eqns_per_node, int mode) {
    // get the field amplitude and quantum noise matrices for
    // the lens
    int i;
    for (i = 0; i < inter.num_lenses; i++) {
        debug_msg("\n");
        debug_msg("Component = Lens\n");
        debug_msg("Classical...\n");

        lens_t *lens;
        lens = &(inter.lens_list[i]);

        node_t node1 = inter.node_list[lens->node1_index];
        node_t node2 = inter.node_list[lens->node2_index];

        int is_node1_connected = node1.connect;
        int is_node2_connected = node2.connect;

        lens->node1_amp_dir.fro = is_node1_connected;
        lens->node1_amp_dir.to = ((is_node1_connected) ? 0 : 1);
        lens->node2_amp_dir.fro = is_node2_connected;
        lens->node2_amp_dir.to = ((is_node2_connected) ? 0 : 1);

        // calculate the base index values to the interferometer matrix
        int base_index_1 = lens->node1_reduced_index * num_eqns_per_node
                + is_node1_connected;
        int base_index_2 = lens->node2_reduced_index * num_eqns_per_node
                + is_node2_connected;

        debug_msg("is_node1_connected? = %d, is_node2_connected? = %d\n",
                is_node1_connected, is_node2_connected);
        debug_msg("base_index_1 = %d, base_index_2 = %d\n",
                base_index_1, base_index_2);

        // field amplitudes
        int j;
        for (j = 0; j < inter.num_fields; j++) { // outputs
            int col_index_1 = base_index_1 + 2 * j + 1 - 2 * is_node1_connected;
            int col_index_2 = base_index_2 + 2 * j + 1 - 2 * is_node2_connected;

            int k;
            for (k = 0; k < inter.num_fields; k++) { // frequencies
                int row_index_1 = base_index_1 + 2 * k;
                int row_index_2 = base_index_2 + 2 * k;

                check_lens_indices(lens, STANDARD);

                if (mode == COUNT) {
                    update_element_count(lens->node1_index, lens->node2_index, col_index_1, col_index_2, SPACE);
                } else {
                    if (NOT node1.gnd_node &&
                            NOT node2.gnd_node) {
                        // tell SPARSE which elements are non-null
                        lens->a21[j][k] = get_matrix_elem(row_index_1, col_index_2,
                                STANDARD);
                        lens->a12[j][k] = get_matrix_elem(row_index_2, col_index_1,
                                STANDARD);
                    } else {
                        lens->a21[j][k] = NULL;
                        lens->a12[j][k] = NULL;
                    }
                }
            }
        }

        // quantum noise
        if (inter.qcorr) {
            base_index_1 = lens->node1_index * num_eqns_per_node
                    + is_node1_connected;
            base_index_2 = lens->node2_index * num_eqns_per_node
                    + is_node2_connected;

            debug_msg("\n");
            debug_msg("Quantum noise...\n");
            debug_msg("is_node1_connected? = %d, is_node2_connected? = %d\n",
                    is_node1_connected, is_node2_connected);
            debug_msg("base_index_1 = %d, base_index_2 = %d\n",
                    base_index_1, base_index_2);

            int j;
            for (j = 0; j < inter.num_fields; j++) { // outputs
                int col_index_1 = base_index_1 + 2 * j + 1 - 2 * is_node1_connected;
                int col_index_2 = base_index_2 + 2 * j + 1 - 2 * is_node2_connected;
                debug_msg("col_index_1 = %d, col_index_2 = %d\n",
                        col_index_1, col_index_2);

                int k;
                for (k = 0; k < inter.num_fields; k++) { // frequencies
                    int row_index_1 = base_index_1 + 2 * k;
                    int row_index_2 = base_index_2 + 2 * k;
                    debug_msg("row_index_1 = %d, row_index_2 = %d\n",
                            row_index_1, row_index_2);

                    check_lens_indices(lens, QCORR);

                    // tell SPARSE which elements are non-null
                    lens->q21[j][k] =
                            get_matrix_elem(row_index_1, col_index_2, QCORR);
                    lens->q12[j][k] =
                            get_matrix_elem(row_index_2, col_index_1, QCORR);
                }
            }
        }

        // increment the connectivity information
        if (NOT node1.gnd_node &&
                NOT node2.gnd_node) {
            ++inter.node_list[lens->node1_index].connect;
            ++inter.node_list[lens->node2_index].connect;
        }
    }

}

//! Get field amplitude and quantum noise diode elements

/*!
 * \param num_eqns_per_node The number of equations per interferometer node
 * \param mode Switches between modes: 0=COUNT prescan for CCS storage
 *             1=GETELEMENT get pointer to matrix element
 *
 * \todo qcorr branch untested
 */
void get_diode_matrix_elements(int num_eqns_per_node, int mode) {
    // get the field amplitude and quantum noise matrices for
    // the diodes
    int i;
    for (i = 0; i < inter.num_diodes; i++) {
        debug_msg("\n");
        debug_msg("Component = Diode\n");
        debug_msg("Classical...\n");

        diode_t *diode;
        diode = &(inter.diode_list[i]);

        node_t node1 = inter.node_list[diode->node1_index];
        node_t node2 = inter.node_list[diode->node2_index];

        int is_node1_connected = node1.connect;
        int is_node2_connected = node2.connect;

        diode->node1_amp_dir.fro = is_node1_connected;
        diode->node1_amp_dir.to = ((is_node1_connected) ? 0 : 1);
        diode->node2_amp_dir.fro = is_node2_connected;
        diode->node2_amp_dir.to = ((is_node2_connected) ? 0 : 1);

        // calculate the base index values to the interferometer matrix
        int base_index_1 = diode->node1_reduced_index * num_eqns_per_node
                + is_node1_connected;
        int base_index_2 = diode->node2_reduced_index * num_eqns_per_node
                + is_node2_connected;

        debug_msg("is_node1_connected? = %d, is_node2_connected? = %d\n",
                is_node1_connected, is_node2_connected);
        debug_msg("base_index_1 = %d, base_index_2 = %d\n",
                base_index_1, base_index_2);

        // field amplitudes
        int j;
        for (j = 0; j < inter.num_fields; j++) { // outputs
            int col_index_1 = base_index_1 + 2 * j + 1 - 2 * is_node1_connected;
            int col_index_2 = base_index_2 + 2 * j + 1 - 2 * is_node2_connected;

            int k;
            for (k = 0; k < inter.num_fields; k++) { // frequencies
                int row_index_1 = base_index_1 + 2 * k;
                int row_index_2 = base_index_2 + 2 * k;

                check_diode_indices(diode, STANDARD);

                if (mode == COUNT) {
                    update_element_count(diode->node1_index, diode->node2_index, col_index_1, col_index_2, SPACE);
                } else {
                    if (NOT node1.gnd_node &&
                            NOT node2.gnd_node) {
                        // tell SPARSE which elements are non-null
                        diode->a21[j][k] = get_matrix_elem(row_index_1, col_index_2,
                                STANDARD);
                        diode->a12[j][k] = get_matrix_elem(row_index_2, col_index_1,
                                STANDARD);
                    } else {
                        diode->a12[j][k] = NULL;
                        diode->a21[j][k] = NULL;
                    }
                }
            }
        }

        // quantum noise
        if (inter.qcorr) {
            base_index_1 = diode->node1_index * num_eqns_per_node
                    + is_node1_connected;
            base_index_2 = diode->node2_index * num_eqns_per_node
                    + is_node2_connected;

            debug_msg("\n");
            debug_msg("Quantum noise...\n");
            debug_msg("is_node1_connected? = %d, is_node2_connected? = %d\n",
                    is_node1_connected, is_node2_connected);
            debug_msg("base_index_1 = %d, base_index_2 = %d\n",
                    base_index_1, base_index_2);

            int j;
            for (j = 0; j < inter.num_fields; j++) { // outputs
                int col_index_1 = base_index_1 + 2 * j + 1 - 2 * is_node1_connected;
                int col_index_2 = base_index_2 + 2 * j + 1 - 2 * is_node2_connected;
                debug_msg("col_index_1 = %d, col_index_2 = %d\n",
                        col_index_1, col_index_2);

                int k;
                for (k = 0; k < inter.num_fields; k++) { // frequencies
                    int row_index_1 = base_index_1 + 2 * k;
                    int row_index_2 = base_index_2 + 2 * k;
                    debug_msg("row_index_1 = %d, row_index_2 = %d\n",
                            row_index_1, row_index_2);

                    check_diode_indices(diode, QCORR);

                    // tell SPARSE which elements are non-null
                    diode->q21[j][k] =
                            get_matrix_elem(row_index_1, col_index_2, QCORR);
                    diode->q12[j][k] =
                            get_matrix_elem(row_index_2, col_index_1, QCORR);
                }
            }
        }

        // increment the connectivity information
        if (NOT node1.gnd_node &&
                NOT node2.gnd_node) {
            ++inter.node_list[diode->node1_index].connect;
            ++inter.node_list[diode->node2_index].connect;
        }
    }

}

//! Get field amplitude and quantum noise matrix elements for gratings

/*!
 * \param num_eqns_per_node The number of equations per interferometer node
 * \param mode Switches between modes: 0=COUNT prescan for CCS storage
 *             1=GETELEMENT get pointer to matrix element
 * 
 * \todo qcorr branches untested
 */
void get_grating_matrix_elements(int num_eqns_per_node, int mode) {
    // get the field amplitude and quantum noise matrices for
    // the gratings
    int i;
    for (i = 0; i < inter.num_gratings; i++) {
        debug_msg("\n");
        debug_msg("Component = Grating\n");
        debug_msg("Classical...\n");

        grating_t *grating;
        grating = &(inter.grating_list[i]);

        node_t node1 = inter.node_list[grating->node1_index];
        node_t node2 = inter.node_list[grating->node2_index];
        node_t node3, node4;

        int is_node1_connected = 0;
        int is_node2_connected = 0;
        int is_node3_connected = 0;
        int is_node4_connected = 0;

        int base_index_1 = 0;
        int base_index_2 = 0;
        int base_index_3 = 0;
        int base_index_4 = 0;

        int col_index_1 = 0;
        int col_index_2 = 0;
        int col_index_3 = 0;
        int col_index_4 = 0;

        int row_index_1 = 0;
        int row_index_2 = 0;
        int row_index_3 = 0;
        int row_index_4 = 0;

        switch (grating->num_of_ports) {
                //########################### case 2 ports ################################################
            case 2:
                is_node1_connected = node1.connect;
                is_node2_connected = node2.connect;

                grating->node1_amp_dir.fro = is_node1_connected;
                grating->node1_amp_dir.to = ((is_node1_connected) ? 0 : 1);
                grating->node2_amp_dir.fro = is_node2_connected;
                grating->node2_amp_dir.to = ((is_node2_connected) ? 0 : 1);

                // calculate the base index values to the interferometer matrix
                base_index_1 = grating->node1_reduced_index * num_eqns_per_node
                        + is_node1_connected;
                base_index_2 = grating->node2_reduced_index * num_eqns_per_node
                        + is_node2_connected;

                debug_msg("is_node1_connected? = %d, is_node2_connected? = %d\n",
                        is_node1_connected, is_node2_connected);
                debug_msg("base_index_1 = %d, base_index_2 = %d\n",
                        base_index_1, base_index_2);

                // field amplitudes
                int j;
                for (j = 0; j < inter.num_fields; j++) { // outputs
                    col_index_1 = base_index_1 + 2 * j + 1 - 2 * is_node1_connected;
                    col_index_2 = base_index_2 + 2 * j + 1 - 2 * is_node2_connected;

                    int k;
                    for (k = 0; k < inter.num_fields; k++) { // frequencies
                        row_index_1 = base_index_1 + 2 * k;
                        row_index_2 = base_index_2 + 2 * k;

                        check_grating_indices(grating, STANDARD);

                        if (mode == COUNT) {
                            update_element_count(grating->node1_index, grating->node2_index, col_index_1, col_index_2, MIRROR);
                        } else {
                            if (NOT node1.gnd_node) {
                                // tell SPARSE which elements are non-null
                                grating->a11[j][k] = get_matrix_elem(row_index_1, col_index_1,
                                        STANDARD);
                            } else {
                                grating->a11[j][k] = NULL;
                            }

                            if (NOT node2.gnd_node) {
                                // tell SPARSE which elements are non-null
                                grating->a22[j][k] = get_matrix_elem(row_index_2, col_index_2,
                                        STANDARD);
                            } else {
                                grating->a22[j][k] = NULL;
                            }

                            if (NOT node1.gnd_node &&
                                    NOT node2.gnd_node) {
                                // tell SPARSE which elements are non-null
                                grating->a21[j][k] = get_matrix_elem(row_index_1, col_index_2,
                                        STANDARD);
                                grating->a12[j][k] = get_matrix_elem(row_index_2, col_index_1,
                                        STANDARD);
                            } else {
                                grating->a12[j][k] = NULL;
                                grating->a21[j][k] = NULL;
                            }
                        }
                    }
                }

                // quantum noise
                if (inter.qcorr) {
                    base_index_1 = grating->node1_index * num_eqns_per_node
                            + is_node1_connected;
                    base_index_2 = grating->node2_index * num_eqns_per_node
                            + is_node2_connected;

                    debug_msg("\n");
                    debug_msg("Quantum noise...\n");
                    debug_msg("is_node1_connected? = %d, is_node2_connected? = %d\n",
                            is_node1_connected, is_node2_connected);
                    debug_msg("base_index_1 = %d, base_index_2 = %d\n",
                            base_index_1, base_index_2);

                    int j;
                    for (j = 0; j < inter.num_fields; j++) { // outputs
                        col_index_1 = base_index_1 + 2 * j + 1 - 2 * is_node1_connected;
                        col_index_2 = base_index_2 + 2 * j + 1 - 2 * is_node2_connected;
                        debug_msg("col_index_1 = %d, col_index_2 = %d\n",
                                col_index_1, col_index_2);

                        int k;
                        for (k = 0; k < inter.num_fields; k++) { // frequencies
                            row_index_1 = base_index_1 + 2 * k;
                            row_index_2 = base_index_2 + 2 * k;
                            debug_msg("row_index_1 = %d, row_index_2 = %d\n",
                                    row_index_1, row_index_2);

                            check_grating_indices(grating, QCORR);

                            // tell SPARSE which elements are non-null
                            grating->q11[j][k] =
                                    get_matrix_elem(row_index_1, col_index_1, QCORR);

                            // tell SPARSE which elements are non-null
                            grating->q22[j][k] =
                                    get_matrix_elem(row_index_2, col_index_2, QCORR);

                            // tell SPARSE which elements are non-null
                            grating->q21[j][k] =
                                    get_matrix_elem(row_index_1, col_index_2, QCORR);
                            grating->q12[j][k] =
                                    get_matrix_elem(row_index_2, col_index_1, QCORR);
                        }
                    }
                }

                // increment the connectivity information
                if (NOT node1.gnd_node) {
                    ++inter.node_list[grating->node1_index].connect;
                }

                if (NOT node2.gnd_node) {
                    ++inter.node_list[grating->node2_index].connect;
                }

                break;
                //#################### case 3 ##########################
            case 3:
                node3 = inter.node_list[grating->node3_index];

                is_node1_connected = inter.node_list[grating->node1_index].connect;
                is_node2_connected = inter.node_list[grating->node2_index].connect;
                is_node3_connected = inter.node_list[grating->node3_index].connect;

                grating->node1_amp_dir.fro = is_node1_connected;
                grating->node1_amp_dir.to = ((is_node1_connected) ? 0 : 1);
                grating->node2_amp_dir.fro = is_node2_connected;
                grating->node2_amp_dir.to = ((is_node2_connected) ? 0 : 1);
                grating->node3_amp_dir.fro = is_node3_connected;
                grating->node3_amp_dir.to = ((is_node3_connected) ? 0 : 1);

                // calculate the base index values to the interferometer matrix
                base_index_1 = grating->node1_reduced_index * num_eqns_per_node
                        + is_node1_connected;
                base_index_2 = grating->node2_reduced_index * num_eqns_per_node
                        + is_node2_connected;
                base_index_3 = grating->node3_reduced_index * num_eqns_per_node
                        + is_node3_connected;

                debug_msg("is_node1_connected? = %d, is_node2_connected? = %d, "
                        "is_node3_connected? = %d\n",
                        is_node1_connected, is_node2_connected, is_node3_connected);
                debug_msg("base_index_1 = %d, base_index_2 = %d, base_index_3 = %d\n",
                        base_index_1, base_index_2, base_index_3);

                // field amplitudes
                for (j = 0; j < inter.num_fields; j++) { // outputs
                    col_index_1 = base_index_1 + 2 * j + 1 - 2 * is_node1_connected;
                    col_index_2 = base_index_2 + 2 * j + 1 - 2 * is_node2_connected;
                    col_index_3 = base_index_3 + 2 * j + 1 - 2 * is_node3_connected;

                    int k;
                    for (k = 0; k < inter.num_fields; k++) { // frequencies
                        row_index_1 = base_index_1 + 2 * k;
                        row_index_2 = base_index_2 + 2 * k;
                        row_index_3 = base_index_3 + 2 * k;

                        check_grating_indices(grating, STANDARD);

                        if (mode == COUNT) {
                            update_element_count(grating->node1_index, grating->node2_index, col_index_1, col_index_2, MIRROR);
                            update_element_count(grating->node1_index, grating->node3_index, col_index_1, col_index_3, SPACE);
                            update_element_count(grating->node2_index, grating->node3_index, col_index_2, col_index_3, SPACE);
                            update_element_count(grating->node2_index, grating->node3_index, col_index_2, col_index_3, GRATING);
                        } else {
                            if (NOT node1.gnd_node &&
                                    NOT node2.gnd_node) {
                                // tell SPARSE which elements are non-null
                                grating->a21[j][k] = get_matrix_elem(row_index_1, col_index_2,
                                        STANDARD);
                                grating->a12[j][k] = get_matrix_elem(row_index_2, col_index_1,
                                        STANDARD);
                            } else {
                                grating->a12[j][k] = NULL;
                                grating->a21[j][k] = NULL;
                            }

                            if (NOT node1.gnd_node &&
                                    NOT node3.gnd_node) {
                                // tell SPARSE which elements are non-null
                                grating->a31[j][k] =
                                        get_matrix_elem(row_index_1, col_index_3, STANDARD);
                                grating->a13[j][k] =
                                        get_matrix_elem(row_index_3, col_index_1, STANDARD);
                            } else {
                                grating->a13[j][k] = NULL;
                                grating->a31[j][k] = NULL;
                            }

                            if (NOT node2.gnd_node &&
                                    NOT node3.gnd_node) {
                                // tell SPARSE which elements are non-null
                                grating->a32[j][k] =
                                        get_matrix_elem(row_index_2, col_index_3, STANDARD);
                                grating->a23[j][k] =
                                        get_matrix_elem(row_index_3, col_index_2, STANDARD);
                            } else {
                                grating->a32[j][k] = NULL;
                                grating->a23[j][k] = NULL;
                            }

                            if (NOT node1.gnd_node) {
                                // tell SPARSE which elements are non-null
                                grating->a11[j][k] =
                                        get_matrix_elem(row_index_1, col_index_1, STANDARD);
                            } else {
                                grating->a11[j][k] = NULL;
                            }

                            if (NOT node2.gnd_node) {
                                // tell SPARSE which elements are non-null
                                grating->a22[j][k] =
                                        get_matrix_elem(row_index_2, col_index_2, STANDARD);
                            } else {
                                grating->a22[j][k] = NULL;
                            }

                            if (NOT node3.gnd_node) {
                                // tell SPARSE which elements are non-null
                                grating->a33[j][k] =
                                        get_matrix_elem(row_index_3, col_index_3, STANDARD);
                            } else {
                                grating->a33[j][k] = NULL;
                            }
                        }
                    }
                }

                // quantum noise terms
                if (inter.qcorr) {
                    base_index_1 = grating->node1_index * num_eqns_per_node
                            + is_node1_connected;
                    base_index_2 = grating->node2_index * num_eqns_per_node
                            + is_node2_connected;
                    base_index_3 = grating->node2_index * num_eqns_per_node
                            + is_node3_connected;

                    debug_msg("\n");
                    debug_msg("Quantum noise...\n");
                    debug_msg("is_node1_connected? = %d, is_node2_connected? = %d, "
                            "is_node3_connected? = %d\n",
                            is_node1_connected, is_node2_connected, is_node3_connected);
                    debug_msg("base_index_1 = %d, base_index_2 = %d, base_index_3 = %d\n",
                            base_index_1, base_index_2, base_index_3);

                    int j;
                    for (j = 0; j < inter.num_fields; j++) { // outputs
                        col_index_1 = base_index_1 + 2 * j + 1 - 2 * is_node1_connected;
                        col_index_2 = base_index_2 + 2 * j + 1 - 2 * is_node2_connected;
                        col_index_3 = base_index_3 + 2 * j + 1 - 2 * is_node3_connected;
                        debug_msg("col_index_1 = %d, col_index_2 = %d, "
                                "col_index_3 = %d\n",
                                col_index_1, col_index_2, col_index_3);

                        int k;
                        for (k = 0; k < inter.num_fields; k++) { // frequencies
                            // these assignments are the same as the field amplitudes
                            row_index_1 = base_index_1 + 2 * k;
                            row_index_2 = base_index_2 + 2 * k;
                            row_index_3 = base_index_3 + 2 * k;
                            debug_msg("row_index_1 = %d, row_index_2 = %d, "
                                    "row_index_3 %d\n",
                                    row_index_1, row_index_2, row_index_3);

                            check_grating_indices(grating, QCORR);

                            // tell SPARSE which elements are non-null
                            grating->q21[j][k] =
                                    get_matrix_elem(row_index_1, col_index_2, QCORR);
                            grating->q12[j][k] =
                                    get_matrix_elem(row_index_2, col_index_1, QCORR);

                            if (grating->num_of_ports > 2) {
                                // tell SPARSE which elements are non-null
                                grating->q31[j][k] =
                                        get_matrix_elem(row_index_1, col_index_3, QCORR);
                                grating->q13[j][k] =
                                        get_matrix_elem(row_index_3, col_index_1, QCORR);

                                // tell SPARSE which elements are non-null
                                grating->q32[j][k] =
                                        get_matrix_elem(row_index_2, col_index_3, QCORR);
                                grating->q23[j][k] =
                                        get_matrix_elem(row_index_3, col_index_2, QCORR);

                                if (grating->num_of_ports > 3) {
                                    // Was passiert hier denn???
                                } else {
                                    // tell SPARSE which elements are non-null
                                    grating->q11[j][k] =
                                            get_matrix_elem(row_index_1, col_index_1, QCORR);

                                    // tell SPARSE which elements are non-null
                                    grating->q22[j][k] =
                                            get_matrix_elem(row_index_2, col_index_2, QCORR);

                                    // tell SPARSE which elements are non-null
                                    grating->q33[j][k] =
                                            get_matrix_elem(row_index_3, col_index_3, QCORR);
                                }
                            }
                        }
                    }
                }

                // increment the connectivity information
                if (NOT node1.gnd_node) {
                    ++inter.node_list[grating->node1_index].connect;
                }

                if (NOT node2.gnd_node) {
                    ++inter.node_list[grating->node2_index].connect;
                }

                if (NOT node3.gnd_node) {
                    ++inter.node_list[grating->node3_index].connect;
                }

                break;
                //################################################ 4 ports #######################################
            case 4:
                node3 = inter.node_list[grating->node3_index];
                node4 = inter.node_list[grating->node4_index];

                is_node1_connected = inter.node_list[grating->node1_index].connect;
                is_node2_connected = inter.node_list[grating->node2_index].connect;
                is_node3_connected = inter.node_list[grating->node3_index].connect;
                is_node4_connected = inter.node_list[grating->node4_index].connect;

                grating->node1_amp_dir.fro = is_node1_connected;
                grating->node1_amp_dir.to = ((is_node1_connected) ? 0 : 1);
                grating->node2_amp_dir.fro = is_node2_connected;
                grating->node2_amp_dir.to = ((is_node2_connected) ? 0 : 1);
                grating->node3_amp_dir.fro = is_node3_connected;
                grating->node3_amp_dir.to = ((is_node3_connected) ? 0 : 1);
                grating->node4_amp_dir.fro = is_node4_connected;
                grating->node4_amp_dir.to = ((is_node4_connected) ? 0 : 1);

                // calculate the base index values to the interferometer matrix
                base_index_1 = grating->node1_reduced_index * num_eqns_per_node
                        + is_node1_connected;
                base_index_2 = grating->node2_reduced_index * num_eqns_per_node
                        + is_node2_connected;
                base_index_3 = grating->node3_reduced_index * num_eqns_per_node
                        + is_node3_connected;
                base_index_4 = grating->node4_reduced_index * num_eqns_per_node
                        + is_node4_connected;

                debug_msg("is_node1_connected? = %d, is_node2_connected? = %d, "
                        "is_node3_connected? = %d, is_node4_connected? = %d\n",
                        is_node1_connected, is_node2_connected,
                        is_node3_connected, is_node4_connected);
                debug_msg("base_index_1 = %d, base_index_2 = %d, "
                        "base_index_3 = %d, base_index_4 = %d\n",
                        base_index_1, base_index_2, base_index_3, base_index_4);

                // field amplitudes
                for (j = 0; j < inter.num_fields; j++) { // output
                    col_index_1 = base_index_1 + 2 * j + 1 - 2 * is_node1_connected;
                    col_index_2 = base_index_2 + 2 * j + 1 - 2 * is_node2_connected;
                    col_index_3 = base_index_3 + 2 * j + 1 - 2 * is_node3_connected;
                    col_index_4 = base_index_4 + 2 * j + 1 - 2 * is_node4_connected;

                    int k;
                    for (k = 0; k < inter.num_fields; k++) { // frequencies
                        row_index_1 = base_index_1 + 2 * k;
                        row_index_2 = base_index_2 + 2 * k;
                        row_index_3 = base_index_3 + 2 * k;
                        row_index_4 = base_index_4 + 2 * k;

                        check_grating_indices(grating, STANDARD);

                        if (mode == COUNT) {
                            update_element_count(grating->node1_index, grating->node2_index, col_index_1, col_index_2, SPACE);
                            update_element_count(grating->node1_index, grating->node3_index, col_index_1, col_index_3, SPACE);
                            update_element_count(grating->node2_index, grating->node4_index, col_index_2, col_index_4, SPACE);
                            update_element_count(grating->node3_index, grating->node4_index, col_index_3, col_index_4, SPACE);
                        } else {
                            if (NOT node1.gnd_node &&
                                    NOT node2.gnd_node) {

                                // tell SPARSE which elements are non-null
                                grating->a21[j][k] = get_matrix_elem(row_index_1, col_index_2,
                                        STANDARD);
                                grating->a12[j][k] = get_matrix_elem(row_index_2, col_index_1,
                                        STANDARD);
                            } else {
                                grating->a12[j][k] = NULL;
                                grating->a21[j][k] = NULL;
                            }

                            if (NOT node1.gnd_node &&
                                    NOT node3.gnd_node) {
                                // tell SPARSE which elements are non-null
                                grating->a31[j][k] = get_matrix_elem(row_index_1, col_index_3,
                                        STANDARD);
                                grating->a13[j][k] = get_matrix_elem(row_index_3, col_index_1,
                                        STANDARD);
                            } else {
                                grating->a13[j][k] = NULL;
                                grating->a31[j][k] = NULL;
                            }

                            if (NOT node2.gnd_node &&
                                    NOT node4.gnd_node) {
                                // tell SPARSE which elements are non-null
                                grating->a42[j][k] = get_matrix_elem(row_index_2, col_index_4,
                                        STANDARD);
                                grating->a24[j][k] = get_matrix_elem(row_index_4, col_index_2,
                                        STANDARD);
                            } else {
                                grating->a24[j][k] = NULL;
                                grating->a42[j][k] = NULL;
                            }

                            if (NOT node3.gnd_node &&
                                    NOT node4.gnd_node) {
                                // tell SPARSE which elements are non-null
                                grating->a43[j][k] = get_matrix_elem(row_index_3, col_index_4,
                                        STANDARD);
                                grating->a34[j][k] = get_matrix_elem(row_index_4, col_index_3,
                                        STANDARD);
                            } else {
                                grating->a34[j][k] = NULL;
                                grating->a43[j][k] = NULL;
                            }
                        }
                    }
                }
                // quantum noise
                if (inter.qcorr) {
                    base_index_1 = grating->node1_index * num_eqns_per_node
                            + is_node1_connected;
                    base_index_2 = grating->node2_index * num_eqns_per_node
                            + is_node2_connected;
                    base_index_3 = grating->node3_index * num_eqns_per_node
                            + is_node3_connected;
                    base_index_4 = grating->node4_index * num_eqns_per_node
                            + is_node4_connected;

                    debug_msg("\n");
                    debug_msg("Quantum noise...\n");
                    debug_msg("is_node1_connected? = %d, is_node2_connected? = %d, "
                            "is_node3_connected? = %d, is_node4_connected? = %d\n",
                            is_node1_connected, is_node2_connected,
                            is_node3_connected, is_node4_connected);
                    debug_msg("base_index_1 = %d, base_index_2 = %d, "
                            "base_index_3 = %d, base_index_4 = %d\n",
                            base_index_1, base_index_2, base_index_3, base_index_4);

                    int j;
                    for (j = 0; j < inter.num_fields; j++) { // outputs
                        col_index_1 = base_index_1 + 2 * j + 1 - 2 * is_node1_connected;
                        col_index_2 = base_index_2 + 2 * j + 1 - 2 * is_node2_connected;
                        col_index_3 = base_index_3 + 2 * j + 1 - 2 * is_node3_connected;
                        col_index_4 = base_index_4 + 2 * j + 1 - 2 * is_node4_connected;
                        debug_msg("col_index_1 = %d, col_index_2 = %d, "
                                "col_index_3 = %d, col_index_4 = %d\n",
                                col_index_1, col_index_2,
                                col_index_3, col_index_4);

                        int k;
                        for (k = 0; k < inter.num_fields; k++) { // frequencies
                            row_index_1 = base_index_1 + 2 * k;
                            row_index_2 = base_index_2 + 2 * k;
                            row_index_3 = base_index_3 + 2 * k;
                            row_index_4 = base_index_4 + 2 * k;
                            debug_msg("row_index_1 = %d, row_index_2 = %d, "
                                    "row_index_3 = %d, row_index_4 = %d\n",
                                    row_index_1, row_index_2,
                                    row_index_3, row_index_4);

                            check_grating_indices(grating, QCORR);

                            // tell SPARSE which elements are non-null
                            grating->q21[j][k] =
                                    get_matrix_elem(row_index_1, col_index_2, QCORR);
                            grating->q12[j][k] =
                                    get_matrix_elem(row_index_2, col_index_1, QCORR);

                            // tell SPARSE which elements are non-null
                            grating->q31[j][k] =
                                    get_matrix_elem(row_index_1, col_index_3, QCORR);
                            grating->q13[j][k] =
                                    get_matrix_elem(row_index_3, col_index_1, QCORR);

                            // tell SPARSE which elements are non-null
                            grating->q42[j][k] =
                                    get_matrix_elem(row_index_2, col_index_4, QCORR);
                            grating->q24[j][k] =
                                    get_matrix_elem(row_index_4, col_index_2, QCORR);

                            // tell SPARSE which elements are non-null
                            grating->q43[j][k] =
                                    get_matrix_elem(row_index_3, col_index_4, QCORR);
                            grating->q34[j][k] =
                                    get_matrix_elem(row_index_4, col_index_3, QCORR);
                        }
                    }
                }

                // increment the connectivity information
                if (NOT node1.gnd_node) {
                    ++inter.node_list[grating->node1_index].connect;
                }

                if (NOT node2.gnd_node) {
                    ++inter.node_list[grating->node2_index].connect;
                }

                if (NOT node3.gnd_node) {
                    ++inter.node_list[grating->node3_index].connect;
                }

                if (NOT node4.gnd_node) {
                    ++inter.node_list[grating->node4_index].connect;
                }

                break;
            default:
                bug_error("grating: wrong number of ports");
        }
    }

}

// gouy phase richtig??

//! Fill a sparse matrix

/*!
 * \param f frequency of light field to fill
 * \param matrix_type the type of matrix to fill (standard or quantum)
 */
void spa_fill(double f, int matrix_type) {
    // sanity checks on inputs
    assert(matrix_type == STANDARD ||
            matrix_type == QCORR);
    // what bounds should f have?  if any?

    if (!options.klu) {
        init_sparse_matrix_diags(matrix_type);
    }

    // fill the matrix with elements from the various kinds of components
    fill_matrix_mirror_elements(f);
    fill_matrix_beamsplitter_elements(f);
    fill_matrix_grating_elements();
    fill_matrix_space_elements(f);
    fill_matrix_squeezer_elements();
    fill_matrix_modulator_elements();
    fill_matrix_lens_elements();
    fill_matrix_diode_elements();

    if (inter.debug & 16384) {
        dump_matrix_space_elements(stdout);
        dump_matrix_mirror_elements(stdout);
        dump_matrix_beamsplitter_elements(stdout);
    }
}

//! Initialise the diagonals of the sparse matrices

/*!
 * \param matrix_type Type of matrix to initialise STANDARD or QCORR
 */
void init_sparse_matrix_diags(int matrix_type) {
    // set up diagonals for interferometer matrix
    int matrix_index;
    switch (matrix_type) {
        case STANDARD:
            spClear(spa.spamat);
            for (matrix_index = 0;
                    matrix_index < inter.num_eqns;
                    matrix_index++) {
                *spa.diag[matrix_index] = complex_1;
            }
            break;
        case QCORR:
            spClear(spa.Qspamat);
            for (matrix_index = 0;
                    matrix_index < inter.num_quantum_eqns;
                    matrix_index++) {
                *spa.Qdiag[matrix_index] = complex_1;
            }
            break;
        default:
            bug_error("no such type");
    }
}

//! Fill relevant sparse matrix elements with mirror component terms

/*!
 * \param f frequency of light field to fill
 */
void fill_matrix_mirror_elements(double f) {
    // fill matrix for mirrors
    int mirror_index;
    for (mirror_index = 0; mirror_index < inter.num_mirrors; mirror_index++) {
        mirror_t *mirror;
        mirror = &(inter.mirror_list[mirror_index]);

        double r = sqrt(mirror->R);
        double t = sqrt(mirror->T);
        double tuning = 2.0 * mirror->phi * RAD * (1.0 + f / inter.f0);
        double c0 = r * cos(tuning);
        double s0 = r * sin(tuning);

        // field amplitudes
        int j, k;



        for (j = 0; j < inter.num_fields; j++) {
            for (k = 0; k < inter.num_fields; k++) {

                if (mirror->a11[j][k] != NULL) {
                    (mirror->a11[j][k])->re = -c0;
                    (mirror->a11[j][k])->im = -s0;

                    // just testing adf 07.07.2010
                    // confirms that a11 is the coupling from n1 into n1
                    // mirror->a11[j][k]->re=0.0;
                    // mirror->a11[j][k]->im=0.0;

                    *(mirror->a11[j][k]) =
                            z_by_z(*(mirror->a11[j][k]), mirror->kmn.k11[j][k]);
                }

                if (mirror->a12[j][k] != NULL) {
                    (mirror->a12[j][k])->re = 0.0;
                    (mirror->a12[j][k])->im = -t;
                    (mirror->a21[j][k])->re = 0.0;
                    (mirror->a21[j][k])->im = -t;

                    *(mirror->a12[j][k]) =
                            z_by_z(*(mirror->a12[j][k]), mirror->kmn.k12[j][k]);
                    *(mirror->a21[j][k]) =
                            z_by_z(*(mirror->a21[j][k]), mirror->kmn.k21[j][k]);
                }

                if (mirror->a22[j][k] != NULL) {
                    (mirror->a22[j][k])->re = -c0;
                    (mirror->a22[j][k])->im = s0;

                    *(mirror->a22[j][k]) =
                            z_by_z(*(mirror->a22[j][k]), mirror->kmn.k22[j][k]);
                }
            }
        }

        // quantum noise
        if (inter.qcorr) {
            int j, k;
            for (j = 0; j < inter.num_fields; j++) { // outputs
                for (k = 0; k < inter.num_fields; k++) { // frequencies
                    if (mirror->q11[j][k] != NULL) {
                        (mirror->q11[j][k])->re = -c0;
                        (mirror->q11[j][k])->im = -s0;
                    }

                    if (mirror->q12[j][k] != NULL) {
                        (mirror->q12[j][k])->re = 0.0;
                        (mirror->q12[j][k])->im = -t;
                    }

                    if (mirror->q21[j][k] != NULL) {
                        (mirror->q21[j][k])->re = 0.0;
                        (mirror->q21[j][k])->im = -t;
                    }

                    if (mirror->q22[j][k] != NULL) {
                        (mirror->q22[j][k])->re = -c0;
                        (mirror->q22[j][k])->im = s0;
                    }
                }
            }
        }

    }
}


//! Dump relevant sparse matrix elements for mirrors

/*!
 *
 */
void dump_matrix_mirror_elements(FILE *fp) {
    int mirror_index;

    // make sure the file pointer isn't null
    assert(fp != NULL);

    fprintf(fp, "dumping mirror matrices\n");

    for (mirror_index = 0; mirror_index < inter.num_mirrors; mirror_index++) {
        mirror_t mirror;
        mirror = inter.mirror_list[mirror_index];
        fprintf(fp, "mirror %d, name: %s\n", mirror_index, mirror.name);

        int j, k;

        for (j = 0; j < inter.num_fields; j++) {
            for (k = 0; k < inter.num_fields; k++) {
                int n1, m1, n2, m2;

                get_tem_modes_from_field_index(&n1, &m1, j);
                get_tem_modes_from_field_index(&n2, &m2, k);

                fprintf(fp, "(%d, %d) [%d %d -> %d %d]\n", j, k, n1, m1, n2, m2);
                if (mirror.a11[j][k] != NULL) {
                    fprintf(fp, "a11: %s, ", complex_form15(*(mirror.a11[j][k])));
                }
                if (mirror.a22[j][k] != NULL) {
                    fprintf(fp, "a22: %s, ", complex_form15(*(mirror.a22[j][k])));
                }
                if (mirror.a12[j][k] != NULL) {
                    fprintf(fp, "a12: %s, ", complex_form15(*(mirror.a12[j][k])));
                }
                if (mirror.a21[j][k] != NULL) {
                    fprintf(fp, "a21: %s\n", complex_form15(*(mirror.a21[j][k])));
                }
            }
        }
    }
}


//! Dump relevant sparse matrix elements for mirrors

/*!
 *
 */
void dump_matrix_beamsplitter_elements(FILE *fp) {
    int bs_index;

    // make sure the file pointer isn't null
    assert(fp != NULL);

    fprintf(fp, "dumping beamsplitter matrices\n");

    for (bs_index = 0; bs_index < inter.num_beamsplitters; bs_index++) {
        beamsplitter_t bs;
        bs = inter.bs_list[bs_index];
        fprintf(fp, "beamsplitter %d, name: %s\n", bs_index, bs.name);

        int j, k;

        for (j = 0; j < inter.num_fields; j++) {
            for (k = 0; k < inter.num_fields; k++) {
                int n1, m1, n2, m2;

                get_tem_modes_from_field_index(&n1, &m1, j);
                get_tem_modes_from_field_index(&n2, &m2, k);

                fprintf(fp, "(%d, %d) [%d %d -> %d %d]\n", j, k, n1, m1, n2, m2);
                if (bs.a12[j][k] != NULL) {
                    fprintf(fp, "a12: %s, ", complex_form15(*(bs.a12[j][k])));
                }
                if (bs.a21[j][k] != NULL) {
                    fprintf(fp, "a21: %s, ", complex_form15(*(bs.a21[j][k])));
                }
                if (bs.a13[j][k] != NULL) {
                    fprintf(fp, "a13: %s, ", complex_form15(*(bs.a13[j][k])));
                }
                if (bs.a31[j][k] != NULL) {
                    fprintf(fp, "a31: %s, ", complex_form15(*(bs.a31[j][k])));
                }
                if (bs.a34[j][k] != NULL) {
                    fprintf(fp, "a34: %s, ", complex_form15(*(bs.a34[j][k])));
                }
                if (bs.a43[j][k] != NULL) {
                    fprintf(fp, "a43: %s, ", complex_form15(*(bs.a43[j][k])));
                }
                if (bs.a24[j][k] != NULL) {
                    fprintf(fp, "a24: %s, ", complex_form15(*(bs.a24[j][k])));
                }
                if (bs.a42[j][k] != NULL) {
                    fprintf(fp, "a42: %s\n", complex_form15(*(bs.a42[j][k])));
                }
            }
        }
    }
}




//! Dump relevant sparse matrix elements for spaces

/*!
 *
 */
void dump_matrix_space_elements(FILE *fp) {

    int space_index;

    // make sure the file pointer isn't null
    assert(fp != NULL);

    fprintf(fp, "dumping space matrices\n");

    for (space_index = 0; space_index < inter.num_spaces; space_index++) {
        space_t space;
        space = inter.space_list[space_index];
        fprintf(fp, "space %d, name: %s\n", space_index, space.name);

        int j, k;

        for (j = 0; j < inter.num_fields; j++) {
            for (k = 0; k < inter.num_fields; k++) {
                int n1, m1, n2, m2;

                get_tem_modes_from_field_index(&n1, &m1, j);
                get_tem_modes_from_field_index(&n2, &m2, k);

                fprintf(fp, "(%d, %d) [%d %d -> %d %d]\n", j, k, n1, m1, n2, m2);
                if (space.a12[j][k] != NULL) {
                    fprintf(fp, "a12: %s, ", complex_form15(*(space.a12[j][k])));
                }
                if (space.a21[j][k] != NULL) {
                    fprintf(fp, "a21: %s\n", complex_form15(*(space.a21[j][k])));
                }
            }
        }
    }
}



//! Fill relevant sparse matrix elements with beamsplitter component terms

/*!
 * \param f frequency of light field to fill
 */
void fill_matrix_beamsplitter_elements(double f) {
    // fill matrix for beam splitters
    int beamsplitter_index;
    for (beamsplitter_index = 0;
            beamsplitter_index < inter.num_beamsplitters;
            beamsplitter_index++) {
        beamsplitter_t *bs;
        bs = &(inter.bs_list[beamsplitter_index]);

        double r = sqrt(bs->R);
        double t = sqrt(bs->T);
        double alpha = bs->alpha * RAD;
        double phi = 2.0 * bs->phi * RAD * (1.0 + f / inter.f0) * cos(alpha);
        double c0 = r * cos(phi);
        double s0 = r * sin(phi);

        // field amplitudes
        int j, k;

        for (j = 0; j < inter.num_fields; j++) { // outputs
            for (k = 0; k < inter.num_fields; k++) { // frequencies

                if (bs->a12[j][k] != NULL) {
                    (bs->a12[j][k])->re = -c0;
                    (bs->a12[j][k])->im = -s0;
                    (bs->a21[j][k])->re = -c0;
                    (bs->a21[j][k])->im = -s0;
                    //printf("general alpha %f phi %f r %f t %f c0 %f s0 %f\n",alpha, phi, r, t, c0, s0);
                    //printf("before a12, a21 (%d, %d) %s %s\n",j,k,complex_form(*(bs->a12[j][k])), complex_form(*(bs->a21[j][k])));

                    *(bs->a12[j][k]) = z_by_z(*(bs->a12[j][k]), bs->k12[j][k]);
                    *(bs->a21[j][k]) = z_by_z(*(bs->a21[j][k]), bs->k21[j][k]);
                    //printf("after a12, a21 (%d, %d) %s %s\n",j,k,complex_form(*(bs->a12[j][k])), complex_form(*(bs->a21[j][k])));
                }

                if (bs->a13[j][k] != NULL) {
                    (bs->a13[j][k])->re = 0.0;
                    (bs->a13[j][k])->im = -t;
                    (bs->a31[j][k])->re = 0.0;
                    (bs->a31[j][k])->im = -t;

                    *(bs->a13[j][k]) = z_by_z(*(bs->a13[j][k]), bs->k13[j][k]);
                    *(bs->a31[j][k]) = z_by_z(*(bs->a31[j][k]), bs->k31[j][k]);
                    //printf("a13, a31 (%d, %d) %s %s\n",j,k,complex_form(*(bs->a13[j][k])), complex_form(*(bs->a31[j][k])));

                }

                if (bs->a34[j][k] != NULL) {
                    (bs->a34[j][k])->re = -c0;
                    (bs->a34[j][k])->im = s0;
                    (bs->a43[j][k])->re = -c0;
                    (bs->a43[j][k])->im = s0;
                    //printf("before a34, a43 (%d, %d) %s %s\n",j,k,complex_form(*(bs->a34[j][k])), complex_form(*(bs->a43[j][k])));


                    *(bs->a34[j][k]) = z_by_z(*(bs->a34[j][k]), bs->k34[j][k]);
                    *(bs->a43[j][k]) = z_by_z(*(bs->a43[j][k]), bs->k43[j][k]);
                    //printf("after a34, a43 (%d, %d) %s %s\n",j,k,complex_form(*(bs->a34[j][k])), complex_form(*(bs->a43[j][k])));
                }

                if (bs->a24[j][k] != NULL) {
                    (bs->a24[j][k])->re = 0.0;
                    (bs->a24[j][k])->im = -t;
                    (bs->a42[j][k])->re = 0.0;
                    (bs->a42[j][k])->im = -t;

                    *(bs->a24[j][k]) = z_by_z(*(bs->a24[j][k]), bs->k24[j][k]);
                    *(bs->a42[j][k]) = z_by_z(*(bs->a42[j][k]), bs->k42[j][k]);
                    //printf("a24, a42 (%d, %d) %s %s\n",j,k,complex_form(*(bs->a24[j][k])), complex_form(*(bs->a42[j][k])));
                }
            }
        }

        // quantum noise
        if (inter.qcorr) {
            int j, k;
            for (j = 0; j < inter.num_fields; j++) { // outputs
                for (k = 0; k < inter.num_fields; k++) { // frequencies
                    if (bs->q12[j][k] != NULL) {
                        (bs->q12[j][k])->re = -c0;
                        (bs->q12[j][k])->im = -s0;
                    }

                    if (bs->q21[j][k] != NULL) {
                        (bs->q21[j][k])->re = -c0;
                        (bs->q21[j][k])->im = -s0;
                    }

                    if (bs->q13[j][k] != NULL) {
                        (bs->q13[j][k])->re = 0.0;
                        (bs->q13[j][k])->im = -t;
                    }

                    if (bs->q31[j][k] != NULL) {
                        (bs->q31[j][k])->re = 0.0;
                        (bs->q31[j][k])->im = -t;
                    }

                    if (bs->q34[j][k] != NULL) {
                        (bs->q34[j][k])->re = -c0;
                        (bs->q34[j][k])->im = s0;
                    }

                    if (bs->q43[j][k] != NULL) {
                        (bs->q43[j][k])->re = -c0;
                        (bs->q43[j][k])->im = s0;
                    }

                    if (bs->q24[j][k] != NULL) {
                        (bs->q24[j][k])->re = 0.0;
                        (bs->q24[j][k])->im = -t;
                    }

                    if (bs->q42[j][k] != NULL) {
                        (bs->q42[j][k])->re = 0.0;
                        (bs->q42[j][k])->im = -t;
                    }
                }
            }
        }

    }
}

//! Fill relevant sparse matrix elements with grating component terms

/*!
 *
 * \todo qcorr branches untested
 */
void fill_matrix_grating_elements(void) {
    double eta[4] = {0};

    // fill matrix for gratings
    int grating_index;
    for (grating_index = 0; grating_index < inter.num_gratings; grating_index++) {
        grating_t *grating;
        grating = &(inter.grating_list[grating_index]);

        int j, k;
        switch (grating->num_of_ports) {
            case 2:
                for (j = 0; j < 2; j++) {
                    eta[j] = sqrt(grating->eta[j]);
                }

                // field amplitudes
                for (j = 0; j < inter.num_fields; j++) {
                    for (k = 0; k < inter.num_fields; k++) {
                        if (j == k) {
                            if (grating->a12[j][k] != NULL) {
                                (grating->a12[j][k])->re = 0.0;
                                (grating->a12[j][k])->im = -eta[1];
                                (grating->a21[j][k])->re = 0.0;
                                (grating->a21[j][k])->im = -eta[1];
                            }

                            if (grating->a11[j][k] != NULL) {
                                (grating->a11[j][k])->re = -eta[0];
                                (grating->a11[j][k])->im = 0.0;
                            }

                            if (grating->a22[j][k] != NULL) {
                                (grating->a22[j][k])->re = -eta[0];
                                (grating->a22[j][k])->im = 0.0;
                            }
                        } else {
                            if (grating->a12[j][k] != NULL) {
                                *(grating->a12[j][k]) = complex_0;
                                *(grating->a21[j][k]) = complex_0;
                            }

                            if (grating->a11[j][k] != NULL) {
                                *(grating->a11[j][k]) = complex_0;
                            }

                            if (grating->a22[j][k] != NULL) {
                                *(grating->a22[j][k]) = complex_0;
                            }
                        }
                    }
                }

                // quantum noise terms
                if (inter.qcorr) {
                    for (j = 0; j < inter.num_fields; j++) {
                        for (k = 0; k < inter.num_fields; k++) {
                            if (j == k) {
                                if (grating->q12[j][k] != NULL) {
                                    (grating->q12[j][k])->re = 0.0;
                                    (grating->q12[j][k])->im = -eta[1];
                                }

                                if (grating->q21[j][k] != NULL) {
                                    (grating->q21[j][k])->re = 0.0;
                                    (grating->q21[j][k])->im = -eta[1];
                                }

                                if (grating->q11[j][k] != NULL) {
                                    (grating->q11[j][k])->re = -eta[0];
                                    (grating->q11[j][k])->im = 0.0;
                                }

                                if (grating->q22[j][k] != NULL) {
                                    (grating->q22[j][k])->re = -eta[0];
                                    (grating->q22[j][k])->im = 0.0;
                                }
                            } else {
                                if (grating->q12[j][k] != NULL) {
                                    (grating->q12[j][k])->re = 0.0;
                                    (grating->q12[j][k])->im = 0.0;
                                }

                                if (grating->q21[j][k] != NULL) {
                                    (grating->q21[j][k])->re = 0.0;
                                    (grating->q21[j][k])->im = 0.0;
                                }

                                if (grating->q11[j][k] != NULL) {
                                    (grating->q11[j][k])->re = 0.0;
                                    (grating->q11[j][k])->im = 0.0;
                                }

                                if (grating->q22[j][k] != NULL) {
                                    (grating->q22[j][k])->re = 0.0;
                                    (grating->q22[j][k])->im = 0.0;
                                }
                            }
                        }
                    }
                }
                break;
            case 3:
                for (j = 0; j < 4; j++) {
                    eta[j] = sqrt(grating->eta[j]);
                }

                /* coupling matrix :
                 * eta2 exp(i phi2)  eta1 exp(i phi1)  eta0 exp(i phi0) 
                 * eta1 exp(i phi1)  rho0 exp(i phi0)  eta1 exp(i phi1) 
                 * eta0 exp(i phi0)  eta1 exp(i phi1)  eta2 exp(i phi2) 
                 */
                double s3 = sin(grating->phi[0]) * eta[3];
                double c3 = cos(grating->phi[0]) * eta[3];

                double s0 = sin(grating->phi[0]) * eta[0];
                double c0 = cos(grating->phi[0]) * eta[0];

                double s1 = sin(grating->phi[1]) * eta[1];
                double c1 = cos(grating->phi[1]) * eta[1];

                double s2 = sin(grating->phi[2]) * eta[2];
                double c2 = cos(grating->phi[2]) * eta[2];

                // field amplitudes
                for (j = 0; j < inter.num_fields; j++) {
                    for (k = 0; k < inter.num_fields; k++) {
                        if (j == k) {
                            if (grating->a11[j][k] != NULL) {
                                (grating->a11[j][k])->re = -c2;
                                (grating->a11[j][k])->im = -s2;
                            }

                            if (grating->a22[j][k] != NULL) {
                                (grating->a22[j][k])->re = -c2;
                                (grating->a22[j][k])->im = -s2;
                            }

                            if (grating->a33[j][k] != NULL) {
                                (grating->a33[j][k])->re = -c3;
                                (grating->a33[j][k])->im = -s3;
                            }

                            if (grating->a12[j][k] != NULL) {
                                (grating->a12[j][k])->re = -c0;
                                (grating->a12[j][k])->im = -s0;
                                (grating->a21[j][k])->re = -c0;
                                (grating->a21[j][k])->im = -s0;
                            }

                            if (grating->a13[j][k] != NULL) {
                                (grating->a13[j][k])->re = -c1;
                                (grating->a13[j][k])->im = -s1;
                                (grating->a31[j][k])->re = -c1;
                                (grating->a31[j][k])->im = -s1;
                            }

                            if (grating->a32[j][k] != NULL) {
                                (grating->a32[j][k])->re = -c1;
                                (grating->a32[j][k])->im = -s1;
                                (grating->a23[j][k])->re = -c1;
                                (grating->a23[j][k])->im = -s1;
                            }
                        } else {
                            if (grating->a11[j][k] != NULL) {
                                *(grating->a11[j][k]) = complex_0;
                            }

                            if (grating->a22[j][k] != NULL) {
                                *(grating->a22[j][k]) = complex_0;
                            }

                            if (grating->a33[j][k] != NULL) {
                                *(grating->a33[j][k]) = complex_0;
                            }

                            if (grating->a12[j][k] != NULL) {
                                *(grating->a12[j][k]) = complex_0;
                                *(grating->a21[j][k]) = complex_0;
                            }

                            if (grating->a13[j][k] != NULL) {
                                *(grating->a13[j][k]) = complex_0;
                                *(grating->a31[j][k]) = complex_0;
                            }

                            if (grating->a23[j][k] != NULL) {
                                *(grating->a23[j][k]) = complex_0;
                                *(grating->a32[j][k]) = complex_0;
                            }
                        }
                    }
                }

                // quantum noise terms
                if (inter.qcorr) {
                    for (j = 0; j < inter.num_fields; j++) {
                        for (k = 0; k < inter.num_fields; k++) {
                            if (j == k) {
                                if (grating->q11[j][k] != NULL) {
                                    (grating->q11[j][k])->re = -c2;
                                    (grating->q11[j][k])->im = -s2;
                                }

                                if (grating->q22[j][k] != NULL) {
                                    (grating->q22[j][k])->re = -c2;
                                    (grating->q22[j][k])->im = -s2;
                                }

                                if (grating->q33[j][k] != NULL) {
                                    (grating->q33[j][k])->re = -c3;
                                    (grating->q33[j][k])->im = -s3;
                                }

                                if (grating->q12[j][k] != NULL) {
                                    (grating->q12[j][k])->re = -c0;
                                    (grating->q12[j][k])->im = -s0;
                                }

                                if (grating->q21[j][k] != NULL) {
                                    (grating->q21[j][k])->re = -c0;
                                    (grating->q21[j][k])->im = -s0;
                                }

                                if (grating->q13[j][k] != NULL) {
                                    (grating->q13[j][k])->re = -c1;
                                    (grating->q13[j][k])->im = -s1;
                                }

                                if (grating->q31[j][k] != NULL) {
                                    (grating->q31[j][k])->re = -c1;
                                    (grating->q31[j][k])->im = -s1;
                                }

                                if (grating->q23[j][k] != NULL) {
                                    (grating->q23[j][k])->re = -c1;
                                    (grating->q23[j][k])->im = -s1;
                                }

                                if (grating->q32[j][k] != NULL) {
                                    (grating->q32[j][k])->re = -c1;
                                    (grating->q32[j][k])->im = -s1;
                                }
                            } else {
                                if (grating->q11[j][k] != NULL) {
                                    (grating->q11[j][k])->re = 0.0;
                                    (grating->q11[j][k])->im = 0.0;
                                }

                                if (grating->q22[j][k] != NULL) {
                                    (grating->q22[j][k])->re = 0.0;
                                    (grating->q22[j][k])->im = 0.0;
                                }

                                if (grating->q33[j][k] != NULL) {
                                    (grating->q33[j][k])->re = 0.0;
                                    (grating->q33[j][k])->im = 0.0;
                                }

                                if (grating->q12[j][k] != NULL) {
                                    (grating->q12[j][k])->re = 0.0;
                                    (grating->q12[j][k])->im = 0.0;
                                }

                                if (grating->q21[j][k] != NULL) {
                                    (grating->q21[j][k])->re = 0.0;
                                    (grating->q21[j][k])->im = 0.0;
                                }

                                if (grating->q13[j][k] != NULL) {
                                    (grating->q13[j][k])->re = 0.0;
                                    (grating->q13[j][k])->im = 0.0;
                                }

                                if (grating->q31[j][k] != NULL) {
                                    (grating->q31[j][k])->re = 0.0;
                                    (grating->q31[j][k])->im = 0.0;
                                }

                                if (grating->q23[j][k] != NULL) {
                                    (grating->q23[j][k])->re = 0.0;
                                    (grating->q23[j][k])->im = 0.0;
                                }

                                if (grating->q32[j][k] != NULL) {
                                    (grating->q32[j][k])->re = 0.0;
                                    (grating->q32[j][k])->im = 0.0;
                                }
                            }
                        }
                    }
                }
                break;
            case 4:
                for (j = 0; j < 2; j++) {
                    eta[j] = sqrt(grating->eta[j]);
                }

                // field amplitudes
                for (j = 0; j < inter.num_fields; j++) {
                    for (k = 0; k < inter.num_fields; k++) {
                        if (j == k) {
                            if (grating->a14[j][k] != NULL) {
                                (grating->a14[j][k])->re = -eta[0];
                                (grating->a14[j][k])->im = 0.0;
                                (grating->a41[j][k])->re = -eta[0];
                                (grating->a41[j][k])->im = 0.0;
                            }

                            if (grating->a13[j][k] != NULL) {
                                (grating->a13[j][k])->re = 0.0;
                                (grating->a13[j][k])->im = -eta[1];
                                (grating->a31[j][k])->re = 0.0;
                                (grating->a31[j][k])->im = -eta[1];
                            }

                            if (grating->a42[j][k] != NULL) {
                                (grating->a42[j][k])->re = 0.0;
                                (grating->a42[j][k])->im = -eta[1];
                                (grating->a24[j][k])->re = 0.0;
                                (grating->a24[j][k])->im = -eta[1];
                            }
                        } else {
                            if (grating->a14[j][k] != NULL) {
                                *(grating->a14[j][k]) = complex_0;
                                *(grating->a41[j][k]) = complex_0;
                            }

                            if (grating->a13[j][k] != NULL) {
                                *(grating->a13[j][k]) = complex_0;
                                *(grating->a31[j][k]) = complex_0;
                            }

                            if (grating->a42[j][k] != NULL) {
                                *(grating->a42[j][k]) = complex_0;
                                *(grating->a24[j][k]) = complex_0;
                            }
                        }
                    }
                }

                // quantum noise terms
                if (inter.qcorr) {
                    for (j = 0; j < inter.num_fields; j++) {
                        for (k = 0; k < inter.num_fields; k++) {
                            if (j == k) {
                                if (grating->q14[j][k] != NULL) {
                                    (grating->q14[j][k])->re = -eta[0];
                                    (grating->q14[j][k])->im = 0.0;
                                }

                                if (grating->q41[j][k] != NULL) {
                                    (grating->q41[j][k])->re = -eta[0];
                                    (grating->q41[j][k])->im = 0.0;
                                }

                                if (grating->q13[j][k] != NULL) {
                                    (grating->q13[j][k])->re = 0.0;
                                    (grating->q13[j][k])->im = -eta[1];
                                }

                                if (grating->q31[j][k] != NULL) {
                                    (grating->q31[j][k])->re = 0.0;
                                    (grating->q31[j][k])->im = -eta[1];
                                }

                                if (grating->q24[j][k] != NULL) {
                                    (grating->q24[j][k])->re = 0.0;
                                    (grating->q24[j][k])->im = -eta[1];
                                }

                                if (grating->q42[j][k] != NULL) {
                                    (grating->q42[j][k])->re = 0.0;
                                    (grating->q42[j][k])->im = -eta[1];
                                }
                            } else {
                                if (grating->q14[j][k] != NULL) {
                                    (grating->q14[j][k])->re = 0.0;
                                    (grating->q14[j][k])->im = 0.0;
                                }

                                if (grating->q41[j][k] != NULL) {
                                    (grating->q41[j][k])->re = 0.0;
                                    (grating->q41[j][k])->im = 0.0;
                                }

                                if (grating->q13[j][k] != NULL) {
                                    (grating->q13[j][k])->re = 0.0;
                                    (grating->q13[j][k])->im = 0.0;
                                }

                                if (grating->q31[j][k] != NULL) {
                                    (grating->q31[j][k])->re = 0.0;
                                    (grating->q31[j][k])->im = 0.0;
                                }

                                if (grating->q24[j][k] != NULL) {
                                    (grating->q24[j][k])->re = 0.0;
                                    (grating->q24[j][k])->im = 0.0;
                                }

                                if (grating->q42[j][k] != NULL) {
                                    (grating->q42[j][k])->re = 0.0;
                                    (grating->q42[j][k])->im = 0.0;
                                }
                            }
                        }
                    }
                }
                break;
            default:
                bug_error("gratings, wrong number of ports");
        }
    }
}

//! Fill relevant sparse matrix elements with space component terms

/*!
 * \param f frequency of light field to fill
 *
 * \todo qcorr branch untested
 */
void fill_matrix_space_elements(double f) {
    // fill matrix for spaces
    int space_index;
    for (space_index = 0; space_index < inter.num_spaces; space_index++) {
        space_t *space;
        space = &(inter.space_list[space_index]);

        double omega = TWOPI * f;
        // changed phi from plus to minus, see manual eq 3.32 281005
        double phi = omega * space->n / init.clight * space->L;

        // field amplitudes
        int j, k;
        for (j = 0; j < inter.num_fields; j++) {
            for (k = 0; k < inter.num_fields; k++) {

                int tn, tm;
                get_tem_modes_from_field_index(&tn, &tm, j);

                double gouyx = 0.0;
                double gouyy = 0.0;
                if (inter.set_tem_phase_zero & 2) {
                    gouyx = tn * RAD * space->gouy_x;
                    gouyy = tm * RAD * space->gouy_y;
                } else {
                    gouyx = (tn + 0.5) * RAD * space->gouy_x;
                    gouyy = (tm + 0.5) * RAD * space->gouy_y;
                }

                double c0 = cos(-phi + gouyx + gouyy);
                double s0 = sin(-phi + gouyx + gouyy);

                if (space->a12[j][k] != NULL) {
                    (space->a12[j][k])->re = -c0;
                    (space->a12[j][k])->im = -s0;

                    *(space->a12[j][k]) = z_by_z(*(space->a12[j][k]), space->k12[j][k]);
                }

                if (space->a21[j][k] != NULL) {
                    (space->a21[j][k])->re = -c0;
                    (space->a21[j][k])->im = -s0;

                    *(space->a21[j][k]) = z_by_z(*(space->a21[j][k]), space->k21[j][k]);
                }
            }
        }

        // quantum noise terms
        if (inter.qcorr) {
            int j, k;
            for (j = 0; j < inter.num_fields; j++) {
                for (k = 0; k < inter.num_fields; k++) {

                    int tn, tm;
                    get_tem_modes_from_field_index(&tn, &tm, j);

                    double gouyx = 0.0;
                    double gouyy = 0.0;
                    if (inter.set_tem_phase_zero & 2) {
                        gouyx = tn * RAD * space->gouy_x;
                        gouyy = tm * RAD * space->gouy_y;
                    } else {
                        gouyx = (tn + 0.5) * RAD * space->gouy_x;
                        gouyy = (tm + 0.5) * RAD * space->gouy_y;
                    }
                    double c0 = cos(-phi + gouyx + gouyy);
                    double s0 = sin(-phi + gouyx + gouyy);

                    if (space->q12[j][k] != NULL) {
                        (space->q12[j][k])->re = -c0;
                        (space->q12[j][k])->im = -s0;
                    }

                    if (space->q21[j][k] != NULL) {
                        (space->q21[j][k])->re = -c0;
                        (space->q21[j][k])->im = -s0;
                    }
                }
            }
        }
    }
}

//! Fill relevant sparse matrix elements with squeezer component terms

/*!
 *
 * \todo untested
 */
void fill_matrix_squeezer_elements(void) {
    // fill matrix for squeezers
    int squeezer_index;
    for (squeezer_index = 0;
            squeezer_index < inter.num_squeezers;
            squeezer_index++) {

        squeezer_t *squeezer;
        squeezer = &(inter.squeezer_list[squeezer_index]);

        // the classical part of the squeezer is supposed to work like a space,
        // so set up the classical part with unit elements and in the normal
        // matrix element locations as for a space

        // field amplitudes
        int j, k;
        for (j = 0; j < inter.num_fields; j++) {
            for (k = 0; k < inter.num_fields; k++) {

                if (squeezer->a11[j][k] != NULL) {
                    (squeezer->a11[j][k])->re = 0.0;
                    (squeezer->a11[j][k])->im = 0.0;
                    // need to set the k's here???
                }

                if (squeezer->a22[j][k] != NULL) {
                    (squeezer->a22[j][k])->re = 0.0;
                    (squeezer->a22[j][k])->im = 0.0;
                    // need to set the k's here???
                }

                if (squeezer->a12[j][k] != NULL) {
                    (squeezer->a12[j][k])->re = -1.0;
                    (squeezer->a12[j][k])->im = 0.0;

                    //*(squeezer->a12[j][k]) = z_by_z(*(squeezer->a12[j][k]), squeezer->k12[j][k]);
                }

                if (squeezer->a21[j][k] != NULL) {
                    (squeezer->a21[j][k])->re = -1.0;
                    (squeezer->a21[j][k])->im = 0.0;

                    //*(squeezer->a21[j][k]) = z_by_z(*(squeezer->a21[j][k]), squeezer->k21[j][k]);
                }
            }
        }

        // quantum noise terms
        if (inter.qcorr) {
            // define the squeezer's transformation matrix elements
            double squeeze_factor = squeezer->squeeze_factor;
            double squeeze_angle = squeezer->squeeze_angle;
            printf("squeeze_factor = %g\n", squeeze_factor);
            printf("squeeze_angle = %g\n", squeeze_angle);
            double m11 =
                    cosh(squeeze_factor) + sinh(squeeze_factor) * cos(2.0 * squeeze_angle);
            // note: m12 == m21
            double m12 = sinh(squeeze_factor) * sin(2.0 * squeeze_angle);
            double m22 =
                    cosh(squeeze_factor) - sinh(squeeze_factor) * cos(2.0 * squeeze_angle);

            printf("m11 = %g\n", m11);
            printf("m12 = %g\n", m12);
            printf("m22 = %g\n", m22);

            int j, k;
            for (j = 0; j < inter.num_fields; j++) {
                for (k = 0; k < inter.num_fields; k++) {

                    if (squeezer->q11[j][k] != NULL) {
                        (squeezer->q11[j][k])->re = -m11;
                        (squeezer->q11[j][k])->im = 0.0;
                    }

                    if (squeezer->q22[j][k] != NULL) {
                        (squeezer->q22[j][k])->re = -m22;
                        (squeezer->q22[j][k])->im = 0.0;
                    }

                    if (squeezer->q12[j][k] != NULL) {
                        (squeezer->q12[j][k])->re = -m12;
                        (squeezer->q12[j][k])->im = 0.0;
                    }

                    if (squeezer->q21[j][k] != NULL) {
                        (squeezer->q21[j][k])->re = -m12;
                        (squeezer->q21[j][k])->im = 0.0;
                    }

                }
            }
        }
    }
}

//! Fill relevant sparse matrix elements with modulator component terms

/*!
 *
 * \todo qcorr branch untested
 */
void fill_matrix_modulator_elements(void) {
    // fill matrix for modulators
    int modulator_index;
    for (modulator_index = 0;
            modulator_index < inter.num_modulators;
            modulator_index++) {

        modulator_t *modulator;
        modulator = &(inter.modulator_list[modulator_index]);

        double c0;
        if (modulator->type == MODTYPE_AM) {
            c0 = 1.0 - modulator->modulation_index * 0.5;
        } else {
            c0 = bessn(0, modulator->modulation_index);
        }

        if (modulator->single_sideband_mode == ON) {
            c0 = 1.0 - (1.0 - c0) * 0.5;
        }

        // field amplitudes
        int j, k;
        for (j = 0; j < inter.num_fields; j++) {
            for (k = 0; k < inter.num_fields; k++) {
                if (modulator->a12[j][k] != NULL) {
                    (modulator->a12[j][k])->re = -c0;
                    (modulator->a12[j][k])->im = 0.0;

                    *(modulator->a12[j][k]) =
                            z_by_z(*(modulator->a12[j][k]), modulator->k12[j][k]);
                }

                if (modulator->a21[j][k] != NULL) {
                    (modulator->a21[j][k])->re = -c0;
                    (modulator->a21[j][k])->im = 0.0;

                    *(modulator->a21[j][k]) =
                            z_by_z(*(modulator->a21[j][k]), modulator->k21[j][k]);
                }
            }
        }

        // quantum noise terms
        if (inter.qcorr) {
            int j, k;
            for (j = 0; j < inter.num_fields; j++) {
                for (k = 0; k < inter.num_fields; k++) {
                    if (modulator->q12[j][k] != NULL) {
                        (modulator->q12[j][k])->re = -c0;
                        (modulator->q12[j][k])->im = 0.0;
                    }

                    if (modulator->q21[j][k] != NULL) {
                        (modulator->q21[j][k])->re = -c0;
                        (modulator->q21[j][k])->im = 0.0;
                    }
                }
            }
        }
    }
}

//! Fill relevant sparse matrix elements with lens component terms

/*!
 *
 * \todo qcorr branch untested
 */
void fill_matrix_lens_elements(void) {
    // fill matrix for lenses
    int lens_index;
    for (lens_index = 0; lens_index < inter.num_lenses; lens_index++) {
        lens_t *lens;
        lens = &(inter.lens_list[lens_index]);

        // field amplitudes
        int j, k;
        for (j = 0; j < inter.num_fields; j++) {
            for (k = 0; k < inter.num_fields; k++) {
                if (lens->a12[j][k] != NULL) {
                    (lens->a12[j][k])->re = -1.0;
                    (lens->a12[j][k])->im = 0.0;

                    *(lens->a12[j][k]) = z_by_z(*(lens->a12[j][k]), lens->k12[j][k]);
                }

                if (lens->a21[j][k] != NULL) {
                    (lens->a21[j][k])->re = -1.0;
                    (lens->a21[j][k])->im = 0.0;

                    *(lens->a21[j][k]) = z_by_z(*(lens->a21[j][k]), lens->k21[j][k]);
                }
            }
        }

        // quantum noise terms
        if (inter.qcorr) {
            int j, k;
            for (j = 0; j < inter.num_fields; j++) {
                for (k = 0; k < inter.num_fields; k++) {
                    if (lens->q12[j][k] != NULL) {
                        (lens->q12[j][k])->re = -1.0;
                        (lens->q12[j][k])->im = 0.0;
                    }

                    if (lens->q21[j][k] != NULL) {
                        (lens->q21[j][k])->re = -1.0;
                        (lens->q21[j][k])->im = 0.0;
                    }
                }
            }
        }
    }
}

//! Fill relevant sparse matrix elements with diode component terms

/*!
 *
 * \todo qcorr branch untested
 */
void fill_matrix_diode_elements(void) {
    // fill matrix for diodes
    int diode_index;
    for (diode_index = 0; diode_index < inter.num_diodes; diode_index++) {
        diode_t *diode;
        diode = &(inter.diode_list[diode_index]);

        // field amplitudes
        int j, k;
        for (j = 0; j < inter.num_fields; j++) {
            for (k = 0; k < inter.num_fields; k++) {

                if (diode->a12[j][k] != NULL) {
                    (diode->a12[j][k])->re = -1.0;
                    (diode->a12[j][k])->im = 0.0;

                    *(diode->a12[j][k]) = z_by_z(*(diode->a12[j][k]), diode->k12[j][k]);
                }

                if (diode->a21[j][k] != NULL) {
                    (diode->a21[j][k])->re = -pow(10.0, -diode->S / 20.0);
                    (diode->a21[j][k])->im = 0.0;

                    *(diode->a21[j][k]) = z_by_z(*(diode->a21[j][k]), diode->k21[j][k]);
                }
            }
        }

        // quantum noise terms
        if (inter.qcorr) {
            int j, k;
            for (j = 0; j < inter.num_fields; j++) {
                for (k = 0; k < inter.num_fields; k++) {
                    if (diode->q12[j][k] != NULL) {
                        (diode->q12[j][k])->re = -1.0;
                        (diode->q12[j][k])->im = 0.0;
                    }

                    if (diode->q21[j][k] != NULL) {
                        (diode->q21[j][k])->re = -pow(10.0, -diode->S / 20.0);
                        (diode->q21[j][k])->im = 0.0;
                    }
                }
            }
        }
    }
}

//! Set a value of the RHS vector

/*!
 * \param vector_index The vector index (aka vector element) to set to z_value
 * \param z_value The value to set for the RHS vector element
 * \param matrix_type The type of matrix to set up (STANDARD or QCORR)
 */
void klu_rhs(int vector_index, complex_t z_value, int matrix_type) {
    // sanity checks on inputs
    assert(matrix_type == STANDARD || matrix_type == QCORR);

    // assign the value to the RHS vector element
    switch (matrix_type) {
        case STANDARD:
            // make sure the element to be assigned is within bounds
            assert(vector_index > -1);
            assert(vector_index < inter.num_eqns);

            klu.rhs[2 * vector_index] = z_value.re;
            klu.rhs[2 * vector_index + 1] = z_value.im;
            break;
        case QCORR:
            // make sure the element to be assigned is within bounds
            assert(vector_index > -1);
            assert(vector_index < inter.num_quantum_eqns);

            break;
        default:
            bug_error("no such type");
    }
}


//! Set a value of the RHS vector

/*!
 * \param vector_index The vector index (aka vector element) to set to z_value
 * \param z_value The value to set for the RHS vector element
 * \param matrix_type The type of matrix to set up (STANDARD or QCORR)
 */
void sparse_rhs(int vector_index, complex_t z_value, int matrix_type) {
    // sanity checks on inputs
    assert(matrix_type == STANDARD || matrix_type == QCORR);

    // assign the value to the RHS vector element
    switch (matrix_type) {
        case STANDARD:
            // make sure the element to be assigned is within bounds
            assert(vector_index > -1);
            assert(vector_index < inter.num_eqns);

            spa.sprhs[vector_index] = z_value;
            break;
        case QCORR:
            // make sure the element to be assigned is within bounds
            assert(vector_index > -1);
            assert(vector_index < inter.num_quantum_eqns);

            spa.Qsprhs[vector_index] = z_value;
            break;
        default:
            bug_error("no such type");
    }
}



//! Set a value of the RHS vector

/*!
 * \param vector_index The vector index (aka vector element) to set to z_value
 * \param z_value The value to set for the RHS vector element
 * \param matrix_type The type of matrix to set up (STANDARD or QCORR)
 */
void spa_rhs(int vector_index, complex_t z_value, int matrix_type) {
    // sanity checks on inputs
    assert(matrix_type == STANDARD || matrix_type == QCORR);

    if (options.klu) {
        klu_rhs(vector_index, z_value, matrix_type);
    } else {
        sparse_rhs(vector_index, z_value, matrix_type);
    }
}



//! Clear the RHS vector (reset to zero)

/*!
 * \param matrix_type the type of matrix to clear (standard or quantum)
 */
void spa_sparse_rhs_clear(int matrix_type) {
    int i;
    switch (matrix_type) {
        case STANDARD:
            for (i = 0; i < inter.num_eqns; i++) {
                spa.sprhs[i] = complex_0;
            }
            break;
        case QCORR:
            for (i = 0; i < inter.num_quantum_eqns; i++) {
                spa.Qsprhs[i] = complex_0;
            }
            break;
        default:
            bug_error("no such type");
    }
}

//! Clear the RHS vector (reset to zero) (KLU package)

/*!
 * \param matrix_type the type of matrix to clear (standard or quantum)
 */
void spa_klu_rhs_clear(int matrix_type) {
    int i;
    switch (matrix_type) {
        case STANDARD:
            for (i = 0; i < 2 * inter.num_eqns; i++) {
                klu.rhs[i] = 0.0;
            }
            break;
        case QCORR:
            break;
        default:
            bug_error("no such type");
    }
}

//! Clear the RHS vector (reset to zero)

/*!
 * \param matrix_type the type of matrix to clear (standard or quantum)
 */
void spa_rhs_clear(int matrix_type) {
    // sanity checks on inputs
    assert(matrix_type == STANDARD || matrix_type == QCORR);

    if (options.klu) {
        spa_klu_rhs_clear(matrix_type);
    } else {
        spa_sparse_rhs_clear(matrix_type);
    }
}


//! Factor a sparse matrix

/*!
 * \param matrix_type the type of matrix to factor (standard or quantum)
 */
void spa_klu_factor(int matrix_type) {

    // sanity checks on inputs
    assert(matrix_type == STANDARD || matrix_type == QCORR);

    switch (matrix_type) {
        case STANDARD:
            klu_zl_refactor(klu.col_ptr, klu.row_index, klu.value, Symbolic, Numeric, &Common);
            break;
        case QCORR:
            break;
        default:
            bug_error("no such type");
    }
}

//! Factor a sparse matrix

/*!
 * \param matrix_type the type of matrix to factor (standard or quantum)
 */
void spa_sparse_factor(int matrix_type) {

    // sanity checks on inputs
    assert(matrix_type == STANDARD || matrix_type == QCORR);

    int error_code = 0;
    switch (matrix_type) {
        case STANDARD:
            error_code = spFactor(spa.spamat);
            report_sparse_matrix_error(error_code, 3);
            break;
        case QCORR:
            error_code = spFactor(spa.Qspamat);
            report_sparse_matrix_error(error_code, 4);
            break;
        default:
            bug_error("no such type");
    }
}

//! Factor a sparse matrix

/*!
 * \param matrix_type the type of matrix to factor (standard or quantum)
 */
void spa_factor(int matrix_type) {

    // sanity checks on inputs
    assert(matrix_type == STANDARD || matrix_type == QCORR);

    if (options.klu) {
        spa_klu_factor(matrix_type);
    } else {
        spa_sparse_factor(matrix_type);
    }
}

//! Dump the result of a solve to the screen

/*!
 *
 * \todo untested
 */
void dump_solution(void) {
    printf("\n  sol \n");
    printf(" Abs : Phase \n");
    int node_index;
    for (node_index = 0; node_index < inter.num_nodes; node_index++) {
        printf("%s = %s , %s   -     %s , %s\n", node_print(node_index),
                long_form(zabs(spa.spsol[2 * node_index]), 12),
                long_form(zdeg(spa.spsol[2 * node_index]), 4),
                long_form(zabs(spa.spsol[2 * node_index + 1]), 12),
                long_form(zdeg(spa.spsol[2 * node_index + 1]), 4)
                );
    }

    printf("\n Re : Im \n");
    for (node_index = 0; node_index < inter.num_nodes; node_index++) {
        printf("%s = %s , %s   -     %s , %s\n", node_print(node_index),
                long_form(spa.spsol[2 * node_index].re, 11),
                long_form(spa.spsol[2 * node_index].im, 11),
                long_form(spa.spsol[2 * node_index + 1].re, 11),
                long_form(spa.spsol[2 * node_index + 1].im, 11)
                );
    }
}

//! Dump the solution vector to the screen or to file

/*!
 * \param matrix_type the type of matrix to dump: STANDARD or QCORR
 * \param output_stream the stream to send to: stdout, file, file_append
 *
 * \todo untested
 */
void dump_solution_vector(int matrix_type, char *output_stream) {
    FILE *fp;

    switch (matrix_type) {
        case STANDARD:
            // decide on output stream
            if (strcmp(output_stream, "stdout") == 0) {
                fp = stdout;
                //debug_msg("Using stdout\n");
            } else if (strcmp(output_stream, "file") == 0) {
                fp = fopen("dump_solution_standard.dat", "w");
                //debug_msg("Using file\n");
            } else if (strcmp(output_stream, "file_append") == 0) {
                fp = fopen("dump_solution_standard.dat", "a");
                //debug_msg("Using file append\n");
            } else {
                // print a warning and just put stuff to stdout
                warn("Not a valid output stream %s.  Using stdout\n", output_stream);
                fp = stdout;
                debug_msg("Using fallback option (stdout)\n");
            }

            fprintf(fp, "\n Classical interferometer solution vector\n");
            fprintf(fp, "Length: %d\n", inter.num_eqns);
            fprintf(fp, "Real | Imaginary\n");
            int equation_index;
            for (equation_index = 0;
                    equation_index < inter.num_eqns;
                    equation_index++) {

                fprintf(fp, "spsol(%d) = " SOLN_FMT " | " SOLN_FMT "\n",
                        equation_index,
                        spa.spsol[equation_index].re, spa.spsol[equation_index].im);
            }

            // close things nicely if necessary
            if (strcmp(output_stream, "file") == 0 ||
                    strcmp(output_stream, "file_append") == 0) {
                fclose(fp);
            }
            break;

        case QCORR:
            // decide on output stream
            if (strcmp(output_stream, "stdout") == 0) {
                //debug_msg("Using stdout\n");
                fp = stdout;
            } else if (strcmp(output_stream, "file") == 0) {
                fp = fopen("dump_solution_qcorr.dat", "w");
                //debug_msg("Using file\n");
            } else if (strcmp(output_stream, "file_append") == 0) {
                fp = fopen("dump_solution_qcorr.dat", "a");
                //debug_msg("Using file append\n");
            } else {
                // print a warning and just put stuff to stdout
                warn("Not a valid output stream %s.  Using stdout\n", output_stream);
                fp = stdout;
                debug_msg("Using fallback option (stdout)\n");
            }

            fprintf(fp, "\n Quantum interferometer solution vector\n");
            fprintf(fp, "Length: %d\n", inter.num_quantum_eqns);
            for (equation_index = 0;
                    equation_index < inter.num_quantum_eqns;
                    equation_index++) {

                fprintf(fp, "spsol(%d) = " SOLN_FMT " | " SOLN_FMT "\n",
                        equation_index,
                        spa.Qspsol[equation_index].re, spa.Qspsol[equation_index].im);
            }

            // close things nicely if necessary
            if (strcmp(output_stream, "file") == 0 ||
                    strcmp(output_stream, "file_append") == 0) {
                fclose(fp);
            }
            break;
    }
}

//! Dump the RHS vector to the screen
/*
  void dump_rhs_vector(void)
  {
  int i;

  printf("\n rhs \n");
  printf("\n Re : Im \n");
  for (i=0; i<inter.num_nodes; i++) {
  printf("%s = %s , %s   -     %s , %s\n", node_print(i),
  long_form(spa.sprhs[2*i].re, 11),
  long_form(spa.sprhs[2*i].im, 11),
  long_form(spa.sprhs[2*i + 1].re, 11),
  long_form(spa.sprhs[2*i + 1].im, 11)
  );
  }
  }
 */

//! Dump the RHS vector to the screen or to file

/*!
 * \param matrix_type the type of matrix to dump: STANDARD or QCORR
 * \param output_stream the stream to send to: stdout, file, file_append
 *
 * \todo untested
 */
void dump_rhs_vector(int matrix_type, char *output_stream) {
    FILE *fp;

    switch (matrix_type) {
        case STANDARD:
            // decide on output stream
            if (strcmp(output_stream, "stdout") == 0) {
                fp = stdout;
                //debug_msg("Using stdout\n");
            } else if (strcmp(output_stream, "file") == 0) {
                fp = fopen("dump_rhs_vector_standard.dat", "w");
                //debug_msg("Using file\n");
            } else if (strcmp(output_stream, "file_append") == 0) {
                fp = fopen("dump_rhs_vector_standard.dat", "a");
                //debug_msg("Using file append\n");
            } else {
                // print a warning and just put stuff to stdout
                warn("Not a valid output stream %s.  Using stdout\n", output_stream);
                fp = stdout;
                debug_msg("Using fallback option (stdout)\n");
            }

            fprintf(fp, "\n Classical interferometer rhs vector\n");
            fprintf(fp, "Length: %d\n", inter.num_eqns);
            fprintf(fp, "Real | Imaginary\n");
            int j;
            for (j = 0; j < inter.num_eqns; j++) {
                fprintf(fp, RHS_FMT " | " RHS_FMT "\n",
                        spa.sprhs[j].re, spa.sprhs[j].im);
            }

            // close things nicely if necessary
            if (strcmp(output_stream, "file") == 0 ||
                    strcmp(output_stream, "file_append") == 0) {
                fclose(fp);
            }
            break;

        case QCORR:
            // decide on output stream
            if (strcmp(output_stream, "stdout") == 0) {
                //debug_msg("Using stdout\n");
                fp = stdout;
            } else if (strcmp(output_stream, "file") == 0) {
                fp = fopen("dump_rhs_vector_qcorr.dat", "w");
                //debug_msg("Using file\n");
            } else if (strcmp(output_stream, "file_append") == 0) {
                fp = fopen("dump_rhs_vector_qcorr.dat", "a");
                //debug_msg("Using file append\n");
            } else {
                // print a warning and just put stuff to stdout
                warn("Not a valid output stream %s.  Using stdout\n", output_stream);
                fp = stdout;
                debug_msg("Using fallback option (stdout)\n");
            }

            fprintf(fp, "\n Quantum interferometer rhs vector\n");
            fprintf(fp, "Length: %d\n", inter.num_quantum_eqns);
            for (j = 0; j < inter.num_quantum_eqns; j++) {
                fprintf(fp, RHS_FMT " | " RHS_FMT "\n",
                        spa.Qsprhs[j].re, spa.Qsprhs[j].im);
            }

            // close things nicely if necessary
            if (strcmp(output_stream, "file") == 0 ||
                    strcmp(output_stream, "file_append") == 0) {
                fclose(fp);
            }
            break;
    }
}

//! Dump the matrix to file or to the screen (use sparingly!!!)

/*!
 * \param matrix_type the type of matrix to dump: STANDARD or QCORR
 * \param output_stream the stream to send to: stdout, file, file_append
 *
 * \todo untested
 */
void dump_matrix(int matrix_type, char *output_stream) {
    FILE *fp;

    // sanity checks on inputs
    assert(matrix_type == STANDARD || matrix_type == QCORR);

    switch (matrix_type) {
        case STANDARD:
            // decide on output stream
            if (strcmp(output_stream, "stdout") == 0) {
                //debug_msg("Using stdout\n");
                fp = stdout;
            } else if (strcmp(output_stream, "file") == 0) {
                fp = fopen("dump_matrix_standard.dat", "w");
                //debug_msg("Using file\n");
            } else if (strcmp(output_stream, "file_append") == 0) {
                fp = fopen("dump_matrix_standard.dat", "a");
                //debug_msg("Using file append\n");
            } else {
                // print a warning and just put stuff to stdout
                warn("Not a valid output stream %s.  Using stdout\n", output_stream);
                debug_msg("Using fallback option (stdout)\n");
                fp = stdout;
            }

            assert(fp != NULL);
            // print out the matrix
            fprintf(fp, "\n Classical interferometer matrix\n");
            fprintf(fp, "Size: %d x %d\n", inter.num_eqns, inter.num_eqns);
            fprintf(fp, "\n Real part\n");
            int k, j;
            for (k = 0; k < inter.num_eqns; k++) {
                for (j = 0; j < inter.num_eqns - 1; j++) {
                    fprintf(fp, MAT_FMT " ", get_matrix_elem(j, k, matrix_type)->re);
                }
                fprintf(fp, MAT_FMT " | " MAT_FMT "\n",
                        get_matrix_elem(inter.num_eqns - 1, k, matrix_type)->re, spa.sprhs[k].re);
            }
            fprintf(fp, "\n Imaginary part\n");
            for (k = 0; k < inter.num_eqns; k++) {
                for (j = 0; j < inter.num_eqns - 1; j++) {
                    fprintf(fp, MAT_FMT " ", get_matrix_elem(j, k, matrix_type)->im);
                }
                fprintf(fp, MAT_FMT " | " MAT_FMT "\n",
                        get_matrix_elem(inter.num_eqns - 1, k, matrix_type)->im, spa.sprhs[k].im);
            }

            // close things nicely if necessary
            if (strcmp(output_stream, "file") == 0 ||
                    strcmp(output_stream, "file_append") == 0) {
                fclose(fp);
            }
            break;

        case QCORR:
            // decide on output stream
            if (strcmp(output_stream, "stdout") == 0) {
                fp = stdout;
                //debug_msg("Using stdout\n");
            } else if (strcmp(output_stream, "file") == 0) {
                fp = fopen("dump_matrix_qcorr.dat", "w");
                //debug_msg("Using file\n");
            } else if (strcmp(output_stream, "file_append") == 0) {
                fp = fopen("dump_matrix_qcorr.dat", "a");
                //debug_msg("Using file append\n");
            } else {
                // print a warning and just put stuff to stdout
                warn("Not a valid output stream %s.  Using stdout\n", output_stream);
                fp = stdout;
                debug_msg("Using fallback option (stdout)\n");
            }

            assert(fp != NULL);
            // print out the matrix
            fprintf(fp, "\n Quantum interferometer matrix\n");
            fprintf(fp, "Size: %d x %d\n",
                    inter.num_quantum_eqns, inter.num_quantum_eqns);
            fprintf(fp, "\n Real part\n");
            for (k = 0; k < inter.num_quantum_eqns; k++) {
                for (j = 0; j < inter.num_quantum_eqns - 1; j++) {
                    fprintf(fp, MAT_FMT " ", get_matrix_elem(j, k, matrix_type)->re);
                }
                fprintf(fp, MAT_FMT " | " MAT_FMT "\n",
                        get_matrix_elem(inter.num_quantum_eqns - 1, k, matrix_type)->re,
                        spa.Qsprhs[k].re);
            }
            fprintf(fp, "\n Imaginary part\n");
            for (k = 0; k < inter.num_quantum_eqns; k++) {
                for (j = 0; j < inter.num_quantum_eqns - 1; j++) {
                    fprintf(fp, MAT_FMT " ", get_matrix_elem(j, k, matrix_type)->im);
                }
                fprintf(fp, MAT_FMT " | " MAT_FMT "\n",
                        get_matrix_elem(inter.num_eqns - 1, k, matrix_type)->im,
                        spa.Qsprhs[k].im);
            }

            // close things nicely if necessary
            if (strcmp(output_stream, "file") == 0 ||
                    strcmp(output_stream, "file_append") == 0) {
                fclose(fp);
            }
            break;
    }

}

//! Dump the sparse matrix to file or to the screen

/*!
 * This routine dumps only the sparse information (the non-null) elements of
 * the sparse matrix to file or to screen.  This differs from dump_matrix()
 * in that it dumps *all* of the matrix.
 *
 * \param matrix_type the type of matrix to dump: STANDARD or QCORR
 * \param output_stream the stream to send to: stdout, file, file_append
 *
 * \todo untested
 */
void dump_sparse_matrix(int matrix_type, char *output_stream) {
    FILE *fp;

    // sanity checks on inputs
    assert(matrix_type == STANDARD || matrix_type == QCORR);

    // doesn't make things easy to read unfortunately...
    //! \todo see if any of these dump*() routines can have duplication reduced
    switch (matrix_type) {
        case STANDARD:
            // decide on output stream
            if (strcmp(output_stream, "stdout") == 0) {
                //debug_msg("Using stdout\n");
                fp = stdout;
            } else if (strcmp(output_stream, "file") == 0) {
                fp = fopen("dump_sparse_matrix_standard.dat", "w");
                //debug_msg("Using file\n");
            } else if (strcmp(output_stream, "file_append") == 0) {
                fp = fopen("dump_sparse_matrix_standard.dat", "a");
                //debug_msg("Using file append\n");
            } else {
                // print a warning and just put stuff to stdout
                warn("Not a valid output stream %s.  Using stdout\n", output_stream);
                debug_msg("Using fallback option (stdout)\n");
                fp = stdout;
            }

            // print out the matrix
            fprintf(fp, "\n Classical interferometer matrix\n");
            fprintf(fp, "Size: %d x %d\n", inter.num_eqns, inter.num_eqns);
            fprintf(fp, "\n bla Real part\n");
            int k, j;
            for (k = 0; k < inter.num_eqns; k++) {
                for (j = 0; j < inter.num_eqns; j++) {
                    double *p = spGetElement(spa.spamat, k + 1, j + 1);
                    complex_t *elem_val = (complex_t *) p;
                    if (elem_val->re != 0.0 && elem_val->im != 0.0) {
                        fprintf(fp, "spamat(%d, %d) = " SPA_FMT " | " SPA_FMT "\n",
                                k, j, elem_val->re, elem_val->im);
                    }
                }
            }

            // close things nicely if necessary
            if (strcmp(output_stream, "file") == 0 ||
                    strcmp(output_stream, "file_append") == 0) {
                fclose(fp);
            }
            break;

        case QCORR:
            // decide on output stream
            if (strcmp(output_stream, "stdout") == 0) {
                fp = stdout;
                //debug_msg("Using stdout\n");
            } else if (strcmp(output_stream, "file") == 0) {
                fp = fopen("dump_sparse_matrix_qcorr.dat", "w");
                //debug_msg("Using file\n");
            } else if (strcmp(output_stream, "file_append") == 0) {
                fp = fopen("dump_sparse_matrix_qcorr.dat", "a");
                //debug_msg("Using file append\n");
            } else {
                // print a warning and just put stuff to stdout
                warn("Not a valid output stream %s.  Using stdout\n", output_stream);
                fp = stdout;
                debug_msg("Using fallback option (stdout)\n");
            }

            // print out the matrix
            fprintf(fp, "\n Quantum interferometer matrix\n");
            fprintf(fp, "Size: %d x %d\n",
                    inter.num_quantum_eqns, inter.num_quantum_eqns);
            fprintf(fp, "\n Real part\n");
            for (k = 0; k < inter.num_quantum_eqns; k++) {
                for (j = 0; j < inter.num_quantum_eqns; j++) {
                    double *p = spGetElement(spa.Qspamat, k + 1, j + 1);
                    complex_t *elem_val = (complex_t *) p;
                    if (elem_val->re != 0.0 && elem_val->im != 0.0) {
                        fprintf(fp, "Qspamat(%d, %d) = " SPA_FMT " | " SPA_FMT "\n",
                                k, j, elem_val->re, elem_val->im);
                    }
                }
            }

            // close things nicely if necessary
            if (strcmp(output_stream, "file") == 0 ||
                    strcmp(output_stream, "file_append") == 0) {
                fclose(fp);
            }
            break;
    }

}

//! Check connections in interferometer

/*!
 *
 */

void check_connections(void) {
    int i;

    for (i = 0; i < inter.num_nodes; i++) {
        /*
        if ((inter.node_list[i].connect < 2) && 
            (inter.node_list[i].io == 0) && 
            (spa.sprhs[2*inter.num_fields*i + 1].re == 0) && 
            (spa.sprhs[2*inter.num_fields*i + 1].im == 0) &&
            NOT inter.node_list[i].gnd_node) {
          // removed warning for the moment as I believe this
          // doesnt really concern the user but prevents him/her
          // from seeing the important warnings
          //warn("Warning: node %s less than 2 connections.\n", node_print(i));
        }
         */

        if (inter.node_list[i].connect == 2 && inter.node_list[i].io & 2) {
            warn("Warning: input at node %s inside interferometer.\n",
                    node_print(i));
        }

        if (inter.node_list[i].connect > 2) {
            gerror(" node %s more than 2 connections.\n", node_print(i));
        }

        if ((inter.node_list[i].connect == 0) && (inter.node_list[i].io)) {
            gerror(" input/output at node %s not connected to interferometer.\n",
                    node_print(i));
        }
    }
}

//! Error codes for sparse matrices

/*!
 * \param err error code
 * \param pos location in code where error occurred
 */
void report_sparse_matrix_error(int err, int pos) {
    if (err) {
        fprintf(stderr,
                "\n*** An unexpected error occured in the SPARSE routines:\n");
        switch (err) {
            case 1:
                fprintf(stderr, "Small pivot (error code 1, %d)\n", pos);
                break;
            case 2:
                fprintf(stderr, "Zero on diagonal (error code 2, %d)\n", pos);
                break;
            case 3:
                fprintf(stderr, "Matrix is singular (error code 3, %d)\n", pos);
                break;
            case 4:
                fprintf(stderr, "Not enough memory (error code 4, %d)\n", pos);
                break;
            case 5:
                fprintf(stderr, "Panic (error code 5, %d)\n", pos);
                break;
            default:
                fprintf(stderr, "Unknown error (error code %d, %d)\n", err, pos);
        }
        bug_error(" ");
    } else
        return;
}



//! Update element count for KLU matrix (CCS format)

/*!
 * \param node_index1 index of node1 of component
 * \param node_index1 index of node2 of component (call this function twice for bs)
 * \param colum_index1 matrix index referring to node1 entry
 * \param colum_index2 matrix index referring to node2 entry
 * \param mode which type of updating should be performed (i.e. reflection+ transmission - or - transmission only)
 *
 * mode MIRROR:  n1->n2, n2->n1, n1->n1, n2->n2
 * mode SPACE:   n1->n2, n2->n1
 * mode GRATING: n2->n2 (this just filles the last connections after update has been called with MIRROR+SPACE already)
 *
 */
void update_element_count(int node_index1, int node_index2, int row_index_1, int row_index_2, int mode) {
    node_t node1, node2;
    node1 = inter.node_list[node_index1];
    node2 = inter.node_list[node_index2];

    switch (mode) {
        case MIRROR:
            if (NOT node1.gnd_node) {
                klu.col_ptr[row_index_1]++;
            }
            if (NOT node2.gnd_node) {
                klu.col_ptr[row_index_2]++;
            }
        case SPACE:
            if ((NOT node1.gnd_node) && (NOT node2.gnd_node)) {
                klu.col_ptr[row_index_1]++;
                klu.col_ptr[row_index_2]++;
            }
            break;
        case GRATING:
            if (NOT node2.gnd_node) {
                klu.col_ptr[row_index_2]++;
            }
            break;
        default:
            bug_error("incorrect mode");
    }
}

//! Reset connect counts for all nodes

/*!
 *
 */
void reset_connect_count(void) {
    int i;
    for (i = 0; i < inter.num_nodes; i++) {
        inter.node_list[i].connect = 0;
    }
}

/*
 * Local variables:
 * c-indentation-style: bsd
 * c-basic-offset: 2
 * indent-tabs-mode: nil
 * End:
 *
 * vim: tabstop=2 expandtab shiftwidth=2:
 */
