// $Id: kat_aa.c 3374 2011-09-12 16:29:43Z dzb610 $

/*!  
 * \file kat_aa.c 
 * \brief Routines for the automatic alignment of the interferometer
 * 
 * @section Copyright notice
 *
 *  This file is part of the interferometer simulation Finesse
 *  http://www.gwoptics.org/finesse
 *
 *  Copyright (C) 1999 onwards Andreas Freise
 *  with parts of the code written by Daniel Brown, Paul Cochrane
 *  and Gerhard Heinzel.
 *
 *  This program is free software; you can redistribute it and/or modify it under
 *  the terms of the GNU General Public License version 3 as published
 *  by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 *  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *  See the GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along with
 *  this library; if not, write to the Free Software Foundation, Inc., 59 Temple Place,
 *  Suite 330, Boston, MA 02111-1307 USA
 */

#include "kat.h"
#include "kat_inline.c"
#include "kat_io.h"
#include "kat_fortran.h"
#include "kat_optics.h"
#include "kat_aux.h"
#include "kat_dump.h"
#include "kat_aa.h"
#include "kat_check.h"
#include <gsl/gsl_errno.h>
#include <gsl/gsl_spline.h>
//#include <sys/sysinfo.h>
#include <cuba.h>
#include <stdlib.h>
#include <pthread.h>

extern init_variables_t init;
extern interferometer_t inter;
extern options_t options;
extern local_var_t vlocal;
extern FILE *fp_log;
extern mirror_kmn_t tmap;

extern const complex_t complex_i; //!< sqrt(-1) or 0 + i
extern const complex_t complex_1; //!< 1 but in complex space: 1 + 0i
extern const complex_t complex_m1; //!< -1 but in complex space: -1 + 0i
extern const complex_t complex_0; //!< 0 but in complex space: 0 + 0i

mirror_kmn_t tmp;

pthread_mutex_t mxinterpfile = PTHREAD_MUTEX_INITIALIZER;

u_nm_accel_t *acc_11_nr1_1, *acc_11_nr1_2;
u_nm_accel_t *acc_22_nr2_1, *acc_22_nr2_2;
u_nm_accel_t *acc_21_nr2_1, *acc_21_nr1_2;
u_nm_accel_t *acc_12_nr1_1, *acc_12_nr2_2;

const ABCD_t Unity = {1.0, 0.0, 0.0, 1.0}; //!< The identity matrix

int n_index; //!< node index; number of nodes in beam trace

void allocate_temporary_memory_for_maps() {

    mirror_kmn_alloc(&tmp);
    mirror_kmn_alloc(&tmap);

}

void allocate_kmn_accel_memory() {
    // don't allocate if we are not using higher order modes
    if (inter.tem_is_set) {
        int max_m, max_n;
        get_tem_modes_from_field_index(&max_n, &max_m, inter.num_fields - 1);
        max_m = max_n; // max m is no the m is also the max_n. e.g. we get n=1 m=2 for maxtem 2

        // make sure they are null before allocating new memory
        assert(acc_11_nr1_1 == NULL);
        assert(acc_11_nr1_2 == NULL);
        assert(acc_12_nr1_1 == NULL);
        assert(acc_12_nr2_2 == NULL);
        assert(acc_21_nr2_1 == NULL);
        assert(acc_21_nr1_2 == NULL);
        assert(acc_22_nr2_1 == NULL);
        assert(acc_22_nr2_2 == NULL);

        // allocate memory for all the accelerators
        acc_11_nr1_1 = u_nm_accel_alloc(max_n, max_m);
        acc_11_nr1_2 = u_nm_accel_alloc(max_n, max_m);
        acc_22_nr2_1 = u_nm_accel_alloc(max_n, max_m);
        acc_22_nr2_2 = u_nm_accel_alloc(max_n, max_m);
        acc_12_nr1_1 = u_nm_accel_alloc(max_n, max_m);
        acc_12_nr2_2 = u_nm_accel_alloc(max_n, max_m);
        acc_21_nr2_1 = u_nm_accel_alloc(max_n, max_m);
        acc_21_nr1_2 = u_nm_accel_alloc(max_n, max_m);
    }
}


//! Compute coupling coefficients between two beams

/*!
 * Each coefficient describes the coupling of the TEM_n1m1 mode into the
 * TEM_n2m2 mode as a function of the misalignments (gamma_x, gamma_y). In
 * TEM_nm, n gives the mode number in the horizontal plane and m the number
 * for the vertical modes. qx gives the beam parameter in the horizontal
 * plane, qy for the vertical modes. Please note that the notation of
 * gamma_x and gamma_y is not intuitive: gamma_x describes the rotation
 * around the vertical axis and gamma_y the rotation around a horizontal
 * axis through the optical component.

 * Utility functions called by the routines below:
 * - cconj(z) : return complex conjugate of z
 * - ceq(z1, z2) : check whether two complex numbers z1, z2 are equal
 * - z_by_z(z1, z2) : return multiplication of two complex numbers
 * - z_pl_z(z1, z2) : return sum of two complex numbers
 * - z_by_x(z, x) : multiply complex number by a real number
 * - inv_complex(z) : return the inverse of a complex number
 * - zsqrt(z) : complex square root of z
 * - zr(q) : compute Rayleigh range from Gaussian beam parameter q
 * - w0_size(q, nr) : compute beam waist size from beam parameter q and 
 *                   refrective index nr
 * - pow_complex(z, n1, n2) : return z**(n1/n2) with n1, n2 as integers
 * - msign(n) : return -1.0**n, with n as an integer
 * - fac(n) : return factorial of n
 *
 * \param n1 mode number of beam 1
 * \param m1 mode number of beam 1
 * \param n2 mode number of beam 2
 * \param m2 mode number of beam 2
 * \param qx1 Gaussian beam q parameter in x-plane for beam 1
 * \param qy1 Gaussian beam q parameter in y-plane for beam 1
 * \param qx2 Gaussian beam q parameter in x-plane for beam 2
 * \param qy2 Gaussian beam q parameter in y-plane for beam 2
 * \param gamma_x misalignment in x-plane 
 * \param gamma_y misalignment in y-plane
 * \param nr refractive index
 *
 * \see Test_k_mnmn()
 */
complex_t k_mnmn(int n1, int m1, int n2, int m2, complex_t qx1, complex_t qy1,
        complex_t qx2, complex_t qy2, double gamma_x, double gamma_y,
        double nr) {
    complex_t z, z1, z2, ztmp;

    // sanity checks on inputs
    // the mode numbers shouldn't be less than zero
    assert(n1 >= 0);
    assert(n2 >= 0);
    assert(m1 >= 0);
    assert(m2 >= 0);
    // the refractive index shouldn't be zero or less
    assert(nr > 0.0);

    // if the misalignment is zero and ...
    if ((gamma_x == 0) && (gamma_y == 0)) {
        // if the Gaussian beam parameters are equal 
        // (i.e. there is no mode mismatch) ...
        if (ceq(qx1, qx2) && ceq(qy1, qy2)) {
            // then return 1 for coupling into the same mode number and zero 
            // for all other couplings
            if (n1 == n2 && m1 == m2) {
                return (complex_1);
            } else {
                return (complex_0);
            }
        }
    }

    // check for inconsistent mode numbers
    if (n1 < 0 || n2 < 0 || m1 < 0 || m2 < 0) {
        bug_error("negative mode numbers");
    }

    // inter.kmn is a global setting for chosing the type of algorithm that 
    // is used for computing the coupling coefficients (mostly for debugging)

    // The sign of gamma_y has to be different between the numeric algorithm
    // and the Bayer-Helms formula. This is probaly due to a wrong definition from
    // me when implementing angles.

    /*
        if ((inter.kmn & 2 && !(gamma_x == 0.0) && !(gamma_y == 0.0)) ||
                (inter.kmn & 4 && !(gamma_x == 0.0 && gamma_y == 0.0)) || inter.kmn & 8) {
     */
    // added new kmn debug for old routine
    if (inter.kmn & INT_BAYER_HELMS_OLD) {
        if (inter.kmn & 1 && !options.quiet) { // TODO should be replaced by ENUM too?
            z1 = k_mm(n1, n2, qx1, qx2, gamma_x, nr);
            z2 = k_mm(m1, m2, qy1, qy2, -gamma_y, nr); // gamma + oder - ? 251004
            ztmp = z_by_z(z1, z2);

            message("%d %d %d %d | qx1 %s qy1 %s qx2 %s qy2 %s | "
                    "gammax %g, gammay %g) ",
                    n1, m1, n2, m2, complex_form(qx1), complex_form(qy1),
                    complex_form(qx2), complex_form(qy2), gamma_x, -gamma_y);
        }

        // do a numerical integration on a grid
        // with adaptive integration routine `dcuhre':
        z = faltung_adapt(n1, n2, m1, m2, qx1, qy1, qx2, qy2, gamma_x, gamma_y, nr);

        // compare Bayer-Helms with numericly derived coefficients
        if (inter.kmn & 1 && !options.quiet) {
            message("\nNumeric:     %s\nBayer-Helms: %s\n", complex_form(z),
                    complex_form(ztmp));
            if (capprox(z, ztmp)) {
                message("--> OK   ------\n");
            } else {
                message("--> DIFF ++++++\n");
            }
        }
    } else {
        // use coupling coefficients as in F. Bayer-Helms (see below)
        if (n1 == vlocal.kn1 && n2 == vlocal.kn2 && ceq(vlocal.kqx1, qx1)
                && ceq(vlocal.kqx2, qx2) && (gamma_x == vlocal.kgammax)
                && (nr == vlocal.knr)) {
            z1 = vlocal.zk1;
        } else {
            z1 = k_mm(n1, n2, qx1, qx2, gamma_x, nr);
            vlocal.zk1 = z1;
            vlocal.kn1 = n1;
            vlocal.kn2 = n2;
            vlocal.kqx1 = qx1;
            vlocal.kqx2 = qx2;
            vlocal.kgammax = gamma_x;
            vlocal.knr = nr;
        }

        // -gammy_y, probably because of my definition of rotation as right-handed 
        // (around y axis)
        z2 = k_mm(m1, m2, qy1, qy2, -gamma_y, nr);
        z = z_by_z(z1, z2);

        if (inter.kmn & 1 && !options.quiet) {
            message("%d %d %d %d | qx1 %s qy1 %s qx2 %s qy2 %s | "
                    "gammax %g, gammay %g) ",
                    n1, m1, n2, m2,
                    complex_form(qx1), complex_form(qy1),
                    complex_form(qx2), complex_form(qy2), gamma_x, -gamma_y);
            message("\nBayer-Helms: %s\n", complex_form(z));
        }
    }

    // `reverse Gouy phase': I found out that on changing from one TEM 
    // base system to another I have to turn back the Gouy phase with 
    // respect to the old beam parameter and add the Gouy phase with 
    // respect to the new beam parameter. This is a required because of 
    // the way the coupling coefficients are defined here and the way 
    // FINESSE handles the Gouy phase: the Gouy phase is added explicitly 
    // to the complex amplitude coefficiants in every `space' whereas the 
    // coupling coefficients are derived using a formula in which the Gouy 
    // phase is implicitly given by spacial distribution.
    // 22/02/2012 This will be moved to after the kmn merging - Daniel
    z = rev_gouy(z, n1, m1, n2, m2, qx1, qx2, qy1, qy2);

    return (z);
}

// same as above but withoyt rev_gouy, to be used with new set_k_mirror function

complex_t k_mnmn_new(int n1, int m1, int n2, int m2, complex_t qx1, complex_t qy1,
        complex_t qx2, complex_t qy2, double gamma_x, double gamma_y,
        double nr) {
    complex_t z, z1, z2, ztmp;

    // sanity checks on inputs
    // the mode numbers shouldn't be less than zero
    assert(n1 >= 0);
    assert(n2 >= 0);
    assert(m1 >= 0);
    assert(m2 >= 0);
    // the refractive index shouldn't be zero or less
    assert(nr > 0.0);

    // if the misalignment is zero and ...
    if ((gamma_x == 0) && (gamma_y == 0)) {
        // if the Gaussian beam parameters are equal 
        // (i.e. there is no mode mismatch) ...
        if (ceq(qx1, qx2) && ceq(qy1, qy2)) {
            // then return 1 for coupling into the same mode number and zero 
            // for all other couplingsf
            if (n1 == n2 && m1 == m2) {
                return (complex_1);
            } else {
                return (complex_0);
            }
        }
    }

    // check for inconsistent mode numbers
    if (n1 < 0 || n2 < 0 || m1 < 0 || m2 < 0) {
        bug_error("negative mode numbers");
    }

    // inter.kmn is a global setting for chosing the type of algorithm that 
    // is used for computing the coupling coefficients (mostly for debugging)

    // The sign of gamma_y has to be different between the numeric algorithm
    // and the Bayer-Helms formula. This is probaly due to a wrong definition from
    // me when implementing angles.

    /*
        if ((inter.kmn & 2 && !(gamma_x == 0.0) && !(gamma_y == 0.0)) ||
                (inter.kmn & 4 && !(gamma_x == 0.0 && gamma_y == 0.0)) || inter.kmn & 8) {
     */
    // added new kmn debug for old routine
    if (inter.kmn & INT_BAYER_HELMS_OLD) {
        if (inter.kmn & 1 && !options.quiet) {
            z1 = k_mm(n1, n2, qx1, qx2, gamma_x, nr);
            z2 = k_mm(m1, m2, qy1, qy2, -gamma_y, nr); // gamma + oder - ? 251004
            ztmp = z_by_z(z1, z2);

            message("%d %d %d %d | qx1 %s qy1 %s qx2 %s qy2 %s | "
                    "gammax %g, gammay %g) ",
                    n1, m1, n2, m2, complex_form(qx1), complex_form(qy1),
                    complex_form(qx2), complex_form(qy2), gamma_x, -gamma_y);
        }

        // do a numerical integration on a grid
        // with adaptive integration routine `dcuhre':
        z = faltung_adapt(n1, n2, m1, m2, qx1, qy1, qx2, qy2, gamma_x, gamma_y, nr);

        // compare Bayer-Helms with numericly derived coefficients
        if (inter.kmn & 1 && !options.quiet) {
            message("\nNumeric:     %s\nBayer-Helms: %s\n", complex_form(z),
                    complex_form(ztmp));
            if (capprox(z, ztmp)) {
                message("--> OK   ------\n");
            } else {
                message("--> DIFF ++++++\n");
            }
        }
    } else {
        // use coupling coefficients as in F. Bayer-Helms (see below)
        if (n1 == vlocal.kn1 && n2 == vlocal.kn2 && ceq(vlocal.kqx1, qx1)
                && ceq(vlocal.kqx2, qx2) && (gamma_x == vlocal.kgammax)
                && (nr == vlocal.knr)) {
            z1 = vlocal.zk1;
        } else {
            z1 = k_mm(n1, n2, qx1, qx2, gamma_x, nr);
            vlocal.zk1 = z1;
            vlocal.kn1 = n1;
            vlocal.kn2 = n2;
            vlocal.kqx1 = qx1;
            vlocal.kqx2 = qx2;
            vlocal.kgammax = gamma_x;
            vlocal.knr = nr;
        }

        // -gammy_y, probably because of my definition of rotation as right-handed 
        // (around y axis)
        z2 = k_mm(m1, m2, qy1, qy2, -gamma_y, nr);
        z = z_by_z(z1, z2);

        if (inter.kmn & 1 && !options.quiet) {
            message("%d %d %d %d | qx1 %s qy1 %s qx2 %s qy2 %s | "
                    "gammax %g, gammay %g) ",
                    n1, m1, n2, m2,
                    complex_form(qx1), complex_form(qy1),
                    complex_form(qx2), complex_form(qy2), gamma_x, -gamma_y);
            message("\nBayer-Helms: %s\n", complex_form(z));
        }
    }

    // `reverse Gouy phase': I found out that on changing from one TEM 
    // base system to another I have to turn back the Gouy phase with 
    // respect to the old beam parameter and add the Gouy phase with 
    // respect to the new beam parameter. This is a required because of 
    // the way the coupling coefficients are defined here and the way 
    // FINESSE handles the Gouy phase: the Gouy phase is added explicitly 
    // to the complex amplitude coefficiants in every `space' whereas the 
    // coupling coefficients are derived using a formula in which the Gouy 
    // phase is implicitly given by spacial distribution.
    // 22/02/2012 This will be moved to after the kmn merging - Daniel
    //z = rev_gouy(z, n1, m1, n2, m2, qx1, qx2, qy1, qy2);

    return (z);
}



//! Coupling coefficients for TEM modes

/*!
 * F.  Bayer-Helms `Coupling coefficients of an incident wave and modes
 * of an spherical optical resonator in the case of mismatching and
 * misalignment', Appl. Optics Vol23 No9 1369-1380 (1984).
 * There is also a report (in German) which gives a more detailed
 * derivation: `Ableitung der Kopplunskoeffizienten fuer die Eigenfunktionen
 * eines sphaerischen optischen Resonators bei Fehlanpassung und 
 * Fehljustierung der einfallenden Welle' PTB Bericht PTB-Me-43 (1983)
 * 
 * k_mm ist for one degree of freedom (x or y) only. m1 and m2 are the 
 * mode numbers of the incoming and outgoing beam, q1 and q2 the 
 * respective Gaussian beam parameters, gamma is the misalignment angle 
 * between the beams(!) and nr the index of refraction of the medium in 
 * which the coupling happens. The latter is needed to compute the diffraction 
 * angle from the Gaussian beam parameter.
 * 
 * The function calls S_u and S_g (see below). The names of parameters and
 * variables are chosen such that they are similar to the notation used in the
 * paper.
 *
 * \param m1 mode number for incoming beam
 * \param m2 mode number for outgoing beam
 * \param q1 Gaussian beam parameter for incoming beam
 * \param q2 Gaussian beam parameter for outgoing beam
 * \param gamma misalignment angle between the beams
 * \param nr refractive index
 */
complex_t k_mm(int m1, int m2, complex_t q1, complex_t q2,
        double gamma, double nr) {
    double Theta, K0;
    double gg0, z2z0;
    double zr;
    double out1;
    complex_t K;
    complex_t XS, X, FS, F;
    complex_t ex;
    complex_t k1, x1, x2;
    complex_t out2, out3;

    // sanity checks on inputs
    // mode numbers bounds; they should not be less than zero
    assert(m1 >= 0);
    assert(m2 >= 0);

    // refractive index should be > 0 
    assert(nr > 0.0);

    if (ceq(q1, q2)) {
        K = complex_0;
        K0 = 0.0;
        k1 = complex_1;
        FS = complex_0;
        F = complex_0;
    } else {
        zr = 1.0 / q2.im;
        K.re = 0.5 * (q1.im - q2.im) * zr;
        K.im = 0.5 * (q1.re - q2.re) * zr;
        K0 = 2 * K.re;
        k1 = inv_complex(zsqrt(z_pl_z(complex_1, cconj(K))));
        FS = z_by_x(K, 1.0 / (2 * (1.0 + K0)));
        F = z_by_x(cconj(K), 0.5);
    }   

    // I think this theta variable is the denominator that comes from bayer-helms equation 12
    Theta = w0_size(q2, nr) / z_r(q2);

    //printf("\nw0=%g\n",w0_size(q2, nr) );
    //printf("zr=%g\n",z_r(q2));

    if (gamma == 0) {
        x1 = x2 = XS = X = ex = complex_0;
        gg0 = 0.0;
    } else {
        //    gg0 = -1.0 * gamma / Theta; // sin (gamma) ???

        //See equation 12, gamma/gamma^hat_0
        gg0 = gamma / Theta; // changed sign of gamma 010708

        //See equation 19 bayer-helms, z^hat_2 / z^hat_0
        z2z0 = z_q(q2) / z_r(q2);

        x1 = z_m_z(co(z2z0, 0), complex_i); // inner bracket of X^hat eq 19
        x2 = z_pl_z(co(z2z0, 0), ci(z_pl_z(complex_1, z_by_x(cconj(K), 2)))); //inner bracket of X eq 19

        /*
        printf("\nx1=%s\n",complex_form15(x1));
        printf("x2=%s\n",complex_form15(x2));
        printf("k1=%s\n",complex_form15(k1));
        printf("gg0=%g\n",gg0);
        printf("gamma=%g\n",gamma);
        printf("Theta=%g\n",Theta);
         */

        // multiply by k1 which is (1+K*)^-0.5
        // also multiply inner brackets calculated above by gamma/gamma^hat_0 factor
        // TODO it appears as though we are missing minus signs according to equation 19 bayer-helms?
        XS = z_by_z(k1, z_by_x(x1, gg0)); //X^hat eq 19
        X = z_by_z(k1, z_by_x(x2, gg0)); //X eq 19
        ex = z_by_x(z_by_z(XS, X), 0.5);
        //printf("ex=%s\n",complex_form15(ex));
    }

    out1 = msign(m2);
    out2 = z_by_x(pow_complex(z_pl_z(complex_1, cconj(K)), -m1 - m2 - 1, 2),
            sqrt(fac(m1) * fac(m2) * pow(1.0 + K0, m1 + 0.5)));

    complex_t sg = S_g(X, XS, F, FS, m1, m2);
    complex_t su = S_u(X, XS, F, FS, m1, m2);

    out3 = z_by_xphr(z_by_z(out2, z_m_z(sg, su)), out1 * exp(-ex.re), -ex.im);

    return (out3);
}

//! 

/*!
 * 
 * \param X ???
 * \param XS ???
 * \param F ???
 * \param FS ???
 * \param m1 ???
 * \param m2 ???
 *
 * \see Test_S_u()
 */
complex_t S_u(complex_t X, complex_t XS, complex_t F, complex_t FS,
        int m1, int m2) {
    int s3, s4;
    complex_t nom1, nom2;
    double denom1, denom2;
    complex_t z, z1, z2;

    if (m1 == 0 || m2 == 0) {
        return (complex_0);
    }

    if (m1 == 1 || m1 == 2) {
        s3 = 0;
    } else {
        if (odd(m1 - 1)) {
            s3 = (int) ((m1 - 2) / 2);
        } else {
            s3 = (int) ((m1 - 1) / 2);
        }
    }

    if (m2 == 1 || m2 == 2) {
        s4 = 0;
    } else {
        if (odd(m2 - 1)) {
            s4 = (int) ((m2 - 2) / 2);
        } else {
            s4 = (int) ((m2 - 1) / 2);
        }
    }

    z = complex_0;
    int i, j;
    for (i = 0; i <= s3; i++) {
        for (j = 0; j <= s4; j++) {
            nom1 = z_by_z(pow_complex(XS, m1 - 2 * i - 1, 1),
                    pow_complex(X, m2 - 2 * j - 1, 1));
            denom1 = fac(m1 - 2 * i - 1) * fac(m2 - 2 * j - 1);
            z1 = z_by_x(nom1, msign(i) / denom1);
            z2 = complex_0;
            int k;
            for (k = 0; k <= nmin(i, j); k++) {
                nom2 = z_by_z(pow_complex(FS, i - k, 1), pow_complex(F, j - k, 1));
                denom2 = fac(2 * k + 1) * fac(i - k) * fac(j - k);
                z2 = z_pl_z(z2, z_by_x(nom2, msign(k) / denom2));
            }
            z = z_pl_z(z, z_by_z(z1, z2));
        }
    }

    return (z);
}

//! 

/*! 
 * 
 * \param X ???
 * \param XS ???
 * \param F ???
 * \param FS ???
 * \param m1 ???
 * \param m2 ???
 *
 * \see Test_S_g()
 */
complex_t S_g(complex_t X, complex_t XS, complex_t F, complex_t FS,
        int m1, int m2) {
    int s1, s2;
    complex_t z, z1, z2;
    complex_t nom1, nom2;
    double denom1, denom2;

    if (m1 == 0 || m1 == 1) {
        s1 = 0;
    } else {
        if (odd(m1)) {
            s1 = (int) ((m1 - 1) / 2);
        } else {
            s1 = (int) (m1 / 2);
        }
    }

    if (m2 == 0 || m2 == 1) {
        s2 = 0;
    } else {
        if (odd(m2)) {
            s2 = (int) ((m2 - 1) / 2);
        } else {
            s2 = (int) (m2 / 2);
        }
    }

    z = complex_0;
    int i, j;
    for (i = 0; i <= s1; i++) {
        for (j = 0; j <= s2; j++) {
            nom1 =
                    z_by_z(pow_complex(XS, m1 - 2 * i, 1), pow_complex(X, m2 - 2 * j, 1));
            denom1 = fac(m1 - 2 * i) * fac(m2 - 2 * j);
            z1 = z_by_x(nom1, msign(i) / denom1);
            z2 = complex_0;
            int k;
            for (k = 0; k <= nmin(i, j); k++) {
                nom2 = z_by_z(pow_complex(FS, i - k, 1), pow_complex(F, j - k, 1));
                denom2 = fac(2 * k) * fac(i - k) * fac(j - k);
                z2 = z_pl_z(z2, z_by_x(nom2, msign(k) / denom2));
            }
            z = z_pl_z(z, z_by_z(z1, z2));
        }
    }

    return (z);
}

//! Adjust the phase of a coupling coefficient with respect to the Gouy phases. 

/*!
 * Used in k_mnmn because in FINESSE the Gouy phase is added explicitly to the
 * amplitude coefficients in a `space' where as the coupling coefficients
 * are derived using a formula in which the Gouy phase resides in the
 * equation for the spatial profile.
 *
 * \param k coupling coefficient between beams
 * \param n1 mode number of beam 1
 * \param m1 mode number of beam 1
 * \param n2 mode number of beam 2
 * \param m2 mode number of beam 2
 * \param qx1 Gaussian q parameter in x-plane for beam 1
 * \param qx2 Gaussian q parameter in x-plane for beam 2
 * \param qy1 Gaussian q parameter in y-plane for beam 1
 * \param qy2 Gaussian q parameter in y-plane for beam 2
 *
 * \see Test_rev_gouy()
 */
complex_t rev_gouy(complex_t k, int n1, int m1, int n2, int m2, complex_t qx1,
        complex_t qx2, complex_t qy1, complex_t qy2) {
    complex_t zt;
    double phase;

    // sanity checks on inputs
    // [n-m][1-2] should be >= 0
    assert(m1 >= 0);
    assert(m2 >= 0);
    assert(n1 >= 0);
    assert(n2 >= 0);

    /*
  if (inter.set_tem_phase_zero & 2) {
    phase = n1 * gouy(qx1) + m1 * gouy(qy1) - (n2 * gouy(qx2) + m2 * gouy(qy2));
  }
  else {
        }
     */
    // 180208: changes this becasue apparently the phase must be always
    // computed from the full Gouy phase, see bug Jerome200108
    phase = ((n1 + 0.5) * gouy(qx1) + (m1 + 0.5) * gouy(qy1))
            -((n2 + 0.5) * gouy(qx2) + (m2 + 0.5) * gouy(qy2));

    zt = z_by_phr(k, -1.0 * phase);

    return (zt);
}

void mirror_kmn_free(mirror_kmn_t* kmn) {
    assert(kmn != NULL);

    if (!IS_MIRROR_KMN_ALLOCD(kmn))
        bug_error("We have tried to free memory that has not been allocated");

    int num_fields = (int) (inter.tem + 1) * (inter.tem + 2) / 2;
    int field_index;

    for (field_index = 0; field_index < num_fields; field_index++) {
        free(kmn->k11[field_index]);
        free(kmn->k12[field_index]);
        free(kmn->k21[field_index]);
        free(kmn->k22[field_index]);
        // remembering to NULL any loose pointers...
        kmn->k11[field_index] = NULL;
        kmn->k21[field_index] = NULL;
        kmn->k12[field_index] = NULL;
        kmn->k22[field_index] = NULL;
    }

    free(kmn->k11);
    free(kmn->k12);
    free(kmn->k21);
    free(kmn->k22);

    kmn->k11 = NULL;
    kmn->k21 = NULL;
    kmn->k12 = NULL;
    kmn->k22 = NULL;
}

void mirror_kmn_alloc(mirror_kmn_t* kmn) {
    if (kmn == NULL)
        bug_error("mirror_kmn_t pointer is nulled");

    if (IS_MIRROR_KMN_ALLOCD(kmn))
        bug_error("We have tried to allocate memory to a mirror_kmn when it is already allocated, that or we have not initialised the pointers to be NULL");

    int num_fields = (int) (inter.tem + 1) * (inter.tem + 2) / 2;
    int field_index;

    // coupling coefficients
    kmn->k11 = (complex_t **) calloc(num_fields + 1, sizeof (complex_t *));
    kmn->k12 = (complex_t **) calloc(num_fields + 1, sizeof (complex_t *));
    kmn->k21 = (complex_t **) calloc(num_fields + 1, sizeof (complex_t *));
    kmn->k22 = (complex_t **) calloc(num_fields + 1, sizeof (complex_t *));

    if (kmn->k11 == NULL ||
            kmn->k12 == NULL ||
            kmn->k21 == NULL ||
            kmn->k22 == NULL) {
        bug_error("memory allocation for surface map data failed (3)");
    }

    for (field_index = 0; field_index < num_fields; field_index++) {
        // coupling coefficients
        kmn->k11[field_index] = (complex_t *) calloc(num_fields + 1, sizeof (complex_t));
        kmn->k12[field_index] = (complex_t *) calloc(num_fields + 1, sizeof (complex_t));
        kmn->k21[field_index] = (complex_t *) calloc(num_fields + 1, sizeof (complex_t));
        kmn->k22[field_index] = (complex_t *) calloc(num_fields + 1, sizeof (complex_t));

        if (kmn->k11[field_index] == NULL ||
                kmn->k12[field_index] == NULL ||
                kmn->k21[field_index] == NULL ||
                kmn->k22[field_index] == NULL) {
            bug_error("memory allocation for surface map data failed (4)");
        }

        kmn->k11[field_index][field_index] = complex_1;
        kmn->k12[field_index][field_index] = complex_1;
        kmn->k21[field_index][field_index] = complex_1;
        kmn->k22[field_index][field_index] = complex_1;
    }
}

// multiply A*B then put result in new matrix

void mirror_kmn_matrix_mult(mirror_kmn_t* A, mirror_kmn_t* B, mirror_kmn_t *result) {
    assert(result != NULL && A != NULL && B != NULL);
    assert(result->k11 != NULL && A->k11 != NULL && B->k11 != NULL);
    assert(result->k12 != NULL && A->k12 != NULL && B->k12 != NULL);
    assert(result->k21 != NULL && A->k21 != NULL && B->k21 != NULL);
    assert(result->k22 != NULL && A->k22 != NULL && B->k22 != NULL);

    int num_fields = (int) (inter.tem + 1) * (inter.tem + 2) / 2;
    int l, n, m;

    mirror_kmn_t *tmp;

    // if A or B are just identity matrices then just copy them into result
    if (A->IsIdentities) {
        if (B != result)
            mirror_kmn_matrix_copy(B, result);

    } else if (B->IsIdentities) {
        if (A != result)
            mirror_kmn_matrix_copy(A, result);

    } else {
        // if result != A | B then no need to use a temporary matrix to store result 
        // just bung it in the result matrix
        if ((result == A) || (result == B))
            tmp = &tmap;
        else
            tmp = result;

        for (n = 0; n < num_fields; n++) {
            for (m = 0; m < num_fields; m++) {
                // initialise to 0 here just incase some other data is there
                tmp->k11[n][m] = complex_0;
                tmp->k12[n][m] = complex_0;
                tmp->k21[n][m] = complex_0;
                tmp->k22[n][m] = complex_0;

                for (l = 0; l < num_fields; l++) {
                    tmp->k11[n][m] = z_pl_z(tmp->k11[n][m], z_by_z(A->k11[n][l], B->k11[l][m]));
                    tmp->k12[n][m] = z_pl_z(tmp->k12[n][m], z_by_z(A->k12[n][l], B->k12[l][m]));
                    tmp->k21[n][m] = z_pl_z(tmp->k21[n][m], z_by_z(A->k21[n][l], B->k21[l][m]));
                    tmp->k22[n][m] = z_pl_z(tmp->k22[n][m], z_by_z(A->k22[n][l], B->k22[l][m]));
                }
            }
        }

        if ((result == A) || (result == B))
            mirror_kmn_matrix_copy(&tmap, result);
    }
}

// Should copy one matrix into another

void mirror_kmn_matrix_copy(mirror_kmn_t* src, mirror_kmn_t* dest) {
    assert(src != NULL);
    assert(dest != NULL);

    int num_fields = (int) (inter.tem + 1) * (inter.tem + 2) / 2;
    int i;

    for (i = 0; i < num_fields; i++) {
        memcpy(dest->k11[i], src->k11[i], num_fields * sizeof (complex_t));
        memcpy(dest->k12[i], src->k12[i], num_fields * sizeof (complex_t));
        memcpy(dest->k21[i], src->k21[i], num_fields * sizeof (complex_t));
        memcpy(dest->k22[i], src->k22[i], num_fields * sizeof (complex_t));
    }
}

// sets matrix to identity matrix

void mirror_kmn_matrix_ident(mirror_kmn_t* M) {
    assert(M != NULL);
    int num_fields = (int) (inter.tem + 1) * (inter.tem + 2) / 2;
    int i, j;

    for (i = 0; i < num_fields; i++) {
        for (j = 0; j < num_fields; j++) {
            M->k11[i][j] = (i == j) ? complex_1 : complex_0;
            M->k12[i][j] = (i == j) ? complex_1 : complex_0;
            M->k21[i][j] = (i == j) ? complex_1 : complex_0;
            M->k22[i][j] = (i == j) ? complex_1 : complex_0;
        }
    }
}

/**
 * This function is for internal use only by compute_bayer_helms_kmn. It is just
 * to reuse similar debug code and loops.
 * 
 * @param qxt
 * @param qyt
 * @param qxt2
 * @param qyt2
 * @param kmn	Pointer to the complex_t** to store the matrix
 * @param gamma_x misalignment about x
 * @param gamma_y misalignment about y
 * @param nr_out The outgoing mediums refractive index
 * @param k_num Sets which k to calculate: k11=1, k22=2, k12=3, k21=4
 */
void _compute_bayer_helms_sub_call(complex_t qxt, complex_t qyt, complex_t qxt2
        , complex_t qyt2, complex_t **kmn, double gamma_x, double gamma_y
        , double nr_out, KNM_MIRROR_NODE_DIRECTION_t k_num) {
    int n, m, n1, n2, m1, m2;
    complex_t kx, ky;
    double kmx, kmy;
    char num[3];
    
    if ((inter.trace & 64) || (inter.debug & 32)) {
        switch (k_num) {
            case K11:
                strcpy(num, "11");
                break;
            case K22:
                strcpy(num, "22");
                break;
            case K12:
                strcpy(num, "12");
                break;
            case K21:
                strcpy(num, "21");
                break;
            default:
                bug_error("Incorrect k_num, %i, provided for _compute_bayer_helms_sub_call", k_num);
        }
    }

    if (inter.trace & 64) {
        if (!ceq(qxt, qxt2) || !ceq(qyt, qyt2)) {
            mode_match(qxt, qxt2, &kx, &kmx);
            mode_match(qyt, qyt2, &ky, &kmy);
            message(" r%s: Kx=%s \n      Ky=%s \n      Kmx %9.3g Kmy %9.3g\n",
                    num, complex_form(kx), complex_form(ky), kmx, kmy);
        }
    }

    if (inter.debug & 32) {
        message("k%s", num);
    }

    for (n = 0; n < inter.num_fields; n++) {
        for (m = 0; m < inter.num_fields; m++) {
            get_tem_modes_from_field_index(&n1, &m1, n);
            get_tem_modes_from_field_index(&n2, &m2, m);

            if (n1 == NOT_FOUND || m1 == NOT_FOUND ||
                    n2 == NOT_FOUND || m2 == NOT_FOUND) {
                bug_error("notfound 1");
            }

            // TODO note the 'new' in the function call below. k_mnMN and k_mnmn_new should merge again later
            kmn[n][m] = k_mnmn_new(n1, m1, n2, m2, qxt, qyt, qxt2,
                    qyt2, gamma_x, gamma_y, nr_out);

            if (inter.debug & 32 && !ceq(complex_0, kmn[n][m])) {
                message(" [%d, %d]=%s, ", n, m, complex_form(kmn[n][m]));
            }
        }
        if (inter.debug & 32) {
            message("\n   ");
        }
    }
    if (inter.debug & 32) {
        message("\n");
    }
}
// q values passed to this should have been turned depending on the direction of
// incoming and outgoing beam

void calculate_mirror_qt_qt2(KNM_MIRROR_NODE_DIRECTION_t KNM, mirror_t *mirror, complex_t qx1,
        complex_t qy1, complex_t qx2, complex_t qy2, double nr1, double nr2, int n1idx, int n2idx, int cidx) {

    ABCD_t trans1;

    switch (KNM) {
        case K11:
            component_matrix(&trans1, cidx, n1idx, n1idx, TANGENTIAL);
            mirror->kmn_q.qxt1_11 = q1_q2(trans1, qx1, nr1, nr1);
            component_matrix(&trans1, cidx, n1idx, n1idx, SAGITTAL);
            mirror->kmn_q.qyt1_11 = q1_q2(trans1, qy1, nr1, nr1);
            mirror->kmn_q.qxt2_11 = cminus(cconj(qx1));
            mirror->kmn_q.qyt2_11 = cminus(cconj(qy1));
            break;
        case K22:
            component_matrix(&trans1, cidx, n2idx, n2idx, TANGENTIAL);
            mirror->kmn_q.qxt1_22 = q1_q2(trans1, cminus(cconj(qx2)), nr2, nr2);
            component_matrix(&trans1, cidx, n2idx, n2idx, SAGITTAL);
            mirror->kmn_q.qyt1_22 = q1_q2(trans1, cminus(cconj(qy2)), nr2, nr2);
            mirror->kmn_q.qxt2_22 = qx2;
            mirror->kmn_q.qyt2_22 = qy2;
            break;
        case K12:
            component_matrix(&trans1, cidx, n1idx, n2idx, TANGENTIAL);
            mirror->kmn_q.qxt1_12 = q1_q2(trans1, qx1, nr1, nr2);
            component_matrix(&trans1, cidx, n1idx, n2idx, SAGITTAL);
            mirror->kmn_q.qyt1_12 = q1_q2(trans1, qy1, nr1, nr2);
            mirror->kmn_q.qxt2_12 = qx2;
            mirror->kmn_q.qyt2_12 = qy2;
            break;
        case K21:
            component_matrix(&trans1, cidx, n2idx, n1idx, TANGENTIAL);
            mirror->kmn_q.qxt1_21 = q1_q2(trans1, cminus(cconj(qx2)), nr2, nr1);
            component_matrix(&trans1, cidx, n2idx, n1idx, SAGITTAL);
            mirror->kmn_q.qyt1_21 = q1_q2(trans1, cminus(cconj(qy2)), nr2, nr1);
            mirror->kmn_q.qxt2_21 = cminus(cconj(qx1));
            mirror->kmn_q.qyt2_21 = cminus(cconj(qy1));
            break;
        default:
            bug_error("calculate_qt_qt2 cannot handle a KNM value of %i\n", KNM);
    }
}

/**
 * This function takes a mirror object and a set or parameters and computes the 
 * misalignment and mismatched coupling coefficient matrix for the mirror. The 
 * results are places in the mirror object in its kmn_bayer_helms member. Returns
 * an integer which is flagged depending on which components are mismatched.
 * 
 * @param params Parameters used to calculate the kmn
 * @param mirror Mirror to compute kmn for
 * 
 * @return Returns true if the Bayer-Helms kmn should be calculated in the integral
 */
bool compute_bayer_helms_kmn(mirror_t *mirror, double nr1, double nr2) {

    // check if we should integrate the bayer_helms kmn's. If we do there is no
    // need to calculate the rest of this function
    if (inter.kmn & INT_BAYER_HELMS) {
        gerror("Integrating the Bayer-Helms coefficients using the Cuba or Riemann integrators is not implemented yet\n");
        return true;
    } 
    else {
        mirror->kmn_bayer_helms.IsIdentities = false;
    }
    
    // Allocate some memory for computing bayer-helms kmn matrix
    if (!IS_MIRROR_KMN_ALLOCD(&(mirror->kmn_bayer_helms)))
        mirror_kmn_alloc(&(mirror->kmn_bayer_helms));
    
    if (inter.debug && !options.quiet)
       message(" * Calculating Bayer-Helms Knm analytically for mirror %s\n", mirror->name);

    mirror_kmn_t *kmn_BH = &(mirror->kmn_bayer_helms);

    if (mirror->K11_calc) {
        _compute_bayer_helms_sub_call(mirror->kmn_q.qxt1_11
                , mirror->kmn_q.qyt1_11
                , mirror->kmn_q.qxt2_11
                , mirror->kmn_q.qyt2_11
                , kmn_BH->k11
                , BETA_SCALE * mirror->beta_x, BETA_SCALE * mirror->beta_y
                , nr1, K11);
    }

    if (mirror->K22_calc) {
        // here we use -beta!
        _compute_bayer_helms_sub_call(mirror->kmn_q.qxt1_22
                , mirror->kmn_q.qyt1_22
                , mirror->kmn_q.qxt2_22
                , mirror->kmn_q.qyt2_22
                , kmn_BH->k22
                , BETA_SCALE * mirror->beta_x, -BETA_SCALE * mirror->beta_y
                , nr2, K22);
    }

    // check if the nodes actually go anywhere, if not there is not point
    // computing coupling coefficients
    if (mirror->K12_21_calc) {

        _compute_bayer_helms_sub_call(mirror->kmn_q.qxt1_12
                , mirror->kmn_q.qyt1_12
                , mirror->kmn_q.qxt2_12
                , mirror->kmn_q.qyt2_12
                , kmn_BH->k12, 0, 0, nr2, K12);

        _compute_bayer_helms_sub_call(mirror->kmn_q.qxt1_21
                , mirror->kmn_q.qyt1_21
                , mirror->kmn_q.qxt2_21
                , mirror->kmn_q.qyt2_21
                , kmn_BH->k21, 0, 0, nr1, K21);

    }

    return false;
}

bool compute_aperture_kmn(mirror_t *mirror) {

    //check if an aperture has even been defined
    if (mirror->r_aperture > 0) {
        // check if we should integrate the aperture coupling coefficients rather
        // than computing them analytically
        if (inter.kmn & INT_APERTURE) {

            // if using any of the Riemann methods, these currently do not support 
            // aperture integration without using a map
            if ((init.mapintmethod == RIEMANN_SUM_NEW) || (init.mapintmethod == RIEMANN_SUM_OLD)) {
                warn("Aperture calculation using Riemann integrators is not currently supported.\n"
                     "   Please use an aperture map or a Cuba integrating routine.\n");
                // return false as with these methods no aperture coupling coefficient
                // can be integrated.
                return false;
            } else {
                return true;
            }

        } else {
            // otherwise we need to calculate the aperture coupling coefficient matrix here analytically
            warn("Analytic aperture coupling coefficient calculation not implemented yet,\n"
                 "   use kmn command to calculate apertures by integration.\n");

            if (inter.debug && !options.quiet)
               message(" * Calculating aperture Knm analytically\n");

            mirror_kmn_t *map = &(mirror->kmn_aperture);
            // as we will be calculating aperture kmn we need to allocate the memory for it
            // remember that free also needs to be called later
            if (!IS_MIRROR_KMN_ALLOCD(map))
                mirror_kmn_alloc(map);

            map->IsIdentities = false;
            // return false to state that no aperture integration is needed
            return false;
        }

        // no aperture integration is needed as there is no aperture...
        return false;
    }

    return false;
}


//! Set the respective coupling coefficients for the given mirror

/*!
 * \param mirror_index index of mirror to set the coupling coefficient
 *
 * \return parameter describing the mismatch (if any)
 *
 * \see Test_set_k_mirror()
 */
int set_k_mirror(int mirror_index) {
    int component_index;
    int node1_index, node2_index;
    int mismatch, astigmatism;
    char buf[LINE_LEN];

    double nr1 = 1.0, nr2 = 1.0;

    complex_t qx1, qy1, qx2, qy2;
    mirror_t *mirror;

    node_t node1, node2;

    // sanity check on input
    // mirror index should be in the range 0 <= mirror_index < inter.num_mirrors
    assert(mirror_index >= 0);
    assert(mirror_index < inter.num_mirrors);

    mirror = &(inter.mirror_list[mirror_index]);
    component_index = get_overall_component_index(MIRROR, mirror_index);

    node1_index = mirror->node1_index;
    node2_index = mirror->node2_index;

    node1 = inter.node_list[node1_index];
    node2 = inter.node_list[node2_index];
   
    mismatch = 0;
    astigmatism = 0;

    if (NOT node1.gnd_node) {
        nr1 = *node1.n;
        
        if (node1.component_index == component_index) {
            // reverse q so that beam comes _to_ the mirror from node1
            qx1 = cminus(cconj(node1.qx));
            qy1 = cminus(cconj(node1.qy));
        } else {
            qx1 = node1.qx;
            qy1 = node1.qy;
        }
    }else{
        qx1.re = 0;
        qx1.im = 0;
        qy1.re = 0;
        qy1.im = 0;
    }

    if (NOT node2.gnd_node) {
        nr2 = *node2.n;
        
        if (node2.component_index == component_index) {
            qx2 = node2.qx;
            qy2 = node2.qy;
        } else {
            // reverse q so that beam goes _from_ the mirror to node2,
            // this requires to use -beta for the reflection node2 -> node2 
            qx2 = cminus(cconj(node2.qx));
            qy2 = cminus(cconj(node2.qy));
        }
    }else{
        qx2.re = 0;
        qx2.im = 0;
        qy2.re = 0;
        qy2.im = 0;
    }
        

    // calculate the individual q values for each KNM and if we should even
    // bother calculating them    
    if (NOT node1.gnd_node) {
        mirror->K11_calc = true;
        calculate_mirror_qt_qt2(K11, mirror, qx1, qy1, qx2, qy2, nr1, nr2, node1_index, node2_index, component_index);
        
        if (!ceq(mirror->kmn_q.qxt1_11, mirror->kmn_q.qxt2_11)
                || !ceq(mirror->kmn_q.qyt1_11, mirror->kmn_q.qyt2_11))
            mismatch = mismatch | 1;

        if (!ceq(mirror->kmn_q.qxt1_11, mirror->kmn_q.qyt1_11)
                || !ceq(mirror->kmn_q.qxt2_11, mirror->kmn_q.qyt2_11))
            astigmatism = astigmatism | 1;
    } else 
        mirror->K11_calc = false;

    if (NOT node2.gnd_node) {
        mirror->K22_calc = true;
        calculate_mirror_qt_qt2(K22, mirror, qx1, qy1, qx2, qy2, nr1, nr2, node1_index, node2_index, component_index);
        
        if (!ceq(mirror->kmn_q.qxt1_22, mirror->kmn_q.qxt2_22)
                || !ceq(mirror->kmn_q.qyt1_22, mirror->kmn_q.qyt2_22))
            mismatch = mismatch | 2;

        if (!ceq(mirror->kmn_q.qxt1_22, mirror->kmn_q.qyt1_22)
                || !ceq(mirror->kmn_q.qxt2_22, mirror->kmn_q.qyt2_22))
            astigmatism = astigmatism | 2;
    } 
    else {
        mirror->K22_calc = false;
    }

    // here we also check if the mirror is 100% reflective, if so no
    // transmission is possible so dont calculate it
    if (NOT node1.gnd_node && NOT node2.gnd_node) {
        mirror->K12_21_calc = true;

        calculate_mirror_qt_qt2(K12, mirror, qx1, qy1, qx2, qy2, nr1, nr2, node1_index, node2_index, component_index);
        calculate_mirror_qt_qt2(K21, mirror, qx1, qy1, qx2, qy2, nr1, nr2, node1_index, node2_index, component_index);
        
        if (!ceq(mirror->kmn_q.qxt1_12, mirror->kmn_q.qxt2_12)
                || !ceq(mirror->kmn_q.qyt1_12, mirror->kmn_q.qyt2_12))
            mismatch = mismatch | 4;

        if (!ceq(mirror->kmn_q.qxt1_12, mirror->kmn_q.qyt1_12)
                || !ceq(mirror->kmn_q.qxt2_12, mirror->kmn_q.qyt2_12))
            astigmatism = astigmatism | 4;

        if (!ceq(mirror->kmn_q.qxt1_21, mirror->kmn_q.qxt2_21)
                || !ceq(mirror->kmn_q.qyt1_21, mirror->kmn_q.qyt2_21))
            mismatch = mismatch | 8;

        if (!ceq(mirror->kmn_q.qxt1_21, mirror->kmn_q.qyt1_21)
                || !ceq(mirror->kmn_q.qxt2_21, mirror->kmn_q.qyt2_21))
            astigmatism = astigmatism | 8;

    } 
    else {   
        mirror->K12_21_calc = false;
    }
    
    if (inter.debug & 128){
       message(" * Mirror %s, Kmn to be calculated: ",mirror->name);
        if (mirror->K11_calc){
           message("K11 ");
        }
        if (mirror->K12_21_calc){
           message("K12_21 ");
        }
        if (mirror->K22_calc){
           message("K22 ");
        }
       message("\n");
    }
        
    // flagged integer that keeps track of which effects need to be integrated.
    unsigned int integrating = 0;

    if(inter.debug && !options.quiet)
       message(" * x-axis value %.2e --------------------------------\n",
                *inter.x1.xaxis);
    
    // TODO: This computes the bayer-helms coefficients or it analytically computes
    // them using the old integration routine. Ideally, later once the rest
    // has been tested, this will be moved over to use the Cuba library. Especially
    // as the old fortran integrator does not run on windows!
    integrating = integrating | (1 & compute_bayer_helms_kmn(mirror, nr1, nr2));

    if (!(inter.kmn & INT_BAYER_HELMS) && mirror->map_merged.save_kmn_matrices) {
        sprintf(buf, "%s_bayerhelms", mirror->map_merged.filename);
        write_kmn_to_matrix_file(buf, &(mirror->kmn_bayer_helms));
    }

    // So far above we have calculated the bayer helms coefficients for misalignments
    // and mode mismatch. We now need to calculate the aperture map coupling coefficients
    integrating = integrating | 2 * (1 & compute_aperture_kmn(mirror));

    if (!(inter.kmn & INT_APERTURE) && mirror->map_merged.save_kmn_matrices) {
        sprintf(buf, "%s_aperture", mirror->map_merged.filename);
        write_kmn_to_matrix_file(buf, &(mirror->kmn_aperture));
    }

    // In the kat_init.c in prepare_maps() we merged all the maps together to
    // form an absolute and phase map for reflection and transmission. This now
    // needs to be integrated over to calculate the kmn's. However to reduce the
    // amount of integrating needed we can also compute all the above effects
    // (aperture, curvature, misalignment, etc.) in the same integral if they
    // have not been calculated using some analytic technique. So this
    // is not just called when maps are present.
    compute_kmn_integral(mirror, qx1, qy1, qx2, qy2, nr1, nr2, integrating, mismatch, astigmatism);

    // We now should have multiple coupling coefficient matrices for each effect
    // which now need to be matrix multiplied together to compute the total 
    // coupling matrix.

    //set tmp matrix to identity
    mirror_kmn_matrix_ident(&tmp);

    // if we have calculated aperture kmn analytically do matrix multiplication
    if ((mirror->r_aperture > 0) && !(inter.kmn & INT_APERTURE))
        mirror_kmn_matrix_mult(&(mirror->kmn_aperture), &tmp, &tmp);

    // here we need to check if:
    //  - anything has had to be integrated
    //  - any maps are present
    // Basically anything that has altered the map_merged.kmn variable that also
    // includes effects that are integrated instead of calculated analytically
    if ((integrating > 0) || mirror->map_merged.kmn_calculated)
        mirror_kmn_matrix_mult(&(mirror->map_merged.kmn), &tmp, &tmp);

    // now finally merge the bayer helms coefficients with the others
    // and stick the final result in the mirror coefficients
    mirror_kmn_matrix_mult(&tmp, &(mirror->kmn_bayer_helms), &(mirror->kmn));

    if ((inter.kmn & PRINT_COEFFS) && mirror->map_merged.save_kmn_matrices) {
        sprintf(buf, "%s_merged", mirror->map_merged.filename);
        write_kmn_to_matrix_file(buf, &(mirror->kmn));
    }

    int n, m, n1, n2, m1, m2;
    int num_fields = (int) (inter.tem + 1) * (inter.tem + 2) / 2;
    int turn;

    // Here we adjust the odd reflected horizontal modes. We do this because on
    // reflection, just like in a mirror, the horizontal direction is flipped
    // left to right. We also reverse the extra gouy phase present and rescale the
    // phase.


    mirror_kmn_q_t *kq = &(mirror->kmn_q);

    for (n = 0; n < num_fields; n++) {
        for (m = 0; m < num_fields; m++) {
            get_tem_modes_from_field_index(&n1, &m1, n);
            get_tem_modes_from_field_index(&n2, &m2, m);


            // 1) the old map code had wrong qx1/qx2 for k12/k21 at this place and I assume
            // this was copied over. Fixed this now.
            // 2) the old code uses the `turned' q's to compute thre reverse gouy, so
            // maybe this is needed here as well
            mirror->kmn.k11[n][m] = rev_gouy(mirror->kmn.k11[n][m], n1, m1, n2, m2, kq->qxt1_11, kq->qxt2_11, kq->qyt1_11, kq->qyt2_11);
            mirror->kmn.k12[n][m] = rev_gouy(mirror->kmn.k12[n][m], n1, m1, n2, m2, kq->qxt1_12, kq->qxt2_12, kq->qyt1_12, kq->qyt2_12);
            mirror->kmn.k21[n][m] = rev_gouy(mirror->kmn.k21[n][m], n1, m1, n2, m2, kq->qxt1_21, kq->qxt2_21, kq->qyt1_21, kq->qyt2_21);
            mirror->kmn.k22[n][m] = rev_gouy(mirror->kmn.k22[n][m], n1, m1, n2, m2, kq->qxt1_22, kq->qxt2_22, kq->qyt1_22, kq->qyt2_22);

        }
    }



    double p11 = zphase(mirror->kmn.k11[0][0]);
    double p12 = zphase(mirror->kmn.k12[0][0]);
    double p21 = zphase(mirror->kmn.k21[0][0]);
    double p22 = zphase(mirror->kmn.k22[0][0]);

    for (n = 0; n < num_fields; n++) {
        for (m = 0; m < num_fields; m++) {
            get_tem_modes_from_field_index(&n1, &m1, n);
            get_tem_modes_from_field_index(&n2, &m2, m);
            if (inter.set_tem_phase_zero & 1) {
                mirror->kmn.k11[n][m] = z_by_phr(mirror->kmn.k11[n][m], -p11);
                mirror->kmn.k12[n][m] = z_by_phr(mirror->kmn.k12[n][m], -p12);
                mirror->kmn.k21[n][m] = z_by_phr(mirror->kmn.k21[n][m], -p21);
                mirror->kmn.k22[n][m] = z_by_phr(mirror->kmn.k22[n][m], -p22);
            }
            turn = turnit(n2);

            mirror->kmn.k11[n][m] = z_by_x(mirror->kmn.k11[n][m], turn);
            mirror->kmn.k22[n][m] = z_by_x(mirror->kmn.k22[n][m], turn);
        }
    }



    if ((inter.kmn & PRINT_COEFFS) && mirror->map_merged.save_kmn_matrices) {
        sprintf(buf, "%s_mergedalt", mirror->map_merged.filename);
        write_kmn_to_matrix_file(buf, &(mirror->kmn));
    }

    return (mismatch);
}

//! Checks direction of gaussian beam parameter component and is in the 
//! direction of the beamplitter from the node.

/*!
 *  \param node Node of interest
 *  \param qx Gaussian parameter x component
 *  \param qy Gaussian parameter y component
 *  \param node_idx Index of node
 *  \param comp_idx Index of component
 */
void check_q_direction(node_t* node, complex_t* qx, complex_t* qy, int node_idx, int comp_idx) {
    if (NOT node->gnd_node) {
        node = &inter.node_list[node_idx];
        //if node 2 or 4 the opposite needs to be done
        if (node->component_index == comp_idx && (node_idx % 2)) {
            // reverse q so that beam comes _to_ the bs from node1
            *qx = cminus(cconj(node->qx));
            *qy = cminus(cconj(node->qy));
        } else {
            *qx = node->qx;
            *qy = node->qy;
        }
    }
}

//! Set the coupling coefficients for the given beam splitter

/*!
 * \param bs_index the beam splitter index
 *
 * \see Test_set_k_beamsplitter()
 */
int set_k_beamsplitter(int bs_index) {
    int err;
    int component;
    int node1_index, node2_index, node3_index, node4_index;
    int n1, m1, n2, m2;
    int mismatch;
    double nr1, nr2;
    complex_t kx, ky;
    double kmx, kmy;
    double co1, co2;
    double phase = 0.0;

    ABCD_t trans1;
    complex_t qx1, qy1, qx2, qy2, qx3, qy3, qx4, qy4;
    complex_t qxt, qyt;
    complex_t qxt2, qyt2;
    beamsplitter_t *bs;

    node_t node1, node2, node3, node4;

    // sanity checks on input
    // bs_index should be in the range 0 <= bs_index < inter.num_beamsplitters
    assert(bs_index >= 0);
    assert(bs_index < inter.num_beamsplitters);

    bs = &(inter.bs_list[bs_index]);
    component = get_overall_component_index(BEAMSPLITTER, bs_index);
    node1_index = bs->node1_index;
    node2_index = bs->node2_index;
    node3_index = bs->node3_index;
    node4_index = bs->node4_index;

    nr1 = bs->ir1;
    nr2 = bs->ir2;

    node1 = inter.node_list[node1_index];
    node2 = inter.node_list[node2_index];
    node3 = inter.node_list[node3_index];
    node4 = inter.node_list[node4_index];

    /*
      check_q_direction(&node1, &qx1, &qy1, node1_index, component);
      check_q_direction(&node2, &qx2, &qy2, node2_index, component);
      check_q_direction(&node3, &qx3, &qy3, node3_index, component);
      check_q_direction(&node4, &qx4, &qy4, node4_index, component);
     */

    if (NOT node1.gnd_node) {
        node1 = inter.node_list[node1_index];
        if (node1.component_index == component) {
            // reverse q so that beam comes _to_ the bs from node1
            qx1 = cminus(cconj(node1.qx));
            qy1 = cminus(cconj(node1.qy));
        } else {
            qx1 = node1.qx;
            qy1 = node1.qy;
        }
    }

    if (NOT node2.gnd_node) {
        node2 = inter.node_list[node2_index];
        if (node2.component_index == component) {
            qx2 = node2.qx;
            qy2 = node2.qy;
        } else {
            // reverse q so that beam goes _from_ the bs to node2
            qx2 = cminus(cconj(node2.qx));
            qy2 = cminus(cconj(node2.qy));
        }
    }

    if (NOT node3.gnd_node) {
        node3 = inter.node_list[node3_index];
        if (node3.component_index == component) {
            // reverse q so that beam comes _to_ the bs from node3
            qx3 = cminus(cconj(node3.qx));
            qy3 = cminus(cconj(node3.qy));
        } else {
            qx3 = node3.qx;
            qy3 = node3.qy;
        }
    }

    if (NOT node4.gnd_node) {
        node4 = inter.node_list[node4_index];
        if (node4.component_index == component) {
            qx4 = node4.qx;
            qy4 = node4.qy;
        } else {
            // reverse q so that beam goes _from_ the bs to node4
            qx4 = cminus(cconj(node4.qx));
            qy4 = cminus(cconj(node4.qy));
        }
    }


    co1 = cos(bs->alpha * RAD);
    co2 = cos(asin(sin(bs->alpha * RAD) * nr1 / nr2));

    // If the index of refraction is not the same on both sides the angle
    // will also be different. Need to calculate what the new misalignment
    // angle will be considering the change in index for transmission only

    double gamma_x_n1_n2 = 0; //BETA_SCALE * asin(sin(bs->beta_x) * (nr1/nr2));
    double gamma_y_n1_n2 = 0; //BETA_SCALE * asin(sin(bs->beta_y) * (nr1/nr2));
    double gamma_x_n2_n1 = 0; //BETA_SCALE * asin(sin(bs->beta_x) * (nr2/nr1));
    double gamma_y_n2_n1 = 0; //BETA_SCALE * asin(sin(bs->beta_y) * (nr2/nr1));

    if (inter.debug & 32
            || (inter.trace & 64 && (bs->rebuild > 1 || !inter.setup))) {
        message("%s :\n", bs->name);
    }

    mismatch = 0; // Initialise mismatch

    // Compute reflections on the 12 side of the beamsplitter
    if (NOT node1.gnd_node && NOT node2.gnd_node) {
        // reflection12
        err = component_matrix(&trans1, component, node1_index, node2_index,
                TANGENTIAL);
        if (!err) {
            qxt = q1_q2(trans1, qx1, nr1, nr1);
            err = component_matrix(&trans1, component, node1_index, node2_index,
                    SAGITTAL);
            qyt = q1_q2(trans1, qy1, nr1, nr1);

            qxt2 = qx2;
            qyt2 = qy2;

            if (!ceq(qxt, qxt2) || !ceq(qyt, qyt2)) {
                mismatch = mismatch | 1;
            }

            if (inter.trace & 64) {
                if (!ceq(qxt, qxt2) || !ceq(qyt, qyt2)) {
                    mode_match(qxt, qxt2, &kx, &kmx);
                    mode_match(qyt, qyt2, &ky, &kmy);
                    message(" r12: Kx=%s \n      Ky=%s \n      Kmx %9.3g Kmy %9.3g\n",
                            complex_form(kx), complex_form(ky), kmx, kmy);
                    //printf(" r12: Kmx %9.3g Kmy %9.3g \n      kx=%s ky=%s\n", kmx, kmy, complex_form(kx), complex_form(ky));
                }
            }

            if (inter.debug & 32) {
                message("k12");
            }

            if (inter.set_tem_phase_zero & 1) {
                phase = zphase(k_mnmn(0, 0, 0, 0, qxt, qyt, qxt2, qyt2,
                        BETA_SCALE * bs->beta_x, BETA_SCALE * co1 * bs->beta_y, nr1));
            } else {
                phase = 0.0;
            }


            //printf("qx1 %s qx2 %s qy1 %s qy2 %s\n",complex_form(qxt),complex_form(qxt2),complex_form(qyt),complex_form(qyt2)); 
            //printf("betax %f, co1 %f betay %f nr1 %f\n",bs->beta_x, co1, bs->beta_y, nr1);

            double turn;
            int n, m;
            for (n = 0; n < inter.num_fields; n++) {
                for (m = 0; m < inter.num_fields; m++) {
                    get_tem_modes_from_field_index(&n1, &m1, n);
                    get_tem_modes_from_field_index(&n2, &m2, m);

                    if (n1 == NOT_FOUND || m1 == NOT_FOUND ||
                            n2 == NOT_FOUND || m2 == NOT_FOUND) {
                        bug_error("notfound 1");
                    }

                    bs->k12[n][m] = z_by_phr(k_mnmn(n1, m1, n2, m2, qxt, qyt, qxt2,
                            qyt2, BETA_SCALE * bs->beta_x,
                            BETA_SCALE * co1 * bs->beta_y, nr1), -phase);

                    turn = turnit(n2);
                    bs->k12[n][m] = z_by_x(bs->k12[n][m], turn);


                    if (inter.debug & 32 && !ceq(complex_0, bs->k12[n][m])) {
                        message(" [%d, %d]=%s, ", n, m, complex_form(bs->k12[n][m]));
                    }
                }

                if (inter.debug & 32) {
                    message("\n   ");
                }
            }

            if (inter.debug & 32) {
                message("\n");
            }
        }

        // reflection21
        err = component_matrix(&trans1, component, node2_index, node1_index,
                TANGENTIAL);
        if (!err) {
            qxt = q1_q2(trans1, cminus(cconj(qx2)), nr1, nr1);
            err = component_matrix(&trans1, component, node2_index, node1_index,
                    SAGITTAL);
            qyt = q1_q2(trans1, cminus(cconj(qy2)), nr1, nr1);

            qxt2 = cminus(cconj(qx1));
            qyt2 = cminus(cconj(qy1));

            if (!ceq(qxt, qxt2) || !ceq(qyt, qyt2)) {
                mismatch = mismatch | 1;
            }

            if (inter.trace & 64) {
                if (!ceq(qxt, qxt2) || !ceq(qyt, qyt2)) {
                    mode_match(qxt, qxt2, &kx, &kmx);
                    mode_match(qyt, qyt2, &ky, &kmy);
                    message(" r21: Kx=%s \n      Ky=%s \n      Kmx %9.3g Kmy %9.3g\n",
                            complex_form(kx), complex_form(ky), kmx, kmy);
                }
            }

            if (inter.debug & 32) {
                message("k21");
            }

            if (inter.set_tem_phase_zero & 1) {
                phase = zphase(k_mnmn(0, 0, 0, 0, qxt, qyt, qxt2, qyt2,
                        BETA_SCALE * bs->beta_x, BETA_SCALE * co1 * bs->beta_y, nr1));
            } else {
                phase = 0.0;
            }

            double turn;
            int n, m;

            for (n = 0; n < inter.num_fields; n++) {
                for (m = 0; m < inter.num_fields; m++) {
                    get_tem_modes_from_field_index(&n1, &m1, n);
                    get_tem_modes_from_field_index(&n2, &m2, m);

                    if (n1 == NOT_FOUND || m1 == NOT_FOUND ||
                            n2 == NOT_FOUND || m2 == NOT_FOUND) {
                        bug_error("notfound 2");
                    }

                    //Calculate coupling coeffcient from node 1 to 2 for hermite gauss mode nm
                    //also rotate phase by -phase calculated previously, not sure why...
                    bs->k21[n][m] = z_by_phr(k_mnmn(n1, m1, n2, m2, qxt, qyt, qxt2,
                            qyt2, BETA_SCALE * bs->beta_x,
                            BETA_SCALE * co1 * bs->beta_y, nr1), -phase);
                    turn = turnit(n2);
                    bs->k21[n][m] = z_by_x(bs->k21[n][m], turn);

                    if (inter.debug & 32 && !ceq(complex_0, bs->k21[n][m])) {
                        message(" [%d, %d]=%s, ", n, m, complex_form(bs->k21[n][m]));
                    }
                }

                if (inter.debug & 32) {
                    message("\n   ");
                }
            }
            if (inter.debug & 32) {
                message("\n");
            }
        }
    }

    if (NOT node3.gnd_node && NOT node4.gnd_node) {
        // reflection34
        err = component_matrix(&trans1, component, node3_index, node4_index,
                TANGENTIAL);
        if (!err) {
            qxt = q1_q2(trans1, qx3, nr2, nr2);
            err = component_matrix(&trans1, component, node3_index, node4_index,
                    SAGITTAL);
            qyt = q1_q2(trans1, qy3, nr2, nr2);

            qxt2 = qx4;
            qyt2 = qy4;
            if (!ceq(qxt, qxt2) || !ceq(qyt, qyt2)) {
                mismatch = mismatch | 2;
            }

            if (inter.trace & 64) {
                if (!ceq(qxt, qxt2) || !ceq(qyt, qyt2)) {
                    mode_match(qxt, qxt2, &kx, &kmx);
                    mode_match(qyt, qyt2, &ky, &kmy);
                    message(" r34: Kx=%s \n      Ky=%s \n      Kmx %9.3g Kmy %9.3g\n",
                            complex_form(kx), complex_form(ky), kmx, kmy);
                }
            }

            if (inter.debug & 32) {
                message("k34");
            }

            if (inter.set_tem_phase_zero & 1) {
                phase = zphase(k_mnmn(0, 0, 0, 0, qxt, qyt, qxt2, qyt2,
                        BETA_SCALE * bs->beta_x, -co2 * BETA_SCALE * bs->beta_y, nr2));
            } else {
                phase = 0.0;
            }

            double turn;
            int n, m;
            for (n = 0; n < inter.num_fields; n++) {
                for (m = 0; m < inter.num_fields; m++) {
                    get_tem_modes_from_field_index(&n1, &m1, n);
                    get_tem_modes_from_field_index(&n2, &m2, m);

                    if (n1 == NOT_FOUND || m1 == NOT_FOUND ||
                            n2 == NOT_FOUND || m2 == NOT_FOUND) {
                        bug_error("notfound 3");
                    }
                    bs->k34[n][m] = z_by_phr(k_mnmn(n1, m1, n2, m2, qxt, qyt, qxt2,
                            qyt2, BETA_SCALE * bs->beta_x,
                            -co2 * BETA_SCALE * bs->beta_y, nr2), -phase);

                    turn = turnit(n2);
                    bs->k34[n][m] = z_by_x(bs->k34[n][m], turn);



                    if (inter.debug & 32 && !ceq(complex_0, bs->k34[n][m])) {
                        message(" [%d, %d]=%s, ", n, m, complex_form(bs->k34[n][m]));
                    }
                }

                if (inter.debug & 32) {
                    message("\n   ");
                }
            }

            if (inter.debug & 32) {
                message("\n");
            }
        }

        // reflection43
        err = component_matrix(&trans1, component, node4_index, node3_index,
                TANGENTIAL);
        if (!err) {
            qxt = q1_q2(trans1, cminus(cconj(qx4)), nr2, nr2);
            err = component_matrix(&trans1, component, node4_index, node3_index,
                    SAGITTAL);
            qyt = q1_q2(trans1, cminus(cconj(qy4)), nr2, nr2);

            qxt2 = cminus(cconj(qx3));
            qyt2 = cminus(cconj(qy3));
            if (!ceq(qxt, qxt2) || !ceq(qyt, qyt2)) {
                mismatch = mismatch | 2;
            }

            if (inter.trace & 64) {
                if (!ceq(qxt, qxt2) || !ceq(qyt, qyt2)) {
                    mode_match(qxt, qxt2, &kx, &kmx);
                    mode_match(qyt, qyt2, &ky, &kmy);
                    message(" r43: Kx=%s \n      Ky=%s \n      Kmx %9.3g Kmy %9.3g\n",
                            complex_form(kx), complex_form(ky), kmx, kmy);
                }
            }

            if (inter.debug & 32) {
                message("k43");
            }

            if (inter.set_tem_phase_zero & 1) {
                phase = zphase(k_mnmn(0, 0, 0, 0, qxt, qyt, qxt2, qyt2,
                        BETA_SCALE * bs->beta_x, -BETA_SCALE * co2 * bs->beta_y, nr2));
            } else {
                phase = 0.0;
            }

            double turn;
            int n, m;
            for (n = 0; n < inter.num_fields; n++) {
                for (m = 0; m < inter.num_fields; m++) {
                    get_tem_modes_from_field_index(&n1, &m1, n);
                    get_tem_modes_from_field_index(&n2, &m2, m);
                    if (n1 == NOT_FOUND || m1 == NOT_FOUND ||
                            n2 == NOT_FOUND || m2 == NOT_FOUND) {
                        bug_error("notfound 4");
                    }
                    bs->k43[n][m] = z_by_phr(k_mnmn(n1, m1, n2, m2, qxt, qyt, qxt2,
                            qyt2, BETA_SCALE * bs->beta_x,
                            -BETA_SCALE * co2 * bs->beta_y, nr2), -phase);

                    turn = turnit(n2);
                    bs->k43[n][m] = z_by_x(bs->k43[n][m], turn);


                    if (inter.debug & 32 && !ceq(complex_0, bs->k43[n][m])) {
                        message(" [%d, %d]=%s, ", n, m, complex_form(bs->k43[n][m]));
                    }
                }

                if (inter.debug & 32) {
                    message("\n   ");
                }
            }

            if (inter.debug & 32) {
                message("\n");
            }
        }
    }

    if (NOT node1.gnd_node &&
            NOT node3.gnd_node) {
        // transmission13
        err = component_matrix(&trans1, component, node1_index, node3_index,
                TANGENTIAL);
        if (!err) {
            qxt = q1_q2(trans1, qx1, nr1, nr2);
            err = component_matrix(&trans1, component, node1_index, node3_index,
                    SAGITTAL);
            qyt = q1_q2(trans1, qy1, nr1, nr2);

            qxt2 = cminus(cconj(qx3));
            qyt2 = cminus(cconj(qy3));

            if (!ceq(qxt, qxt2) || !ceq(qyt, qyt2)) {
                mismatch = mismatch | 4;
            }

            if (inter.trace & 64) {
                if (!ceq(qxt, qxt2) || !ceq(qyt, qyt2)) {
                    mode_match(qxt, qxt2, &kx, &kmx);
                    mode_match(qyt, qyt2, &ky, &kmy);
                    //printf(" t13: Kmx %9.3g Kmy %9.3g \n      kx=%s ky=%s\n", kmx, kmy, complex_form(kx), complex_form(ky));
                    message(" t13: Kx=%s \n      Ky=%s \n      Kmx %9.3g Kmy %9.3g\n",
                            complex_form(kx), complex_form(ky), kmx, kmy);
                }
            }

            if (inter.debug & 32) {
                message("k13");
            }

            if (inter.set_tem_phase_zero & 1) {
                phase = zphase(k_mnmn(0, 0, 0, 0, qxt, qyt, qxt2, qyt2, gamma_x_n1_n2, gamma_y_n1_n2, nr2));
            } else {
                phase = 0.0;
            }

            int n, m;
            for (n = 0; n < inter.num_fields; n++) {
                for (m = 0; m < inter.num_fields; m++) {
                    get_tem_modes_from_field_index(&n1, &m1, n);
                    get_tem_modes_from_field_index(&n2, &m2, m);
                    if (n1 == NOT_FOUND || m1 == NOT_FOUND ||
                            n2 == NOT_FOUND || m2 == NOT_FOUND) {
                        bug_error("notfound 5");
                    }
                    // beta = 0.0 fuer transmission
                    bs->k13[n][m] = z_by_phr(k_mnmn(n1, m1, n2, m2, qxt, qyt, qxt2,
                            qyt2, gamma_x_n1_n2, gamma_y_n1_n2, nr2), -phase);
                    if (inter.debug & 32 && !ceq(complex_0, bs->k13[n][m])) {
                        message(" [%d, %d]=%s, ", n, m, complex_form(bs->k13[n][m]));
                    }
                }

                if (inter.debug & 32) {
                    message("\n   ");
                }
            }

            if (inter.debug & 32) {
                message("\n");
            }
        }

        // transmission31
        err = component_matrix(&trans1, component, node3_index, node1_index,
                TANGENTIAL);
        if (!err) {
            qxt = q1_q2(trans1, qx3, nr2, nr1); // bug cmcc 260402
            err = component_matrix(&trans1, component, node3_index, node1_index,
                    SAGITTAL);
            qyt = q1_q2(trans1, qy3, nr2, nr1);

            qxt2 = cminus(cconj(qx1));
            qyt2 = cminus(cconj(qy1));
            if (!ceq(qxt, qxt2) || !ceq(qyt, qyt2)) {
                mismatch = mismatch | 4;
            }

            if (inter.trace & 64) {
                if (!ceq(qxt, qxt2) || !ceq(qyt, qyt2)) {
                    mode_match(qxt, qxt2, &kx, &kmx);
                    mode_match(qyt, qyt2, &ky, &kmy);
                    //printf(" t31: Kmx %9.3g Kmy %9.3g \n      kx=%s ky=%s\n", kmx, kmy, complex_form(kx), complex_form(ky));
                    message(" t31: Kx=%s \n      Ky=%s \n      Kmx %9.3g Kmy %9.3g\n",
                            complex_form(kx), complex_form(ky), kmx, kmy);
                }
            }

            if (inter.debug & 32) {
                message("k31");
            }

            if (inter.set_tem_phase_zero & 1) {
                phase = zphase(k_mnmn(0, 0, 0, 0, qxt, qyt, qxt2, qyt2, gamma_x_n2_n1, gamma_y_n2_n1, nr1));
            } else {
                phase = 0.0;
            }

            int n, m;
            for (n = 0; n < inter.num_fields; n++) {
                for (m = 0; m < inter.num_fields; m++) {
                    get_tem_modes_from_field_index(&n1, &m1, n);
                    get_tem_modes_from_field_index(&n2, &m2, m);
                    if (n1 == NOT_FOUND || m1 == NOT_FOUND ||
                            n2 == NOT_FOUND || m2 == NOT_FOUND) {
                        bug_error("notfound 6");
                    }
                    // beta = 0.0 fuer transmission
                    bs->k31[n][m] = z_by_phr(k_mnmn(n1, m1, n2, m2, qxt, qyt, qxt2,
                            qyt2, gamma_x_n2_n1, gamma_y_n2_n1, nr1), -phase);
                    if (inter.debug & 32 && !ceq(complex_0, bs->k31[n][m])) {
                        message(" [%d, %d]=%s, ", n, m, complex_form(bs->k31[n][m]));
                    }
                }

                if (inter.debug & 32) {
                    message("\n   ");
                }
            }

            if (inter.debug & 32) {
                message("\n");
            }
        }
    }

    if (NOT node2.gnd_node &&
            NOT node4.gnd_node) {
        // transmission24
        err = component_matrix(&trans1, component, node2_index, node4_index,
                TANGENTIAL);
        if (!err) {
            qxt = q1_q2(trans1, cminus(cconj(qx2)), nr1, nr2);
            err = component_matrix(&trans1, component, node2_index, node4_index,
                    SAGITTAL);
            qyt = q1_q2(trans1, cminus(cconj(qy2)), nr1, nr2);

            qxt2 = qx4;
            qyt2 = qy4;
            if (!ceq(qxt, qxt2) || !ceq(qyt, qyt2)) {
                mismatch = mismatch | 8;
            }

            if (inter.trace & 64) {
                if (!ceq(qxt, qxt2) || !ceq(qyt, qyt2)) {
                    mode_match(qxt, qxt2, &kx, &kmx);
                    mode_match(qyt, qyt2, &ky, &kmy);
                    //printf(" t24: Kmx %9.3g Kmy %9.3g \n      kx=%s ky=%s\n", kmx, kmy, complex_form(kx), complex_form(ky));
                    message(" t24: Kx=%s \n      Ky=%s \n      Kmx %9.3g Kmy %9.3g\n",
                            complex_form(kx), complex_form(ky), kmx, kmy);
                }
            }

            if (inter.debug & 32) {
                message("k24");
            }

            if (inter.set_tem_phase_zero & 1) {
                phase = zphase(k_mnmn(0, 0, 0, 0, qxt, qyt, qxt2, qyt2, gamma_x_n1_n2, -gamma_y_n1_n2, nr2));
            } else {
                phase = 0.0;
            }

            int n, m;
            for (n = 0; n < inter.num_fields; n++) {
                for (m = 0; m < inter.num_fields; m++) {
                    get_tem_modes_from_field_index(&n1, &m1, n);
                    get_tem_modes_from_field_index(&n2, &m2, m);
                    if (n1 == NOT_FOUND || m1 == NOT_FOUND ||
                            n2 == NOT_FOUND || m2 == NOT_FOUND) {
                        bug_error("notfound 7");
                    }
                    // beta = 0.0 fuer transmission
                    bs->k24[n][m] = z_by_phr(k_mnmn(n1, m1, n2, m2, qxt, qyt, qxt2,
                            qyt2, gamma_x_n1_n2, -gamma_y_n1_n2, nr2), -phase);
                    if (inter.debug & 32 && !ceq(complex_0, bs->k24[n][m])) {
                        message(" [%d, %d]=%s, ", n, m, complex_form(bs->k24[n][m]));
                    }
                }

                if (inter.debug & 32) {
                    message("\n   ");
                }
            }

            if (inter.debug & 32) {
                message("\n");
            }
        }

        // transmission42
        err = component_matrix(&trans1, component, node4_index, node2_index,
                TANGENTIAL);
        if (!err) {
            qxt = q1_q2(trans1, cminus(cconj(qx4)), nr2, nr1);
            err = component_matrix(&trans1, component, node4_index, node2_index,
                    SAGITTAL);
            qyt = q1_q2(trans1, cminus(cconj(qy4)), nr2, nr1);

            qxt2 = qx2;
            qyt2 = qy2;
            if (!ceq(qxt, qxt2) || !ceq(qyt, qyt2)) {
                mismatch = mismatch | 8;
            }

            if (inter.trace & 64) {
                if (!ceq(qxt, qxt2) || !ceq(qyt, qyt2)) {
                    mode_match(qxt, qxt2, &kx, &kmx);
                    mode_match(qyt, qyt2, &ky, &kmy);
                    //printf(" t42: Kmx %9.3g Kmy %9.3g \n      kx=%s ky=%s\n", kmx, kmy, complex_form(kx), complex_form(ky));
                    message(" t42: Kx=%s \n      Ky=%s \n      Kmx %9.3g Kmy %9.3g\n",
                            complex_form(kx), complex_form(ky), kmx, kmy);
                }
            }

            if (inter.debug & 32) {
                message("k42");
            }

            if (inter.set_tem_phase_zero & 1) {
                phase = zphase(k_mnmn(0, 0, 0, 0, qxt, qyt, qxt2, qyt2, gamma_x_n2_n1, -gamma_y_n2_n1, nr1));
            } else {
                phase = 0.0;
            }

            int n, m;
            for (n = 0; n < inter.num_fields; n++) {
                for (m = 0; m < inter.num_fields; m++) {
                    get_tem_modes_from_field_index(&n1, &m1, n);
                    get_tem_modes_from_field_index(&n2, &m2, m);
                    if (n1 == NOT_FOUND || m1 == NOT_FOUND ||
                            n2 == NOT_FOUND || m2 == NOT_FOUND) {
                        bug_error("notfound 8");
                    }
                    // beta = 0.0 fuer transmission
                    bs->k42[n][m] = z_by_phr(k_mnmn(n1, m1, n2, m2, qxt, qyt, qxt2,
                            qyt2, gamma_x_n2_n1, -gamma_y_n2_n1, nr1), -phase);
                    if (inter.debug & 32 && !ceq(complex_0, bs->k42[n][m])) {
                        message(" [%d, %d]=%s, ", n, m, complex_form(bs->k42[n][m]));
                    }
                }

                if (inter.debug & 32) {
                    message("\n   ");
                }
            }

            if (inter.debug & 32) {
                message("\n");
            }
        }
    }

    return (mismatch);
}

//! Set coupling coefficient for a free space element

/*!
 * \param space_index index of free space element
 *
 * \see Test_set_k_space()
 */
int set_k_space(int space_index) {
    int err;
    int component;
    int node1_index, node2_index;
    int n1, m1, n2, m2;
    int mismatch;
    double nr;
    complex_t kx, ky;
    double kmx, kmy;
    double phase = 0.0;

    ABCD_t trans1;
    complex_t qx1, qy1, qx2, qy2;
    complex_t qxt, qyt;
    complex_t qxt2, qyt2;
    space_t *space;

    node_t node1, node2;

    // do sanity check on input
    // space_index should be in the range 0 <= space_index < inter.num_spaces
    assert(space_index >= 0);
    assert(space_index < inter.num_spaces);

    space = &(inter.space_list[space_index]);
    nr = space->n;
    component = get_overall_component_index(SPACE, space_index);

    node1_index = space->node1_index;
    node2_index = space->node2_index;

    node1 = inter.node_list[node1_index];
    node2 = inter.node_list[node2_index];

    //if (NOT node1.gnd_node) {
    //node1 = inter.node_list[node1_index];
    //}

    //if (NOT node2.gnd_node) {
    //node2 = inter.node_list[node2_index];
    //}
    mismatch = 0;

    if (NOT node1.gnd_node) {
        if (node1.component_index == component) {
            // reverse q so that beam comes _to_ the space from node1
            qx1 = cminus(cconj(node1.qx));
            qy1 = cminus(cconj(node1.qy));
        } else {
            qx1 = node1.qx;
            qy1 = node1.qy;
        }
    }

    if (NOT node2.gnd_node) {
        if (node2.component_index == component) {
            qx2 = node2.qx;
            qy2 = node2.qy;
        } else {
            // reverse q so that beam goes _from_ the space to node2
            qx2 = cminus(cconj(node2.qx));
            qy2 = cminus(cconj(node2.qy));
        }
    }

    if (inter.debug & 32 || inter.trace & 64) {
        message("%s :\n", space->name);
    }

    if (NOT node1.gnd_node &&
            NOT node2.gnd_node) {
        // transmission1
        err = component_matrix(&trans1, component, node1_index, node2_index,
                TANGENTIAL);
        if (!err) {
            qxt = q1_q2(trans1, qx1, nr, nr);
            err = component_matrix(&trans1, component, node1_index, node2_index,
                    SAGITTAL);
            qyt = q1_q2(trans1, qy1, nr, nr);

            qxt2 = qx2;
            qyt2 = qy2;
            if (!ceq(qxt, qxt2) || !ceq(qyt, qyt2)) {
                mismatch = mismatch | 4;
            }

            if (inter.trace & 64) {
                if (!ceq(qxt, qxt2) || !ceq(qyt, qyt2)) {
                    mode_match(qxt, qxt2, &kx, &kmx);
                    mode_match(qyt, qyt2, &ky, &kmy);
                    message(" t12: Kx=%s \n      Ky=%s \n      Kmx %9.3g Kmy %9.3g\n",
                            complex_form(kx), complex_form(ky), kmx, kmy);
                }
            }

            if (inter.debug & 32) {
                message("k12");
            }

            if (inter.set_tem_phase_zero & 1) {
                phase = zphase(k_mnmn(0, 0, 0, 0, qxt, qyt, qxt2, qyt2, 0.0, 0.0, nr));
            } else {
                phase = 0.0;
            }

            int n, m;
            for (n = 0; n < inter.num_fields; n++) {
                for (m = 0; m < inter.num_fields; m++) {
                    get_tem_modes_from_field_index(&n1, &m1, n);
                    get_tem_modes_from_field_index(&n2, &m2, m);
                    if (n1 == NOT_FOUND || m1 == NOT_FOUND ||
                            n2 == NOT_FOUND || m2 == NOT_FOUND) {
                        bug_error("notfound 1");
                    }
                    space->k12[n][m] = z_by_phr(k_mnmn(n1, m1, n2, m2, qxt, qyt, qxt2,
                            qyt2, 0.0, 0.0, nr), -phase);
                    if (inter.debug & 32 && !ceq(complex_0, space->k12[n][m])) {
                        message(" [%d, %d]=%s, ", n, m, complex_form(space->k12[n][m]));
                    }
                }

                if (inter.debug & 32) {
                    message("\n   ");
                }
            }

            if (inter.debug & 32) {
                message("\n");
            }
        }

        // transmission2
        err = component_matrix(&trans1, component, node2_index, node1_index,
                TANGENTIAL);
        if (!err) {
            qxt = q1_q2(trans1, cminus(cconj(qx2)), nr, nr);
            err = component_matrix(&trans1, component, node2_index, node1_index,
                    SAGITTAL);
            qyt = q1_q2(trans1, cminus(cconj(qy2)), nr, nr);

            qxt2 = cminus(cconj(qx1));
            qyt2 = cminus(cconj(qy1));
            if (!ceq(qxt, qxt2) || !ceq(qyt, qyt2)) {
                mismatch = mismatch | 8;
            }

            if (inter.trace & 64) {
                if (!ceq(qxt, qxt2) || !ceq(qyt, qyt2)) {
                    mode_match(qxt, qxt2, &kx, &kmx);
                    mode_match(qyt, qyt2, &ky, &kmy);
                    message(" t21: Kx=%s \n      Ky=%s \n      Kmx %9.3g Kmy %9.3g\n",
                            complex_form(kx), complex_form(ky), kmx, kmy);
                }
            }

            if (inter.debug & 32) {
                message("k21");
            }

            if (inter.set_tem_phase_zero & 1) {
                phase = zphase(k_mnmn(0, 0, 0, 0, qxt, qyt, qxt2, qyt2, 0.0, 0.0, nr));
            } else {
                phase = 0.0;
            }

            int n, m;
            for (n = 0; n < inter.num_fields; n++) {
                for (m = 0; m < inter.num_fields; m++) {
                    get_tem_modes_from_field_index(&n1, &m1, n);
                    get_tem_modes_from_field_index(&n2, &m2, m);
                    if (n1 == NOT_FOUND || m1 == NOT_FOUND ||
                            n2 == NOT_FOUND || m2 == NOT_FOUND) {
                        bug_error("notfound 2");
                    }
                    space->k21[n][m] = z_by_phr(k_mnmn(n1, m1, n2, m2, qxt, qyt, qxt2,
                            qyt2, 0.0, 0.0, nr), -phase);
                    if (inter.debug & 32 && !ceq(complex_0, space->k21[n][m])) {
                        message(" [%d, %d]=%s, ", n, m, complex_form(space->k21[n][m]));
                    }
                }

                if (inter.debug & 32) {
                    message("\n   ");
                }
            }

            if (inter.debug & 32) {
                message("\n");
            }
        }
    }

    return (mismatch);
}

//! Set coupling coefficient for a lens

/*!
 * \param lens_index lens index
 *
 * \see Test_set_k_lens()
 */
int set_k_lens(int lens_index) {
    int err;
    int component;
    int node1_index, node2_index;
    int n1, m1, n2, m2;
    int mismatch;
    double nr1, nr2;
    complex_t kx, ky;
    double kmx, kmy;
    double phase = 0.0;

    ABCD_t trans1;
    complex_t qx1, qy1, qx2, qy2;
    complex_t qxt, qyt;
    complex_t qxt2, qyt2;
    lens_t *lens;

    node_t node1, node2;

    // sanity check on input
    // lens index should be in the range 0 <= lens_index < inter.num_lenses
    assert(lens_index >= 0);
    assert(lens_index < inter.num_lenses);

    lens = &(inter.lens_list[lens_index]);
    component = get_overall_component_index(LENS, lens_index);
    node1_index = lens->node1_index;
    node2_index = lens->node2_index;

    nr1 = nr2 = init.n0;

    node1 = inter.node_list[node1_index];
    node2 = inter.node_list[node2_index];

    if (NOT node1.gnd_node) {
        //node1 = inter.node_list[node1_index];
        nr1 = *node1.n;
        if (node1.component_index == component) {
            // reverse q so that beam comes _to_ the lens from node1
            qx1 = cminus(cconj(node1.qx));
            qy1 = cminus(cconj(node1.qy));
        } else {
            qx1 = node1.qx;
            qy1 = node1.qy;
        }
    }

    if (NOT node2.gnd_node) {
        //node2 = inter.node_list[node2_index];
        nr2 = *node2.n;
        if (node2.component_index == component) {
            qx2 = node2.qx;
            qy2 = node2.qy;
        } else {
            // reverse q so that beam goes _from_ the lens to node2
            qx2 = cminus(cconj(node2.qx));
            qy2 = cminus(cconj(node2.qy));
        }
    }

    if (nr1 != nr2) {
        gerror("index of refraction not consistent at lens '%s'\n", lens->name);
    }

    if (inter.debug & 32 || inter.trace & 64) {
        message("%s :\n", lens->name);
    }

    mismatch = 0;

    if (NOT node1.gnd_node &&
            NOT node2.gnd_node) {
        // transmission1
        err = component_matrix(&trans1, component, node1_index, node2_index,
                TANGENTIAL);
        if (!err) {
            qxt = q1_q2(trans1, qx1, nr1, nr2);
            err = component_matrix(&trans1, component, node1_index, node2_index,
                    SAGITTAL);
            qyt = q1_q2(trans1, qy1, nr1, nr2);

            qxt2 = qx2;
            qyt2 = qy2;
            if (!ceq(qxt, qxt2) || !ceq(qyt, qyt2)) {
                mismatch = mismatch | 4;
            }

            if (inter.trace & 64) {
                if (!ceq(qxt, qxt2) || !ceq(qyt, qyt2)) {
                    mode_match(qxt, qxt2, &kx, &kmx);
                    mode_match(qyt, qyt2, &ky, &kmy);
                    message(" t12: Kx=%s \n      Ky=%s \n      Kmx %9.3g Kmy %9.3g\n",
                            complex_form(kx), complex_form(ky), kmx, kmy);
                }
            }

            if (inter.debug & 32) {
                message("k12");
            }

            if (inter.set_tem_phase_zero & 1) {
                phase = zphase(k_mnmn(0, 0, 0, 0, qxt, qyt, qxt2, qyt2, 0.0, 0.0, nr2));
            } else {
                phase = 0;
            }

            int n, m;
            for (n = 0; n < inter.num_fields; n++) {
                for (m = 0; m < inter.num_fields; m++) {
                    get_tem_modes_from_field_index(&n1, &m1, n);
                    get_tem_modes_from_field_index(&n2, &m2, m);
                    if (n1 == NOT_FOUND || m1 == NOT_FOUND ||
                            n2 == NOT_FOUND || m2 == NOT_FOUND) {
                        bug_error("notfound 1");
                    }
                    lens->k12[n][m] = z_by_phr(k_mnmn(n1, m1, n2, m2, qxt, qyt, qxt2,
                            qyt2, 0.0, 0.0, nr2), -phase);
                    if (inter.debug & 32 && !ceq(complex_0, lens->k12[n][m])) {
                        message(" [%d, %d]=%s, ", n, m, complex_form(lens->k12[n][m]));
                    }
                }

                if (inter.debug & 32) {
                    message("\n   ");
                }
            }

            if (inter.debug & 32) {
                message("\n");
            }
        }

        // transmission2
        err = component_matrix(&trans1, component, node2_index, node1_index,
                TANGENTIAL);
        if (!err) {
            qxt = q1_q2(trans1, cminus(cconj(qx2)), nr2, nr1);
            err = component_matrix(&trans1, component, node2_index, node1_index,
                    SAGITTAL);
            qyt = q1_q2(trans1, cminus(cconj(qy2)), nr2, nr1);

            qxt2 = cminus(cconj(qx1));
            qyt2 = cminus(cconj(qy1));
            if (!ceq(qxt, qxt2) || !ceq(qyt, qyt2)) {
                mismatch = mismatch | 4;
            }

            if (inter.trace & 64) {
                if (!ceq(qxt, qxt2) || !ceq(qyt, qyt2)) {
                    mode_match(qxt, qxt2, &kx, &kmx);
                    mode_match(qyt, qyt2, &ky, &kmy);
                    message(" t21: Kx=%s \n      Ky=%s \n      Kmx %9.3g Kmy %9.3g\n",
                            complex_form(kx), complex_form(ky), kmx, kmy);
                }
            }

            if (inter.debug & 32) {
                message("k21");
            }

            if (inter.set_tem_phase_zero & 1) {
                phase = zphase(k_mnmn(0, 0, 0, 0, qxt, qyt, qxt2, qyt2, 0.0, 0.0, nr1));
            } else {
                phase = 0;
            }

            int n, m;
            for (n = 0; n < inter.num_fields; n++) {
                for (m = 0; m < inter.num_fields; m++) {
                    get_tem_modes_from_field_index(&n1, &m1, n);
                    get_tem_modes_from_field_index(&n2, &m2, m);
                    if (n1 == NOT_FOUND || m1 == NOT_FOUND ||
                            n2 == NOT_FOUND || m2 == NOT_FOUND) {
                        bug_error("notfound 2");
                    }
                    lens->k21[n][m] = z_by_phr(k_mnmn(n1, m1, n2, m2, qxt, qyt, qxt2,
                            qyt2, 0.0, 0.0, nr1), -phase);
                    if (inter.debug & 32 && !ceq(complex_0, lens->k21[n][m])) {
                        message(" [%d, %d]=%s, ", n, m, complex_form(lens->k21[n][m]));
                    }
                }

                if (inter.debug & 32) {
                    message("\n   ");
                }
            }

            if (inter.debug & 32) {
                message("\n");
            }
        }
    }

    return (mismatch);
}

//! Set coupling coefficient for a modulator

/*!
 * \param modulator_index modulator index
 *
 * \see Test_set_k_modulator
 */
int set_k_modulator(int modulator_index) {
    int component;
    int node1_index, node2_index;
    int n1, m1, n2, m2;
    int mismatch;
    double nr1, nr2;
    complex_t kx, ky;
    double kmx, kmy;
    double phase = 0.0;

    complex_t qx1, qy1, qx2, qy2;
    complex_t qxt, qyt;
    complex_t qxt2, qyt2;
    modulator_t *modulator;
    node_t node1, node2;

    // sanity check on input
    // modulator index should be in the range 
    // 0 <= modulator_index < inter.num_modulators
    assert(modulator_index >= 0);
    assert(modulator_index < inter.num_modulators);

    modulator = &(inter.modulator_list[modulator_index]);
    component = get_overall_component_index(MODULATOR, modulator_index);
    node1_index = modulator->node1_index;
    node2_index = modulator->node2_index;

    nr1 = nr2 = init.n0;

    node1 = inter.node_list[node1_index];
    node2 = inter.node_list[node2_index];

    if (NOT node1.gnd_node) {
        //node1 = inter.node_list[node1_index];
        nr1 = *node1.n;
        if (node1.component_index == component) {
            // reverse q so that beam comes _to_ the mod from node1
            qx1 = cminus(cconj(node1.qx));
            qy1 = cminus(cconj(node1.qy));
        } else {
            qx1 = node1.qx;
            qy1 = node1.qy;
        }
    }

    if (NOT node2.gnd_node) {
        //node2 = inter.node_list[node2_index];
        nr2 = *node2.n;
        if (node2.component_index == component) {
            qx2 = node2.qx;
            qy2 = node2.qy;
        } else {
            // reverse q so that beam goes _from_ the mod to node2
            qx2 = cminus(cconj(node2.qx));
            qy2 = cminus(cconj(node2.qy));
        }
    }

    if (nr1 != nr2) {
        gerror("index of refraction not consistent at modulator '%s'\n",
                modulator->name);
    }

    if (inter.debug & 32 || inter.trace & 64) {
        message("%s :\n", modulator->name);
    }

    mismatch = 0;

    if (NOT node1.gnd_node &&
            NOT node2.gnd_node) {
        qxt = qx1;
        qyt = qy1;

        qxt2 = qx2;
        qyt2 = qy2;
        if (!ceq(qxt, qxt2) || !ceq(qyt, qyt2)) {
            mismatch = mismatch | 4;
        }

        if (inter.trace & 64) {
            if (!ceq(qxt, qxt2) || !ceq(qyt, qyt2)) {
                mode_match(qxt, qxt2, &kx, &kmx);
                mode_match(qyt, qyt2, &ky, &kmy);
                message(" t12: Kx=%s \n      Ky=%s \n      Kmx %9.3g Kmy %9.3g\n",
                        complex_form(kx), complex_form(ky), kmx, kmy);
            }
        }

        if (inter.debug & 32) {
            message("k12");
        }

        if (inter.set_tem_phase_zero & 1) {
            phase = zphase(k_mnmn(0, 0, 0, 0, qxt, qyt, qxt2, qyt2, 0.0, 0.0, nr2));
        } else {
            phase = 0.0;
        }

        int n, m;
        for (n = 0; n < inter.num_fields; n++) {
            for (m = 0; m < inter.num_fields; m++) {
                get_tem_modes_from_field_index(&n1, &m1, n);
                get_tem_modes_from_field_index(&n2, &m2, m);
                if (n1 == NOT_FOUND || m1 == NOT_FOUND ||
                        n2 == NOT_FOUND || m2 == NOT_FOUND) {
                    bug_error("notfound 1");
                }
                modulator->k12[n][m] =
                        z_by_phr(k_mnmn(n1, m1, n2, m2, qxt, qyt, qxt2, qyt2, 0.0, 0.0, nr2),
                        -phase);
                if (inter.debug & 32 && !ceq(complex_0, modulator->k12[n][m])) {
                    message(" [%d, %d]=%s, ", n, m, complex_form(modulator->k12[n][m]));
                }
            }

            if (inter.debug & 32) {
                message("\n   ");
            }
        }

        if (inter.debug & 32) {
            message("\n");
        }

        // transmission2
        qxt = cminus(cconj(qx2));
        qyt = cminus(cconj(qy2));
        qxt2 = cminus(cconj(qx1));
        qyt2 = cminus(cconj(qy1));
        if (!ceq(qxt, qxt2) || !ceq(qyt, qyt2)) {
            mismatch = mismatch | 4;
        }

        if (inter.trace & 64) {
            if (!ceq(qxt, qxt2) || !ceq(qyt, qyt2)) {
                mode_match(qxt, qxt2, &kx, &kmx);
                mode_match(qyt, qyt2, &ky, &kmy);
                message(" t21: Kx=%s \n      Ky=%s \n      Kmx %9.3g Kmy %9.3g\n",
                        complex_form(kx), complex_form(ky), kmx, kmy);
            }
        }

        if (inter.debug & 32) {
            message("k21");
        }

        if (inter.set_tem_phase_zero & 1) {
            phase = zphase(k_mnmn(0, 0, 0, 0, qxt, qyt, qxt2, qyt2, 0.0, 0.0, nr1));
        } else {
            phase = 0.0;
        }

        for (n = 0; n < inter.num_fields; n++) {
            for (m = 0; m < inter.num_fields; m++) {
                get_tem_modes_from_field_index(&n1, &m1, n);
                get_tem_modes_from_field_index(&n2, &m2, m);
                if (n1 == NOT_FOUND || m1 == NOT_FOUND ||
                        n2 == NOT_FOUND || m2 == NOT_FOUND) {
                    bug_error("notfound 2");
                }
                modulator->k21[n][m] =
                        z_by_phr(k_mnmn(n1, m1, n2, m2, qxt, qyt, qxt2, qyt2,
                        0.0, 0.0, nr1), -phase);
                if (inter.debug & 32 && !ceq(complex_0, modulator->k21[n][m])) {
                    message(" [%d, %d]=%s, ", n, m, complex_form(modulator->k21[n][m]));
                }
            }

            if (inter.debug & 32) {
                message("\n   ");
            }
        }

        if (inter.debug & 32) {
            message("\n");
        }
    }

    return (mismatch);
}

//! Set coupling coefficients for the given diode

/*!
 * \param diode_index diode index
 *
 * \see Test_set_k_diode()
 */
int set_k_diode(int diode_index) {
    int component;
    int node1_index, node2_index;
    int n1, m1, n2, m2;
    int mismatch;
    double nr1, nr2;
    complex_t kx, ky;
    double kmx, kmy;
    double phase = 0.0;

    complex_t qx1, qy1, qx2, qy2;
    complex_t qxt, qyt;
    complex_t qxt2, qyt2;
    diode_t *diode;
    node_t node1, node2;

    // sanity checks on input
    // diode index should be in the range: 0 <= diode_index < inter.num_diodes
    assert(diode_index >= 0);
    assert(diode_index < inter.num_diodes);

    diode = &(inter.diode_list[diode_index]);
    component = get_overall_component_index(DIODE, diode_index);

    node1_index = diode->node1_index;
    node2_index = diode->node2_index;

    nr1 = nr2 = init.n0;

    node1 = inter.node_list[node1_index];
    node2 = inter.node_list[node2_index];

    if (NOT node1.gnd_node) {
        //node1 = inter.node_list[node1_index];
        nr1 = *node1.n;
        if (node1.component_index == component) {
            // reverse q so that beam comes _to_ the diode from node1
            qx1 = cminus(cconj(node1.qx));
            qy1 = cminus(cconj(node1.qy));
        } else {
            qx1 = node1.qx;
            qy1 = node1.qy;
        }
    }

    if (NOT node2.gnd_node) {
        //node2 = inter.node_list[node2_index];
        nr2 = *node2.n;
        if (node2.component_index == component) {
            qx2 = node2.qx;
            qy2 = node2.qy;
        } else {
            // reverse q so that beam goes _from_ the diode to node2
            qx2 = cminus(cconj(node2.qx));
            qy2 = cminus(cconj(node2.qy));
        }
    }

    if (nr1 != nr2) {
        gerror("index of refraction not consistent at diode '%s'\n", diode->name);
    }

    if (inter.debug & 32 || inter.trace & 64) {
        message("%s :\n", diode->name);
    }

    mismatch = 0;

    if (NOT node1.gnd_node &&
            NOT node2.gnd_node) {
        qxt = qx1;
        qyt = qy1;

        qxt2 = qx2;
        qyt2 = qy2;
        if (!ceq(qxt, qxt2) || !ceq(qyt, qyt2)) {
            mismatch = mismatch | 4;
        }

        if (inter.trace & 64) {
            if (!ceq(qxt, qxt2) || !ceq(qyt, qyt2)) {
                mode_match(qxt, qxt2, &kx, &kmx);
                mode_match(qyt, qyt2, &ky, &kmy);
                message(" t12: Kx=%s \n      Ky=%s \n      Kmx %9.3g Kmy %9.3g\n",
                        complex_form(kx), complex_form(ky), kmx, kmy);
            }
        }

        if (inter.debug & 32) {
            message("k12");
        }

        if (inter.set_tem_phase_zero & 1) {
            phase = zphase(k_mnmn(0, 0, 0, 0, qxt, qyt, qxt2, qyt2, 0.0, 0.0, nr2));
        } else {
            phase = 0.0;
        }

        int n, m;
        for (n = 0; n < inter.num_fields; n++) {
            for (m = 0; m < inter.num_fields; m++) {
                get_tem_modes_from_field_index(&n1, &m1, n);
                get_tem_modes_from_field_index(&n2, &m2, m);
                if (n1 == NOT_FOUND || m1 == NOT_FOUND ||
                        n2 == NOT_FOUND || m2 == NOT_FOUND) {
                    bug_error("notfound 1");
                }
                diode->k12[n][m] = z_by_phr(k_mnmn(n1, m1, n2, m2, qxt, qyt, qxt2, qyt2,
                        0.0, 0.0, nr2), -phase);
                if (inter.debug & 32 && !ceq(complex_0, diode->k12[n][m])) {
                    message(" [%d, %d]=%s, ", n, m, complex_form(diode->k12[n][m]));
                }
            }

            if (inter.debug & 32) {
                message("\n   ");
            }
        }

        if (inter.debug & 32) {
            message("\n");
        }

        // transmission2
        qxt = cminus(cconj(qx2));
        qyt = cminus(cconj(qy2));
        qxt2 = cminus(cconj(qx1));
        qyt2 = cminus(cconj(qy1));
        if (!ceq(qxt, qxt2) || !ceq(qyt, qyt2)) {
            mismatch = mismatch | 4;
        }

        if (inter.trace & 64) {
            if (!ceq(qxt, qxt2) || !ceq(qyt, qyt2)) {
                mode_match(qxt, qxt2, &kx, &kmx);
                mode_match(qyt, qyt2, &ky, &kmy);
                message(" t21: Kx=%s \n      Ky=%s \n      Kmx %9.3g Kmy %9.3g\n",
                        complex_form(kx), complex_form(ky), kmx, kmy);
            }
        }

        if (inter.debug & 32) {
            message("k21");
        }

        if (inter.set_tem_phase_zero & 1) {
            phase = zphase(k_mnmn(0, 0, 0, 0, qxt, qyt, qxt2, qyt2, 0.0, 0.0, nr1));
        } else {
            phase = 0.0;
        }

        for (n = 0; n < inter.num_fields; n++) {
            for (m = 0; m < inter.num_fields; m++) {
                get_tem_modes_from_field_index(&n1, &m1, n);
                get_tem_modes_from_field_index(&n2, &m2, m);
                if (n1 == NOT_FOUND || m1 == NOT_FOUND ||
                        n2 == NOT_FOUND || m2 == NOT_FOUND) {
                    bug_error("notfound 2");
                }
                diode->k21[n][m] = z_by_phr(k_mnmn(n1, m1, n2, m2, qxt, qyt, qxt2, qyt2,
                        0.0, 0.0, nr1), -phase);
                if (inter.debug & 32 && !ceq(complex_0, diode->k21[n][m])) {
                    message(" [%d, %d]=%s, ", n, m, complex_form(diode->k21[n][m]));
                }
            }

            if (inter.debug & 32) {
                message("\n   ");
            }
        }

        if (inter.debug & 32) {
            message("\n");
        }
    }

    return (mismatch);
}

//! Compute the mode match between two beam segments 

/*!
 * i.e. between two chosen base systems
 *
 * trace 4: mode mismatch parameters for the initial setup
 *  - Kx and Ky are used internally, but the output of 'trace 4'
 *    computes the mode mismatch with min(z_r, z_r') in the denominator 
 *    (see manual).  This gives a better quantitative number for the 
 *    mode mismatch.
 *  - kmx, kmy are now computed as K**2/(1+K**2) which seems to be more 
 *    useful than the original formula from Bayer-Helms.
 *  - please see the section about these parameters in the manual!
 *
 * \param q1 Gaussian beam parameter
 * \param q2 Gaussian beam parameter
 * \param k array of coupling coefficients
 * \param km array of coupling coefficients
 *
 * \see Test_mode_match()
 */
void mode_match(complex_t q1, complex_t q2, complex_t *k, double *km) {
    double zr;
    complex_t kt;
    double kmt;

    zr = 1.0 / min(q1.im, q2.im);
    kt.re = 0.5 * (q1.im - q2.im) * zr;
    kt.im = 0.5 * (q1.re - q2.re) * zr;
    kmt = zabs(kt) / (1.0 + zabs(kt)); // 22.07.03
    *k = kt;
    *km = kmt;

    /*
     * printf("min: k %s, km %g, ", complex_form(kt), kmt);
     * zr=1.0/q2.im;
     * kt.re=0.5*(q1.im-q2.im)*zr;
     * kt.im=0.5*(q1.re-q2.re)*zr;
     * //kmt=zabs(div_complex(kt, z_pl_z(complex_1, kt)));
     * kmt=zabs(kt)/(1.0+zabs(kt)); // 22.07.03
     * printf("normal  k %s, km %g\n", complex_form(kt), kmt);
     */
}

//! Set k_mnmn for all components

void set_k_all_components(bool rebuilding) {
    int initonly, noinit;
    int mm = 0;
    int mbs = 0; 
    int ms = 0; 
    int ml = 0;

#if DEBUG
    FILE *fp_debug;
    fp_debug = fopen("dump_ks.dat", "w");
#endif

    initonly = noinit = 0;

    if (inter.trace & 4) {  
        message("--- initial mode mismatch:\n");
        if (!(inter.trace & 64)) {
            inter.trace |= 64;
            initonly = 1;
        }
    } else {
        if (inter.trace & 64) {
            noinit = 1;
            inter.trace &= 447;
        }
    }

    int i;
    for (i = 0; i < inter.num_mirrors; i++) {
        mirror_t *mirror = &(inter.mirror_list[i]);
        
        if (!rebuilding || mirror->rebuild)
            mm = set_k_mirror(i);
        else{
            // no known rebuilding, still need to check if the q values have
            // changed
            node_t n1 = inter.node_list[mirror->node1_index];
            node_t n2 = inter.node_list[mirror->node2_index];

            //printf("mirror %s node %s changed %d\n",mirror->name, n1.name, n1.q_changed);
            //printf("mirror %s node %s changed %d\n",mirror->name, n2.name, n2.q_changed);

            if(((!n1.gnd_node) && (n1.q_changed))
                    || ((!n2.gnd_node) && (n2.q_changed)))
                mm = set_k_mirror(i);
        }
            
    }

    for (i = 0; i < inter.num_beamsplitters; i++) {
        beamsplitter_t *bs = &(inter.bs_list[i]);
        if (!rebuilding || bs->rebuild)
            mbs = set_k_beamsplitter(i);
        else{
            // no known rebuilding, still need to check if the q values have
            // changed
            node_t n1 = inter.node_list[bs->node1_index];
            node_t n2 = inter.node_list[bs->node2_index];
            node_t n3 = inter.node_list[bs->node3_index];
            node_t n4 = inter.node_list[bs->node4_index];
            
            if(((!n1.gnd_node) && (n1.q_changed))
                    || ((!n2.gnd_node) && (n2.q_changed))
                    || ((!n3.gnd_node) && (n3.q_changed))
                    || ((!n4.gnd_node) && (n4.q_changed)))
                mm = set_k_beamsplitter(i);
        }
    }

    for (i = 0; i < inter.num_spaces; i++) {
        space_t *s = &inter.space_list[i];
        if (!rebuilding || s->rebuild) 
            ms = set_k_space(i);
        else{
            // no known rebuilding, still need to check if the q values have
            // changed
            node_t n1 = inter.node_list[s->node1_index];
            node_t n2 = inter.node_list[s->node2_index];
            
            if(((!n1.gnd_node) && (n1.q_changed))
                    || ((!n2.gnd_node) && (n2.q_changed)))
                mm = set_k_space(i);
        }
        
        if (inter.trace & 4 && ms) {
            message("Warning: mode mismatch at space %s (%d)\n",
                    inter.space_list[i].name, ms);
        }
    }
    
    for (i = 0; i < inter.num_lenses; i++) {
        lens_t *l = &inter.lens_list[i];
        if (!rebuilding || l->rebuild)
            ml = set_k_lens(i);
        else{
            // no known rebuilding, still need to check if the q values have
            // changed
            node_t n1 = inter.node_list[l->node1_index];
            node_t n2 = inter.node_list[l->node2_index];
            
            if(((!n1.gnd_node) && (n1.q_changed))
                    || ((!n2.gnd_node) && (n2.q_changed)))
                mm = set_k_lens(i);
        }
    }

    for (i = 0; i < inter.num_modulators; i++) {
        modulator_t *mod = &inter.modulator_list[i];
        
        if (!rebuilding || mod->rebuild)
            ml = set_k_modulator(i);
        else{
            // no known rebuilding, still need to check if the q values have
            // changed
            node_t n1 = inter.node_list[mod->node1_index];
            node_t n2 = inter.node_list[mod->node2_index];
            
            if(((!n1.gnd_node) && (n1.q_changed))
                    || ((!n2.gnd_node) && (n2.q_changed)))
                mm = set_k_modulator(i);
        }
    }

    for (i = 0; i < inter.num_diodes; i++) {
        diode_t *d = &inter.diode_list[i];
        
        if (!rebuilding || d->rebuild)
            ml = set_k_diode(i);
        else{
            // no known rebuilding, still need to check if the q values have
            // changed
            node_t n1 = inter.node_list[d->node1_index];
            node_t n2 = inter.node_list[d->node2_index];
            
            if(((!n1.gnd_node) && (n1.q_changed))
                    || ((!n2.gnd_node) && (n2.q_changed)))
                mm = set_k_diode(i);
        }
    }

    if (initonly) {
        inter.trace &= 447;
    }

    if (noinit) {
        inter.trace |= 64;
    }

    if (inter.trace & 4) {
        message("\n");
    }

#if DEBUG
    dump_ks(fp_debug);
    fclose(fp_debug);
#endif

}

//! Set up an ABCD mirror

/*!
 * \param mirror_index index of mirror in mirror list
 *
 * \see Test_set_ABCD_mirror()
 */
void set_ABCD_mirror(int mirror_index) {
    double n, n1, n2;
    mirror_t *mirror;
    node_t node1, node2;

    // sanity checks on input
    // mirror index should be in the range 0 <= mirror_index < inter.num_mirrors
    assert(mirror_index >= 0);
    assert(mirror_index < inter.num_mirrors);


    mirror = &(inter.mirror_list[mirror_index]);

    node1 = inter.node_list[mirror->node1_index];
    node2 = inter.node_list[mirror->node2_index];

    n1 = n2 = 0; // silence compiler warning

    // setting reflection matrix
    if (NOT node1.gnd_node) {
        n1 = *(node1.n);
        if (mirror->Rcx == 0) {
            mirror->qqr1t = Unity;
        } else {
            // n1 added 040803
            mirror->qqr1t = make_ABCD(1.0, 0.0, -2.0 * n1 / mirror->Rcx, 1.0);
        }

        if (mirror->Rcy == 0) {
            mirror->qqr1s = Unity;
        } else {
            // n1 added 040803
            mirror->qqr1s = make_ABCD(1.0, 0.0, -2.0 * n1 / mirror->Rcy, 1.0);
        }
    }

    if (NOT node2.gnd_node) {
        n2 = *(node2.n);
        if (mirror->Rcx == 0) {
            mirror->qqr2t = Unity;
        } else {
            mirror->qqr2t = make_ABCD(1.0, 0.0, 2.0 * n2 / mirror->Rcx, 1.0);
        }

        if (mirror->Rcy == 0) {
            mirror->qqr2s = Unity;
        } else {
            mirror->qqr2s = make_ABCD(1.0, 0.0, 2.0 * n2 / mirror->Rcy, 1.0);
        }
    }

    if (NOT node1.gnd_node &&
            NOT node2.gnd_node) {
        // setting transmission matrix
        n = n2 - n1;
        if (n == 0) {
            mirror->qqtt = Unity;
            mirror->qqts = Unity;
        } else {
            if (mirror->Rcx == 0) {
                mirror->qqtt = Unity;
            } else {
                mirror->qqtt = make_ABCD(1.0, 0.0, n / mirror->Rcx, 1.0);
            }

            if (mirror->Rcy == 0) {
                mirror->qqts = Unity;
            } else {
                mirror->qqts = make_ABCD(1.0, 0.0, n / mirror->Rcy, 1.0);
            }
        }
    }

    if (inter.debug & 256) {
        fprintf(stdout, "mirror : %s ----------------\n", mirror->name);
        fprintf(stdout, "  qqtt\n");
        dump_ABCD(stdout, mirror->qqtt);
        fprintf(stdout, "  qqtt\n");
        dump_ABCD(stdout, mirror->qqtt);
        fprintf(stdout, "  qqtr1t\n");
        dump_ABCD(stdout, mirror->qqr1t);
        fprintf(stdout, "  qqr1s\n");
        dump_ABCD(stdout, mirror->qqr1s);
        fprintf(stdout, "  qqr2t\n");
        dump_ABCD(stdout, mirror->qqr2t);
        fprintf(stdout, "  qqr2s\n");
        dump_ABCD(stdout, mirror->qqr2s);

        // and send the output to the log file
        fprintf(fp_log, "mirror : %s ----------------\n", mirror->name);
        fprintf(fp_log, "  qqtt\n");
        dump_ABCD(fp_log, mirror->qqtt);
        fprintf(fp_log, "  qqtt\n");
        dump_ABCD(fp_log, mirror->qqtt);
        fprintf(fp_log, "  qqtr1t\n");
        dump_ABCD(fp_log, mirror->qqr1t);
        fprintf(fp_log, "  qqr1s\n");
        dump_ABCD(fp_log, mirror->qqr1s);
        fprintf(fp_log, "  qqr2t\n");
        dump_ABCD(fp_log, mirror->qqr2t);
        fprintf(fp_log, "  qqr2s\n");
        dump_ABCD(fp_log, mirror->qqr2s);

    }

    // dump_ABCD(stdout, mirror->qqtt);
    // dump_ABCD(stdout, mirror->qqts);
}

//! Set up an ABCD matrix for a beam splitter

/*!
 * \param bs_index index of beam splitter in bs_list
 *
 * \see Test_set_ABCD_beamsplitter()
 */
void set_ABCD_beamsplitter(int bs_index) {
    int j;
    int node1_index, node2_index, node3_index, node4_index;
    node_t node1, node2, node3, node4;
    double n1, n2;
    double co1, co2, si1, si2, coco;
    double alpha2;
    double ir[2] = {0};

    beamsplitter_t *bs;

    // sanity checks on input
    // beam splitter index should be in the range: 
    // 0 <= bs_index < inter.num_beamsplitters
    assert(bs_index >= 0);
    assert(bs_index < inter.num_beamsplitters);

    bs = &(inter.bs_list[bs_index]);

    node1_index = bs->node1_index;
    node2_index = bs->node2_index;
    node3_index = bs->node3_index;
    node4_index = bs->node4_index;

    node1 = inter.node_list[node1_index];
    node2 = inter.node_list[node2_index];
    node3 = inter.node_list[node3_index];
    node4 = inter.node_list[node4_index];

    n1 = n2 = init.n0;

    j = 0;
    if (NOT node1.gnd_node) {
        //node1 = inter.node_list[node1_index];
        ir[j++] = *node1.n;
    }

    if (NOT node2.gnd_node) {
        //node2 = inter.node_list[node2_index];
        ir[j++] = *node2.n;
    }

    int k;
    for (k = 0; k < j; k++) {
        if (ir[k] != ir[0]) {
            gerror("index of refraction not consistent at beamsplitter '%s'\n",
                    bs->name);
        }
    }

    if (j) {
        n1 = ir[0];
    }

    j = 0;
    if (NOT node3.gnd_node) {
        //node3 = inter.node_list[node3_index];
        ir[j++] = *node3.n;
    }

    if (NOT node4.gnd_node) {
        //node4 = inter.node_list[node4_index];
        ir[j++] = *node4.n;
    }

    for (k = 0; k < j; k++) {
        if (ir[k] != ir[0]) {
            gerror("index of refraction not consistent at beamsplitter '%s'\n",
                    bs->name);
        }
    }

    if (j) {
        n2 = ir[0];
    }

    bs->ir1 = n1;
    bs->ir2 = n2;

    si1 = sin(bs->alpha * RAD);
    si2 = n1 / n2*si1;
    if (fabs(si2) > 1) { // fabs 290402
        gerror("total reflection at beamsplitter %s\n", bs->name);
    }
    alpha2 = DEG * asin(si2);

    co1 = cos(bs->alpha * RAD);
    co2 = cos(alpha2 * RAD);

    if (inter.debug & 256) {
        fprintf(stdout, "Setting ABCD for bs %s\n", bs->name);
        fprintf(stdout, "%s: alpha %g , alpha2 %g\n", bs->name, bs->alpha, alpha2);
        fprintf(stdout, "n1 %g n2 %g c1 %g c2 %g si %g s2 %g\n", n1, n2, co1, co2, si1, si2);
    }

    if (co1 == 0.0 || co2 == 0.0) {
        gerror("beamsplitter '%s': alpha(%s) out of range\n", bs->name,
                double_form(bs->alpha));
    }

    coco = n2 * co2 - n1*co1;

    // setting transmission matrix
    // x = tangential plane, y = sagittal plane

    if (n1 == n2) {
        bs->qqt1t = bs->qqt2t = Unity;
        bs->qqt1s = bs->qqt2s = Unity;
    } else {
        if (bs->Rcx == 0.0) {
            bs->qqt1t = make_ABCD(co2 / co1, 0.0, 0.0, co1 / co2);
            bs->qqt2t = make_ABCD(co1 / co2, 0.0, 0.0, co2 / co1);
        } else {
            bs->qqt1t =
                    make_ABCD(co2 / co1, 0.0, coco / (bs->Rcx * co1 * co2), co1 / co2);
            bs->qqt2t =
                    make_ABCD(co1 / co2, 0.0, coco / (bs->Rcx * co1 * co2), co2 / co1);
        }

        if (bs->Rcy == 0.0) {
            bs->qqt1s = bs->qqt2s = Unity;
        } else {
            bs->qqt1s = bs->qqt2s =
                    make_ABCD(1.0, 0.0, coco / (bs->Rcy * co1 * co2), 1.0);
        }
    }

    // setting reflection matrix
    if (bs->Rcx == 0.0) {
        bs->qqr1t = bs->qqr2t = Unity;
    } else {
        // n1 added 040803
        bs->qqr1t = make_ABCD(1.0, 0.0, -2.0 * n1 / (bs->Rcx * co1), 1.0);
        // n2 added 231003
        bs->qqr2t = make_ABCD(1.0, 0.0, 2.0 * n2 / (bs->Rcx * co2), 1.0);
    }

    if (bs->Rcy == 0.0) {
        bs->qqr1s = bs->qqr2s = Unity;
    } else {
        // n1 added 040803
        bs->qqr1s = make_ABCD(1.0, 0.0, -2.0 * n1 * co1 / bs->Rcy, 1.0);
        // n2 added 231003
        bs->qqr2s = make_ABCD(1.0, 0.0, 2.0 * n2 * co2 / bs->Rcy, 1.0);
    }

    if (inter.debug & 256) {
        fprintf(stdout, "bs : %s ----------------\n", bs->name);
        fprintf(stdout, "  qqt1t\n");
        dump_ABCD(stdout, bs->qqt1t);
        fprintf(stdout, "  qqt1s\n");
        dump_ABCD(stdout, bs->qqt1s);
        fprintf(stdout, "  qqt2t\n");
        dump_ABCD(stdout, bs->qqt2t);
        fprintf(stdout, "  qqt2s\n");
        dump_ABCD(stdout, bs->qqt2s);

        fprintf(stdout, "  qqr1t\n");
        dump_ABCD(stdout, bs->qqr1t);
        fprintf(stdout, "  qqr1s\n");
        dump_ABCD(stdout, bs->qqr1s);
        fprintf(stdout, "  qqr2t\n");
        dump_ABCD(stdout, bs->qqr2t);
        fprintf(stdout, "  qqr2s\n");
        dump_ABCD(stdout, bs->qqr2s);

        // and send the output to the log file
        fprintf(fp_log, "bs : %s ----------------\n", bs->name);
        fprintf(fp_log, "  qqt1t\n");
        dump_ABCD(fp_log, bs->qqt1t);
        fprintf(fp_log, "  qqt1s\n");
        dump_ABCD(fp_log, bs->qqt1s);
        fprintf(fp_log, "  qqt2t\n");
        dump_ABCD(fp_log, bs->qqt2t);
        fprintf(fp_log, "  qqt2s\n");
        dump_ABCD(fp_log, bs->qqt2s);

        fprintf(fp_log, "  qqr1t\n");
        dump_ABCD(fp_log, bs->qqr1t);
        fprintf(fp_log, "  qqr1s\n");
        dump_ABCD(fp_log, bs->qqr1s);
        fprintf(fp_log, "  qqr2t\n");
        dump_ABCD(fp_log, bs->qqr2t);
        fprintf(fp_log, "  qqr2s\n");
        dump_ABCD(fp_log, bs->qqr2s);
    }
}

//! Set up an ABCD matrix for a free space element

/*!
 * \param space_index index of space in space list
 *
 * \see Test_set_ABCD_space()
 */
void set_ABCD_space(int space_index) {
    space_t *space;
    node_t node1, node2;

    // sanity checks on input
    // space index should be in the range: 0 <= space_index < inter.num_spaces
    assert(space_index >= 0);
    assert(space_index < inter.num_spaces);

    space = &(inter.space_list[space_index]);

    node1 = inter.node_list[space->node1_index];
    node2 = inter.node_list[space->node2_index];

    if (NOT node1.gnd_node &&
            NOT node2.gnd_node) {
        space->qq = make_ABCD(1.0, space->L / space->n, 0.0, 1.0);
        //    space->qq = make_ABCD(1.0, space->L, 0.0, 1.0);

        if (inter.debug & 256) {
            fprintf(stdout, "space : %s ----------------\n", space->name);
            fprintf(stdout, "  qq\n");
            dump_ABCD(stdout, space->qq);

            // and send the output to the log file
            fprintf(fp_log, "space : %s ----------------\n", space->name);
            fprintf(fp_log, "  qq\n");
            dump_ABCD(fp_log, space->qq);
        }
    }
}

//! Set up an ABCD matrix for a lens 

/*!
 * \param lens_index index of lens in lens list
 *
 * \see Test_set_ABCD_lens()
 */
void set_ABCD_lens(int lens_index) {
    lens_t *lens;
    node_t node1, node2;

    // sanity checks on input
    // lens index should be in range: 0 <= lens_index < inter.num_lenses
    assert(lens_index >= 0);
    assert(lens_index < inter.num_lenses);

    lens = &(inter.lens_list[lens_index]);

    node1 = inter.node_list[lens->node1_index];
    node2 = inter.node_list[lens->node2_index];


    // if focal length=0, ignore lens
    double f = lens->f;
    if (f == 0.0) {
        warn("focal length is zero, lens '%s' ignored\n", lens->name);
        f = -1.0e23;
    }

    if (NOT node1.gnd_node &&
            NOT node2.gnd_node) {
        lens->qq = make_ABCD(1.0, 0.0, -1.0 / f, 1.0);
    }
}

//! Set up an ABCD matrix for a grating

/*!
 * \param grating_index index of grating in grating list
 *
 * \see Test_set_ABCD_grating()
 */
void set_ABCD_grating(int grating_index) {
    double ca;
    double reRs, reRt; // refers to 1/Rs and 1/Rt
    double M;
    grating_t *grating;

    // sanity checks on input
    // grating index should be in range: 0 <= grating_index < inter.num_gratings
    assert(grating_index >= 0);
    assert(grating_index < inter.num_gratings);

    grating = &(inter.grating_list[grating_index]);

    ca = cos(RAD * grating->alpha);

    // zeroth order is equal for all grating types:
    if (grating->Rcx == 0.0) {
        reRt = 0.0;
    } else {
        reRt = 1.0 / (grating->Rcx * ca);
    }
    if (grating->Rcy == 0.0) {
        reRs = 0.0;
    } else {
        reRs = ca / grating->Rcy;
    }
    M = 1;
    grating->qqt[0][1] = make_ABCD(M, 0.0, -2.0 * reRt, 1.0 / M);
    grating->qqt[1][0] = make_ABCD(M, 0.0, -2.0 * reRt, 1.0 / M);
    grating->qqs[0][1] = make_ABCD(1.0, 0.0, -2.0 * reRs, 1.0);
    grating->qqs[1][0] = make_ABCD(1.0, 0.0, -2.0 * reRs, 1.0);

    grating->qqt[0][0] = make_ABCD(M, 0.0, -2.0 * reRt, 1.0 / M);
    grating->qqt[1][1] = make_ABCD(M, 0.0, -2.0 * reRt, 1.0 / M);
    grating->qqs[0][0] = make_ABCD(1.0, 0.0, -2.0 * reRs, 1.0);
    grating->qqs[1][1] = make_ABCD(1.0, 0.0, -2.0 * reRs, 1.0);


    switch (grating->num_of_ports) {
        case 2:
            break;
        case 3:
            if (grating->Rcx == 0.0) {
                reRt = 0.0;
            } else {
                reRt = (1.0 + ca) / (2.0 * grating->Rcx * ca);
            }
            if (grating->Rcy == 0.0) {
                reRs = 0.0;
            } else {
                reRs = (1.0 + ca) / (2.0 * grating->Rcy);
            }
            M = 1 / ca;
            grating->qqt[0][2] = make_ABCD(M, 0.0, -2.0 * reRt, 1.0 / M);
            grating->qqs[0][2] = make_ABCD(1.0, 0.0, -2.0 * reRs, 1.0);
            M = ca;
            grating->qqt[2][0] = make_ABCD(M, 0.0, -2.0 * reRt, 1.0 / M);
            grating->qqs[2][0] = make_ABCD(1.0, 0.0, -2.0 * reRs, 1.0);

            M = 1 / ca;
            grating->qqt[1][2] = make_ABCD(M, 0.0, -2.0 * reRt, 1.0 / M);
            grating->qqs[1][2] = make_ABCD(1.0, 0.0, -2.0 * reRs, 1.0);
            M = ca;
            grating->qqt[2][1] = make_ABCD(M, 0.0, -2.0 * reRt, 1.0 / M);
            grating->qqs[2][1] = make_ABCD(1.0, 0.0, -2.0 * reRs, 1.0);

            if (grating->Rcx == 0.0) {
                reRt = 0.0;
            } else {
                reRt = 1.0 / grating->Rcx;
            }
            if (grating->Rcy == 0.0) {
                reRs = 0.0;
            } else {
                reRs = 1.0 / grating->Rcy;
            }
            M = 1.0;
            grating->qqt[2][2] = make_ABCD(M, 0.0, -2.0 * reRt, 1.0 / M);
            grating->qqs[2][2] = make_ABCD(1.0, 0.0, -2.0 * reRs, 1.0);


            break;
        case 4:

            break;
        default:
            bug_error("incorrect number of ports");
    }
}

//! Set ABCD matrices for all components

/*!
 *
 * \todo refactor for readability
 *
 * \see Test_set_ABCDs()
 */
void set_ABCDs(void) {
    // mirrors
    int mirror_index;
    for (mirror_index = 0; mirror_index < inter.num_mirrors; mirror_index++) {
        set_ABCD_mirror(mirror_index);
    }

    // beamsplitters
    int bs_index;
    for (bs_index = 0; bs_index < inter.num_beamsplitters; bs_index++) {
        set_ABCD_beamsplitter(bs_index);
    }

    //gratings
    int grating_index;
    for (grating_index = 0; grating_index < inter.num_gratings; grating_index++) {
        set_ABCD_grating(grating_index);
    }

    // modulators
    int modulator_index;
    for (modulator_index = 0;
            modulator_index < inter.num_modulators;
            modulator_index++) {
        modulator_t *modulator;
        modulator = &(inter.modulator_list[modulator_index]);
        modulator->qq = Unity;
    }

    // spaces
    int space_index;
    for (space_index = 0; space_index < inter.num_spaces; space_index++) {
        set_ABCD_space(space_index);
    }

    // lenses
    int lens_index;
    for (lens_index = 0; lens_index < inter.num_lenses; lens_index++) {
        set_ABCD_lens(lens_index);
    }

    // diodes
    int diode_index;
    for (diode_index = 0; diode_index < inter.num_diodes; diode_index++) {
        diode_t *diode;
        diode = &(inter.diode_list[diode_index]);
        diode->qq = Unity;
    }
}

//! Return number of equations for given order of used TEM modes

/*!
 * \param modes the number of TEM modes
 *
 * \see Test_num_eqns_for()
 */
int num_eqns_for(int modes) {
    // modes should be >= 0
    assert(modes >= 0);

    //sum: 1+2+3+4+...+modes+1
    return ((int) ((modes + 1) * (modes + 2) / 2.0));
}

//! Compute the cavity parameters

/*!
 *
 * \todo refactor for readability
 * \todo break up into smaller routines!!!
 *
 * \see Test_get_cavity()
 */
void get_cavity(void) {
    ABCD_t eigen_x, eigen1_x, eigen2_x, temp_x;
    ABCD_t eigen_y, eigen1_y, eigen2_y, temp_y;

    double nr1, nr2;

    complex_t q_x, q_y, q_xt, q_yt;

    if (inter.trace & 2 || inter.trace & 128) {
        message("--- cavity tracing\n");
    }

    int cavity_index;
    for (cavity_index = 0; cavity_index < inter.num_cavities; cavity_index++) {
        cavity_t *cavity;
        cavity = &(inter.cavity_list[cavity_index]);
        cavity->stable = 0;
        int c1_index = 0;
        int c2_index = 0;
        // initialise matrices
        eigen_x = eigen1_x = eigen2_x = temp_x = Unity;
        eigen_y = eigen1_y = eigen2_y = temp_y = Unity;

        int complete;
        int complete2;
        if (get_component_type(cavity->component1_index) == BEAMSPLITTER) {
            complete = trace_cavity(&eigen1_x, &eigen1_y,
                    cavity->component1_index, cavity->node1_index,
                    cavity->component2_index, cavity->node2_index,
                    &c1_index, vlocal.trace_c1);
            int err_x = component_matrix(&temp_x,
                    cavity->component1_index, cavity->node2_index,
                    cavity->node1_index, TANGENTIAL);
            int err_y = component_matrix(&temp_y,
                    cavity->component1_index, cavity->node2_index,
                    cavity->node1_index, SAGITTAL);
            if (!err_x && !err_y) {
                eigen_x = multiply_ABCD(eigen1_x, inv_ABCD(temp_x));
                eigen_y = multiply_ABCD(eigen1_y, inv_ABCD(temp_y));
            } else {
                bug_error("errxy0");
            }
            complete = complete & (!err_x) & (!err_y);
        } else {
            complete = trace_cavity(&eigen1_x, &eigen1_y,
                    cavity->component1_index, cavity->node1_index,
                    cavity->component2_index, cavity->node2_index,
                    &c1_index, vlocal.trace_c1);
            if (inter.debug & 256) {
                fprintf(stdout, "eigen1 (x, y):\n");
                dump_ABCD(stdout, eigen1_x);
                dump_ABCD(stdout, eigen1_y);
            }
            complete2 = trace_cavity(&eigen2_x, &eigen2_y,
                    cavity->component2_index, cavity->node2_index,
                    cavity->component1_index, cavity->node1_index,
                    &c2_index, vlocal.trace_c2);

            if (inter.debug & 256) {
                fprintf(stdout, "eigen2 (x, y):\n");
                dump_ABCD(stdout, eigen2_x);
                dump_ABCD(stdout, eigen2_y);
            }

            int err_x = component_matrix(&temp_x,
                    cavity->component2_index, cavity->node2_index,
                    cavity->node2_index, TANGENTIAL);
            int err_y = component_matrix(&temp_y,
                    cavity->component2_index, cavity->node2_index,
                    cavity->node2_index, SAGITTAL);

            if (inter.debug & 256) {
                fprintf(stdout, "temp2 (x, y):\n");
                dump_ABCD(stdout, temp_x);
                dump_ABCD(stdout, temp_y);
            }

			if (!err_x && !err_y) {
				eigen_x =
						multiply_ABCD(multiply_ABCD(eigen1_x, inv_ABCD(temp_x)), eigen2_x);
				eigen_y =
						multiply_ABCD(multiply_ABCD(eigen1_y, inv_ABCD(temp_y)), eigen2_y);
			} else {
				bug_error("errxy1");
			}

			complete = complete & complete2 & (!err_x) & (!err_y);
			err_x = component_matrix(&temp_x,
					cavity->component1_index, cavity->node1_index,
					cavity->node1_index, TANGENTIAL);
			err_y = component_matrix(&temp_y,
					cavity->component1_index, cavity->node1_index,
					cavity->node1_index, SAGITTAL);

			if (inter.debug & 256) {
				fprintf(stdout, "temp1 (x, y):\n");
				dump_ABCD(stdout, temp_x);
				dump_ABCD(stdout, temp_y);
			}

			if (!err_x && !err_y) {
				eigen_x = multiply_ABCD(eigen_x, inv_ABCD(temp_x));
				eigen_y = multiply_ABCD(eigen_y, inv_ABCD(temp_y));
			} else {
				bug_error("errxy2");
			}
			complete = complete & (!err_x) & (!err_y);
		}

		int m = 0;
		if (complete) {
			// using inversed matrix to reconstruct right order 
			// AB^(-1)=B^(-1) A^(-1)
			eigen_x = inv_ABCD(eigen_x);
			eigen_y = inv_ABCD(eigen_y);
			int m_x = check_ABCD(eigen_x);
			int m_y = check_ABCD(eigen_y);

			if (inter.debug & 128) {
				printf("%s eigenmatrix (x):\n ", cavity->name);
				dump_ABCD(stdout, eigen_x);
				printf("eigenvalue mx=%.15g\n ", (eigen_x.A + eigen_x.D) / 2.0);
				printf(" (y):\n ");
				dump_ABCD(stdout, eigen_y);
				printf("eigenvalue my=%.15g\n ", (eigen_y.A + eigen_y.D) / 2.0);

				// dump the same info to the log file
				fprintf(fp_log, "%s eigenmatrix (x):\n ", cavity->name);
				dump_ABCD(fp_log, eigen_x);
				fprintf(fp_log, "eigenvalue mx=%.15g\n ", (eigen_x.A + eigen_x.D) / 2.0);
				fprintf(fp_log, " (y):\n ");
				dump_ABCD(fp_log, eigen_y);
				fprintf(fp_log, "eigenvalue my=%.15g\n ", (eigen_y.A + eigen_y.D) / 2.0);
			}

			m = m_x | m_y;

			if (inter.trace & 2 || inter.trace & 128) {
				message("cavity %s:\n", cavity->name);
			}

			// Adding final node to list
			if (get_component_type(cavity->component1_index) == BEAMSPLITTER) {
				vlocal.trace_c1[c1_index++] = cavity->node2_index;
			}

			switch (m) {
				case 0: // cavity is stable!
					cavity->stable = 1;
					nr1 = *inter.node_list[cavity->node1_index].n;
					q_x = q_cavity(eigen_x, nr1);
					q_y = q_cavity(eigen_y, nr1);

					if (inter.trace & 2) {
						message(" cavity is stable! Eigenvalues:\n");
						if (ceq(q_x, q_y)) {
							message(" q=%s, w0=%sm z=%sm\n", complex_form(q_x),
									xdouble_form(w0_size(q_x, nr1)), xdouble_form(z_q(q_x)));
						} else {
							message(" qx=%s, w0x=%sm zx=%sm\n", complex_form(q_x),
									xdouble_form(w0_size(q_x, nr1)), xdouble_form(z_q(q_x)));
							message(" qy=%s, w0y=%sm zy=%sm\n", complex_form(q_y),
									xdouble_form(w0_size(q_y, nr1)), xdouble_form(z_q(q_y)));
						}
					}

					if (inter.trace & 128) {
						message(" found the cavity nodes: ");
						int j;
						for (j = 0; j < c1_index; j++) {
							message("%s(%d) ", get_node_name(vlocal.trace_c1[j]),
									vlocal.trace_c1[j]);
						}

						if (get_component_type(cavity->component1_index) == MIRROR) {
							message("\n ... and back: ");
							int j;
							for (j = 0; j < c2_index; j++) {
								message("%s(%d) ", get_node_name(vlocal.trace_c2[j]),
										vlocal.trace_c2[j]);
							}
						}
						message("\n");
					}
					break;
				case 1:
					if (inter.trace & 2) {
						warn(" WARNING!\n"
								" cavity is critical, cannot derive beam parameters\n");
					} else {
						warn("cavity %s is critical, skipping it\n", cavity->name);
					}
					break;
				default: // 2 or 3
					if (inter.trace & 2) {
						warn(" WARNING!\n"
								" cavity is unstable, cannot derive beam parameters\n");
					} else {
						warn("cavity %s is unstable, skipping it\n", cavity->name);
					}
					break;
			}
		} else {
			gerror("could not trace through cavity %s\n", cavity->name);
		}

		// set q values for cavity
		if (get_component_type(cavity->component1_index) == MIRROR) {
			//setting turning node as last node for setting q values
			vlocal.trace_c1[c1_index++] = vlocal.trace_c2[0];
		}

		if (!m) { // set q values if cavity is stable (1)
			set_q(cavity->node1_index, cavity->component1_index, q_x, q_y);
			cavity->qx = q_x;
			cavity->qx = q_y;
			cavity->n = *inter.node_list[cavity->node1_index].n;
		}

		double cavity_power = 1.0;
		double cavity_length = 0.0;
		int type = 0;
		int j;
		for (j = 0; j < c1_index - 1; j++) {
			if (j + 2 > c1_index) { // wrong number of nodes
				gerror("cavity %s: cavity is broken\n", cavity->name);
			}
			int node1_index = vlocal.trace_c1[j];
			int node2_index = vlocal.trace_c1[j + 1];
			if (node1_index == node2_index) {
				gerror("cavity %s: cavity is broken (1)\n", cavity->name);
			}
			int component1_index;
			int component2_index;
			int component3_index;
			int component4_index;
			which_components(node1_index, &component1_index, &component2_index);
			which_components(node2_index, &component3_index, &component4_index);

			//      printf("c1 %s c2 %s c3 %s c4 %s\n",
			//             get_component_name(component1_index), get_component_name(component2_index), get_component_name(component3_index), get_component_name(component4_index)); 
			int component0_index = 0; // silence compiler warning
			if ((component1_index == component3_index ||
					component1_index == component4_index)
					&& component1_index != NOT_FOUND) {
				component0_index = component1_index;
			} else if ((component2_index == component3_index ||
					component2_index == component4_index)
					&& component2_index != NOT_FOUND) {
				component0_index = component2_index;
			} else { // cannot find component
				gerror("cavity %s: cavity is broken (2)\n", cavity->name);
			}

			// getting type and relative comp number for calculating
			// FSR and Finesse of cavity 
			int rcomp = component0_index;
			type = get_component_type_decriment_index(&rcomp);

			switch (type) {
				case MIRROR:
					if ((node1_index == inter.mirror_list[rcomp].node1_index &&
							node2_index == inter.mirror_list[rcomp].node2_index) ||
							(node2_index == inter.mirror_list[rcomp].node1_index &&
							node1_index == inter.mirror_list[rcomp].node2_index)) {
						cavity_power *= inter.mirror_list[rcomp].T;
					} else {
						cavity_power *= inter.mirror_list[rcomp].R;
					}
					break;
				case BEAMSPLITTER:
					if (node1_index == inter.bs_list[rcomp].node1_index &&
							node2_index == inter.bs_list[rcomp].node2_index) {
						cavity_power *= inter.bs_list[rcomp].R;
					} else if (node1_index == inter.bs_list[rcomp].node1_index &&
							node2_index == inter.bs_list[rcomp].node3_index) {
						cavity_power *= inter.bs_list[rcomp].T;
					} else if (node1_index == inter.bs_list[rcomp].node2_index &&
							node2_index == inter.bs_list[rcomp].node1_index) {
						cavity_power *= inter.bs_list[rcomp].R;
					} else if (node1_index == inter.bs_list[rcomp].node2_index &&
							node2_index == inter.bs_list[rcomp].node4_index) {
						cavity_power *= inter.bs_list[rcomp].T;
					} else if (node1_index == inter.bs_list[rcomp].node3_index &&
							node2_index == inter.bs_list[rcomp].node4_index) {
						cavity_power *= inter.bs_list[rcomp].R;
					} else if (node1_index == inter.bs_list[rcomp].node3_index &&
							node2_index == inter.bs_list[rcomp].node1_index) {
						cavity_power *= inter.bs_list[rcomp].T;
					} else if (node1_index == inter.bs_list[rcomp].node4_index &&
							node2_index == inter.bs_list[rcomp].node3_index) {
						cavity_power *= inter.bs_list[rcomp].R;
					} else if (node1_index == inter.bs_list[rcomp].node4_index &&
							node2_index == inter.bs_list[rcomp].node2_index) {
						cavity_power *= inter.bs_list[rcomp].T;
					} else {
						bug_error("cavity FSR: wrong node at BS");
					}
					break;
				case DIODE:
					if (node1_index == inter.diode_list[rcomp].node2_index) {
						// e.g. 20db loss is a factor of 100 in power
						cavity_power *= pow(10., -inter.diode_list[rcomp].S / 10.0);
					}
					break;
				case SPACE:
					cavity_length += inter.space_list[rcomp].L * inter.space_list[rcomp].n;
					if (inter.debug & 256) {
						printf("current cav length %g (added %g*%g from space %s)\n", cavity_length, inter.space_list[rcomp].L, inter.space_list[rcomp].n, inter.space_list[rcomp].name);
					}
					break;
				default:
					break;
			}

			if (inter.debug & 256) {
				message("get_cavity: comp %s \n", get_component_name(component0_index));
				message("cavity loss: %.10g \n", 1.0 - cavity_power);
			}

			if (!m) { // set q values if cavity is stable (2)
				if (!inter.node_list[node1_index].q_is_set) {
					bug_error("q not set");
				}

				// q -> q(-z)=-q* if necessary (reverse beam direction)
				if (inter.node_list[node1_index].component_index == component0_index) {
					if (inter.debug & 256) {
						message("get_cavity: reverse direction\n");
					}
					q_x = cminus(cconj(inter.node_list[node1_index].qx));
					q_y = cminus(cconj(inter.node_list[node1_index].qy));
				} else {
					q_x = inter.node_list[node1_index].qx;
					q_y = inter.node_list[node1_index].qy;
				}

				ABCD_t transform;
				int err = component_matrix(&transform, component0_index,
						node1_index, node2_index, TANGENTIAL);

				if (!err) {
					nr1 = *inter.node_list[node1_index].n;
					nr2 = *inter.node_list[node2_index].n;
					q_xt = q1_q2(transform, q_x, nr1, nr2);
					err = component_matrix(&transform, component0_index,
							node1_index, node2_index, SAGITTAL);
					q_yt = q1_q2(transform, q_y, nr1, nr2);
					set_q(node2_index, component0_index, q_xt, q_yt);
				} else if (inter.debug & 256) {
					message("get_cavity: err -> %s %s\n", get_node_name(node1_index),
							get_node_name(node2_index));
				}
			}
		}

		// final computation of FSR and Finesse of cavity
		int rcomp;

		if (get_component_type(cavity->component1_index) == BEAMSPLITTER) {
			rcomp = get_component_index(cavity->component1_index);
			if (rcomp > inter.num_beamsplitters) {
				bug_error("cavity FSR: wrong bs start comp");
			}
			cavity_power *= inter.bs_list[rcomp].R;
		} else {
			cavity_length *= 2.0;
			cavity_power *= cavity_power;
			rcomp = get_component_index(cavity->component1_index);
			if (rcomp > inter.num_mirrors) {
				bug_error("cavity FSR: wrong m start comp");
			}

			cavity_power *= inter.mirror_list[rcomp].R;
			rcomp = get_component_index(cavity->component2_index);
			if (rcomp > inter.num_mirrors) {
				bug_error("cavity FSR: wrong m stop comp");
			}
			cavity_power *= inter.mirror_list[rcomp].R;
		}

		cavity->finesse = finesse(1.0 - cavity_power);
		cavity->loss = 1.0 - cavity_power;
		cavity->length = cavity_length;
		cavity->FSR = init.clight / cavity_length;
		cavity->FWHM = init.clight / cavity_length / finesse(1.0 - cavity_power);
		cavity->pole = 0.5 * init.clight / cavity_length / finesse(1.0 - cavity_power);

		if (inter.trace & 2) {
			message(" finesse : %g, round-trip power loss: %g\n",
					cavity->finesse, cavity->loss);
			message(" opt. length: %sm, FSR: %sHz\n",
					xdouble_form(cavity->length), xdouble_form(cavity->FSR));
			message(" FWHM: %sHz (pole: %sHz)\n",
					xdouble_form(cavity->FWHM),
					xdouble_form(cavity->pole));

			/*
			message(" finesse : %g, round-trip power loss: %g\n",
					finesse(1.0 - cavity_power), 1.0 - cavity_power);
			message(" opt. length: %sm, FSR: %sHz\n",
					xdouble_form(cavity_length), xdouble_form(init.clight / cavity_length));
			message(" FWHM: %sHz (pole: %sHz)\n",
					xdouble_form(init.clight / cavity_length / finesse(1.0 - cavity_power)),
					xdouble_form(0.5* init.clight / cavity_length / finesse(1.0 - cavity_power)));
			 */
		}
	}
	if (inter.trace & 2) {
		message("\n");
	}
}

//! Trace the cavity (find a path through the nodal net of interferometer)

/*!
 * \param Mlocal_x eigenmatrix of cavity in x-plane
 * \param Mlocal_y eigenmatrix of cavity in y-plane
 * \param component1_index index of first component in trace
 * \param node1_index first node index in trace
 * \param component2_index index of second component in trace
 * \param node2_index second node index in trace
 * \param index index of the current node
 * \param traced array of already found nodes
 *
 * \return an integer indicating if cavity tracing is finished or not
 *
 * \see Test_trace_cavity()
 */
int trace_cavity(ABCD_t *Mlocal_x, ABCD_t *Mlocal_y,
		int component1_index, int node1_index,
		int component2_index, int node2_index,
		int *index, int *traced) {
	int i;
	int index_local;
	int c1, c2, component;
	int node2t, comp2t;
	int node_indices[4] = {0};
	int num_nodes;
	int test;
	int finished;
	int err_x, err_y;

	ABCD_t transform_x, temp_x;
	ABCD_t transform_y, temp_y;

	// sanity checks on inputs
	// component indices should be in the range: 
	// 0 <= component_index < inter.num_components
	assert(component1_index >= 0);
	assert(component2_index >= 0);
	assert(component1_index < inter.num_components);
	assert(component2_index < inter.num_components);
	// node indices should be in the range:
	// 0 <= node_index < inter.num_nodes
	assert(node1_index >= 0);
	assert(node2_index >= 0);
	assert(node1_index < inter.num_nodes);
	assert(node2_index < inter.num_nodes);
	// index should be in the range:
	// 0 <= index < inter.num_nodes
	assert(*index >= 0);
	assert(*index < inter.num_nodes);


	if (inter.debug & 256) {
		message("trace_cavity(): ------- starting new trace\n");
	}

	finished = 0;

	traced[(*index)++] = node1_index;

	comp2t = component2_index;
	node2t = node2_index;

	if (inter.debug & 256) {
		message("trace_cavity(): start comp %s, end comp %s\n",
				get_component_name(component1_index), get_component_name(component2_index));
	}

	//!< \todo c1 and c2 need to be called component1_index and component2_index
	which_components(node1_index, &c1, &c2);

	if (c2 == component1_index) {
		component = c1;
	} else {
		component = c2;
	}

	if (inter.debug & 256) {
		message("trace_cavity(): moving to component %s\n",
				get_component_name(component));
	}

	if (component != NOT_FOUND) {
		which_nodes(component, node1_index, &num_nodes, node_indices);

		i = 0;
		while (i < num_nodes && NOT finished) {
			if (inter.debug & 256) {
				message("trace_cavity(): (%d) node %s\n", i,
						get_node_name(node_indices[i]));
			}

			if (node_indices[i] == node2t) {
				err_x = component_matrix(&transform_x, component,
						node1_index, node_indices[i], TANGENTIAL);
				err_y = component_matrix(&transform_y, component,
						node1_index, node_indices[i], SAGITTAL);

				if (NOT err_x && NOT err_y) {
					// using inversed matrix to reconstruct right order later
					// AB^(-1)=B^(-1) A^(-1)
					temp_x = multiply_ABCD(*Mlocal_x, inv_ABCD(transform_x));
					temp_y = multiply_ABCD(*Mlocal_y, inv_ABCD(transform_y));
					finished = 1;
					*Mlocal_x = temp_x;
					*Mlocal_y = temp_y;

					if (inter.debug & 256) {
						message("trace_cavity(): finished!\n");
					}
				} else {
					bug_error("errxy");
				}
			} else {
				test = 0;
				int j;
				for (j = 0; j < (*index); j++) {
					if (node_indices[i] == traced[j]) {
						test = 1;
					}
				}
				if (NOT test) { // found new node
					if (inter.debug & 256) {
						message("trace_cavity(): new node  %s (index %d)\n",
								get_node_name(node_indices[i]), node_indices[i]);
					}
					err_x = component_matrix(&transform_x, component, node1_index,
							node_indices[i], TANGENTIAL);
					err_y = component_matrix(&transform_y, component, node1_index,
							node_indices[i], SAGITTAL);

					if (NOT err_x && NOT err_y) {
						temp_x = multiply_ABCD(*Mlocal_x, inv_ABCD(transform_x));
						temp_y = multiply_ABCD(*Mlocal_y, inv_ABCD(transform_y));

						if (node_indices[i] == node2t) {
							finished = 1;
						} else {
							index_local = *index;
							finished =
									trace_cavity(&temp_x, &temp_y, component,
									node_indices[i], comp2t, node2t, index, traced);

							if (NOT finished) {
								*index = index_local;
							}
						}

						if (finished) {
							*Mlocal_x = temp_x;
							*Mlocal_y = temp_y;
						}
					} else {
						bug_error("errxy3");
					}
				}
			}
			i++;
		}
	}

	return (finished);
}

//! Set up the trace beam

/*!
 * \param startnode the node at which to start the trace
 *
 * \todo startnode -> start_node_index
 *
 * \see Test_trace_beam()
 */
void trace_beam(int startnode) {
	int j, component1_index, component2_index;

	// sanity checks on input
	// start node should be in the range: 0 <= startnode < inter.num_nodes
	assert(startnode >= 0);
	assert(startnode < inter.num_nodes);

	n_index = 0;
	j = 0;

	if (inter.debug & 32) {
		message("-- Start tracing beam :\n");
	}

	which_components(startnode, &component1_index, &component2_index);

	if (component1_index != NOT_FOUND) {
		set_qbase(component1_index);
	}

	if (component2_index != NOT_FOUND) {
		set_qbase(component2_index);
	}

	if (component1_index != NOT_FOUND) {
		tracing(startnode, component1_index);
	}

	if (component2_index != NOT_FOUND) {
		tracing(startnode, component2_index);
	}

	if (inter.debug & 32) {
		message("-- ... tracing complete\n");
	}

	// with new gnd_node implementation, the dumps are now counted as normal
	// nodes, hence the inter.num_nodes variable will be greater than in the
	// previous implementation.  So, now need to check if the number of nodes
	// traced through PLUS the number of dumps is equal to the total number of
	// nodes, if not, then barf.
	if ((n_index + inter.num_dump_nodes) != inter.num_nodes) {
		fprintf(stderr,
				"Interferometer may be broken;"
				" no parameters were set for the\nfollowing nodes:\n");

		int i;
		for (i = 0; i < inter.num_nodes; i++) {
			if ((NOT inter.node_list[i].q_is_set) && (NOT inter.node_list[i].gnd_node)) {
				fprintf(stderr, "%s ", inter.node_list[i].name);
				j++;
				if (j > 6) {
					j = 0;
					fprintf(stderr, "\n");
				}
			}
		}
		fprintf(stderr, "\n");
		gerror("Could not completely trace beam.\n");
	}
}

//! Perform the full trace of the interferometer

/*!
 * \param startnode the node at which to start tracing
 * \param start_comp the component at which to start tracing
 *
 * \todo startnode -> start_node_index
 * \todo start_comp -> start_component_index
 *
 * \see Test_tracing()
 */
void tracing(int startnode, int start_comp) {
	int component1_index, component2_index, component_index;
	int node_indices[4] = {0};
	int num_nodes;
	int test;
	int testqset;

	// sanity checks on inputs
	// start node should be in the range: 0 <= startnode < inter.num_nodes
	assert(startnode >= 0);
	assert(startnode < inter.num_nodes);
	// start_comp should be in the range:
	// 0 <= start_comp < inter.num_components
	assert(start_comp >= 0);
	assert(start_comp < inter.num_components);

	which_components(startnode, &component1_index, &component2_index);
	if (component2_index == start_comp) {
		component_index = component1_index;
	} else {
		component_index = component2_index;
	}

	test = 0;
	int i;
	for (i = 0; i < n_index; i++) {
		if (vlocal.trace_n[0][i] == startnode) {
			test = 1;
		}
	}

	if (!test) {
		vlocal.trace_n[0][n_index] = startnode;
		vlocal.trace_n[1][n_index] = start_comp;
		vlocal.trace_n[2][n_index] = component_index;
		n_index++;
	}

	if (component_index != NOT_FOUND) {
		which_nodes(component_index, GND_NODE, &num_nodes, node_indices);

		if (num_nodes) { // new nodes
			testqset = 0;
			int i;
			for (i = 0; i < num_nodes; i++) {
				if (inter.node_list[node_indices[i]].q_is_set) {
					testqset++;
				}
			}
			if (testqset == 0) {
				bug_error("missing beam parameter");
			}

			// setting new q values around new component
			set_qbase(component_index);

			for (i = 0; i < num_nodes; i++) {
				if (inter.node_list[node_indices[i]].gnd_node) {
					test = 1;
				} else {
					test = 0;
				}

				int j;
				for (j = 0; j < n_index; j++) {
					if (node_indices[i] == vlocal.trace_n[0][j]) {
						test++;
						if (inter.debug & 256) {
							message("tracing: node %s(%d) already found as number %d\n",
									get_node_name(node_indices[i]), node_indices[i], j);
						}
					}
				}
				if (NOT test) { // found new node
					if (inter.debug & 256) {
						message("tracing: New node %s(%d)\n",
								get_node_name(node_indices[i]), node_indices[i]);
					}

					// recursively look for next nodes/components
					tracing(node_indices[i], component_index);
				}
			}
		}
	}

}

//! Set up the component matrix

/*!
 * \param matrix the matrix to return
 * \param component_index component index for which the matrix should be set up
 * \param node1_index component node index
 * \param node2_index component node index
 * \param tan_sag flag: determines if the return matrix is in sagittal or
 * tangential plane
 *
 * \return the success status of the routine: 1 failure, 0 success.
 *
 * \see Test_component_matrix()
 */
int component_matrix(ABCD_t *matrix, int component_index,
		int node1_index, int node2_index, int tan_sag) {
	int comp_type;
	mirror_t *mirror;
	space_t *space;
	beamsplitter_t *bs;
	modulator_t *modulator;
	diode_t *diode;
	lens_t *lens;
	grating_t *grating;

	comp_type = get_component_type_decriment_index(&component_index);

	*matrix = Unity;

	// sanity checks on inputs
	// component_index should be in the range:
	// 0 <= component_index < inter.num_components
	assert(component_index >= 0);
	assert(component_index < inter.num_components);
	// node[1-2]_index should be in the range:
	// 0 <= node[1-2]_index < inter.num_nodes
	assert(node1_index >= 0);
	assert(node1_index < inter.num_nodes);
	assert(node2_index >= 0);
	assert(node2_index < inter.num_nodes);
	// tan_sag should be either TANGENTIAL or SAGITTAL
	assert(tan_sag == TANGENTIAL || tan_sag == SAGITTAL);

	switch (comp_type) {
		case MIRROR:
			mirror = &(inter.mirror_list[component_index]);
			if ((node1_index != mirror->node1_index &&
					node1_index != mirror->node2_index)
					||
					(node2_index != mirror->node1_index &&
					node2_index != mirror->node2_index)) {
				bug_error("mirror no such node");
			}

			// transmission
			if (node1_index != node2_index) {
				if (tan_sag == TANGENTIAL) {
					*matrix = (mirror->qqtt);
				} else {
					*matrix = (mirror->qqts);
				}
			}// reflection, left to right
			else if (node1_index == mirror->node1_index) {
				if (tan_sag == TANGENTIAL) {
					*matrix = (mirror->qqr1t);
				} else {
					*matrix = (mirror->qqr1s);
				}
			}// reflection, right to left
			else {
				if (tan_sag == TANGENTIAL) {
					*matrix = (mirror->qqr2t);
				} else {
					*matrix = (mirror->qqr2s);
				}
			}
			//dump_ABCD(stdout, *matrix);
			break;
		case BEAMSPLITTER:
			bs = &(inter.bs_list[component_index]);
			if ((node1_index != bs->node1_index &&
					node1_index != bs->node2_index &&
					node1_index != bs->node3_index &&
					node1_index != bs->node4_index)
					||
					(node2_index != bs->node1_index &&
					node2_index != bs->node2_index &&
					node2_index != bs->node3_index &&
					node2_index != bs->node4_index)) {
				bug_error("bs no such node");
			}

			if (node1_index == node2_index) {
				return (1);
			}

			// reflection1
			if ((node1_index == bs->node1_index &&
					node2_index == bs->node2_index)
					||
					(node1_index == bs->node2_index &&
					node2_index == bs->node1_index)) {
				if (tan_sag == TANGENTIAL) {
					*matrix = (bs->qqr1t);
				} else {
					*matrix = (bs->qqr1s);
				}
			}// reflection2
			else if ((node1_index == bs->node3_index &&
					node2_index == bs->node4_index)
					||
					(node1_index == bs->node4_index &&
					node2_index == bs->node3_index)) {
				if (tan_sag == TANGENTIAL) {
					*matrix = (bs->qqr2t);
				} else {
					*matrix = (bs->qqr2s);
				}
			}// transmission to
			else if ((node1_index == bs->node1_index &&
					node2_index == bs->node3_index)
					||
					(node1_index == bs->node2_index &&
					node2_index == bs->node4_index)) {
				if (tan_sag == TANGENTIAL)
					*matrix = (bs->qqt1t);
				else
					*matrix = (bs->qqt1s);
			}// transmission back
            else if ((node1_index == bs->node3_index &&
                    node2_index == bs->node1_index)
                    ||
                    (node1_index == bs->node4_index &&
                    node2_index == bs->node2_index)) {
                if (tan_sag == TANGENTIAL) {
                    *matrix = (bs->qqt2t);
                } else {
                    *matrix = (bs->qqt2s);
                }
            } else {
                return (1);
            }
            break;
        case SPACE:
            space = &(inter.space_list[component_index]);
            if ((node1_index != space->node1_index &&
                    node1_index != space->node2_index)
                    ||
                    (node2_index != space->node1_index &&
                    node2_index != space->node2_index)) {
                bug_error("space no such node");
            }

            if (node1_index == node2_index) {
                return (1);
            }
            *matrix = (space->qq);
            break;
        case MODULATOR:
            modulator = &(inter.modulator_list[component_index]);
            if ((node1_index != modulator->node1_index &&
                    node1_index != modulator->node2_index)
                    ||
                    (node2_index != modulator->node1_index &&
                    node2_index != modulator->node2_index)) {
                bug_error("modulator no such node");
            }

            if (node1_index == node2_index) {
                return (1);
            }
            *matrix = (modulator->qq);
            break;
        case LENS:
            lens = &(inter.lens_list[component_index]);
            if ((node1_index != lens->node1_index &&
                    node1_index != lens->node2_index)
                    ||
                    (node2_index != lens->node1_index &&
                    node2_index != lens->node2_index)) {
                bug_error("lens no such node");
            }

            if (node1_index == node2_index) {
                return (1);
            }
            *matrix = (lens->qq);
            break;
        case DIODE:
            diode = &(inter.diode_list[component_index]);
            if ((node1_index != diode->node1_index &&
                    node1_index != diode->node2_index)
                    ||
                    (node2_index != diode->node1_index &&
                    node2_index != diode->node2_index)) {
                bug_error("diode no such node");
            }

            if (node1_index == node2_index) {
                return (1);
            }
            *matrix = (diode->qq);
			break;
		case GRATING:
			grating = &(inter.grating_list[component_index]);
			switch (grating->num_of_ports) {
				case 2:
					if ((node1_index != grating->node1_index &&
							node1_index != grating->node2_index)
							||
							(node2_index != grating->node1_index &&
							node2_index != grating->node2_index)) {
						bug_error("gr2 no such node");
					}

					// reflection
					if (node1_index == node2_index) {
						if (tan_sag == TANGENTIAL) {
							*matrix = (grating->qqt[0][0]);
						} else {
							*matrix = (grating->qqs[0][0]);
						}
					}// reflection, input to 0th order
					else if (node1_index == grating->node1_index) {
						if (tan_sag == TANGENTIAL) {
							*matrix = (grating->qqt[0][1]);
						} else {
							*matrix = (grating->qqs[0][1]);
						}
					}// reflection, zeroth order into input
					else {
						if (tan_sag == TANGENTIAL) {
							*matrix = (grating->qqt[1][0]);
						} else {
							*matrix = (grating->qqs[1][0]);
						}
					}
					break;
				case 3:
					if ((node1_index != grating->node1_index &&
							node1_index != grating->node2_index &&
							node1_index != grating->node3_index)
							||
							(node2_index != grating->node1_index &&
							node2_index != grating->node2_index &&
							node2_index != grating->node3_index)) {
						bug_error("gr3 no such node");
					}

					// reflection, input to 0th order
					if (node1_index == grating->node1_index) {
						if (node2_index == grating->node1_index) {
							if (tan_sag == TANGENTIAL) {
								*matrix = (grating->qqt[0][0]);
							} else {
								*matrix = (grating->qqs[0][0]);
							}
						} else if (node2_index == grating->node2_index) {
							if (tan_sag == TANGENTIAL) {
								*matrix = (grating->qqt[0][1]);
							} else {
								*matrix = (grating->qqs[0][1]);
							}
						} else {
							if (tan_sag == TANGENTIAL) {
								*matrix = (grating->qqt[0][2]);
							} else {
								*matrix = (grating->qqs[0][2]);
							}
						}
					} else if (node1_index == grating->node2_index) {
						if (node2_index == grating->node1_index) {
							if (tan_sag == TANGENTIAL) {
								*matrix = (grating->qqt[1][0]);
							} else {
								*matrix = (grating->qqs[1][0]);
							}
						} else if (node2_index == grating->node2_index) {
							if (tan_sag == TANGENTIAL) {
								*matrix = (grating->qqt[1][1]);
							} else {
								*matrix = (grating->qqs[1][1]);
							}
						} else {
							if (tan_sag == TANGENTIAL) {
								*matrix = (grating->qqt[1][2]);
							} else {
								*matrix = (grating->qqs[1][2]);
							}
						}
					} else // node1_index = grating-> node3
					{
						if (node2_index == grating->node1_index) {
							if (tan_sag == TANGENTIAL) {
								*matrix = (grating->qqt[2][0]);
							} else {
								*matrix = (grating->qqs[2][0]);
							}
						} else if (node2_index == grating->node2_index) {
							if (tan_sag == TANGENTIAL) {
								*matrix = (grating->qqt[2][1]);
							} else {
								*matrix = (grating->qqs[2][1]);
							}
						} else // node 2 == grating->node3
						{
							if (tan_sag == TANGENTIAL) {
								*matrix = (grating->qqt[2][2]);
							} else {
								*matrix = (grating->qqs[2][2]);
							}
						}
					}
					break;
				case 4:
					break;
			}
			break;
		default:
			bug_error("no such type");
	}
	return (0);
}

//! Set the Gaussian beam parameter for the respective mode.

/*!
 * \param component_index index of component at which to set the Gaussian
 * beam parameter
 *
 * \see Test_set_qbase()
 */
void set_qbase(int component_index) {
	int err;
	int no, node_indices[4] = {0};
	ABCD_t transform;
	complex_t qx, qy, qxt, qyt;
	double nr1, nr2;

	node_t node_i, node_j;

	// sanity checks on input
	// component index should be in the range:
	// 0 <= component_index < inter.num_components
	assert(component_index >= 0);
	assert(component_index < inter.num_components);

	which_nodes(component_index, GND_NODE, &no, node_indices);
	int k;
	for (k = 0; k < 2; k++) {
		int i;
		for (i = 0; i < no; i++) {
			node_i = inter.node_list[node_indices[i]];
			if (node_i.q_is_set) {
				int j;
				for (j = 0; j < no; j++) {
					node_j = inter.node_list[node_indices[j]];
					// set q if q is not yet set
					if (NOT node_j.q_is_set &&
							NOT node_j.gnd_node) { // || get_component_type(component)==SPACE)
						err = component_matrix(&transform, component_index,
								node_indices[i], node_indices[j], TANGENTIAL);
						if (!err) {
							if (inter.debug & 256) {
								printf("set_qbase: q set: %s\n",
										get_node_name(node_indices[j]));
								dump_ABCD(stdout, transform);

								// dump info to log file
								fprintf(fp_log, "set_qbase: q set: %s\n",
										get_node_name(node_indices[j]));
								dump_ABCD(fp_log, transform);
							}
							// q -> q(-z)=-q* if necessary (reverse beam direction)
							if (node_i.component_index == component_index) {
								qx = cminus(cconj(node_i.qx));
								qy = cminus(cconj(node_i.qy));
							} else {
								qx = node_i.qx;
								qy = node_i.qy;
							}

							nr1 = *node_i.n;
							nr2 = *node_j.n;
							qxt = q1_q2(transform, qx, nr1, nr2);
							err = component_matrix(&transform, component_index,
									node_indices[i], node_indices[j], SAGITTAL);
							qyt = q1_q2(transform, qy, nr1, nr2);
							set_q(node_indices[j], component_index, qxt, qyt);
						} else if (inter.debug & 256) {
							message("set_qbase: err -> %s %s\n",
									get_node_name(node_indices[i]),
									get_node_name(node_indices[j]));
						}
					}
				}
			}
		}
	}
}

//! For two given nodes returns the component connecting these nodes

/*!
 * (if the component is a beam splitter, and consequently has four nodes,
 * one only needs to give two out of the four as this is for tracing a path
 * when one walks from node to node).
 * 
 * \param component_type the type of component
 * \param component_index the component index
 * \param node1_index first component node index
 * \param node2_index second component node index
 *
 * \see Test_get_component()
 */
void get_component(int *component_type, int *component_index,
		int node1_index, int node2_index) {
	int tmp_component_index, component1_index, component2_index;
	int component3_index, component4_index;

	// sanity checks on inputs
	// the component index should be in the range:
	// 0 <= component_index < inter.num_components
	assert(*component_index >= 0);
	assert(*component_index < inter.num_components);
	// the node indices should be in the range:
	// 0 <= node_index < inter.num_nodes
	assert(node1_index >= 0);
	assert(node2_index >= 0);
	assert(node1_index < inter.num_nodes);
	assert(node2_index < inter.num_nodes);

	which_components(node1_index, &component1_index, &component2_index);
	which_components(node2_index, &component3_index, &component4_index);

	if (component1_index == component3_index ||
			component1_index == component4_index) {
		tmp_component_index = component1_index;
	} else if (component2_index == component3_index ||
			component2_index == component4_index) {
		tmp_component_index = component2_index;
	} else {
		bug_error("nodes not at the same component");
	}

	*component_type = get_component_type_decriment_index(&tmp_component_index);
	*component_index = tmp_component_index;
}

//! For a given component number return number of nodes and node numbers

/*! 
 * The 'current_node' is used to determine which BS nodes are reachable 
 * for the incident beam. This feature is used for beam and cavity tracing. 
 * In other places you can set current_node to GND_NODE and all four BS 
 * nodes will be returned!
 *
 * \param component_index the component number
 * \param current_node the node we are currently at in the system
 * \param num_nodes the number of nodes
 * \param node_indices the node indices
 *
 * \see Test_which_nodes()
 */
void which_nodes(int component_index, int current_node,
		int *num_nodes, int *node_indices) {
	int k;
	int no;

	// sanity checks on inputs
	// component index should be in the range:
	assert(component_index >= 0);
	assert(component_index < inter.num_components);
	// current node should be in the range:
	assert(current_node == GND_NODE ||
			(current_node >= 0 && current_node < inter.num_nodes));

	k = component_index;

	no = 0;

	//if (inter.debug & 256)
	//  printf("which_nodes component_index: %d\n", component_index);

	if (component_index == NOT_FOUND) { // experimental: node only used once
		no = 0;
	} else if (component_index < inter.num_mirrors) {
		if (component_index < 0) {
			bug_error("component index less than zero");
		}

		mirror_t mirror = inter.mirror_list[component_index];
		node_t node1 = inter.node_list[mirror.node1_index];
		node_t node2 = inter.node_list[mirror.node2_index];

		if (NOT node1.gnd_node) {
			node_indices[no++] = mirror.node1_index;
		}

		if (NOT node2.gnd_node) {
			node_indices[no++] = mirror.node2_index;
		}
	}// experimental: BS returns only the two available nodes
        // depending on the node of the incident field
    else if ((component_index -= inter.num_mirrors) < inter.num_beamsplitters) {
        beamsplitter_t beamsplitter = inter.bs_list[component_index];

        node_t node1 = inter.node_list[beamsplitter.node1_index];
        node_t node2 = inter.node_list[beamsplitter.node2_index];
        node_t node3 = inter.node_list[beamsplitter.node3_index];
        node_t node4 = inter.node_list[beamsplitter.node4_index];

        if (current_node == GND_NODE ||
                inter.node_list[current_node].gnd_node) {
            if (NOT node1.gnd_node) {
                node_indices[no++] = beamsplitter.node1_index;
            }

            if (NOT node2.gnd_node) {
                node_indices[no++] = beamsplitter.node2_index;
            }

            if (NOT node3.gnd_node) {
                node_indices[no++] = beamsplitter.node3_index;
            }

            if (NOT node4.gnd_node) {
                node_indices[no++] = beamsplitter.node4_index;
            }
        } else {
            if (current_node == beamsplitter.node1_index) {
                if (NOT node2.gnd_node) {
                    node_indices[no++] = beamsplitter.node2_index;
                }

                if (NOT node3.gnd_node) {
                    node_indices[no++] = beamsplitter.node3_index;
                }
            }

            if (current_node == beamsplitter.node2_index) {
                if (NOT node1.gnd_node) {
                    node_indices[no++] = beamsplitter.node1_index;
                }

                if (NOT node4.gnd_node) {
                    node_indices[no++] = beamsplitter.node4_index;
                }
            }

            if (current_node == beamsplitter.node3_index) {
                if (NOT node4.gnd_node) {
                    node_indices[no++] = beamsplitter.node4_index;
                }

                if (NOT node1.gnd_node) {
                    node_indices[no++] = beamsplitter.node1_index;
                }
            }

            if (current_node == beamsplitter.node4_index) {
                if (NOT node3.gnd_node) {
                    node_indices[no++] = beamsplitter.node3_index;
                }

                if (NOT node2.gnd_node) {
                    node_indices[no++] = beamsplitter.node2_index;
                }
            }
        }
    } else if ((component_index -= inter.num_beamsplitters) < inter.num_gratings) {
        grating_t grating = inter.grating_list[component_index];

        node_t node1 = inter.node_list[grating.node1_index];
        node_t node2 = inter.node_list[grating.node2_index];
        node_t node3 = inter.node_list[grating.node3_index];
        node_t node4 = inter.node_list[grating.node4_index];

        if (NOT node1.gnd_node) {
            node_indices[no++] = grating.node1_index;
        }
        if (NOT node2.gnd_node) {
            node_indices[no++] = grating.node2_index;
        }
        switch (grating.num_of_ports) {
            case 3:
                if (NOT node3.gnd_node) {
                    node_indices[no++] = grating.node3_index;
                }
            case 4:
                if (NOT node4.gnd_node) {
                    node_indices[no++] = grating.node4_index;
                }
        }
    } else if ((component_index -= inter.num_gratings) < inter.num_spaces) {
        space_t space = inter.space_list[component_index];

        node_t node1 = inter.node_list[space.node1_index];
        node_t node2 = inter.node_list[space.node2_index];

        if (NOT node1.gnd_node) {
            node_indices[no++] = space.node1_index;
        }

        if (NOT node2.gnd_node) {
            node_indices[no++] = space.node2_index;
        }
    } else if ((component_index -= inter.num_spaces) < inter.num_lenses) {
        lens_t lens = inter.lens_list[component_index];

        node_t node1 = inter.node_list[lens.node1_index];
        node_t node2 = inter.node_list[lens.node2_index];

        if (NOT node1.gnd_node) {
            node_indices[no++] = lens.node1_index;
        }

        if (NOT node2.gnd_node) {
            node_indices[no++] = lens.node2_index;
        }
    } else if ((component_index -= inter.num_lenses) < inter.num_diodes) {
        diode_t diode = inter.diode_list[component_index];

        node_t node1 = inter.node_list[diode.node1_index];
        node_t node2 = inter.node_list[diode.node2_index];

        if (NOT node1.gnd_node) {
            node_indices[no++] = diode.node1_index;
        }

        if (NOT node2.gnd_node) {
            node_indices[no++] = diode.node2_index;
        }
    } else if ((component_index -= inter.num_diodes) < inter.num_light_inputs) {
        light_in_t light_in = inter.light_in_list[component_index];

        node_t light_in_node = inter.node_list[light_in.node_index];

        if (NOT light_in_node.gnd_node) {
            node_indices[no++] = light_in.node_index;
        }
    } else if ((component_index -= inter.num_light_inputs) < inter.num_modulators) {
        modulator_t modulator = inter.modulator_list[component_index];

        node_t node1 = inter.node_list[modulator.node1_index];
        node_t node2 = inter.node_list[modulator.node2_index];

        if (NOT node1.gnd_node) {
            node_indices[no++] = modulator.node1_index;
        }

        if (NOT node2.gnd_node) {
            node_indices[no++] = modulator.node2_index;
        }
    } else if ((component_index -= inter.num_modulators) < inter.num_signals) {
        bug_error("which_node 3");
        no = 2;
    } else if ((component_index -= inter.num_signals) < inter.num_outputs) {
        output_data_t output_data = inter.output_data_list[component_index];

        node_t output_data_node = inter.node_list[output_data.node_index];

        if (NOT output_data_node.gnd_node) {
            node_indices[no++] = output_data.node_index;
        }
    } else {
        bug_error("component index outside correct range");
    }

    *num_nodes = no;

    if (inter.debug & 16) {
        if (*num_nodes) {
            message("For component %s(%d) -> nodes:", get_component_name(k), k);
        } else {
            message("no new nodes found here\n");
        }

        int j;
        for (j = 0; j<*num_nodes; j++) {
            message(" %s(%d)", get_node_name(node_indices[j]), node_indices[j]);
        }
        message("\n");
    }
}

//! For given node number sets component numbers with connected components

/*!
 * For a given node number sets the two component numbers with the 
 * connected components to NOT_FOUND if only one component is connected.
 * The result is stored in a global array vlocal.component_trace_list
 *
 * \param node_index the node index to check
 * \param component1_index index of one component connected to node
 * \param component2_index index of other component connected to node
 *
 * \see Test_which_components()
 */
void which_components(int node_index,
        int *component1_index, int *component2_index) {
    // sanity checking on inputs
    // node_index should be in the range 0 <= node_index < inter.num_nodes
    assert(node_index >= 0);
    assert(node_index < inter.num_nodes);

    //  int comp[MAXCOMP] = { 0 };

    if (node_index < 0) {
        bug_error("node index less than zero");
    }

    int index = 0; //!< \todo is this the trace index???
    int object_index = 0;

    int mirror_index;
    for (mirror_index = 0; mirror_index < inter.num_mirrors; mirror_index++) {
        if (inter.mirror_list[mirror_index].node1_index == node_index ||
                inter.mirror_list[mirror_index].node2_index == node_index) {
            vlocal.component_trace_list[index++] = mirror_index;
        }
    }
    object_index += inter.num_mirrors;

    int beamsplitter_index;
    for (beamsplitter_index = 0;
            beamsplitter_index < inter.num_beamsplitters;
            beamsplitter_index++) {
        if (inter.bs_list[beamsplitter_index].node1_index == node_index ||
                inter.bs_list[beamsplitter_index].node2_index == node_index ||
                inter.bs_list[beamsplitter_index].node3_index == node_index ||
                inter.bs_list[beamsplitter_index].node4_index == node_index) {
            vlocal.component_trace_list[index++] = beamsplitter_index + object_index;
        }
    }
    object_index += inter.num_beamsplitters;

    int grating_index;
    for (grating_index = 0; grating_index < inter.num_gratings; grating_index++) {
        switch (inter.grating_list[grating_index].num_of_ports) {
            case 2:
                if (inter.grating_list[grating_index].node1_index == node_index ||
                        inter.grating_list[grating_index].node2_index == node_index) {
                    vlocal.component_trace_list[index++] = grating_index + object_index;
                }
                break;
            case 3:
                if (inter.grating_list[grating_index].node1_index == node_index ||
                        inter.grating_list[grating_index].node2_index == node_index ||
                        inter.grating_list[grating_index].node3_index == node_index) {
                    vlocal.component_trace_list[index++] = grating_index + object_index;
                }
                break;
            case 4:
                if (inter.grating_list[grating_index].node1_index == node_index ||
                        inter.grating_list[grating_index].node2_index == node_index ||
                        inter.grating_list[grating_index].node3_index == node_index ||
                        inter.grating_list[grating_index].node4_index == node_index) {
                    vlocal.component_trace_list[index++] = grating_index + object_index;
                }
                break;
            default:
                bug_error("aa, comptrace: grating - wrong number of ports");
        }
    }
    object_index += inter.num_gratings;

    int space_index;
    for (space_index = 0; space_index < inter.num_spaces; space_index++) {
        if (inter.space_list[space_index].node1_index == node_index ||
                inter.space_list[space_index].node2_index == node_index) {
            vlocal.component_trace_list[index++] = space_index + object_index;
        }
    }
    object_index += inter.num_spaces;

    int squeezer_index;
    for (squeezer_index = 0;
            squeezer_index < inter.num_squeezers;
            squeezer_index++) {
        if (inter.squeezer_list[squeezer_index].node1_index == node_index ||
                inter.squeezer_list[squeezer_index].node2_index == node_index) {
            vlocal.component_trace_list[index++] = squeezer_index + object_index;
        }
    }
    object_index += inter.num_squeezers;

    int lens_index;
    for (lens_index = 0; lens_index < inter.num_lenses; lens_index++) {
        if (inter.lens_list[lens_index].node1_index == node_index ||
                inter.lens_list[lens_index].node2_index == node_index) {
            vlocal.component_trace_list[index++] = lens_index + object_index;
        }
    }
    object_index += inter.num_lenses;

    int diode_index;
    for (diode_index = 0; diode_index < inter.num_diodes; diode_index++) {
        if (inter.diode_list[diode_index].node1_index == node_index ||
                inter.diode_list[diode_index].node2_index == node_index) {
            vlocal.component_trace_list[index++] = diode_index + object_index;
        }
    }
    object_index += inter.num_diodes;

    int light_input_index;
    for (light_input_index = 0;
            light_input_index < inter.num_light_inputs;
            light_input_index++) {
        if (inter.light_in_list[light_input_index].node_index == node_index) {
            vlocal.component_trace_list[index++] = light_input_index + object_index;
        }
    }
    object_index += inter.num_light_inputs;

    int modulator_index;
    for (modulator_index = 0;
            modulator_index < inter.num_modulators;
            modulator_index++) {
        if (inter.modulator_list[modulator_index].node1_index == node_index ||
                inter.modulator_list[modulator_index].node2_index == node_index) {
            vlocal.component_trace_list[index++] = modulator_index + object_index;
        }
    }
    //  j+=inter.num_modulators+inter.num_signals;
    object_index += inter.num_modulators; // 080805

    if (index > 2 || index < 1) {
        if (index < 1) {
            fprintf(stderr, "at node %s !!", get_node_name(node_index));
            bug_error("no component at node");
        }
        gerror("node %s connected to %s components!\n", get_node_name(node_index),
                int_form(index));
    }

    *component1_index = vlocal.component_trace_list[0];
    if (index == 2) {
        *component2_index = vlocal.component_trace_list[1];
    } else {
        *component2_index = NOT_FOUND;
    }

    if (inter.debug & 16) {
        message("For node %s(%d) -> components: %s(%d) and %s(%d)\n",
                get_node_name(node_index), node_index,
                get_component_name(*component1_index), *component1_index,
                get_component_name(*component2_index), *component2_index);
    }
}

//! Returns a given node's name

/*!
 * \param node_index the node index
 *
 * \see Test_get_node_name()
 */
char *get_node_name(int node_index) {
    // sanity checks on input
    // node index should be in the range: 0 <= node_index < inter.num_nodes
    assert(node_index >= 0);
    assert(node_index < inter.num_nodes);

    if (node_index < 0) {
        message("node_index:%d\n", node_index);
        bug_error("node index less than zero");
    }

    if (node_index >= inter.num_nodes) {
        message("node_index:%d\n", node_index);
        bug_error("node index greater than number of nodes");
    }

    return duplicate_string(inter.node_list[node_index].name);
}

//! Sets Gaussian beam parameter (different situation to set_qbase)

/*!
 * \param node_index the node index to set
 * \param component_index the component index
 * \param qx Gaussian beam parameter in x-plane
 * \param qy Gaussian beam parameter in y-plane
 *
 * \see set_qbase
 * \see Test_set_q()
 */
void set_q(int node_index, int component_index, complex_t qx, complex_t qy) {
    int component1_index, component2_index;
    int type, no;
    int num_nodes, node_indices[2] = {0};
    double nr1, nr2;
    complex_t qxt, qyt;
    complex_t qxt2, qyt2;
    node_t *node;
    space_t space;

    // make sure that the arguments are sane:
    // the node_index should be greater than, or equal to, zero
    assert(node_index >= 0);
    // and less than the total number of nodes in the interferometer
    assert(node_index < inter.num_nodes);
    // the component index should be >= 0
    assert(component_index >= 0);
    // and less than the total number of components in the interferometer
#ifndef NDEBUG
    if (component_index >= inter.num_components) {
        message("set_q(): component_index >= inter.num_components\n"
                "component_index = %d, inter.num_components = %d\n",
                component_index, inter.num_components);
    }
#endif
    assert(component_index < inter.num_components);

    node = &(inter.node_list[node_index]);
        
    
    if ((!(node->q_is_set)) ||
            (node->q_is_set && (!ceq(node->qx, qx) || !ceq(node->qy, qy)))) {
        if (node->q_is_set) {
            if (inter.trace & 4 || inter.trace & 32) {
                message("q value already set at node %s, overwritten!\n",
                        node_print(node_index));

                if (inter.debug & 256) {
                    message("(old qx %s qy %s , new qx %s qy %s)\n",
                            complex_form(node->qx), complex_form(node->qy),
                            complex_form(qx), complex_form(qy));
                }
            }
        }

        // set q value
        if (inter.debug & 256) {
            message("--- setting q for node %s\n", get_node_name(node_index));
        }
        node->qx = qx;
        node->qy = qy;
        node->component_index = component_index;
        node->q_is_set = true;
        node->q_changed = true;

        // now check whether a space is connected and set second
        // q value if necessary
        which_components(node_index, &component1_index, &component2_index);

        // first component
        no = component1_index;
        type = get_component_type_decriment_index(&no);
        if (type == SPACE) {
            space = inter.space_list[no];
            which_nodes(component1_index, GND_NODE, &num_nodes, node_indices);
            if (num_nodes > 2) {
                bug_error("too many nodes");
            }

            if (num_nodes == 2) {
                if (node_indices[0] == node_index) {
                    node = &(inter.node_list[node_indices[1]]);
                } else if (node_indices[1] == node_index) {
                    node = &(inter.node_list[node_indices[0]]);
                } else {
                    bug_error("wrong node 1");
                }
            }

            nr1 = *inter.node_list[node_index].n;
            nr2 = *(node->n);
            // if first component is initial component then reverse q
            if (component1_index == component_index) {
                qxt = q1_q2(space.qq, cminus(cconj(qx)), nr1, nr2);
                qyt = q1_q2(space.qq, cminus(cconj(qy)), nr1, nr2);
            } else {
                qxt = q1_q2(space.qq, qx, nr1, nr2);
                qyt = q1_q2(space.qq, qy, nr1, nr2);
            }

            if (node->q_is_set) {
                // if set q comes not out of first comp then reverse q
                if (node->component_index != component1_index) {
                    qxt2 = cminus(cconj(node->qx));
                    qyt2 = cminus(cconj(node->qy));
                } else {
                    qxt2 = node->qx;
                    qyt2 = node->qy;
                }

    
                if (!ceq(qxt2, qxt) || !ceq(qyt2, qyt)) {
                    if (inter.trace & 32) {
                        message("   overwriting second q value at space %s\n", space.name);
                    }
                    node->qx = qxt;
                    node->qy = qyt;
                    node->component_index = component1_index;
                    node->q_is_set = true;
                    node->q_changed = true;
                    
                    if (inter.debug & 256) {
                        message("   1 :qx %s, qy %s\n", complex_form(node->qx),
                                complex_form(node->qy));
                    }
                }
            } else {
                if (inter.trace & 32) {
                    message("   setting second q value at space %s\n", space.name);
                }
                node->qx = qxt;
                node->qy = qyt;
                node->component_index = component1_index;
                node->q_is_set = true;

                /*
                 * if (inter.debug & 256)
                 * {
                 * dump_ABCD(stdout, space.qq);
                 * printf("   ... and qx %s qy %s\n", complex_form(qx), complex_form(qy));
                 * printf("   2 :qx %s, qy %s\n", complex_form(n->qx), complex_form(n->qy));
                 * }
                 */
            }
        }

        // second component
        no = component2_index;
        type = get_component_type_decriment_index(&no);
        if (type == SPACE) {
            space = inter.space_list[no];
            which_nodes(component2_index, GND_NODE, &num_nodes, node_indices);
            if (num_nodes > 2) {
                bug_error("too many nodes");
            }

            if (num_nodes == 2) {
                if (node_indices[0] == node_index) {
                    node = &(inter.node_list[node_indices[1]]);
                } else if (node_indices[1] == node_index) {
                    node = &(inter.node_list[node_indices[0]]);
                } else {
                    bug_error("wrong node 1");
                }
            }

            nr1 = *inter.node_list[node_index].n;
            nr2 = *(node->n);
            // if second component is initial component then reverse q
            if (component2_index == component_index) {
                qxt = q1_q2(space.qq, cminus(cconj(qx)), nr1, nr2);
                qyt = q1_q2(space.qq, cminus(cconj(qy)), nr1, nr2);
            } else {
                qxt = q1_q2(space.qq, qx, nr1, nr2);
                qyt = q1_q2(space.qq, qy, nr1, nr2);
            }

            if (node->q_is_set) {
                // if set q does not come out of second comp then reverse q
                if (node->component_index != component2_index) {
                    qxt2 = cminus(cconj(node->qx));
                    qyt2 = cminus(cconj(node->qy));
                } else {
                    qxt2 = node->qx;
                    qyt2 = node->qy;
                }

                if (!ceq(qxt2, qxt) || !ceq(qyt2, qyt)) {
                    if (inter.trace & 32) {
                        message("   overwriting second q value at space %s\n", space.name);
                    }
                    node->qx = qxt;
                    node->qy = qyt;
                    node->component_index = component2_index;
                    node->q_is_set = true;

                    if (inter.debug & 256) {
                        message("   1 :qx %s, qy %s\n", complex_form(node->qx),
                                complex_form(node->qy));
                    }
                }
            } else {
                if (inter.trace & 32) {
                    message("   setting second q value at space %s\n", space.name);
                }
                node->qx = qxt;
                node->qy = qyt;
                node->component_index = component2_index;
                node->q_is_set = true;

                /*
                 * if (inter.debug & 256)
                 * {
                 * dump_ABCD(stdout, space.qq);
                 * printf("   ... and qx %s qy %s\n", complex_form(qx), complex_form(qy));
                 * printf("   2 :qx %s, qy %s\n", complex_form(n->qx), complex_form(n->qy));
                 * }
                 */
            }
        }
    } else {
        if (inter.debug & 256) {
            message("--- q already set for node %s\n", get_node_name(node_index));
        }
    }
}

FILE * ipfile;

/**
 * Merged map interpolation involves interpolating both the amplitude and phase
 * of a merged map. Depending on which coupling coefficient you want to 
 * calculate you need to choose the knum argument correctly: this determines
 * whether to use reflection or transmission parts.
 * 
 * This function applies the refractive index to the phase maps.
 * 
 * This method also deals with providing the A and phi in what should be the
 * outgoing beams (from a mirror surface) coordinate system. So the mirror map
 * is applied to a mirror looking at it's surface along the beam axis. Positive
 * values correspond to a surface that is closer, negative is further away.
 * So the beam is integrated in the outgoing modes coordinate system, this
 * causes many headaches and requires careful thinking. The best way (I find) is
 * to consider a tilted surface as is considered next.
 * 
 * So below x,y,z are the maps coordinates, x',y',z' are the outgoing modes 
 * coordinates, this only applies to phase maps remember,
 * 
 *       x'          \                            
 *        ^           \<---map (tilt in this case)  y' o--> z' 
 *    K11 |            \                               | 
 *        |             \                              v  K12
 *   z'<--o y'  n1 z <---o y   n2                      x'
 *                       |\                         
 *        x'             | \                        y' o--> z' 
 *   K21  ^              |  \                          |
 *        |              |   \                         v  K22
 *   z'<--o y'           v                             x'
 *                       x
 * 
 * So from the above diagram we should see that K11 and K21 has x' = -x and 
 * z' = z whereas K22 and K12 has x' = x and z' = -z (the z'=-z corresponds to 
 * a phase sign flip). n1->n2 and n2->n1 should see 0 phase if refractive index 
 * is the same on both sides. You have to think consider when nr2 != nr1 to 
 * realise why it also works for K12 and K21.
 * 
 * The above should hold true so long as the applied phase map is not purely a
 * transmission map. A transmission only phase map should combine all the various
 * complex layers in a mirror that might exist: coatings, substrate, refractive
 * index variations, etc.; such a map does not portray a surface height variation.
 * This requires us to treat transmission phase maps differently. Now the phase
 * should not be sign flipped for K12 and K22; you can justify this by assuming
 * a transmission phase map means the surface is flat across the mirror, thus no
 * difference in travel in nr1 or nr2 needs to be considered.
 * 
 * @param[in] map	Merged map that will be used for interpolation
 * @param[in] knum	specifies which knm to calculate: 0=k11, 1=k12, 2=k21, 3=k22
 * @param[in] xx	x coorindate to interpolate to
 * @param[in] yy	y cordinate to interpolate to
 * @param[out] A	interpolated amplitude
 * @param[out] phi	interpolated phase
 * @param[in] nr1   refractive index of node 1 side
 * @param[in] nr2   refractive index of node 2 side
 */
void interpolate_merged_map(surface_merged_map_t *map, int knum, const double px
        , const double py, double *const A, double *const phi, double nr1
        , double nr2, double angle) {

    assert(map != NULL);

    int N = map->cols;
    int M = map->rows;
    double *x = map->x;
    double *y = map->y;
    double **z_abs = NULL, **z_phs = NULL;
    int x0 = map->x0;
    int y0 = map->y0;
    double dx = map->xstep;
    double dy = map->ystep;

    int INTERP_SIZE = map->interpolation_size;
    interpolation_method_t INTERP_METHOD = map->interpolation_method;

    assert(INTERP_SIZE % 2 == 1); //Check interpolation grid size is an odd number
    assert(INTERP_SIZE >= 3); //Ensure that 3 or more points are used for
    //cubic spline

    switch (knum) {
        case K11:
        case K22:
            z_abs = map->r_abs;
            z_phs = map->r_phs;
            break;
        case K12:
        case K21:
            z_abs = map->t_abs;
            z_phs = map->t_phs;
            break;
    }

    int INTERP_ULIM = (INTERP_SIZE - 1) / 2;
    int INTERP_LLIM = -(int) (INTERP_SIZE - 1) / 2;

    gsl_spline *splinex = NULL, *spliney = NULL;

    if (INTERP_METHOD != NEAREST) {
        splinex = gsl_spline_alloc((INTERP_METHOD == 1) ? gsl_interp_cspline : gsl_interp_linear, INTERP_SIZE);
        spliney = gsl_spline_alloc((INTERP_METHOD == 1) ? gsl_interp_cspline : gsl_interp_linear, INTERP_SIZE);
    }

    double xx = px;
    double yy = py;
    
    // From diagram above we need to invert x'
    if ((knum == K21) || (knum == K11))
        xx *= -1;

    // TODO the direction of the rotation needs testing more rigorously 
    // not sure where the -1 should be + or - depends on the angle direction
    // convention used
    if(angle != 0.0)
        rotate_vector(&xx, &yy, -angle);
    
    int u = round(xx / dx + x0);
    int v = round(yy / dy + y0);

    if (u < 0 || u > N - 1 || v < 0 || v > M - 1) {
        //if we are out of the bounds of the map then we need to return either
        //an unaffected beam or an apertured one. Assume for now unaffected beam. 
        *A = 1;
        *phi = 0;
    } else {

        if (u + INTERP_LLIM < 0) u = -INTERP_LLIM;
        if (u + INTERP_ULIM >= N) u = N - 1 - INTERP_ULIM;
        if (v + INTERP_LLIM < 0) v = -INTERP_LLIM;
        if (v + INTERP_ULIM >= M) v = M - 1 - INTERP_ULIM;

        if (INTERP_METHOD == NEAREST) {
            // if we are doing nearest neighbour interpolation then we just
            // use the u,v values above
            // update: we should be using the matlab convention of z[rows][cols]
            // which means z[y][x]
            *A = z_abs[v][u];
            *phi = z_phs[v][u];

        } else {

            double interpA[INTERP_SIZE];
            double interpPhi[INTERP_SIZE];
            int n, __u, __v, a;

            for (a = INTERP_LLIM, n = 0; a <= INTERP_ULIM; a++, n++) {
                //For some reason I cant see, the z variable indexes need to be swapped around
                // z[x][y] (which it should be) -> z[y][x]
                // Update: the reason for this is because of the matlab cconvetion is used
                // to store the map data z[rows][cols]
              
                __v = v + a;
                __u = u + INTERP_LLIM;

                if ((xx < x[0]) || (xx > x[N - 1])){
                    interpA[n] = 1;
                    interpPhi[n] = 0;
                }else{   
                    gsl_spline_init(splinex, &x[__u], &z_abs[__v][__u], INTERP_SIZE);
                    interpA[n] = gsl_spline_eval(splinex, xx, NULL);
                    
                    gsl_spline_init(splinex, &x[__u], &z_phs[__v][__u], INTERP_SIZE);
                    interpPhi[n] = gsl_spline_eval(splinex, xx, NULL);
                }   
            }

            if ((yy < y[0]) || (yy > y[M - 1])) {
                *A = 1;
                *phi = 0;
            } else {
                gsl_spline_init(spliney, &y[v + INTERP_LLIM], &interpA[0], INTERP_SIZE);
                *A = gsl_spline_eval(spliney, yy, NULL);
                gsl_spline_init(spliney, &y[v + INTERP_LLIM], &interpPhi[0], INTERP_SIZE);
                *phi = gsl_spline_eval(spliney, yy, NULL);
            }
        }
    }
    
    // as the refractive index  of the medium either side of a mirror could
    // change in the simulation we must apply the refractive index terms of the
    // phase here
    switch (knum) {
        case K11:
            *phi *= nr1;
            break;
        case K12:
        case K21:
            // if the phase map is transmission only we consider the surface profile
            // to be flat thus no difference because of nr1 and nr2 should be seen
            if(!map->phaseIsOnlyTransmission)
                *phi *= (nr2-nr1);
            break;
        case K22:
            *phi *= nr2;
            break;
    }
    
    //if we are only applying a transmission phase then the phase should be same
    //whether we go from n1->n2 or n2->n1 as the T phase represents a "thickness"
    //of the component not a surface displacement so phase sign flip not needed.
    if(!map->phaseIsOnlyTransmission){
        // From the diagram above we see that z' needs to be inverted
        // or that we just invert the sign of the phase, same thing
        if ((knum == K12) || (knum == K22))
            (*phi) *= -1;
    }
    
    if (INTERP_METHOD != NEAREST) {
        gsl_spline_free(splinex);
        gsl_spline_free(spliney);
    }

    // if a interpolation output data file is open then we should output some 
    // data to it.
    if (ipfile != NULL){
        //mutex needed here to stop threads fighting over writing to the file
        pthread_mutex_lock(&mxinterpfile);
        
        switch(knum){
            case K11: fprintf(ipfile,"11 ");break;
            case K12: fprintf(ipfile,"12 ");break;
            case K21: fprintf(ipfile,"21 ");break;
            case K22: fprintf(ipfile,"22 ");break;
        }
        
        fprintf(ipfile,"%.15e %.15e %.15e %.15e\n",xx,yy,*A,*phi);
        
        fflush(ipfile); // need this flush as large files seem to get corrupted otherwise
        pthread_mutex_unlock(&mxinterpfile);
    }
}

// This initialises a map to be used with GSL interpolation
// The gsl_map_interpol_free must also be called to free memory

void gsl_map_interpol_init_mm(surface_merged_map_t *map) {

    map->x = (double*) malloc(sizeof (double) *map->cols);
    map->y = (double*) malloc(sizeof (double) *map->rows);

    if (map->x == NULL || map->y == NULL) {
        bug_error("Could not allocate enough memory for interpolation routine. rows=%d cols=%d map=%s"
                , map->rows, map->cols, map->name);
    }

    int n;

    for (n = 0; n < map->cols; n++) {
        map->x[n] = (n - map->x0) * map->xstep;
    }

    for (n = 0; n < map->rows; n++) {
        map->y[n] = (n - map->y0) * map->ystep;
    }
}

void gsl_map_interpol_free_mm(surface_merged_map_t *map) {
    free(map->x);
    free(map->y);

    map->x = NULL;
    map->y = NULL;
}

/**
 * Merged map alternative
 * 
 * @param ndim
 * @param xx
 * @param ncomp
 * @param ff
 * @param userdata
 * @return 
 */
static int integrand_cuba_serial(const int *ndim, const double xx[], const int *ncomp, double ff[], void *userdata) {
    assert(*ndim == 2);
    assert(*ncomp == 2);
    // get rid of compiler warnings
    (void) ndim;
    (void) ncomp;

    knm_map_integration_params_t p = *((struct knm_map_integration_params*) userdata);

#if DDEBUG
    if (p.knum == K11 && !p.K11_calc) bug_error("Should not be computing K11\n");
    if (p.knum == K12 && !p.K12_calc) bug_error("Should not be computing K12\n");
    if (p.knum == K21 && !p.K21_calc) bug_error("Should not be computing K21\n");
    if (p.knum == K22 && !p.K22_calc) bug_error("Should not be computing K22\n");
#endif

    assert(p.merged_map != NULL);

    double x, y;
    double r = (p.xmin[0] + xx[0] * (p.xmax[0] - p.xmin[0]));
    double t = (p.xmin[1] + xx[1] * (p.xmax[1] - p.xmin[1]));

    // if polar coordinates have been used then we need to convert
    // r and theta - which are x and y respectively from above
    // - into Cartesian coordinates
    if (p.polar_limits_used) {
        x = r * cos(t);
        y = r * sin(t);
    } else {
        x = r;
        y = t;
    }

    complex_t u1, u2, f;
    double A, phi;

    if (p.usingMap)
        interpolate_merged_map(p.merged_map, p.knum, x, y, &A, &phi, p.nr1, p.nr2, p.mirror->angle);
    else {
        A = 1;
        phi = 0;
    }

    // get the hermite-gaussian base that we are coupling from and to
    switch (p.knum) {
        case K11:
            u1 = u_nm_fast_x(x, y, p.n1, p.m1, p.acc_11_nr1_1);
            u2 = u_nm_fast_x(x, y, p.n2, p.m2, p.acc_11_nr1_2);
            break;
        case K12:
            u1 = u_nm_fast_x(x, y, p.n1, p.m1, p.acc_12_nr1_1);
            u2 = u_nm_fast_x(x, y, p.n2, p.m2, p.acc_12_nr2_2);
            break;
        case K21:
            u1 = u_nm_fast_x(x, y, p.n1, p.m1, p.acc_21_nr2_1);
            u2 = u_nm_fast_x(x, y, p.n2, p.m2, p.acc_21_nr1_2);
            break;
        case K22:
            u1 = u_nm_fast_x(x, y, p.n1, p.m1, p.acc_22_nr2_1);
            u2 = u_nm_fast_x(x, y, p.n2, p.m2, p.acc_22_nr2_2);
            break;
        default:
            bug_error("Unacceptable knum input");
            return -1;
    }

    f = z_by_zc(z_by_xphr(u1, A, phi), u2);

    switch (p.knum) {
        case K11:
            f = z_by_z(f, p.znm1c1);
            break;
        case K12:
            f = z_by_z(f, p.znm1c2);
            break;
        case K21:
            f = z_by_z(f, p.znm2c1);
            break;
        case K22:
            f = z_by_z(f, p.znm2c2);
            break;
    }

    ff[0] = f.re;
    ff[1] = f.im;

    // need to apply extra r term if integrating in polar coordinates
    if (p.polar_limits_used) {
        ff[0] *= r;
        ff[1] *= r;
    }

    return 0;
}

static int integrand_cuba_para(const int *ndim, const double xx[], const int *ncomp, double ff[], void *userdata) {
    assert(*ndim == 2);
    assert(*ncomp > 0);

    (void) ndim;
    (void) ncomp;

    knm_map_integration_params_t p = *((knm_map_integration_params_t*) userdata);

    double x, y;
    double r = (p.xmin[0] + xx[0] * (p.xmax[0] - p.xmin[0]));
    double t = (p.xmin[1] + xx[1] * (p.xmax[1] - p.xmin[1]));

    // if polar coordinates have been used then we need to convert
    // r and theta - which are x and y respectively from above
    // - into Cartesian coordinates
    if (p.polar_limits_used) {
        x = r * cos(t);
        y = r * sin(t);
    } else {
        x = r;
        y = t;
    }

    double A11, phi11;
    double A12, phi12;
    double A21, phi21;
    double A22, phi22;

    complex_t u11_1, u11_2, u12_1, u12_2, u21_1, u21_2, u22_1, u22_2;

    if (p.usingMap) {
        // Unless I am missing something it now appears we have to do alot more 
        // interpolating using the merged_map, as we have to go over amplitude
        // and reflection for each KNM.
        if (p.K11_calc) interpolate_merged_map(p.merged_map, K11, x, y, &A11, &phi11, p.nr1, p.nr2, p.mirror->angle);
        if (p.K12_calc) interpolate_merged_map(p.merged_map, K12, x, y, &A12, &phi12, p.nr1, p.nr2, p.mirror->angle);
        if (p.K21_calc) interpolate_merged_map(p.merged_map, K21, x, y, &A21, &phi21, p.nr1, p.nr2, p.mirror->angle);
        if (p.K22_calc) interpolate_merged_map(p.merged_map, K22, x, y, &A22, &phi22, p.nr1, p.nr2, p.mirror->angle);
    } else {
        A11 = A12 = A21 = A22 = 1.0;
        phi11 = phi12 = phi21 = phi22 = 0.0;
    }

    int i = 0; // keeps track of which integral we are calculating

    if (p.K11_calc) {
        u11_1 = u_nm_fast_x(x, y, p.n1, p.m1, p.acc_11_nr1_1);
        u11_2 = u_nm_fast_x(x, y, p.n2, p.m2, p.acc_11_nr1_2);
        complex_t f11 = z_by_z(z_by_zc(z_by_xphr(u11_1, A11, phi11), u11_2), p.znm1c1);

        ff[i++] = f11.re;
        ff[i++] = f11.im;
    }

    if (p.K12_calc) {
        u12_1 = u_nm_fast_x(x, y, p.n1, p.m1, p.acc_12_nr1_1);
        u12_2 = u_nm_fast_x(x, y, p.n2, p.m2, p.acc_12_nr2_2);
        complex_t f12 = z_by_z(z_by_zc(z_by_xphr(u12_1, A12, phi12), u12_2), p.znm1c2);

        ff[i++] = f12.re;
        ff[i++] = f12.im;
    }

    if (p.K21_calc) {
        u21_2 = u_nm_fast_x(x, y, p.n2, p.m2, p.acc_21_nr1_2);
        u21_1 = u_nm_fast_x(x, y, p.n1, p.m1, p.acc_21_nr2_1);
        complex_t f21 = z_by_z(z_by_zc(z_by_xphr(u21_1, A21, phi21), u21_2), p.znm2c1);

        ff[i++] = f21.re;
        ff[i++] = f21.im;
    }

    if (p.K22_calc) {
        u22_1 = u_nm_fast_x(x, y, p.n1, p.m1, p.acc_22_nr2_1);
        u22_2 = u_nm_fast_x(x, y, p.n2, p.m2, p.acc_22_nr2_2);
        complex_t f22 = z_by_z(z_by_zc(z_by_xphr(u22_1, A22, phi22), u22_2), p.znm2c2);

        ff[i++] = f22.re;
        ff[i++] = f22.im;
    }

    // need to apply extra r term if integrating in polar coordinates
    if (p.polar_limits_used) {
        for (i = 0; i<*ncomp; i++) {
            ff[i] *= r;
        }
    }

    return 0;
}

void do_para_cuhre_map_int(knm_map_integration_params_t *p, complex_t *result11, complex_t *result12,
        complex_t *result21, complex_t *result22) {

    int NCOMP = 0;
    int nregions, neval, fail;
    double integral[8] = {0}, error[8] = {0}, prob[8] = {0};

    // determine how many integrals we need to calculate
    if (p->K11_calc) NCOMP += 2;
    if (p->K12_calc) NCOMP += 2;
    if (p->K21_calc) NCOMP += 2;
    if (p->K22_calc) NCOMP += 2;

    
    if (NCOMP != 0) {

        get_int_limit(p, 0); //calculate the map integration limits
            
        Cuhre(2, NCOMP, integrand_cuba_para, p,
                0, init.abserr, 0,
                0, init.maxintcuba, 13,
                &nregions, &neval, &fail, integral, error, prob);

        if (inter.debug && neval >= init.maxintcuba) {
            warn("maxintcuba reached for %d %d->%d %d (evals: %d) Errors: \n", p->n1, p->m1);
            int i=0;

            if (p->K11_calc) {
                printf("K11: %.3g+-i", error[i++]);
                printf("%.3g ", error[i++]);
            }
            if (p->K12_calc) {
                printf("K12: %.3g+-i", error[i++]);
                printf("%.3g ", error[i++]);
            }
            if (p->K21_calc) {
                printf("K21: %.3g+-i", error[i++]);
                printf("%.3g ", error[i++]);
            }
            if (p->K22_calc) {
                printf("K22: %.3g+-i", error[i++]);
                printf("%.3g ", error[i++]);
            }
            printf("\n");
        }

        if (inter.debug && p->merged_map->show_kmn_neval && !options.quiet)
            printf("   - %i%i->%i%i calculated with %i evaluations\n", p->n1, p->m1, p->n2, p->m2, neval);
    }

    //This variable is needed for the INT_VAL macro
    double J = (p->xmax[0] - p->xmin[0]) * (p->xmax[1] - p->xmin[1]);
    int i = 0;

    if (p->K11_calc) {
        result11->re = INT_VAL(i++);
        result11->im = INT_VAL(i++);
    } else {
        *result11 = complex_0;// DELTA_FUNC(p->n1, p->n2, p->m1, p->m2);
    }

    if (p->K12_calc) {
        result12->re = INT_VAL(i++);
        result12->im = INT_VAL(i++);
    } else {
        *result12 = complex_0;//DELTA_FUNC(p->n1, p->n2, p->m1, p->m2);
    }

    if (p->K21_calc) {
        result21->re = INT_VAL(i++);
        result21->im = INT_VAL(i++);
    } else {
        *result21 = complex_0;//DELTA_FUNC(p->n1, p->n2, p->m1, p->m2);
    }

    if (p->K22_calc) {
        result22->re = INT_VAL(i++);
        result22->im = INT_VAL(i++);
    } else {
        *result22 = complex_0;//DELTA_FUNC(p->n1, p->n2, p->m1, p->m2);
    }
}

void do_serial_cuhre_map_int(knm_map_integration_params_t *p, complex_t *result11, complex_t *result12,
        complex_t *result21, complex_t *result22) {
    const int NCOMP = 2;
    int nregions, neval, fail;
    double integral[NCOMP], error[NCOMP], prob[NCOMP];
    int tneval = 0;

    //Trying to do each integral separately to see which ones are causing problems
    //need to switch back to doing them all at once later
    int i;
    for (i = K11; i <= K22; i++) {
        //check if we are meant to be calculating this KMN
        if ((i == K11 && !p->K11_calc) || (i == K22 && !p->K22_calc)
                || (i == K12 && !p->K12_calc) || (i == K21 && !p->K21_calc)) {
            // if we are not doing this KMN then save the result as 1, this
            // just makes sure something sensible is in the KMN matrix
            switch (i) {
                case K11:
                    *result11 = complex_0;//DELTA_FUNC(p->n1, p->n2, p->m1, p->m2);
                    break;
                case K12:
                    *result12 = complex_0;//DELTA_FUNC(p->n1, p->n2, p->m1, p->m2);
                    break;
                case K21:
                    *result21 = complex_0;//DELTA_FUNC(p->n1, p->n2, p->m1, p->m2);
                    break;
                case K22:
                    *result22 = complex_0;//DELTA_FUNC(p->n1, p->n2, p->m1, p->m2);
                    break;
            }
        } else {
            // We are meant to be calculating KNM
            p->knum = i;
            
            get_int_limit(p, i); //calculate the map integration limits
            
            Cuhre(2, NCOMP, integrand_cuba_serial, p,
                    0, init.abserr, 0,
                    0, init.maxintcuba, 13,
                    &nregions, &neval, &fail, integral, error, prob);

            tneval += neval;

            if (inter.debug && neval > init.maxintcuba) {
                char kstr[4];

                switch (i) {
                    case K11:
                        strcpy(kstr, "k11"); 
                        break;
                    case K12:
                        strcpy(kstr, "k12"); 
                        break;
                    case K21:
                        strcpy(kstr, "k21"); 
                        break;
                    case K22:
                        strcpy(kstr, "k22"); 
                        break;
                }

                warn("maxintcuba reached for %d %d->%d %d (evals: %d) Errors: %s: %.3g+-i%.3g\n", p->n1, p->m1
                        , p->n2, p->m2, neval
                        , kstr, error[0], error[1]);
            }

            //This variable is needed for the INT_VAL macro
            double J = (p->xmax[0] - p->xmin[0]) * (p->xmax[1] - p->xmin[1]);

            switch (i) {
                case K11:
                    result11->re = INT_VAL(0);
                    result11->im = INT_VAL(1);
                    break;
                case K12:
                    result12->re = INT_VAL(0);
                    result12->im = INT_VAL(1);
                    break;
                case K21:
                    result21->re = INT_VAL(0);
                    result21->im = INT_VAL(1);
                    break;
                case K22:
                    result22->re = INT_VAL(0);
                    result22->im = INT_VAL(1);
                    break;
            }
        }
    }

    if (inter.debug && p->merged_map->show_kmn_neval && !options.quiet)
        printf("   - %i%i->%i%i calculating with %i evaluations\n", p->n1, p->m1, p->n2, p->m2, tneval);
}

/*
void do_riemann_sum_old_int(knm_map_integration_params_t *userdata, complex_t *result11, complex_t *result12,
        complex_t *result21, complex_t *result22) {
    
    bug_error("Old Riemann not available");
    
    knm_map_integration_params_t p = *((knm_map_integration_params_t*) userdata);
    surface_merged_map_t *map = p.merged_map;
    double x, y, x_store;
    complex_t u11, u12, u21, u22;
    int i, j;
    int _i = map->cols - 1;

    double A11, phi11;
    double A12, phi12;
    double A21, phi21;
    double A22, phi22;

    int intops = 0;

    for (i = 0; i < map->cols; i++) {
        x_store = x = (i - map->x0) * map->xstep;

        for (j = 0; j < map->rows; j++) {
            y = (j - map->y0) * map->ystep;
            x = x_store;

            if (p.usingMap) {
                A11 = map->r_abs[i][j];
                A12 = map->t_abs[i][j];
                A21 = map->t_abs[_i - i][j];
                A22 = map->r_abs[_i - i][j];
                phi11 = map->r_phs[i][j];
                phi12 = map->t_phs[i][j];
                phi21 = -map->t_phs[_i - i][j];
                phi22 = -map->r_phs[_i - i][j];
            } else {
                A11 = A12 = A21 = A22 = 1.0;
                phi11 = phi12 = phi21 = phi22 = 0.0;
            }

            u11 = u_nm(p.n1, p.m1, p.qx1, p.qy1, x, y, p.nr1);
            u12 = u_nm(p.n2, p.m2, p.qx1, p.qy1, x, y, p.nr1);
            u21 = u_nm(p.n1, p.m1, p.qx2, p.qy2, -1.0 * x, y, p.nr2);
            u22 = u_nm(p.n2, p.m2, p.qx2, p.qy2, -1.0 * x, y, p.nr2);

            *result11 = z_pl_z(*result11, z_by_zc(z_by_xphr(u11, A11, phi11), u12));
            *result12 = z_pl_z(*result12, z_by_zc(z_by_xphr(u11, A12, phi12), u22));
            *result21 = z_pl_z(*result21, z_by_zc(z_by_xphr(u21, A21, phi21), u12));
            *result22 = z_pl_z(*result22, z_by_zc(z_by_xphr(u21, A22, phi22), u22));

            intops++;
        }
    }

    if (inter.debug && p.merged_map->show_kmn_neval && !options.quiet)
        printf("   - %i%i->%i%i calculating with %i evaluations\n", p.n1, p.m1, p.n2, p.m2, intops);

    *result11 = z_by_x(*result11, map->xstep * map->ystep);
    *result12 = z_by_x(*result12, map->xstep * map->ystep);
    *result21 = z_by_x(*result21, map->xstep * map->ystep);
    *result22 = z_by_x(*result22, map->xstep * map->ystep);
  }
 */



void do_riemann_sum_new_int(knm_map_integration_params_t *userdata, complex_t *result11, complex_t *result12,
        complex_t *result21, complex_t *result22) {

    knm_map_integration_params_t p = *((knm_map_integration_params_t*) userdata);
    surface_merged_map_t *map = p.merged_map;
    double x, y;
    complex_t u11_1 = complex_0;
    complex_t u12_1 = complex_0;
    complex_t u21_1 = complex_0;
    complex_t u22_1 = complex_0;
    complex_t u11_2 = complex_0;
    complex_t u12_2 = complex_0;
    complex_t u21_2 = complex_0;
    complex_t u22_2 = complex_0;
    
    int i, j;
    
    get_int_limit(&p, 0); //calculate the map integration limits

    int ndx = (int) ((p.xmax[0] - p.xmin[0]) / map->xstep);
    int ndy = (int) ((p.xmax[1] - p.xmin[1]) / map->ystep);

    double dx = (p.xmax[0] - p.xmin[0]) / (double) ndx;
    double dy = (p.xmax[1] - p.xmin[1]) / (double) ndy;
    double da = dx*dy;

    double A11, phi11;
    double A12, phi12;
    double A21, phi21;
    double A22, phi22;

    if (ndx == 0 || ndy == 0) {
        gerror("Riemann sum: integration limits are smaller than the step size. Please use a map with a higher resolution.\n");
    }
    
    if (false & inter.debug && !options.quiet) {
        printf(" Riemann sum statistics: dx=%e dy=%e ndx=%d ndy=%d da=%e\n", dx, dy, ndx, ndy, da);
        printf("                         x range: %e %e  y range: %e %e\n", p.xmin[0], p.xmax[0], p.xmin[1], p.xmax[1]);
        printf("                         x step: %e  y step: %e\n", map->xstep, map->ystep);
    }

    int intops = 0;

    for (i = 0; i <= ndx; i++) {
        x = p.xmin[0] + i * dx;

        for (j = 0; j <= ndy; j++) {
            y = p.xmin[1] + j * dy;

            if (p.usingMap) {
                // Unless I am missing something it now appears we have to do alot more 
                // interpolating using the merged_map, as we have to go over amplitude
                // and reflection for each KNM.
                if (p.K11_calc) interpolate_merged_map(p.merged_map, K11, x, y, &A11, &phi11, p.nr1, p.nr2, p.mirror->angle);
                if (p.K12_calc) interpolate_merged_map(p.merged_map, K12, x, y, &A12, &phi12, p.nr1, p.nr2, p.mirror->angle);
                if (p.K21_calc) interpolate_merged_map(p.merged_map, K21, x, y, &A21, &phi21, p.nr1, p.nr2, p.mirror->angle);
                if (p.K22_calc) interpolate_merged_map(p.merged_map, K22, x, y, &A22, &phi22, p.nr1, p.nr2, p.mirror->angle);
            } else {
                A11 = A12 = A21 = A22 = 1.0;
                phi11 = phi12 = phi21 = phi22 = 0.0;
            }

            if (p.K11_calc) {
                u11_1 = u_nm_fast_x(x, y, p.n1, p.m1, p.acc_11_nr1_1);
                u11_2 = u_nm_fast_x(x, y, p.n2, p.m2, p.acc_11_nr1_2);
                *result11 = z_pl_z(*result11, z_by_z(z_by_zc(z_by_xphr(u11_1, A11, phi11), u11_2), p.znm1c1));
            }    
            
            if (p.K12_calc){                
                u12_1 = u_nm_fast_x(x, y, p.n1, p.m1, p.acc_12_nr1_1);
                u12_2 = u_nm_fast_x(x, y, p.n2, p.m2, p.acc_12_nr2_2);
                *result12 = z_pl_z(*result12, z_by_z(z_by_zc(z_by_xphr(u12_1, A12, phi12), u12_2), p.znm1c2));
            }
            
            if (p.K21_calc) {
                u21_1 = u_nm_fast_x(x, y, p.n1, p.m1, p.acc_21_nr2_1);
                u21_2 = u_nm_fast_x(x, y, p.n2, p.m2, p.acc_21_nr1_2);
                *result21 = z_pl_z(*result21, z_by_z(z_by_zc(z_by_xphr(u21_1, A21, phi21), u21_2), p.znm2c1));
            }
            
            if (p.K22_calc) {
                u22_1 = u_nm_fast_x(x, y, p.n1, p.m1, p.acc_22_nr2_1);
                u22_2 = u_nm_fast_x(x, y, p.n2, p.m2, p.acc_22_nr2_2);
                *result22 = z_pl_z(*result22, z_by_z(z_by_zc(z_by_xphr(u22_1, A22, phi22), u22_2), p.znm2c2));
            }

            intops++;
        }
    }

    if (inter.debug && p.merged_map->show_kmn_neval && !options.quiet)
        printf("   - %i%i->%i%i calculating with %i evaluations\n", p.n1, p.m1, p.n2, p.m2, intops);

    complex_t delta = complex_0;//DELTA_FUNC(p.n1, p.n2, p.m1, p.m2);

    //if we are not calculating return 1 just to ensure something sensible is there
    if (p.K11_calc) {
        *result11 = z_by_x(*result11, da);
    } else {
        *result11 = delta;
    }
    if (p.K12_calc) {
        *result12 = z_by_x(*result12, da);
    } else {
        *result12 = delta;
    };
    if (p.K21_calc) {
        *result21 = z_by_x(*result21, da);
    } else {
        *result21 = delta;
    };
    if (p.K22_calc) {
        *result22 = z_by_x(*result22, da);
    } else {
        *result22 = delta;
    };
}

void calc_limit_ws(double *wx1,double *wx2,double *wy1,double *wy2, KNM_MIRROR_NODE_DIRECTION_t knum, knm_map_integration_params_t *p){
    mirror_kmn_q_t *kq = p->kmnq;
    
    switch(knum){
        case K11:
            *wx1 = w_size(kq->qxt1_11, p->nr1);
            *wx2 = w_size(kq->qyt1_11, p->nr1);            
            *wy1 = w_size(kq->qxt2_11, p->nr1);
            *wy2 = w_size(kq->qyt2_11, p->nr1);
            break;
        case K12:
            *wx1 = w_size(kq->qxt1_12, p->nr1);
            *wx2 = w_size(kq->qyt1_12, p->nr2);            
            *wy1 = w_size(kq->qxt2_12, p->nr1);
            *wy2 = w_size(kq->qyt2_12, p->nr2);
            break;
        case K21:
            *wx1 = w_size(kq->qxt1_21, p->nr2);
            *wx2 = w_size(kq->qyt1_21, p->nr1);            
            *wy1 = w_size(kq->qxt2_21, p->nr2);
            *wy2 = w_size(kq->qyt2_21, p->nr1);
            break;
        case K22:
            *wx1 = w_size(kq->qxt1_22, p->nr2);
            *wx2 = w_size(kq->qyt1_22, p->nr2);            
            *wy1 = w_size(kq->qxt2_22, p->nr2);
            *wy2 = w_size(kq->qyt2_22, p->nr2);
            break;
    }
}

// Scale the integration limits slightly depending on the beam mode
// as the beam tends to get larger at higher modes. If a mirror aperture
// has been specified then polar coordinates are used for the integration
// so r and theta limits are set.
//
// If the Riemann integration method is being used then cartesian coords
// are used by default. Polar limits are only set when using the new integration
// methods Cuba parallel and serial.
//
// if the beam is much smaller than the aperture, cartesian coordinates are used.
void get_int_limit(knm_map_integration_params_t *p, int knum) {
    if (p->aperture_radius < 0)
        bug_error("get_int_limit(*p): Mirror aperture cannot be negative\n");

    double wx1=0,wx2=0,wy1=0,wy2=0;
    double _wx1,_wx2,_wy1,_wy2;
    
    switch(knum){
        case 0:
            // used by cuba parallel, find the largest w values out of all the
            // different 
            if(p->K11_calc){
                calc_limit_ws(&_wx1,&_wx2,&_wy1,&_wy2,K11,p);
                wx1 = max(wx1,_wx1);
                wx2 = max(wx2,_wx2);
                wy1 = max(wy1,_wy1);
                wy2 = max(wy2,_wy2);                
            }
            if(p->K12_calc){
                calc_limit_ws(&_wx1,&_wx2,&_wy1,&_wy2,K12,p);
                wx1 = max(wx1,_wx1);
                wx2 = max(wx2,_wx2);
                wy1 = max(wy1,_wy1);
                wy2 = max(wy2,_wy2);                
            }
            if(p->K21_calc){
                calc_limit_ws(&_wx1,&_wx2,&_wy1,&_wy2,K21,p);
                wx1 = max(wx1,_wx1);
                wx2 = max(wx2,_wx2);
                wy1 = max(wy1,_wy1);
                wy2 = max(wy2,_wy2);                
            }
            if(p->K22_calc){
                calc_limit_ws(&_wx1,&_wx2,&_wy1,&_wy2,K22,p);
                wx1 = max(wx1,_wx1);
                wx2 = max(wx2,_wx2);
                wy1 = max(wy1,_wy1);
                wy2 = max(wy2,_wy2);                
            }
            break;
        case K11:
            calc_limit_ws(&wx1,&wx2,&wy1,&wy2,K11,p);
            break;
        case K12:
            calc_limit_ws(&wx1,&wx2,&wy1,&wy2,K12,p);
            break;
        case K21:
            calc_limit_ws(&wx1,&wx2,&wy1,&wy2,K21,p);
            break;
        case K22:
            calc_limit_ws(&wx1,&wx2,&wy1,&wy2,K22,p);
            break;
        default:
            bug_error("Cannot handle a knum input of %i\n",knum);
    }
    
    p->xmax[0] = 5 * max(sqrt(p->n1 + 0.5) * wx1, sqrt(p->n2 + 0.5) * wx2);
    p->xmin[0] = -p->xmax[0];
    p->xmax[1] = 5 * max(sqrt(p->m1 + 0.5) * wy1, sqrt(p->m2 + 0.5) * wy2);
    p->xmin[1] = -p->xmax[1];

    double r = p->aperture_radius;
    
    // if using the Riemann just stick to the Cartesian coordinates
    // if r == 0 then the mirror is infinitely big
    // if r > max(...) then the beam size is smaller than the aperture so Cartesian will do
    // also check if we should be integrating to calculate the aperture with inter.knm, if we're not
    // then there is no need to use polar coordinates as polar coordinates are only used for aperture
    // calculations.
    if ((p->merged_map->integration_method == RIEMANN_SUM_OLD 
            || p->merged_map->integration_method == RIEMANN_SUM_NEW)
            || !(inter.kmn & INT_APERTURE)
            || ((r == 0) || (r > min(p->xmax[0], p->xmax[1])))) {
        p->polar_limits_used = false;
    } else {
        //[0] is r, [1] is theta
        p->xmin[0] = 0;
        p->xmin[1] = 0;
        p->xmax[0] = r;
        p->xmax[1] = TWOPI;
        p->polar_limits_used = true;
    }
    
    unsigned int *usermessages = &(p->merged_map->usermessages);
    
    if(inter.debug && !options.quiet && !(*usermessages & USING_POLAR)){
        if(p->polar_limits_used)
           message(" * Polar coordinates used to integrate map %s\n",p->merged_map->name);
        else
           message(" * Cartesian coordinates used to integrate map %s\n",p->merged_map->name);
            
        *usermessages = *usermessages | USING_POLAR;
    }
    
    if(p->usingMap && !(*usermessages & MAP_TOO_SMALL)){        
        surface_merged_map_t *map = p->merged_map;
        // here we check to see if the map size is smaller than the limits. if so
        // you could probably do with a bigger map
        if(p->polar_limits_used){
            if((map->cols * map->xstep < abs(p->xmax[0] - p->xmin[0])) 
                    || (map->rows * map->ystep < abs(p->xmax[0] - p->xmin[0]))){
                warn("The map %s is smaller than the dimensions of the integration limits\n"
                     "   (r range=%e, phi range=%e). You should use a bigger map\n"
                     "   or a smaller beam for more accurate results.\n"
                        , p->merged_map->filename
                        , p->xmax[0] - p->xmin[0]
                        , p->xmax[1] - p->xmin[1]);
                
                *usermessages = *usermessages | MAP_TOO_SMALL;
            }
        }else{
            if( ((map->cols-map->x0) * map->xstep < p->xmax[0]) 
             || ((map->x0-map->cols) * map->xstep > p->xmin[0]) 
             || ((map->rows-map->x0) * map->ystep < p->xmax[1]) 
             || ((map->x0-map->rows) * map->ystep > p->xmin[1]) ){
                warn("The map %s is smaller than the dimensions of the integration limits\n"
                     "   (x range=%e, y range=%e, map width=%e map height=%e). You should\n"
                     "   use a bigger map or a smaller beam for more accurate results.\n"
                        , p->merged_map->filename
                        , p->xmax[0] - p->xmin[0]
                        , p->xmax[1] - p->xmin[1]
                        , map->cols * map->xstep
                        , map->rows * map->ystep);
            
                
                *usermessages = *usermessages | MAP_TOO_SMALL;
            }
        }
    }
    
    if(p->usingMap && !(*usermessages & MAP_RES_TOO_LOW )
            && p->merged_map->integration_method == RIEMANN_SUM_NEW){
        if((min(wx1,wx2)/p->merged_map->xstep < 5.0) || (min(wy1,wy2)/p->merged_map->ystep < 5.0) ){
            warn("The map %s might not have a high enough resolution. Currently\n"
              "   the beam size samples less than 10 points on the map. This can\n"
              "   cause large errors when using higher orders.\n", p->merged_map->name);
            *usermessages |= MAP_RES_TOO_LOW;
        }
    }
}
    

void compute_kmn_integral(mirror_t *mirror, complex_t qx1, complex_t qy1,
        complex_t qx2, complex_t qy2, double nr1, double nr2, int integrating, int mismatch, int astigmatism) {
    
    surface_merged_map_t *map = &(mirror->map_merged);
    mirror_kmn_t *kmn = &(mirror->map_merged.kmn);

    // A quick check is needed to determine if any maps have actually been applied.
    // if so and there is no other integrating needed we need not continue.
    // Otherwise we need to allocate some memory to store the results in.
    if (map->noMapPresent && !integrating) {
        return;
    } else
        if (!IS_MIRROR_KMN_ALLOCD(kmn))
            mirror_kmn_alloc(kmn);

    // We now need to determine whether we actually need to compute the integral
    // again. As this method can be called multiple times if one of the parameters
    // is varied by the x-axis.
    bool recompute = false;

    recompute = recompute | (map->_nr1 != nr1);
    recompute = recompute | (map->_nr2 != nr2);

    recompute = recompute | (map->_angle != mirror->angle);
    
    recompute = recompute | (map->_maxtem < inter.tem);

    recompute = recompute | !ceq(qx1, map->_qx1);
    recompute = recompute | !ceq(qx2, map->_qx2);
    recompute = recompute | !ceq(qy1, map->_qy1);
    recompute = recompute | !ceq(qy2, map->_qy2);
    
    recompute = recompute | (init.abserr != map->_abserr);
    recompute = recompute | (init.relerr != map->_relerr);
    
    recompute = recompute | (init.lambda != map->_lambda);
    
    // check the current methods against the previously used methods
    recompute = recompute | (map->interpolation_method != map->_intermeth);
    recompute = recompute | (map->interpolation_size != map->_intersize);
    recompute = recompute | (map->integration_method != map->_intmeth);

    recompute = recompute | ((inter.kmn & INT_APERTURE) && (mirror->r_aperture != map->_r_aperture));

    recompute = recompute | ((inter.kmn & INT_BAYER_HELMS) && (map->_xbeta != mirror->beta_x));
    recompute = recompute | ((inter.kmn & INT_BAYER_HELMS) && (map->_ybeta != mirror->beta_y));

    int start_tem = 0;

    if (!recompute) {
        // So far everything should be the same, the last comparison to make
        // is between the maximum number of calculated modes, if more are requested
        // start off where we were, if not we have the previous ones already
        if (map->_maxtem >= inter.tem) {
            if (inter.debug && !options.quiet)message(" * Did not need to integrate any knm's for mirror %s\n", mirror->name);
            return;
        } else {
            start_tem = map->_maxtem + 1; // start from the next maxtem
            if (inter.debug && !options.quiet)message(" * Starting from tem %i on mirror %s\n", start_tem, mirror->name);
        }
    }

    bool usingMap = !(map->noMapPresent);

    if (inter.debug && !options.quiet) {
       message(" * Effects integrated for mirror %s:\n", mirror->name);

        if (usingMap)message("    - Merged maps\n");
        if (integrating & 1)message("    - Bayer-Helms\n");
        if (integrating & 2)message("    - Aperture\n");
        if (integrating & 4)message("    - Curvature\n");

       message(" * Coupling coefficients to compute for %s:", mirror->name);
        if (mirror->K11_calc)message(" K11");
        if (mirror->K12_21_calc)message(" K12");
        if (mirror->K12_21_calc)message(" K21");
        if (mirror->K22_calc)message(" K22");
       message("\n");
    }

    // boolean value that states whether we should calculate kmn using knm without
    // having to do the whole integral and just by applying some phase factor
    int calc_knm_transpose = (init.calc_knm_transpose && (mismatch == 0) && (astigmatism == 0));

    if (inter.debug && calc_knm_transpose == 1 && !options.quiet && !(map->usermessages & KMN_TRANSPOSE)) {
       message(" * Calculating kmn matrix lower triangle using phase factor\n"
               "   and upper triangle. See kat.ini calc_knm_transpose variable.\n");
        map->usermessages |= KMN_TRANSPOSE;
    }

    int max_m, max_n;
    get_tem_modes_from_field_index(&max_n, &max_m, inter.num_fields - 1);
    max_m = max_n; // max m is no the m is also the max_n. e.g. we get n=1 m=2 for maxtem 2

    mirror_kmn_q_t *kq = &(mirror->kmn_q);
    // TODO This might not have to be called every time if the q's don't change
    
    // K11
    if (mirror->K11_calc) {
        u_nm_accel_get(acc_11_nr1_1, max_n, max_m, kq->qxt1_11, kq->qyt1_11, nr1);
        u_nm_accel_get(acc_11_nr1_2, max_n, max_m, kq->qxt2_11, kq->qyt2_11, nr1);
    }
    // K22
    if (mirror->K22_calc) {
        u_nm_accel_get(acc_22_nr2_1, max_n, max_m, kq->qxt1_22, kq->qyt1_22, nr2);
        u_nm_accel_get(acc_22_nr2_2, max_n, max_m, kq->qxt2_22, kq->qyt2_22, nr2);
    }
    // K12
    if (mirror->K12_21_calc) {
        u_nm_accel_get(acc_12_nr1_1, max_n, max_m, kq->qxt1_12, kq->qyt1_12, nr1);
        u_nm_accel_get(acc_12_nr2_2, max_n, max_m, kq->qxt2_12, kq->qyt2_12, nr2);
        // K21
        u_nm_accel_get(acc_21_nr2_1, max_n, max_m, kq->qxt1_21, kq->qyt1_21, nr2);
        u_nm_accel_get(acc_21_nr1_2, max_n, max_m, kq->qxt2_21, kq->qyt2_21, nr1);
    }

    knm_map_integration_params_t param;
    param.usingMap = usingMap;
    param.merged_map = map;
    param.mirror = mirror;
    param.nr1 = nr1;
    param.nr2 = nr2;
    param.acc_11_nr1_1 = (mirror->K11_calc) ? acc_11_nr1_1 : NULL;
    param.acc_11_nr1_2 = (mirror->K11_calc) ? acc_11_nr1_2 : NULL;
    param.acc_12_nr1_1 = (mirror->K12_21_calc) ? acc_12_nr1_1 : NULL;
    param.acc_12_nr2_2 = (mirror->K12_21_calc) ? acc_12_nr2_2 : NULL;
    param.acc_21_nr2_1 = (mirror->K12_21_calc) ? acc_21_nr2_1 : NULL;
    param.acc_21_nr1_2 = (mirror->K12_21_calc) ? acc_21_nr1_2 : NULL;
    param.acc_22_nr2_1 = (mirror->K22_calc) ? acc_22_nr2_1 : NULL;
    param.acc_22_nr2_2 = (mirror->K22_calc) ? acc_22_nr2_2 : NULL;
    param.kmnq = &(mirror->kmn_q);
    param.K11_calc = mirror->K11_calc;
    param.K12_calc = mirror->K12_21_calc;
    param.K21_calc = mirror->K12_21_calc;
    param.K22_calc = mirror->K22_calc;

    param.aperture_radius = mirror->r_aperture;
    param.betax = mirror->beta_x;
    param.betay = mirror->beta_y;
    param.integrating = integrating;

    complex_t result11, result12, result21, result22;

    result11 = complex_0;
    result12 = complex_0;
    result21 = complex_0;
    result22 = complex_0;

        //This sets the number of cores to use for cuba library
    if (mirror->map_merged.integration_method == CUBA_CUHRE_SERIAL 
            || mirror->map_merged.integration_method == CUBA_CUHRE_PARA) {
        char buf[ERR_MSG_LEN];
#ifdef __CYGWIN__
        sprintf(buf, "CUBACORES=1");
#else
        sprintf(buf, "CUBACORES=%d", init.cuba_numprocs);
#endif
        putenv(buf);
    }
    
    if (inter.debug && !options.quiet && !(map->usermessages & INT_METHODS)) {
        map->usermessages |= INT_METHODS;
        
       message(" * Mirror integration - %s\n", mirror->name);
       message("   - Using ");

        switch (map->integration_method) {
        case RIEMANN_SUM_NEW:
           message("Riemann integrator\n");
            break;
        case RIEMANN_SUM_OLD:
           message("old Riemann integrator\n");
            break;
        case CUBA_CUHRE_PARA:
           message("parallel cubature integrator\n");
            break;
        case CUBA_CUHRE_SERIAL:
           message("serial cubature integrator\n");
        }
        
       message("   - Using %i core%s for integration\n", init.cuba_numprocs, ((init.cuba_numprocs == 1) ? "" : "s"));
        
        if(usingMap){
           message("   - Using ");

            switch (map->interpolation_method) {
                case SPLINE:message("spline");
                    break;
                case LINEAR:message("linear");
                    break;
                case NEAREST:message("nearest-neighbour");
                    break;
            }

           message(" interpolation, kernel size %i\n",mirror->map_merged.interpolation_size);
        }
    }

    int num_coeffs = inter.num_fields * inter.num_fields;
    int current_coeff = 1;
    //fprintf(stderr, " * Progress integrating map %s: \n", map->name);

    time_t starttime = time(NULL);

    // if we need to use GSL interpolation we need to allocate some memory
    // and do some calculations beforehand, don't bother with this if using Riemann old
    // as it doesn't user interpolation
    if (usingMap && (map->integration_method != RIEMANN_SUM_OLD) && (map->interpolation_method != NEAREST)) {
        gsl_map_interpol_init_mm(param.merged_map);
    }

    int n, m, n1, m1, n2, m2;

    // Iterate over each mode to calculate k_n1,m1,n2,m2
    for (n = 0; n < inter.num_fields; n++) {
        for (m = 0; m < inter.num_fields; m++) {
            print_progress_and_time(num_coeffs, current_coeff, starttime);
            current_coeff++;

            // If we are using the knm to calc kmn then we do not need to bother
            // with doing all the integrating for any of the lower triangle.
            if (!calc_knm_transpose || (n >= m)) {

                //Transform linear mode index system into actual TEM_NM mode
                get_tem_modes_from_field_index(&n1, &m1, n);
                get_tem_modes_from_field_index(&n2, &m2, m);

                // There is no need to calculate kmn's that have been loaded
                // or calculated before
                if (n1 + m1 >= start_tem || n2 + m2 >= start_tem) {
                    param.m1 = m1;
                    param.m2 = m2;
                    param.n1 = n1;
                    param.n2 = n2;

                    // compute some constants that are n1,n2,m1,m2 dependant
                    if (param.K11_calc) param.znm1c1 = z_by_zc(z_by_z(acc_11_nr1_1->acc_n->prefac[n1], acc_11_nr1_1->acc_m->prefac[m1]),
                            z_by_z(acc_11_nr1_2->acc_n->prefac[n2], acc_11_nr1_2->acc_m->prefac[m2]));
                    if (param.K12_calc) param.znm1c2 = z_by_zc(z_by_z(acc_12_nr1_1->acc_n->prefac[n1], acc_12_nr1_1->acc_m->prefac[m1]),
                            z_by_z(acc_12_nr2_2->acc_n->prefac[n2], acc_12_nr2_2->acc_m->prefac[m2]));
                    if (param.K21_calc) param.znm2c1 = z_by_zc(z_by_z(acc_21_nr2_1->acc_n->prefac[n1], acc_21_nr2_1->acc_m->prefac[m1]),
                            z_by_z(acc_21_nr1_2->acc_n->prefac[n2], acc_21_nr1_2->acc_m->prefac[m2]));
                    if (param.K22_calc) param.znm2c2 = z_by_zc(z_by_z(acc_22_nr2_1->acc_n->prefac[n1], acc_22_nr2_1->acc_m->prefac[m1]),
                            z_by_z(acc_22_nr2_2->acc_n->prefac[n2], acc_22_nr2_2->acc_m->prefac[m2]));

                    if (map->save_interp_file){
                        if(inter.debug && !options.quiet && !(map->usermessages & INTERP_SAVE)){
                           message(" * Writing interpolated data for each kmn for mirror %s, switch off for many kmn!!!\n",mirror->name);
                            map->usermessages = map->usermessages | INTERP_SAVE;
                        }
                        
                        char ipfn[FILENAME_MAX];
                        sprintf(ipfn,"%s_%i%i_%i%i.interp",mirror->name,n1,m1,n2,m2);
                        open_file_to_write_ascii(ipfn, &ipfile);
                    }

                    switch (map->integration_method) {
                        case RIEMANN_SUM_NEW:
                            do_riemann_sum_new_int(&param, &result11, &result12, &result21, &result22);
                            break;
                        case RIEMANN_SUM_OLD:
                            bug_error("Old Riemann not available");                            
                            //do_riemann_sum_old_int(&param, &result11, &result12, &result21, &result22);
                            break;
                        case CUBA_CUHRE_SERIAL:
                            do_serial_cuhre_map_int(&param, &result11, &result12, &result21, &result22);
                            break;
                        case CUBA_CUHRE_PARA:
                            do_para_cuhre_map_int(&param, &result11, &result12, &result21, &result22);
                            break;
                        default:
                            bug_error("Could not handle a map integration of type %d", map->integration_method);
                    }

                    if (map->save_interp_file){
                        fflush(ipfile);
                        fclose(ipfile);
                    }

                    map->kmn.k11[n][m] = result11;
                    map->kmn.k12[n][m] = result12;
                    map->kmn.k21[n][m] = result21;
                    map->kmn.k22[n][m] = result22;

                    // If we have a mode matched beam we are able to calculate the transposed
                    // elements by applying a simple phase factor
                    if (calc_knm_transpose && !(n1 == n2 && m1 == m2)) {
                        // TODO what qx/qy parameter to use here the input or output one?
                        double phi = 2.0 * gouy(qx1) * (n2 - n1 + m2 - m1);
                        map->kmn.k11[m][n] = z_by_phr(result11, phi);
                        map->kmn.k12[m][n] = z_by_phr(result12, phi);
                        map->kmn.k21[m][n] = z_by_phr(result21, phi);
                        map->kmn.k22[m][n] = z_by_phr(result22, phi);
                    }

                    if (inter.debug & 64) {
                        message("%s: (%d %d) -> (%d %d) %s %s %s %s\n", map->name, n1, m1, n2, m2, complex_form(map->kmn.k11[n][m]), complex_form(map->kmn.k12[n][m]), complex_form(map->kmn.k21[n][m]), complex_form(map->kmn.k22[n][m]));
                    }
                }
            }
        }
    }

    // if interpolation was used then free the memory
    if (usingMap && (map->integration_method != RIEMANN_SUM_OLD)
            && (map->interpolation_method != NEAREST))
        gsl_map_interpol_free_mm(param.merged_map);

    // we now store the values used to calculate the current knm so later we can 
    // determine whether they need recalculating
    map->kmn_calculated = true;
    map->_maxtem = inter.tem;
    map->_lambda = init.lambda;
    map->_nr1 = nr1;
    map->_nr2 = nr2;
    map->_qx1 = qx1;
    map->_qy1 = qy1;
    map->_qx2 = qx2;
    map->_qy2 = qy2;
    map->_angle = mirror->angle;
    map->_r_aperture = mirror->r_aperture;
    map->_xbeta = mirror->beta_x;
    map->_ybeta = mirror->beta_y;
    map->_abserr = init.abserr;
    map->_relerr = init.relerr;
    map->_intermeth = map->interpolation_method;
    map->_intersize = map->interpolation_size;
    map->_intmeth = map->integration_method;
    
    if (map->save_to_file)
        write_mirror_kmn_file(mirror, false);

    if ((!(mirror->map_merged.noMapPresent) || integrating) && mirror->map_merged.save_kmn_matrices) {
        char buf[LINE_LEN];
        sprintf(buf, "%s_integrated", map->filename);
        write_kmn_to_matrix_file(buf, &(mirror->map_merged.kmn));
    }
}

void check_map_size(surface_map_t *map, complex_t qx1, complex_t qy1, complex_t qx2, complex_t qy2, double nr1, double nr2) {
    double w1, w2, r0x, r0y;
    double diam_beam, diam_map;

    // Some emprically derived typical sizes for beam integration
    w1 = w_size(qx1, nr1);
    w2 = w_size(qx2, nr2);
    r0x = max(sqrt(inter.tem + 0.5) * w1, sqrt(inter.tem + 0.5) * w2);

    w1 = w_size(qy1, nr1);
    w2 = w_size(qy2, nr2);
    r0y = max(sqrt(inter.tem + 0.5) * w1, sqrt(inter.tem + 0.5) * w2);

    diam_beam = 2 * max(r0x, r0y);
    //fprintf(stdout,"cols = %d, center =%g\n",map->cols, map->x0);
    r0x = min(abs(map->cols - (map->x0 + 0.5)), abs(map->x0 + 0.5)) * map->xstep;
    r0y = min(abs(map->cols - (map->y0 + 0.5)), abs(map->y0 + 0.5)) * map->ystep;
    diam_map = 2 * min(r0x, r0y);

    if (diam_map < 4 * diam_beam) {
        warn("map %s: map size (%g) is smaller than the recommended size (%g).\n",
                map->name, diam_map, 4 * diam_beam);
    }

    if (inter.debug & 64) {
        //      message("Estimated sizes: max beam diameter: %g m, min map length: %g m\n",diam_beam,diam_map);
        if (diam_map > 6 * diam_beam) {
            warn("map %s the map has a much larger physical size (%g) than the beam size (%g)\n",
                    map->name, diam_map, diam_beam);
        }
    }
}


/*
 * Local variables:
 * c-indentation-style: bsd
 * c-basic-offset: 2
 * indent-tabs-mode: nil
 * End:
 *
 * vim: tabstop=2 expandtab shiftwidth=2:
 */
